/*
 * spline.cpp
 * 
 * Copyright 2013 Vova <vova@lexx>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 * 
 * 
 */

#include "spline.h"
#include <string.h>
#include <math.h>
#include <TVectorD.h>
#include <TMatrixD.h>
#include <TDecompSVD.h>
#include <TDecompQRH.h>
#include <TProfile.h>
#include <TProfile2D.h>
#include <QVector>
#include <QtAlgorithms>
#include "jsonparser.h"
#include <QDebug>

//#define USE_EIGEN
#ifdef USE_EIGEN
#include <eigen3/Eigen/Dense>
#endif

// TODO:
// - Consider encalsulating this fuctionality in classes derived from BSpline and TPSpline

void write_bspline3_json(Bspline3 *bs, QJsonObject &json)
{
    int nint = bs->GetNint();
    QJsonArray data;

    json["xmin"] = bs->GetXmin();
    json["xmax"] = bs->GetXmax();
    json["intervals"] = nint;
    for (int i=0; i<nint+3; i++)
        data.append(bs->GetCoef(i));
    json["data"] = data;
}

void write_bspline3_json(Bspline3 *bs, std::ofstream *out)
{
    *out << "{" << std::endl;
    *out << "\"xmin\": " << bs->GetXmin() << "," << std::endl;
    *out << "\"xmax\": " << bs->GetXmax() << "," << std::endl;
    *out << "\"intervals\": " << bs->GetNint() << "," << std::endl;
    *out << "\"data\": [" << std::endl;
    int nint = bs->GetNint();
    for (int i=0; i<nint+3; i++)
        *out << bs->GetCoef(i) << ", ";
    *out << std::endl;
    *out << "]" << std::endl;
    *out << "}" << std::endl;
}

void write_tpspline3_json(TPspline3 *ts, QJsonObject &json)
{
    int nintx = ts->GetNintX();
    int ninty = ts->GetNintY();
    QJsonArray datax, datay;

    json["xmin"] = ts->GetXmin();
    json["xmax"] = ts->GetXmax();
    json["ymin"] = ts->GetYmin();
    json["ymax"] = ts->GetYmax();
    json["xintervals"] = nintx;
    json["yintervals"] = ninty;

    for (int i=0; i<ninty+3; i++) {
        for (int j=0; j<nintx+3; j++)
            datax.append(ts->GetCoef(i*(nintx+3)+j));
        datay.append(datax);
        for (int j=0; j<nintx+3; j++)   // couldn't find a shorter way
            datax.removeLast();         // to clear a QJsonArray
    }
    json["data"] = datay;
}

Bspline3 *read_bspline3_json(QJsonObject &json)
{
    JsonParser parser(json);
    double xmin, xmax;
    int n_int;
    QJsonArray data;
    QVector <double> coef;

    parser.ParseObject("xmin", xmin);
    parser.ParseObject("xmax", xmax);
    parser.ParseObject("intervals", n_int);
    parser.ParseObject("data", data);
    parser.ParseArray(data, coef);
    if (parser.GetError() || coef.size()<n_int+3)
        return NULL;

    Bspline3 *bs = new Bspline3(xmin, xmax, n_int);
    bs->SetCoef(coef.data());
    return bs;
}

TPspline3 *read_tpspline3_json(QJsonObject &json)
{
    JsonParser parser(json);
    double xmin, xmax, ymin, ymax;
    int nintx, ninty;
    QJsonArray data;
    QVector <double> coef, onerow;
    QVector <QJsonArray> rows;

    parser.ParseObject("xmin", xmin);
    parser.ParseObject("xmax", xmax);
    parser.ParseObject("ymin", ymin);
    parser.ParseObject("ymax", ymax);
    parser.ParseObject("xintervals", nintx);
    parser.ParseObject("yintervals", ninty);
    if (!parser.ParseObject("data", data))
        return NULL;
    if (!parser.ParseArray(data, rows))
        return NULL;
    if (rows.size()<ninty+3) {
        qDebug() << rows.size() << " rows";
        return NULL;
    }
    for (int i=0; i<ninty+3; i++) {
        onerow.clear();
        parser.ParseArray(rows[i], onerow);
        if (onerow.size()<nintx+3)
            return NULL;
        coef += onerow;
    }
    if (parser.GetError())
        return NULL;

    TPspline3 *ts = new TPspline3(xmin, xmax, nintx, ymin, ymax, ninty);
    ts->SetCoef(coef.data());

    return ts;
}

#ifdef USE_PUGIXML
void write_bspline3_xml(Bspline3 *bs, std::ofstream *out)
{
    *out << "<bspline3>" << std::endl;
    *out << "<xmin>" << bs->GetXmin() << "</xmin>" << std::endl;
    *out << "<xmax>" << bs->GetXmax() << "</xmax>" << std::endl;
    *out << "<intervals>" << bs->GetNint() << "</intervals>" << std::endl;
    *out << "<data>" << std::endl;
    int nint = bs->GetNint();
    for (int i=0; i<nint+3; i++)
        *out << bs->GetCoef(i) << " ";
    *out << std::endl;
    *out << "</data>" << std::endl;
    *out << "</bspline3>" << std::endl;
}

Bspline3 *read_bspline3_xml(pugi::xml_node node)
{
    double xmin = node.child("xmin").text().as_double();
    double xmax = node.child("xmax").text().as_double();
    int n_int = node.child("intervals").text().as_int();
    Bspline3 *bs = new Bspline3(xmin, xmax, n_int);
    std::string data(node.child("data").text().get());
    std::istringstream datastream(data);

    double *c = new double[n_int+3];
    for (int i=0; i<n_int+3; i++)
        datastream >> c[i];
    bs->SetCoef(c);
    delete[] c;

    return bs;
}

void write_tpspline3_xml(TPspline3 *ts, std::ofstream *out)
{
    *out << "<tpspline3>" << std::endl;
    *out << "<xmin>" << ts->GetXmin() << "</xmin>" << std::endl;
    *out << "<xmax>" << ts->GetXmax() << "</xmax>" << std::endl;
    *out << "<ymin>" << ts->GetYmin() << "</ymin>" << std::endl;
    *out << "<ymax>" << ts->GetYmax() << "</ymax>" << std::endl;
    *out << "<xintervals>" << ts->GetNintX() << "</xintervals>" << std::endl;
    *out << "<yintervals>" << ts->GetNintY() << "</yintervals>" << std::endl;

    int nintx = ts->GetNintX();
    int ninty = ts->GetNintY();
    *out << "<data>" << std::endl;
    for (int i=0; i<ninty+3; i++) {
        for (int j=0; j<nintx+3; j++)
            *out << ts->GetCoef(i*(nintx+3)+j) << " ";
        *out << std::endl;
    }
    *out << "</data>" << std::endl;
    *out << "</tpspline3>" << std::endl;
}

TPspline3 *read_tpspline3_xml(pugi::xml_node node)
{
    double xmin = node.child("xmin").text().as_double();
    double xmax = node.child("xmax").text().as_double();
    int nintx = node.child("xintervals").text().as_int();
    double ymin = node.child("ymin").text().as_double();
    double ymax = node.child("ymax").text().as_double();
    int ninty = node.child("yintervals").text().as_int();
    TPspline3 *ts = new TPspline3(xmin, xmax, nintx, ymin, ymax, ninty);
    std::string data(node.child("data").text().get());
    std::istringstream datastream(data);

    double *c = new double[(nintx+3)*(ninty+3)];
    for (int i=0; i<ninty+3; i++)
        for (int j=0; j<nintx+3; j++)
            datastream >> c[i*(nintx+3)+j];
    ts->SetCoef(c);
    delete[] c;

    return ts;
}
#endif

#ifdef USE_EIGEN
using Eigen::MatrixXd;
using Eigen::VectorXd;
using Eigen::JacobiSVD;
double fit_bspline3(Bspline3 *bs, int npts, double *datax, double *datay, bool even)
{
	int nbas = bs->GetNbas(); 	// numbes of basis functions
	Bool_t success;
    int expand = even ? 2 : 0;
		
// linear least squares	
// we've got npts (+1 if forcing even) equations with nbas unknowns	
    MatrixXd A(npts+expand, nbas);
    VectorXd y(npts+expand);

// fill the equations	
	for (int i=0; i<npts; i++) {
		y(i) = datay[i];
		for (int k=0; k<nbas; k++)
			A(i, k) = bs->Basis(datax[i], k);
	}

// force the function even by setting 
// the derivative to 0 at x=0 in the last equation
	if (even) {
		double strength = ((double)npts)/nbas;
        double xmax = bs->GetXmax();
		y(npts) = 0.;
		for (int k=0; k<nbas; k++)
			A(npts, k) = bs->BasisDrv(0., k)*strength;
    // also constrain the tail
        y(npts+1) = 0.;
        for (int k=0; k<nbas; k++)
            A(npts+1, k) = bs->BasisDrv(xmax-0.0001, k)*strength;
	}
		
// solve the system
    JacobiSVD <MatrixXd> svd(A, Eigen::ComputeThinU | Eigen::ComputeThinV);
    std::cout << "Singular values: " << svd.singularValues();
    VectorXd c_svd = svd.solve(y);
//	if (!success) {
//		printf("fit_bspline3: SVD failed\n");
//		return -1.;
//	}

// fill bs from the solution vector
	double *c = new double[nbas];
	for (int i=0; i<nbas; i++)
		c[i] = c_svd(i);
	bs->SetCoef(c);
	delete[] c;
	
    VectorXd residual = A*c_svd - y;
    return sqrt(residual.squaredNorm());
}
#else

double fit_bspline3(Bspline3 *bs, int npts, double *datax, double *datay, double *weight, bool even)
{
    int nbas = bs->GetNbas(); 	// number of basis functions
    Bool_t success;
    int expand = even ? 2 : 0;

// linear least squares
// we've got npts (+1 if forcing even) equations with nbas unknowns
    TMatrixD A(npts+expand, nbas);
    TVectorD y(npts+expand);

// fill the equations
    for (int i=0; i<npts; i++) {
        if (weight) {
            y(i) = datay[i]*weight[i];
            for (int k=0; k<nbas; k++)
                A(i, k) = bs->Basis(datax[i], k)*weight[i];
        } else {
            y(i) = datay[i];
            for (int k=0; k<nbas; k++)
                A(i, k) = bs->Basis(datax[i], k);
        }
    }

// force the function even by setting
// the derivative to 0 at x=0 in the last equation
    if (even) {
        double strength;
        if (weight) {
            double sum=0.;
            for (int i=0; i<npts; i++)
                sum += weight[i];
            strength = sum/nbas;
        } else {
            strength = ((double)npts)/nbas;
        }
        double xmax = bs->GetXmax();
        y(npts) = 0.;
        for (int k=0; k<nbas; k++)
            A(npts, k) = bs->BasisDrv(0., k)*strength;
    // also constrain the tail
        y(npts+1) = 0.;
        for (int k=0; k<nbas; k++)
            A(npts+1, k) = bs->BasisDrv(xmax-0.0001, k)*strength;
    }
/*
    for (int i=0; i<npts+expand; i++) {
        QDebug debug = qDebug();
        for (int j=0; j<nbas; j++) {
            debug << A(i,j) << ", ";
        }
    }
*/
// solve the system
    TDecompSVD svd(A);
    const TVectorD c_svd = svd.Solve(y, success);
    if (!success) {
        qDebug() << "fit_bspline3: SVD: Can't solve, the system is singular";
        return -1.;
    }

//    TVectorD sig = svd.GetSig();
//    qDebug() << "Singular values: " << sig(0) << " ... " << sig(nbas-1);

// fill bs from the solution vector
    double *c = new double[nbas];
    for (int i=0; i<nbas; i++)
        c[i] = c_svd(i);
    bs->SetCoef(c);
    delete[] c;

    TVectorD residual = A*c_svd - y;
    return sqrt(residual.Norm2Sqr());
}

// wrapper for compatibility
double fit_bspline3(Bspline3 *bs, int npts, double *datax, double *datay, bool even)
{
    return fit_bspline3(bs, npts, datax, datay, NULL, even);
}

#endif


int get_data_from_profile(TProfile *h, std::vector <double> *x, std::vector <double> *y,
                          std::vector <double> *w, std::vector <double> *rms)
{
    int nbins = h->GetNbinsX(); // number of bins

    for (int ix=1; ix<=nbins; ix++) {
        int nent = h->GetBinEntries(ix); // number of entries in the bin ix
        if ( nent < 1)
            continue;
        x->push_back(h->GetXaxis()->GetBinCenter(ix));
        y->push_back(h->GetBinContent(ix));
        w->push_back(nent);
        rms->push_back(h->GetBinError(ix) * sqrt((double)nent));
    }
    return w->size();
}

int get_data_from_profile2D(TProfile2D *h, std::vector <double> *x, std::vector <double> *y,
                            std::vector <double> *z, std::vector <double> *w, std::vector <double> *rms)
{
    int nbinsx = h->GetNbinsX(); // number of bins
    int nbinsy = h->GetNbinsY(); // number of bins
    int weight;

    for (int ix=1; ix<=nbinsx; ix++)
      for (int iy=1; iy<=nbinsy; iy++) {
        if ((weight = h->GetBinEntries(h->GetBin(ix,iy))) < -1)
            continue;
        x->push_back(h->GetXaxis()->GetBinCenter(ix));
        y->push_back(h->GetYaxis()->GetBinCenter(iy));
        z->push_back(h->GetBinContent(ix, iy));
        w->push_back((double)weight);
        rms->push_back(h->GetBinError(ix, iy) * sqrt((double)weight));
    }
    return w->size();
}

double fit_bspline3_grid(Bspline3 *bs, int npts, double *datax, double *datay, bool even, bool takeroot)
{
    // number of intervals can't be less then number of basis functions in the spline
    // we set it (somewhat arbitrarily) to 4 times number of intervals in the spline
    int nint = bs->GetNint()*4;

    //qDebug() << "Fit bslpine grid, takeroot = " << takeroot;

    TProfile *h1 = new TProfile("bs3aux", "", nint, bs->GetXmin(), bs->GetXmax());
    for (int i=0; i<npts; i++)
        h1->Fill(datax[i], datay[i]);

    std::vector <double> x, y, w, rms;
    int ngood = get_data_from_profile(h1, &x, &y, &w, &rms);
    delete h1;
    if (takeroot) {
        for (int i=0; i<(int)y.size(); i++){
            y[i] = sqrt(y[i]);
//            qDebug() << "Point " << i << ": " << x[i] << ", " << y[i];
        }
    }
    return fit_bspline3(bs, ngood, &x[0], &y[0], even);
}

#ifdef USE_EIGEN
double fit_tpspline3(TPspline3 *bs, int npts, double *datax, double *datay, double *dataz, double *weight, bool even)
{
    int nbas = bs->GetNbas(); 	// numbes of basis functions
    int nbasx = bs->GetBSX()->GetNbas();
    int nbasy = bs->GetBSY()->GetNbas();
    Bool_t success;

    std::vector <double> vz;
    std::vector <double> vrow;
    std::vector <std::vector <double> > vmat;

// linear least squares

    // ========== fill the equations ===========

    for (int i=0; i<npts; i++) {
        int w = weight ? weight[i] : 1.;
        if (w>0.5) { // normal point
// f(x,y) = z
            vz.push_back(dataz[i]*w);
            for (int k=0; k<nbas; k++)
                vrow.push_back(bs->Basis(datax[i], datay[i], k)*w);
            vmat.push_back(vrow);
            vrow.clear();
        } else { // empty point - force derivatives to zero
// df/dx = 0
            vz.push_back(0.);
            for (int k=0; k<nbas; k++)
                vrow.push_back(bs->BasisDrvX(datax[i], datay[i], k));
            vmat.push_back(vrow);
            vrow.clear();
// df/dy = 0
            vz.push_back(0.);
            for (int k=0; k<nbas; k++)
                vrow.push_back(bs->BasisDrvY(datax[i], datay[i], k));
            vmat.push_back(vrow);
            vrow.clear();
        }
    }
    // if required, force the function even by setting
    // the derivative d/dx to 0 along x=0 axis
    if (even) {
        double strength = ((double)npts)/nbas;
        for (int iy=0; iy<nbasy; iy++) {
            vz.push_back(0.);
            vrow.assign(nbas, 0.);
            for (int ix=0; ix<nbasx; ix++)
                vrow[iy*nbasx+ix] = bs->GetBSX()->BasisDrv(0., ix)*strength;
            vmat.push_back(vrow);
            vrow.clear();
        }
    // also tell it NOT wiggle the tail by setting d/dx to 0 at x=xmax
    // (experimental feature)
        double xmax = bs->GetXmax();
        for (int iy=0; iy<nbasy; iy++) {
            vz.push_back(0.);
            vrow.assign(nbas, 0.);
            for (int ix=0; ix<nbasx; ix++)
                vrow[iy*nbasx+ix] = bs->GetBSX()->BasisDrv(xmax-0.001, ix)*strength;
            vmat.push_back(vrow);
            vrow.clear();
        }
    }

// ========= solve the system using ROOT machinery =========

// we've got vz.size() equations with nbas unknowns
    MatrixXd A(vz.size(), nbas);
    VectorXd z(vz.size());

    for (int irow=0; irow<vz.size(); irow++) {
        z(irow) = vz[irow];
        for (int icol=0; icol<nbas; icol++)
            A(irow, icol) = vmat[irow][icol];
    }

    JacobiSVD <MatrixXd> svd(A, Eigen::ComputeThinU | Eigen::ComputeThinV);
    std::cout << "Singular values: " << svd.singularValues();
    VectorXd c_svd = svd.solve(z);

//    if (!success) {
//        printf("fit_bspline3: SVD failed\n");
//        return -1.;
//    }

// fill bs from the solution vector
    double *c = new double[nbas];
    for (int i=0; i<nbas; i++)
        c[i] = c_svd(i);
    bs->SetCoef(c);
    delete[] c;

    VectorXd residual = A*c_svd - z;
    return sqrt(residual.squaredNorm());
}
#else
double fit_tpspline3(TPspline3 *bs, int npts, double *datax, double *datay, double *dataz, double *weight, bool even)
{
    int nbas = bs->GetNbas(); 	// numbes of basis functions
    int nbasx = bs->GetBSX()->GetNbas();
    int nbasy = bs->GetBSY()->GetNbas();
    Bool_t success;

// a dirty wrap hack. removed. Raimundo
//turns out it's really dirty, and breaks things
    /*if (bs->GetWrapY()) {
        nbasy -= 3;
        nbas -= nbasx*3;
    }*/

    std::vector <double> vz;
    std::vector <double> vrow;
    std::vector <std::vector <double> > vmat;
    /*static*/ std::vector <std::vector <double> > vcache;
//    /*static*/ TDecompSVD svd;
    //TDecompQRH svd;
    TDecompSVD svd; //Slower but works... (probably because after-treatment wasn't correct with QRH). Raimundo

// linear least squares

    // ========== fill the equations ===========

    for (int i=0; i<npts; i++) {
        int w = weight ? weight[i] : 1.;
        if (w>0.5) { // normal point
// f(x,y) = z
            vz.push_back(dataz[i]*w);
            for (int k=0; k<nbas; k++)
                vrow.push_back(bs->Basis(datax[i], datay[i], bs->GetWrapped(k))*w);
            vmat.push_back(vrow);
            vrow.clear();
        }
/* This might be a good idea for polar fitting when a large contigous chunk
 * of data is missing, but in other cases may cause unnecessary oscillations.
 * Commented out for now, may convert later in an option just for polar case */
// Turns out its a bad idea... NOT Commented. Raimundo
        else { // empty point - force derivatives to zero
// df/dx = 0
            vz.push_back(0.);
            for (int k=0; k<nbas; k++)
                vrow.push_back(bs->BasisDrvX(datax[i], datay[i], bs->GetWrapped(k)));
            vmat.push_back(vrow);
            vrow.clear();
// df/dy = 0
            vz.push_back(0.);
            for (int k=0; k<nbas; k++)
                vrow.push_back(bs->BasisDrvY(datax[i], datay[i], bs->GetWrapped(k)));
            vmat.push_back(vrow);
            vrow.clear();
        }
    }
    // if required, force the function even by setting
    // the derivative d/dx to 0 along x=0 axis
    if (even) {
        double strength = ((double)npts)/nbas;
        for (int iy=0; iy<nbasy; iy++) {
            vz.push_back(0.);
            vrow.assign(nbas, 0.);
            for (int ix=0; ix<nbasx; ix++)
                vrow[iy*nbasx+ix] = bs->GetBSX()->BasisDrv(0., ix)*strength;
            vmat.push_back(vrow);
            vrow.clear();
        }
    // also tell it NOT wiggle the tail by setting d/dx to 0 at x=xmax
    // (experimental feature)
        double xmax = bs->GetXmax();
        for (int iy=0; iy<nbasy; iy++) {
            vz.push_back(0.);
            vrow.assign(nbas, 0.);
            for (int ix=0; ix<nbasx; ix++)
                vrow[iy*nbasx+ix] = bs->GetBSX()->BasisDrv(xmax-0.001, ix)*strength;
            vmat.push_back(vrow);
            vrow.clear();
        }
    }

// ========= solve the system using ROOT machinery =========

// we've got vz.size() equations with nbas unknowns
    TMatrixD A(vz.size(), nbas);
    TVectorD z(vz.size());


// compare the matrix with one saved from the last call (vcache)
// if they are equal we'll save a lot of time by reusing SVD
    bool use_cache = false;
/*
    if (vmat.size() == vcache.size()) {
        for (unsigned int i=0; i<vmat.size(); i++)
            if (!(vmat[i] ==vcache[i]))
                break;
        use_cache = true;
    }
*/
// matrix changed: save the new one to vcache and do SVD
    if (!use_cache) {
        vcache.clear();
        vcache.resize(vmat.size());
        for (unsigned int irow=0; irow<vz.size(); irow++) {
//            z(irow) = vz[irow];
            vcache[irow].resize(vmat[irow].size());
            for (int icol=0; icol<nbas; icol++) {
                A(irow, icol) = vcache[irow][icol] = vmat[irow][icol];
            }
        }
        svd.SetMatrix(A);
    }

// update the right-hand side
    for (unsigned int irow=0; irow<vz.size(); irow++)
        z(irow) = vz[irow];

// solve
    const TVectorD c_svd = svd.Solve(z, success);
    if (!success) {
        fprintf(stderr, "fit_bspline3: SVD failed\n");
        return -1.;
    }

    //TVectorD sig = svd.GetSig();
    //fprintf(stderr, "Singular values: %fl ... %lf\n", sig(0), sig(nbas-1));

// fill bs from the solution vector
    //nbas = bs->GetNbas(); // restore in case it was mangled by wrap hack
    double *c = new double[nbas];
    for (int i=0; i<nbas; i++)
        c[i] = c_svd(bs->GetWrapped(i));    // duplicate wrapped knots if necessary
    bs->SetCoef(c);
    delete[] c;

    TVectorD residual = A*c_svd - z;
    double resid = sqrt(residual.Norm2Sqr());
    return resid;
}
#endif

/*
double fit_tpspline3(TPspline3 *bs, int npts, double *datax, double *datay, double *dataz, double *weight, bool even)
{
	int nbas = bs->GetNbas(); 	// numbes of basis functions
    int nbasx = bs->GetBSX()->GetNbas();
    int nbasy = bs->GetBSY()->GetNbas();
	Bool_t success;
    int zeros = 0;
    if (weight)
        for (int i=0; i<npts; i++)
            if (weight[i]<0.5)
                zeros++;

    printf("zeros: %d\n", zeros);
    int expand = even ? nbasy : 0;
    expand += zeros;

// linear least squares
// we've got npts (+nbasy if forcing even) equations with nbas unknowns
    TMatrixD A(npts+expand, nbas);
    TVectorD z(npts+expand);

    std::vector <double> vx, vy;

// fill the equations	
	for (int i=0; i<npts; i++) {
        int w = weight ? weight[i] : 1.;
        if (w>0.5) {
            z(i) = dataz[i]*w;
            for (int k=0; k<nbas; k++)
                A(i, k) = bs->Basis(datax[i], datay[i], k)*w;
        } else {
            z(i) = 0.;
            for (int k=0; k<nbas; k++)
                A(i, k) = bs->BasisDrvX(datax[i], datay[i], k);
            vx.push_back(datax[i]);
            vy.push_back(datay[i]);
        }
    }

    for (int j=0; j<zeros; j++) {
        int i = npts + j;
        z(i) = 0.;
        for (int k=0; k<nbas; k++)
            A(i, k) = bs->BasisDrvY(vx[i], vy[i], k);
    }

// force the function even by setting
// the derivative d/dx to 0 along x=0 axis
    if (even) {
        double strength = ((double)npts)/nbas;
        for (int iy=0; iy<nbasy; iy++) {
            z(npts+zeros+iy) = 0.;
            for (int ix=0; ix<nbasx; ix++)
                A(npts+zeros+iy, iy*nbasx+ix) = bs->GetBSX()->BasisDrv(0., ix)*strength;
        }
    }

// solve the system
	TDecompSVD svd(A);
	const TVectorD c_svd = svd.Solve(z, success);
	if (!success) {
		printf("fit_bspline3: SVD failed\n");
		return -1.;
	} 

// fill bs from the solution vector
	double *c = new double[nbas];
	for (int i=0; i<nbas; i++)
		c[i] = c_svd(i);
	bs->SetCoef(c);
	delete[] c;
	
	TVectorD residual = A*c_svd - z;
	return sqrt(residual.Norm2Sqr());
}
*/
double fit_tpspline3_grid(TPspline3 *bs, int npts, double *datax, double *datay, double *dataz, bool even)
{
    // number of intervals can't be less then number of basis functions in the spline
    // we set it (somewhat arbitrarily) to 4 times number of intervals in the spline
    int nintx = bs->GetNintX()*2;
    int ninty = bs->GetNintY()*2;
//    printf ("nx = %d, ny = %d\n", nintx, ninty);
//    printf ("ymin = %lf, ymax = %lf\n", bs->GetYmin(), bs->GetYmax());

    TProfile2D *h1 = new TProfile2D("bs32Daux", "", nintx, bs->GetXmin(), bs->GetXmax(),
                                    ninty, bs->GetYmin(), bs->GetYmax());
    for (int i=0; i<npts; i++)
        h1->Fill(datax[i], datay[i], dataz[i]);

//    h1->DrawPanel();
    std::vector <double> x, y, z, w, rms;
    int ngood = get_data_from_profile2D(h1, &x, &y, &z, &w, &rms);
//    for (int j=0; j<w.size(); j++)
//        printf("%f, %f, %f, %f\n", x[j], y[j], z[j], w[j]);
    delete h1;
    return fit_tpspline3(bs, ngood, &x[0], &y[0], &z[0], &w[0], even);
}






