#ifndef CLASSES_H
#define CLASSES_H

#include <QString>
#include <QVector>
#include <QtGui>

//Root
#include "TVector3.h"
#include "TH1D.h"
#include "TPolyMarker3D.h"

class TGeoMaterial;
class TGeoMedium;

struct EnergyDepositionCell  //element of the EnergyVector QVector
    {
        double r[3];  // x,y,z -coordinates of the cell
        double cellLength; //cell length for charged particles
        double time;
        double dE; //energy deposited in this cell
        int ParticleId;          // index of the particle form the particle list
        int MaterialId;         //  index of the material
        int index;  //particle number - need for visualization if tracking several particles with same id - index continuosly increases throughout the EnergyVector
        int eventId; //for EnergyVector import mode (to be done) - event number
    };

enum ParticleType {_gamma_ = 0, _charged_, _neutron_};

class ParticleDescriptor //properties of the particles
{
public:
  ParticleDescriptor(QString Name, ParticleType Type, int Charge, double Mass)
  {ParticleName = Name; type = Type, charge = Charge; mass = Mass;}
  ParticleDescriptor(){ParticleName = "undefined", type = _charged_, charge = 1, mass = 666;}

  QString ParticleName;
  ParticleType type; //gamma, charged, neutron
  double mass;
  int charge;

  void writeToJson(QJsonObject &json) {json["Name"]=ParticleName;json["Type"]=type;json["Mass"]=mass;json["Charge"]=charge;}
  void readFromJson(QJsonObject &json) {ParticleName=json["Name"].toString();type=static_cast<ParticleType>(json["Type"].toInt());mass=json["Mass"].toDouble();charge=json["Charge"].toInt();}
};

class ParticleOnStack //element of the particle stack
{
  public:
    ParticleOnStack(int ID, double x, double y, double z, double vx, double vy, double vz, double T, double E,  int PrSec = 0)
    {Id = ID; flagSecondary = PrSec;r[0]=x;r[1]=y;r[2]=z; v[0]=vx;v[1]=vy;v[2]=vz; t = T; Energy = E;}
    ParticleOnStack(){;}

    int Id;
    int flagSecondary; //0-primary, 1 - secondary
    double r[3];  //starting point
    double v[3];  //starting vector
    double t;     //time on start
    double Energy; //staring energy

    ParticleOnStack* clone()
    {
       return new ParticleOnStack(Id, r[0], r[1], r[2], v[0], v[1], v[2], t, Energy, flagSecondary);
    }
};

struct One_Photon
{
    double r[3]; //position
    double v[3]; //direction (must be already normalized to unit vector!!!)
    double time; //time stamp
    double wavelength;
    int WaveIndex;
    int scint_type; //1 - primary //2 - secondary // 0 - ?
    bool SkipThisPhoton; //flag to skip this photon due to e.g. direction check

    One_Photon() {SkipThisPhoton = false;}
};

struct MaterialHistoryStructure
{
    MaterialHistoryStructure(int mat, double en, double dis) {MaterialId = mat; DepositedEnergy = en; Distance = dis;}
    MaterialHistoryStructure(){;}
    int MaterialId;
    double DepositedEnergy;
    double Distance;   
};

struct EventHistoryStructure
{
    int ParticleId;
    int index; // this is particle index! - "serial number" of the particle
    int flagSecondary;
    float dx, dy, dz; //direction
    int Termination;    
    //1 - escaped
    //2 - all energy dissipated
    //3 - photoeffect
    //4 - compton
    //5 - capture
    //6 - error - undefined termination
    //7 - created outside defined geometry
    //8 - found untrackable material

    QVector<MaterialHistoryStructure> Deposition;
};

struct GeneratedPhotonsHistoryStructure
{
    int event;
    int index; //# t distinguish between particles with the same ParticleId (e.g. during the same event)
    int ParticleId;
    int MaterialId;
    double Energy;
    int PrimaryPhotons;
    int SecondaryPhotons;
};

struct GammaStructure
{
    double energy;
    TVector3 direction;
};

//===========================================
struct PositionEnergyRecord
{
    double r[3];
    double energy;
};

class PositionEnergyBuffer
{
 private:
    int numRecords;

 public:
    PositionEnergyRecord *rec; //pointer to array of position energy records

    PositionEnergyBuffer() { rec = new PositionEnergyRecord[1]; numRecords = 1;}   //by default 1 point!
    ~PositionEnergyBuffer(){ delete [] rec; }

    //public functions
    inline PositionEnergyRecord &operator[](int i)  //get record number i
      { return rec[i]; }
    inline int size() {return numRecords;}          //get number of points
    inline void Reinitialize(int size)              //resize (data is undefined!)
      {
         delete [] rec;
         rec = new PositionEnergyRecord[size];
         numRecords = size;
      }
    void AddPoint(double X, double Y, double Z, double Energy) //add new point to the record
      {
        PositionEnergyRecord *recOld = rec;

        rec = new PositionEnergyRecord[numRecords+1];
        for (int i=0; i<numRecords; i++)
          {
            for (int j=0; j<3; j++) rec[i].r[j] = recOld[i].r[j];
            rec[i].energy = recOld[i].energy;
          }        
        delete [] recOld;
        rec[numRecords].r[0] = X; rec[numRecords].r[1] = Y; rec[numRecords].r[2] = Z;
        rec[numRecords].energy = Energy;
        numRecords++;
      }
    void AddPoint(double* R, double Energy) //add new point to the record
      {
        PositionEnergyRecord *recOld = rec;

        rec = new PositionEnergyRecord[numRecords+1];
        for (int i=0; i<numRecords; i++)
          {
            for (int j=0; j<3; j++) rec[i].r[j] = recOld[i].r[j];
            rec[i].energy = recOld[i].energy;
          }
        delete [] recOld;
        rec[numRecords].r[0] = R[0]; rec[numRecords].r[1] = R[1]; rec[numRecords].r[2] = R[2];
        rec[numRecords].energy = Energy;
        numRecords++;
      }
};

struct BaseScanAndReconStructure
{   
   PositionEnergyBuffer Points;  //collection of points (elements contain:    position, energy)
   bool GoodEvent; //true - good event  (for scan - bad event = noise event)
};

struct ScanStructure : BaseScanAndReconStructure
{
  //pos+energy+Good  - see Base
  int EventType; //type of noise event

  int ScintType; //0 - unknown, 1 - primary, 2 - secondary
  double zStop;   //if secondary, r[2] is start and zStop is stop (cover SecScint z-width)

  ScanStructure() {GoodEvent = true;}  //by default there is one point //good event by defaut
};

struct ReconstructionStruct : BaseScanAndReconStructure
{
  //pos+energy+Good  - see Base
  double chi2;
  bool ReconstructionOK; // false = reconstruction failed
  int EventId; //serial number, should be kept if a part of data is purged

    //CoG info
  double xCoG, yCoG, zCoG;
  int iPMwithMaxSignal;

    //for all filters: 0 - not set, 1 - pass, 2 - fail
  char FilterEventNumber;
  char FilterCutOff;
  char FilterSumCutOff;
  char FilterEnergy;
  char FilterLoadedEnergy;
  char FilterChi2;
  char FilterPosition;
  char FilterCorrelation;
  char FilterkNN;

  ReconstructionStruct() {xCoG=0; yCoG=0; zCoG=0; iPMwithMaxSignal=0; chi2=0; }  //by default there is one point
  void CopyTo(ReconstructionStruct* target) //copy all data to another ReconstructionStructure
    {
      if (target == this) return;
      int targetSize = target->Points.size();
      int thisSize = Points.size();
      if (targetSize != thisSize) target->Points.Reinitialize(thisSize);
      for (int i=0; i<thisSize; i++)
        {
          target->Points[i].energy = Points[i].energy;
          target->Points[i].r[0] = Points[i].r[0];
          target->Points[i].r[1] = Points[i].r[1];
          target->Points[i].r[2] = Points[i].r[2];
        }
      target->xCoG = xCoG;
      target->yCoG = yCoG;
      target->zCoG = zCoG;
      target->iPMwithMaxSignal = iPMwithMaxSignal;
      target->GoodEvent = GoodEvent;
      target->EventId = EventId;
      target->chi2 = chi2;
      target->ReconstructionOK = ReconstructionOK;
      target->FilterEventNumber = FilterEventNumber;
      target->FilterCutOff = FilterCutOff;
      target->FilterSumCutOff = FilterSumCutOff;
      target->FilterEnergy = FilterEnergy;
      target->FilterLoadedEnergy = FilterLoadedEnergy;
      target->FilterChi2 = FilterChi2;
      target->FilterPosition = FilterPosition;
      target->FilterCorrelation = FilterCorrelation;
      target->FilterkNN = FilterkNN;
    }

  void report()
  {
    qDebug() << "----------------";
    qDebug() << "Event id:"<<EventId;
    qDebug() << "Points:"<<Points.size();
    for (int i=0; i<Points.size(); i++)
      {
        qDebug() << "   XYZ:"<<Points[i].r[0]<<Points[i].r[1]<<Points[i].r[2];
        qDebug() << "   Energy:"<<Points[i].energy;
      }
    qDebug() << "CoG XYZ"<<xCoG<< yCoG<< zCoG<<" max signal PM:"<<iPMwithMaxSignal;
    qDebug() << "Chi2:"<<chi2;
    qDebug() << "Good?"<<GoodEvent;
    qDebug() << "Reconstruction OK?"<<ReconstructionOK;
    qDebug() << "Filters:";
    qDebug() << "  FilterEventNumber"<<(int)FilterEventNumber;
    qDebug() << "  FilterCutOff"<<(int)FilterCutOff;
    qDebug() << "  FilterSumCutOff"<<(int)FilterSumCutOff;
    qDebug() << "  FilterEnergy"<<(int)FilterEnergy;
    qDebug() << "  FilterLoadedEnergy"<<(int)FilterLoadedEnergy;
    qDebug() << "  FilterChi2"<<(int)FilterChi2;
    qDebug() << "  FilterPosition"<<(int)FilterPosition;
    qDebug() << "  FilterCorrelation"<<(int)FilterCorrelation;
    qDebug() << "  FilterkNN"<<(int)FilterkNN;
  }
};

struct DotsTGeoStruct
{
  double r[3]; //position 
  DotsTGeoStruct(){;}
  DotsTGeoStruct(const double* r){this->r[0]=r[0]; this->r[1]=r[1];this->r[2]=r[2];}
  DotsTGeoStruct(double x, double y, double z){this->r[0]=x; this->r[1]=y;this->r[2]=z;}
};

class GeoMarkerClass : public TPolyMarker3D
{
public:
  QString Type; //Source, Nodes, Recon, Scan

  GeoMarkerClass(QString type, int style, int size, Color_t color) {configure(type, style, size, color);}
  GeoMarkerClass() {}
  void configure(QString type, int style, int size, Color_t color) {Type = type; SetMarkerStyle(style); SetMarkerSize(size); SetMarkerColor(color);}
};

struct ScanFloodStructure
{
  QString Description;
  bool Active;
  double Probability;
  double AverageValue;
  double Spread;
};

struct DetectorExampleStructure
{
  int MainType;
  int SubType;
  QString Description;
  QString filename;

  DetectorExampleStructure(const int mainT, const int subT, const QString file, const QString desc){Description = desc; MainType = mainT; filename = file; SubType = subT;}
  DetectorExampleStructure(){;}
};

struct TrackNodeStruct
{
  Double_t R[3];
  Double_t Time;

  TrackNodeStruct(const Double_t* r, Double_t time) {R[0]=r[0]; R[1]=r[1]; R[2]=r[2]; Time=time;}
  TrackNodeStruct(Double_t* r, Double_t time) {R[0]=r[0]; R[1]=r[1]; R[2]=r[2]; Time=time;}
  TrackNodeStruct(Double_t x, Double_t y, Double_t z, Double_t time) {R[0]=x; R[1]=y; R[2]=z; Time=time;}
  TrackNodeStruct(){}
};

class TrackHolderClass
{
public:
  QVector<TrackNodeStruct> Nodes;
  Color_t Color;
  Int_t Width;
  Int_t UserIndex;  

  TrackHolderClass(Int_t index, Color_t color = kBlack) {UserIndex = index; Color = color;}
  TrackHolderClass() {UserIndex = 0; Color = kBlack;}
};

#endif // CLASSES_H
