#include "opticaloverrideclass.h"
#include "TMath.h"

#include "phscatclaudiomodel.h"
#include "scatteronmetal.h"

void BasicOpticalOverride::RandomDir(TRandom2 *RandGen, double *PhotonDir)
{
  //Sphere function of Root:
  double a=0, b=0, r2=1;
  while (r2 > 0.25)
    {
      a  = RandGen->Rndm() - 0.5;
      b  = RandGen->Rndm() - 0.5;
      r2 =  a*a + b*b;
    }
  PhotonDir[2] = ( -1.0 + 8.0 * r2 );
  double scale = 8.0 * TMath::Sqrt(0.25 - r2);
  PhotonDir[0] = a*scale;
  PhotonDir[1] = b*scale;
}

OpticalOverrideClass::OpticalOverrideResultEnum BasicOpticalOverride::calculate(TRandom2 *RandGen, double *PhotonDir, const double *NormalVector, int /*iWave*/)
{
  LastError = "";
  double rnd = RandGen->Rndm();

  // surface loss?
  rnd -= probLoss;
  if (rnd<0)
    {
      // qDebug()<<"Override: surface loss!";
      Status = Absorption;
      return Absorbed;
    }

  // specular reflection?
  rnd -= probRef;
  if (rnd<0)
    {
      // qDebug()<<"Override: specular reflection!";
        //rotating the vector: K = K - 2*(NK)*N
      double NK = NormalVector[0]*PhotonDir[0]; NK += NormalVector[1]*PhotonDir[1];  NK += NormalVector[2]*PhotonDir[2];
      PhotonDir[0] -= 2.0*NK*NormalVector[0]; PhotonDir[1] -= 2.0*NK*NormalVector[1]; PhotonDir[2] -= 2.0*NK*NormalVector[2];
      Status = SpikeReflection;
      return Back;
    }

  // scattering?
  rnd -= probDiff;
  if (rnd<0)
    {
      // qDebug()<<"scattering triggered";
      switch (scatterModel)
        {
        case 0: //4Pi scattering
          // qDebug()<<"4Pi scatter";
          RandomDir(RandGen, PhotonDir);
          // qDebug()<<"New direction:"<<K[0]<<K[1]<<K[2];

          //enering new volume or backscattering?
          //normal is in the positive direction in respect to the original direction!
          if (PhotonDir[0]*NormalVector[0] + PhotonDir[1]*NormalVector[1] + PhotonDir[2]*NormalVector[2] < 0)
            {
              // qDebug()<<"   scattering back";              
              return Back;
            }
          // qDebug()<<"   continuing to the next volume";
          Status = Transmission;
          return Forward;

        case 1: //2Pi lambertian, remaining in the same volume (back scattering)
          {
            // qDebug()<<"2Pi lambertian scattering backward";
            double norm2;
            do
              {
                RandomDir(RandGen, PhotonDir);
                PhotonDir[0] -= NormalVector[0]; PhotonDir[1] -= NormalVector[1]; PhotonDir[2] -= NormalVector[2];
                norm2 = PhotonDir[0]*PhotonDir[0] + PhotonDir[1]*PhotonDir[1] + PhotonDir[2]*PhotonDir[2];
              }
            while (norm2 < 0.000001);

            double normInverted = 1.0/TMath::Sqrt(norm2);
            PhotonDir[0] *= normInverted; PhotonDir[1] *= normInverted; PhotonDir[2] *= normInverted;
            // qDebug()<<"Photon dir after scatter:"<<K[0]<<K[1]<<K[2];
            Status = LambertianReflection;
            return Back;
          }
        case 2: //2Pi lambertian, scattering to the next volume
          {
            // qDebug()<<"2Pi lambertian scattering forward";
            double norm2;
            do
              {
                RandomDir(RandGen, PhotonDir);
                PhotonDir[0] += NormalVector[0]; PhotonDir[1] += NormalVector[1]; PhotonDir[2] += NormalVector[2];
                norm2 = PhotonDir[0]*PhotonDir[0] + PhotonDir[1]*PhotonDir[1] + PhotonDir[2]*PhotonDir[2];
              }
            while (norm2 < 0.000001);

            double normInverted = 1.0/TMath::Sqrt(norm2);
            PhotonDir[0] *= normInverted; PhotonDir[1] *= normInverted; PhotonDir[2] *= normInverted;
            // qDebug()<<"Photon dir after scatter:"<<K[0]<<K[1]<<K[2];
            Status = Transmission;
            return Forward;
          }
        }
    }

  // overrides NOT triggered - what is left is covered by Fresnel in the tracker code
  // qDebug()<<"Overrides did not trigger, using fresnel";
  Status = Transmission;
  return NotTriggered;
}

void BasicOpticalOverride::printConfiguration(int /*iWave*/)
{
  qDebug() << "-------Configuration:-------";
  qDebug() << "Absorption fraction:"<<probLoss;
  qDebug() << "Specular fraction:"<<probRef;
  qDebug() << "Scatter fraction:"<<probDiff;
  qDebug() << "Scatter model (4Pi/LambBack/LambForward):"<<scatterModel;
  qDebug() << "----------------------------";
}

QString BasicOpticalOverride::getReportLine()
{
  QString s = "to " + (*MatCollection)[MatTo]->name;
  QString s1;
  s += "->";

  double prob = probRef+probDiff+probLoss;
  if (prob == 0) return s + " To be defined"; //not defined - shown during configuration phase only
  double probFresnel = 1.0 - prob;  

  if (probLoss>0)
  {
      s1.setNum(probLoss*100.);
      s += " Loss:"+s1+"%";
  }
  if (probRef>0)
  {
      s1.setNum(probRef*100.);
      s += " Spec:"+s1+"%";
  }
  if (probDiff>0)
  {
      s1.setNum(probDiff*100.);
      switch( scatterModel )
        {
        case 0:
          s += " Scat(4Pi):";
          break;
        case 1:
          s += " Scat(Lamb_B):";
          break;
        case 2:
          s += " Scat(Lamb_F):";
          break;
        }
      s += s1+"%";
  }
  s1.setNum(probFresnel*100.);
  s += " Fresnel:"+s1+"%";
  return s;
}

bool BasicOpticalOverride::writeToJson(QJsonObject &json)
{
  json["Model"] = getType();
  json["MatFrom"] = MatFrom;
  json["MatTo"] = MatTo;

  json["Abs"]  = probLoss;
  json["Spec"] = probRef;
  json["Scat"] = probDiff;
  json["ScatMode"] = scatterModel;

  return true;
}

bool BasicOpticalOverride::readFromJson(QJsonObject &json)
{
  QString type = json["Model"].toString();
  if (type != getType()) return false; //file for wrong model!

  probLoss = json["Abs"].toDouble();
  probRef =  json["Spec"].toDouble();
  probDiff = json["Scat"].toDouble();
  scatterModel = json["ScatMode"].toInt();
  return true;
}

OpticalOverrideClass *OpticalOverrideFactory(QString model, MaterialCollectionClass *MatCollection, int MatFrom, int MatTo)
{
   if (model == "Simplistic_model")
     return new BasicOpticalOverride(MatCollection, MatFrom, MatTo);
   else if (model == "Claudio_Model_V1")
     return new PhScatClaudioModelV1(MatCollection, MatFrom, MatTo);
   else if (model == "Claudio_Model_V2")
     return new PhScatClaudioModelV2(MatCollection, MatFrom, MatTo);
   else if (model == "Claudio_Model_V2d1")
     return new PhScatClaudioModelV2d1(MatCollection, MatFrom, MatTo);
   else if (model == "Claudio_Model_V2d2")
     return new PhScatClaudioModelV2d2(MatCollection, MatFrom, MatTo);
   else if (model == "Claudio_Model") //compatibility
     return new PhScatClaudioModelV2(MatCollection, MatFrom, MatTo);
   else if (model == "DielectricToMetal")
     return new ScatterOnMetal(MatCollection, MatFrom, MatTo);

   return NULL; //undefined override type!
}
