//ANTS2 modules
#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "geometrywindowclass.h"
#include "graphwindowclass.h"
#include "lrfwindow.h"
#include "reconstructionwindow.h"
#include "sensorlrfs.h"
#include "pms.h"
#include "materialinspectorwindow.h"
#include "outputwindow.h"
#include "guiutils.h"
#include "windownavigatorclass.h"
#include "exampleswindow.h"
#include "detectoraddonswindow.h"
#include "checkupwindowclass.h"
#include "CorrelationFilters.h"
#include "materialcollectionclass.h"
#include "genericscriptwindowclass.h"
#include "detectorclass.h"
#include "simulationmanager.h"
#include "reconstructionmanagerclass.h"
#include "globalsettingsclass.h"
#include "globalsettingswindowclass.h"

#ifdef ANTS_FANN
#include "neuralnetworksmodule.h"
#include "neuralnetworkswindow.h"
#endif

//Root
#include "TGaxis.h"
#include "TApplication.h"
#include "TMultiGraph.h"

MainWindow::MainWindow(DetectorClass *Detector, EventsDataClass *EventsDataHub, TApplication *RootApp, ReconstructionManagerClass *ReconstructionManager, TmpObjHubClass *TmpHub) :
    QMainWindow(), Detector(Detector), EventsDataHub(EventsDataHub), RootApp(RootApp),
    ReconstructionManager(ReconstructionManager), TmpHub(TmpHub),
    ui(new Ui::MainWindow)
{
    //qDebug() << "-> Main window constructor started";

    //Inits
    GeometryWindow = 0;
    GraphWindow = 0;
    Owindow = 0;
#ifdef ANTS_FANN
    NNwindow = 0;    
#endif
    MIwindow = 0;
    WindowNavigator = 0;
    ELwindow = 0;
    DAwindow = 0;
    CheckUpWindow = 0;
    GainWindow = 0;  
    GenScriptWindow = 0;
    GlobSetWindow = 0;
    SimManager = 0;
    PreviousNumberPMs = 0;    
    tmpPMtype = 0;
    ScanFloodNoiseProbability = 0;
    TMPmultiGraph = 0;
    hist1 = 0;
    hist2 = 0;
    hist2D = 0;
    histScan = 0;
    histSecScint = 0;
    msBox = 0;
    BulkUpdate = false;
    ShutDown = false;
    TriggerForbidden = false;
    ShowTop = false;
    ColorByMaterial = false;
    GeometryDrawDisabled = false;
    fSimDataNotSaved = false;
    ShowExamplesOnStart = true; //Show examples window on start
    fDemoMode = false;
    //aliases to use in GUI
    MaterialCollection = Detector->MaterialCollection; // just an alias
    PMs = Detector->PMs;                               // just an alias
    SensLRF = Detector->SensLRF;                       // just an alias    

    //qDebug()<<"-> Creating Global settings object";
    GlobSet = new GlobalSettingsClass(); //also sets up directories and loads ANTS2 settings(config) if found

    //qDebug()<<"-> Creating user interface for the main window";
    ui->setupUi(this);
    this->setFixedSize(this->size());
    this->move(10,30); //default position
    MainWindow::initDetectorSandwich(); //create layers and link GUI controls
    int minVer = ANTS2_MINOR;
    QString miv = QString::number(minVer);
    if (miv.length() == 1) miv = "0"+miv;
    int majVer = ANTS2_MAJOR;
    QString mav = QString::number(majVer);
    setWindowTitle("ANTS2_v"+mav+"."+miv);

    //adding Context menus
    ui->lwLoadedEventsFiles->setContextMenuPolicy(Qt::CustomContextMenu);
    QObject::connect(ui->lwLoadedEventsFiles, SIGNAL(customContextMenuRequested(const QPoint&)),
        this, SLOT(LoadEventsListContextMenu(const QPoint&)));


    //setting up particle sources container class - just for GUI
    //RandGen = new TRandom2();
    //RandGen->SetSeed();
    //qDebug() << "-> Random generator created";
    ParticleSources.connectRandomGenerator(Detector->RandGen);
    ParticleSources.clear();
    ParticleSources.connectDataContainers(&Detector->ParticleCollection, MaterialCollection); //RandGen is connected after it is initialized!
    MainWindow::on_pbUpdateSourcesIndication_clicked();
    //qDebug() << "->Container for particle sources created and configured";

    //interface windows
    //qDebug()<<"->Creating Examples Window";
    QWidget* w = new QWidget();
    ELwindow = new ExamplesWindow(w, this);
    ELwindow->move(100,100);
    ELwindow->setShowExamplesOnStartStatus(ShowExamplesOnStart);
    //qDebug()<<"->Creating Detector Add-ons window";  //created as child window, no delete on mainwin close!
    DAwindow = new DetectorAddOnsWindow(this, Detector);
    DAwindow->move(50,50);
    //qDebug()<<"->Creating Material Inspector Window";
    w = new QWidget();
    MIwindow = new MaterialInspectorWindow(w, this, Detector);
    MIwindow->move(50,50);
    //qDebug()<<"->Creating Output Window";
    w = new QWidget();
    Owindow = new OutputWindow(w, this, EventsDataHub);
    Owindow->move(600,580);
    //qDebug()<<"->Creating Reconstruction Window";
    w = new QWidget();
    Rwindow = new ReconstructionWindow(w, this, EventsDataHub);
    Rwindow->move(20,250);
    //qDebug()<<"->Creating LRF Window";
    w = new QWidget();
    lrfwindow = new LRFwindow(w, this, EventsDataHub);
    lrfwindow->move(25,25);
#ifdef ANTS_FANN
    //qDebug()<<"->Creating NeuralNetworks Window";
    w = new QWidget();
    NNwindow = new NeuralNetworksWindow(w, this, EventsDataHub);
    NNwindow->move(25,25);
    QObject::connect(Rwindow,SIGNAL(cb3DreconstructionChanged(bool)),NNwindow,SLOT(onReconstruct3D(bool)));
    QObject::connect(Rwindow,SIGNAL(cbReconstructEnergyChanged(bool)),NNwindow,SLOT(onReconstructE(bool)));
#endif
    //qDebug()<<"->Creating check-up window"; //created as child window, no delete on mainwin close!
    CheckUpWindow = new CheckUpWindowClass(this, Detector);
    CheckUpWindow->move(50,50);
    //qDebug()<<"->Creating settings window";
    GlobSetWindow = new GlobalSettingsWindowClass(this);
    GlobSetWindow->move(50,50);
    //qDebug()<<"->Creating window navigator";
    w = new QWidget();
    WindowNavigator = new WindowNavigatorClass(w, this);
    WindowNavigator->move(700,50);
    //qDebug()<<"->Creating Graph Window";
    w = new QWidget();
    GraphWindow = new GraphWindowClass(w, this);
    GraphWindow->move(25,25);
    //qDebug()<<"->Creating Geometry Window";
    w = new QWidget();
    GeometryWindow = new GeometryWindowClass(w, this);
    GeometryWindow->move(25,25);
    //qDebug()<<"-> All windows created";

    //root update cycle
    RootUpdateTimer = new QTimer(this);
    RootUpdateTimer->setInterval(100);
    QObject::connect(RootUpdateTimer, SIGNAL(timeout()), this, SLOT(timerTimeout()));
    RootUpdateTimer->start();
    //qDebug()<<"->Timer to refresh Root events started";

    QObject::connect(EventsDataHub, SIGNAL(loaded(int, int)), this, SLOT(updateLoaded(int, int)));

    QObject::connect(SensLRF, SIGNAL(SensorLRFsReadySignal(bool)), Rwindow, SLOT(LRF_ModuleReadySlot(bool)));
    QObject::connect(SensLRF, SIGNAL(SensorLRFsReadySignal(bool)), lrfwindow, SLOT(LRF_ModuleReadySlot(bool)));
    QObject::connect(SensLRF, SIGNAL(SensorLRFsReadySignal(bool)), this, SLOT(LRF_ModuleReadySlot(bool)));
    QObject::connect(SensLRF, SIGNAL(ProgressReport(int)), lrfwindow, SLOT(onProgressReportReceived(int)));

    QObject::connect(ReconstructionManager, SIGNAL(UpdateReady(int,double)), Rwindow, SLOT(RefreshOnTimer(int,double)));
    QObject::connect(Rwindow, SIGNAL(StopRequested()), ReconstructionManager, SLOT(requestStop()));
#ifdef ANTS_FANN
    QObject::connect(ReconstructionManager->ANNmodule,SIGNAL(status(QString)),NNwindow,SLOT(status(QString)));
    QObject::connect(NNwindow,SIGNAL(train_stop()),ReconstructionManager->ANNmodule,SLOT(train_stop()));
#endif

    SimManager = new SimulationManager(Detector, EventsDataHub);
    QObject::connect(SimManager, SIGNAL(updateReady(int, double)), this, SLOT(RefreshPhotSimOnTimer(int, double))); //Simulation interface refresh/update stuff
    QObject::connect(SimManager, SIGNAL(simulationFinished()), this, SLOT(simulationFinished())); //Simulation finished
    QObject::connect(this, SIGNAL(StopRequested()), SimManager, SLOT(requestStop()), Qt::DirectConnection); //Stop button pressed
    SimManager->moveToThread(&simThread); //Move to background thread, as it always runs synchronously even in MT
    QObject::connect(&simThread, SIGNAL(started()), SimManager, SLOT(simulate())); //Connect thread to simulation start
    simTimerGuiUpdate.setInterval(1000);
    QObject::connect(&simTimerGuiUpdate, SIGNAL(timeout()), SimManager, SLOT(updateGui()), Qt::DirectConnection);
    //qDebug() << "->Simulation manager module created";

    DoNotUpdateGeometry = false; //control

    //qDebug()<<"->Creating default detector configuration";
    MainWindow::startupDetector();

    //Environment
    on_cbTimeResolved_toggled(ui->cbTimeResolved->isChecked());
    on_cbWaveResolved_toggled(ui->cbWaveResolved->isChecked());
    on_cbAngularSensitive_toggled(ui->cbAngularSensitive->isChecked());
    on_cbAreaSensitive_toggled(ui->cbAreaSensitive->isChecked());
    ui->fLoadProgress->setVisible(false);
    TGaxis::SetMaxDigits(3);  //Global setting for cern Root graphs!
     //Script window geometry
    ScriptWinX = this->x() + 100;
    ScriptWinY = this->y() + 100;
    ScriptWinW = 350;
    ScriptWinH = 380;

    //GUI updates
    //installing validators for edit boxes
      //double
    QDoubleValidator* dv = new QDoubleValidator(this);
    dv->setNotation(QDoubleValidator::ScientificNotation);
    QList<QLineEdit*> list = this->findChildren<QLineEdit *>();
    foreach(QLineEdit *w, list) if (w->objectName().startsWith("led")) w->setValidator(dv);
      //int
    QIntValidator* iv  = new QIntValidator(this);
    iv->setBottom(0);
    foreach(QLineEdit *w, list) if (w->objectName().startsWith("lei")) w->setValidator(iv);
      //styles
    QString styleGrey = "QLabel { background-color: #E0E0E0; }";  //setting grey background color
    ui->laBackground->setStyleSheet(styleGrey);
    ui->laBackground_2->setStyleSheet(styleGrey);
    //QString styleWhite = "QLabel { background-color: #FbFbFb; }";  //setting grey background color
    //ui->laBackgroundPhotonSource->setStyleSheet(styleWhite);
    QString styleWhiteF = "QFrame { background-color: #FbFbFb; }";
    ui->fAdvOptionsSim->setStyleSheet(styleWhiteF);
      //fonts
    QFont ff = ui->tabwidMain->tabBar()->font();
    ff.setBold(true);
    ui->tabwidMain->tabBar()->setFont(ff);
      //pixmaps
    ui->labReloadRequired->setPixmap(Rwindow->RedIcon.pixmap(16,16));
    ui->labAdvancedOn->setPixmap(Rwindow->YellowIcon.pixmap(16,16));
    ui->labAdvancedOn->setVisible(false);
      //misc gui inits
    ui->swPMTvsSiPM->setCurrentIndex(ui->cobPMdeviceType->currentIndex());
    MainWindow::on_pbRefreshPMproperties_clicked(); //indication of PM properties
    ui->tabWidget->setCurrentIndex(0);
    MainWindow::on_cobPMampGainModel_currentIndexChanged(0);
    MainWindow::on_pbElUpdateIndication_clicked();
    ui->cbEnableSPePHS->setChecked(false);
    ui->cbEnableElNoise->setChecked(false);
    ui->cbEnableADC->setChecked(false);
    ui->twSourcePhotonsParticles->setCurrentIndex(0);
    MainWindow::PointSource_InitTabWidget();
    QList<QWidget*> invis;
    invis << ui->cobS1 << ui->cobS2 << ui->cobS3 << ui->cobS4<<ui->cobS5<<ui->cobSS<<ui->cobLW
     << ui->ledS1 << ui->ledS2 << ui->ledS3 << ui->ledS4<<ui->ledS5<<ui->ledSS<<ui->ledLW
     << ui->pbUpdateSandwich << ui->pbRefreshParticles << ui->pbOverride << ui->pbRefreshOverrides
     << ui->pbRefreshStack << ui->pbShowPMsArrayRegularData << ui->pbRefreshPMArrayData << ui->pbUpdateElectronics
     << ui->pbRefreshPMproperties << ui->pbUpdatePMproperties << ui->pbRefreshMaterials
     << ui->pbGunRefreshparticles << ui->fPolygon << ui->pbIndPMshowInfo << ui->pbUpdateToFixedZ
     << ui->pbUpdateToFullCustom << ui->pbElUpdateIndication << ui->pbElUpdatePMampGainProperties
     << ui->pbInitializeScanFloodNoise << ui->pbUpdateScanFloodTabWidget << ui->fScanFloodTotProb
     << ui->fSecondaryScintLoadProfile << ui->pbUpdateSources << ui->pbUpdateSourcesIndication
     << ui->fLinkedParticle << ui->sbPMtype << ui->fUpperLowerArrays << ui->sbPMtypeForGroup
     << ui->pbRebuildDetector << ui->fReloadRequired << ui->pbYellow << ui->pbGDML;
    for (int i=0;i<invis.length();i++) invis[i]->setVisible(false);
    if (ui->twSingleScan->currentIndex() == 0) ui->cbScanFloodLimitToPrimaryScint->setVisible(false);
    else ui->cbScanFloodLimitToPrimaryScint->setVisible(true);
    QList<QWidget*> disables;
    disables << ui->fGunParticle << ui->fGunMultipleEvents
     << ui->fPreprocessing << ui->fFixedDir << ui->fFixedWorldSize;
    for (int i=0;i<disables.length();i++) disables[i]->setEnabled(false);
    ui->fWaveOptions->setEnabled(ui->cbWaveResolved->isChecked());
    ui->fWaveTests->setEnabled(ui->cbWaveResolved->isChecked());
    ui->fTime->setEnabled(ui->cbTimeResolved->isChecked());
    ui->fPointSource_Wave->setEnabled(ui->cbWaveResolved->isChecked());
    ui->fPointSource_Time->setEnabled(ui->cbTimeResolved->isChecked());
    ui->swPointSourceWaveTime->setCurrentIndex(ui->cobDirectlyOrFromMaterial->currentIndex());
    ui->fAngular->setEnabled(ui->cbAngularSensitive->isChecked());    
    ui->fScanSecond->setEnabled(ui->cbSecondAxis->isChecked());
    ui->fScanThird->setEnabled(ui->cbThirdAxis->isChecked());

    //change font size for all windows
    if (this->font().pointSize() != GlobSet->FontSize) setFontSizeAllWindows(GlobSet->FontSize);

    if (GlobSet->SaveLoadWindows)
      MainWindow::on_actionLoad_positions_and_status_of_all_windows_triggered();
    ui->actionSave_Load_windows_status_on_Exit_Init->setChecked(GlobSet->SaveLoadWindows);

    Owindow->InitWindow();
    Owindow->resize(Owindow->width()+1, Owindow->height());
    Owindow->resize(Owindow->width()-1, Owindow->height());

    Rwindow->InitWindow();

    WindowNavigator->show();

    //qDebug()<<"->Showing geometry";
    GeometryWindow->show();
    GeometryWindow->resize(GeometryWindow->width()+1, GeometryWindow->height());
    GeometryWindow->resize(GeometryWindow->width()-1, GeometryWindow->height());
    QThread::msleep(50);
    MainWindow::ShowGeometry(false);

    ui->cobScatteringModel->setCurrentIndex(1); //default to Lambertian back
    //MainWindow::on_pbRefreshOverrides_clicked();  //already in load detector
    MainWindow::updateCOBsWithPMtypeNames();
    //qDebug()<<"->Initialization complete";
}
