#include "interfacetocheckerscript.h"
#include "mainwindow.h"
#include "eventsdataclass.h"


InterfaceToInteractionScript::InterfaceToInteractionScript(MainWindow* MW, EventsDataClass* EventsDataHub)
  : MW(MW), EventsDataHub(EventsDataHub)
{  
  fOwnEventsHistory = false;

  H["ClearStack"] = "Clear particle stack";
  H["AddParticleToStack"] = "Add a particle (or several identical particles) to the stack";
  H["TrackStack"] = "Track all particles from the stack";
  H["ReleaseMemory"] = "Deletes EnergyVector and EventsHistory containers - gui tools will not work, but more memory is available during script processing";

  H["count"] = "Returns the number of available particle records";

  H["termination"] = "Returns termination type (int) for the given particle record";
  //1 - escaped 2 - all energy dissipated 3 - photoeffect 4 - compton 5 - capture
  //6 - error - undefined termination 7 - created outside defined geometry 8 - found untrackable material
  H["terminationStr"] = "Returns termination type (string) for the given pareticle record";
  H["dX"] = "Returns x-component of the particle direction vector";
  H["dY"] = "Returns x-component of the particle direction vector";
  H["dZ"] = "Returns x-component of the particle direction vector";
  H["particleId"] = "Returns the ID of the particle, used by the material collection";
  H["sernum"] = "Returns serial number of the particle";
  H["isSecondary"] = "Returns true if the particle is secondary (created during tracking of the particle stack)";

  // MaterialCrossed - info from DepositionHistory
  H["MaterialsCrossed_count"] = "Returns the number of materials crossed by the particle during tracking";
  H["MaterialsCrossed_matId"] = "Returns material ID (used by material collection) for the given material index";
  H["MaterialsCrossed_energy"] = "Returns the total deposited energy inside the given (by index) material";
  H["MaterialsCrossed_distance"] = "Returns the total distance travelled inside the given (by index) material";

  // Deposition - info from EnergyVector
  H["Deposition_countMaterials"] = "Returns the number of material records for the given particle";
  H["Deposition_matId"] = "Returns material ID for the given particle and material index";
  H["Deposition_startX"] = "Returns the start X position for the given particle inside the material with the given index";
  H["Deposition_startY"] = "Returns the start Y position for the given particle inside the material with the given index";
  H["Deposition_startZ"] = "Returns the start Z position for the given particle inside the material with the given index";
    // +subindex - node index
  H["Deposition_countNodes"] = "Returns the number of deposition nodes for the given particle inside the material with the given index";
  H["Deposition_X"] = "Returns the X position for the given particle, material index and node index";
  H["Deposition_Y"] = "Returns the Y position for the given particle, material index and node index";
  H["Deposition_Z"] = "Returns the Z position for the given particle, material index and node index";
  H["Deposition_dE"] = "Returns the deposited energy for the given particle, material index and node index";
  H["Deposition_dL"] = "Returns the length of the node with the given particle, material index and node index";
}

void InterfaceToInteractionScript::ClearStack()
{
  qDebug() << "Clear stack triggered";
  MW->on_pbClearAllStack_clicked();
  qDebug() << "Done";
}

void InterfaceToInteractionScript::AddParticleToStack(int particleID, double X, double Y, double Z,
                                                      double dX, double dY, double dZ,
                                                      double Time, double Energy,
                                                      int numCopies)
{
  MW->ParticleStack.reserve(MW->ParticleStack.size() + numCopies);
  for (int i=0; i<numCopies; i++)
  {
     ParticleOnStack *tmp = new ParticleOnStack(particleID,   X, Y, Z,   dX, dY, dZ,  Time, Energy);
     MW->ParticleStack.append(tmp);
  }
  MW->on_pbRefreshStack_clicked();
}

void InterfaceToInteractionScript::TrackStack()
{
  qDebug() << "Track stack triggered";
  clearPR();
  qDebug() << "  PR cleared";
  MW->on_pbTrackStack_clicked();
  qDebug() << "  Stack tracked";
  populateParticleRecords();
  qDebug() << "  Particle records populated";
}

void InterfaceToInteractionScript::ReleaseMemory() {clearEVandEH();}

void InterfaceToInteractionScript::populateParticleRecords()
{
  fOwnEventsHistory = false; //only if clearEVandEH is called the objects owns EventsHistory

  if (EventsDataHub->EventHistory.isEmpty())
    {
      abort("EventHistory is empty!");
      return;
    }
  if (MW->EnergyVector.isEmpty())
    {
      abort("EnergyVector is empty!");
      return;
    }

  int indexEV = 0;
  for (int i=0; i<EventsDataHub->EventHistory.size(); i++)
    {
      ParticleRecord pr;

      // migrating EventHistory
      pr.History = EventsDataHub->EventHistory.at(i); //only address!

      // adding info from EnergyVector
      int sernum = pr.History->index;
        //needs at least one entry in EnergyVector with the same serial number
      while (indexEV<MW->EnergyVector.size() && MW->EnergyVector.at(indexEV)->index == sernum)
        {          
          MaterialRecord mr;
          // Material ID
          mr.MatId = MW->EnergyVector.at(indexEV)->MaterialId;
          // Start position
          double* r = MW->EnergyVector.at(indexEV)->r;
          double  l = MW->EnergyVector.at(indexEV)->cellLength;
          mr.StartPosition[0] = r[0] - l * pr.History->dx;
          mr.StartPosition[1] = r[1] - l * pr.History->dy;
          mr.StartPosition[2] = r[2] - l * pr.History->dz;
          // Deposition info
          do
            {
              DepoNode depo;
              // position
              depo.R[0] = MW->EnergyVector.at(indexEV)->r[0];
              depo.R[1] = MW->EnergyVector.at(indexEV)->r[1];
              depo.R[2] = MW->EnergyVector.at(indexEV)->r[2];
              // energy
              depo.Energy = MW->EnergyVector.at(indexEV)->dE;
              // length
              depo.CellLength = MW->EnergyVector.at(indexEV)->cellLength;
              // adding to the record
              mr.ByMaterial.append(depo);
              indexEV++;              
            }
          while(indexEV<MW->EnergyVector.size() && MW->EnergyVector.at(indexEV)->index==sernum && MW->EnergyVector.at(indexEV)->MaterialId==mr.MatId);
          pr.Deposition.append(mr);
        }
      PR.append(pr);
    }
}

void InterfaceToInteractionScript::clearEVandEH()
{
  fOwnEventsHistory = true;
  EventsDataHub->EventHistory.clear(); //contains only objects address, migrated!

  for (int i=0; i<MW->EnergyVector.size(); i++) delete MW->EnergyVector[i];
  MW->EnergyVector.clear();
}

void InterfaceToInteractionScript::clearPR()
{
  if (EventsDataHub->EventHistory.size()>0 || fOwnEventsHistory)
    {
      for (int i=0; i<PR.size(); i++) delete PR[i].History;
    }
  EventsDataHub->EventHistory.clear();

  PR.clear();

  for (int i=0; i<MW->EnergyVector.size(); i++) delete MW->EnergyVector[i];
  MW->EnergyVector.clear();
}

int InterfaceToInteractionScript::termination(int i)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return -1;
    }
  return PR.at(i).History->Termination;
}

QString InterfaceToInteractionScript::terminationStr(int i)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return "Attempt to address non-existent particle";
    }
  switch (PR.at(i).History->Termination)
    {
    case 1:  return "escaped";
    case 2:  return "all energy dissipated";
    case 3:  return "photoeffect";
    case 4:  return "compton";
    case 5:  return "neutron capture";
    case 6:  return "bug detected!";
    case 7:  return "was created outside of the world";
    case 8:  return "tracking forbidden";
    default: return "unknown termination";
    }
}

double InterfaceToInteractionScript::dX(int i)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return std::numeric_limits<double>::quiet_NaN();
    }
  return PR.at(i).History->dx;
}

double InterfaceToInteractionScript::dY(int i)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return std::numeric_limits<double>::quiet_NaN();
    }
  return PR.at(i).History->dy;
}

double InterfaceToInteractionScript::dZ(int i)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return std::numeric_limits<double>::quiet_NaN();
    }
  return PR.at(i).History->dz;
}

int InterfaceToInteractionScript::particleId(int i)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return -1;
    }
  return PR.at(i).History->ParticleId;
}

int InterfaceToInteractionScript::sernum(int i)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return -1;
    }
  return PR.at(i).History->index;
}

int InterfaceToInteractionScript::isSecondary(int i)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return false;
    }
  return PR.at(i).History->flagSecondary;
}

int InterfaceToInteractionScript::MaterialsCrossed_count(int i)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return -1;
    }
  return PR.at(i).History->Deposition.size();
}

int InterfaceToInteractionScript::MaterialsCrossed_matId(int i, int m)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return -1;
    }
  if (m<0 || m>PR.at(i).History->Deposition.size()-1)
    {
      abort("Attempt to address non-existent material");
      return -1;
    }
  return PR.at(i).History->Deposition.at(m).MaterialId;
}

int InterfaceToInteractionScript::MaterialsCrossed_energy(int i, int m)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return -1;
    }
  if (m<0 || m>PR.at(i).History->Deposition.size()-1)
    {
      abort("Attempt to address non-existent material");
      return -1;
    }
  return PR.at(i).History->Deposition.at(m).DepositedEnergy;
}

int InterfaceToInteractionScript::MaterialsCrossed_distance(int i, int m)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return -1;
    }
  if (m<0 || m>PR.at(i).History->Deposition.size()-1)
    {
      abort("Attempt to address non-existent material");
      return -1;
    }
  return PR.at(i).History->Deposition.at(m).Distance;
}

int InterfaceToInteractionScript::Deposition_countMaterials(int i)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return -1;
    }
  return PR.at(i).Deposition.size();
}

int InterfaceToInteractionScript::Deposition_matId(int i, int m)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return -1;
    }
  if (m<0 || m>PR.at(i).Deposition.size()-1)
    {
      abort("Attempt to address non-existent material in deposition");
      return -1;
    }

  return PR.at(i).Deposition.at(m).MatId;
}

double InterfaceToInteractionScript::Deposition_startX(int i, int m)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return std::numeric_limits<double>::quiet_NaN();
    }
  if (m<0 || m>PR.at(i).Deposition.size()-1)
    {
      abort("Attempt to address non-existent material in deposition");
      return std::numeric_limits<double>::quiet_NaN();
    }

  return PR.at(i).Deposition.at(m).StartPosition[0];
}

double InterfaceToInteractionScript::Deposition_startY(int i, int m)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return std::numeric_limits<double>::quiet_NaN();
    }
  if (m<0 || m>PR.at(i).Deposition.size()-1)
    {
      abort("Attempt to address non-existent material in deposition");
      return std::numeric_limits<double>::quiet_NaN();
    }

  return PR.at(i).Deposition.at(m).StartPosition[1];
}

double InterfaceToInteractionScript::Deposition_startZ(int i, int m)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return std::numeric_limits<double>::quiet_NaN();
    }
  if (m<0 || m>PR.at(i).Deposition.size()-1)
    {
      abort("Attempt to address non-existent material in deposition");
      return std::numeric_limits<double>::quiet_NaN();
    }

  return PR.at(i).Deposition.at(m).StartPosition[2];
}

int InterfaceToInteractionScript::Deposition_countNodes(int i, int m)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return -1;
    }
  if (m<0 || m>PR.at(i).Deposition.size()-1)
    {
      abort("Attempt to address non-existent material in deposition");
      return -1;
    }

  return PR.at(i).Deposition.at(m).ByMaterial.size();
}

double InterfaceToInteractionScript::Deposition_X(int i, int m, int n)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return std::numeric_limits<double>::quiet_NaN();
    }
  if (m<0 || m>PR.at(i).Deposition.size()-1)
    {
      abort("Attempt to address non-existent material in deposition");
      return std::numeric_limits<double>::quiet_NaN();
    }
  if (n<0 || n>PR.at(i).Deposition.at(m).ByMaterial.size()-1)
    {
      abort("Attempt to address non-existent deposition node");
      return std::numeric_limits<double>::quiet_NaN();
    }
  return PR.at(i).Deposition.at(m).ByMaterial.at(n).R[0];
}

double InterfaceToInteractionScript::Deposition_Y(int i, int m, int n)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return std::numeric_limits<double>::quiet_NaN();
    }
  if (m<0 || m>PR.at(i).Deposition.size()-1)
    {
      abort("Attempt to address non-existent material in deposition");
      return std::numeric_limits<double>::quiet_NaN();
    }
  if (n<0 || n>PR.at(i).Deposition.at(m).ByMaterial.size()-1)
    {
      abort("Attempt to address non-existent deposition node");
      return std::numeric_limits<double>::quiet_NaN();
    }
  return PR.at(i).Deposition.at(m).ByMaterial.at(n).R[1];
}

double InterfaceToInteractionScript::Deposition_Z(int i, int m, int n)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return std::numeric_limits<double>::quiet_NaN();
    }
  if (m<0 || m>PR.at(i).Deposition.size()-1)
    {
      abort("Attempt to address non-existent material in deposition");
      return std::numeric_limits<double>::quiet_NaN();
    }
  if (n<0 || n>PR.at(i).Deposition.at(m).ByMaterial.size()-1)
    {
      abort("Attempt to address non-existent deposition node");
      return std::numeric_limits<double>::quiet_NaN();
    }
  return PR.at(i).Deposition.at(m).ByMaterial.at(n).R[2];
}

double InterfaceToInteractionScript::Deposition_dE(int i, int m, int n)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return std::numeric_limits<double>::quiet_NaN();
    }
  if (m<0 || m>PR.at(i).Deposition.size()-1)
    {
      abort("Attempt to address non-existent material in deposition");
      return std::numeric_limits<double>::quiet_NaN();
    }
  if (n<0 || n>PR.at(i).Deposition.at(m).ByMaterial.size()-1)
    {
      abort("Attempt to address non-existent deposition node");
      return std::numeric_limits<double>::quiet_NaN();
    }
  return PR.at(i).Deposition.at(m).ByMaterial.at(n).Energy;
}

double InterfaceToInteractionScript::Deposition_dL(int i, int m, int n)
{
  if (i<0 || i>PR.size()-1)
    {
      abort("Attempt to address non-existent particle");
      return std::numeric_limits<double>::quiet_NaN();
    }
  if (m<0 || m>PR.at(i).Deposition.size()-1)
    {
      abort("Attempt to address non-existent material in deposition");
      return std::numeric_limits<double>::quiet_NaN();
    }
  if (n<0 || n>PR.at(i).Deposition.at(m).ByMaterial.size()-1)
    {
      abort("Attempt to address non-existent deposition node");
      return std::numeric_limits<double>::quiet_NaN();
    }
  return PR.at(i).Deposition.at(m).ByMaterial.at(n).CellLength;
}


//-------------------------------------------
InterfaceToCheckerScript::InterfaceToCheckerScript(QVector<EnergyDepositionCell *> *EnergyVector, QVector<EventHistoryStructure *> *EventHistory)
  : EnergyVector(EnergyVector), EventHistory(EventHistory)
{
  //Directly addressing EnergyVector (ev)"
  H["ev_count"] = "Returns number of records in the EnergyVector";
  H["ev_X"] = "Returns X position of the ith element of the EnergyVector";
  H["ev_Y"] = "Returns Y position of the ith element of the EnergyVector";
  H["ev_Z"] = "Returns Z position of the ith element of the EnergyVector";
  H["ev_dE"] = "Returns energy deposition for the ith element of the EnergyVector";
  H["ev_dL"] = "Returns length of the ith element of the EnergyVector";
  H["ev_pId"] = "Returns particle ID for the ith element of the EnergyVector";
  H["ev_mId"] = "Returns material ID for the ith element of the EnergyVector";
  H["ev_pIndex"] = "Returns particle serial number for the ith element of the EnergyVector";

  //Particle Record (pr) - related: helper object to address EnergyVector
  H["pr_init"] = "Builds ParticleRecord starting from index=position in EnergyVecor.\n"
                 "Returns index in EnergyVector for the next particle";
  H["pr_pId"] = "Returns particle Id for the current particle record";
  H["pr_dX"] = "Returns X-component of the particle's direction for the current ParticleRecord";
  H["pr_dY"] = "Returns Y-component of the particle's direction for the current ParticleRecord";
  H["pr_dZ"] = "Returns Z-component of the particle's direction for the current ParticleRecord";
  H["pr_firstEV"] = "Returns the start position in the EnergyVector for the current ParticleRecord";
  H["pr_lastEV"] = "Returns the last position in the EnergyVector for the current ParticleRecord";
  H["pr_lengthEV"] = "Returns the number of records in the EnergyVector for the current ParticleRecord";

  //The current PR contains material record (MR), which simplifies access to EV
  //by material layers where the particle deposited energy
  H["pr_matCount"] = "Returns number of MaterialRecords in the current ParticleRecord";
  H["pr_mr_mId"] = "Returns material ID for the given material index";
  H["pr_mr_firstEV"] = "Returns first index in EnergyVector for the given material index";
  H["pr_mr_lastEV"] = "Returns last index in EnergyVector for the given material index";
  H["pr_mr_lengthEV"] = "Returns the length in EnergyVector for the given material index";
  H["pr_mr_x0"] = "Returns X where the particle entered the material with the given material index";
  H["pr_mr_y0"] = "Returns Y where the particle entered the material with the given material index";
  H["pr_mr_z0"] = "Returns Z where the particle entered the material with the given material index";

  //EventHistory processing functions
  H["eh_count"] = "Return number of records in EventHistory";
  H["eh_pId"] = "Returns particle ID for the given EventHistory record";
  H["eh_pIndex"] = "Returns particle serial number or the given EventHistory record";
  H["eh_Secondary"] = "Returns secondary flag for the given EventHistory record";
  H["eh_dX"] = "Returns X distance for the given EventHistory record";
  H["eh_dY"] = "Returns Y distance for the given EventHistory record";
  H["eh_dZ"] = "Returns Z distance for the given EventHistory record";
  H["eh_termination"] = "Returns termination type (int) for the given EventHistory record";
  H["eh_terminationStr"] = "Returns termination type (string) for the given EventHistory record";
  //1 - escaped 2 - all energy dissipated 3 - photoeffect 4 - compton 5 - capture
  //6 - error - undefined termination 7 - created outside defined geometry 8 - found untrackable material
  H["eh_depo_count"] = "Retuns number of deposition records in the given EventHistory record";
  H["eh_depo_matId"] = "Returns material ID for the given EventHistory record and deposition record index";
  H["eh_depo_dE"] = "Returns total energy depsoition for the given EventHistory record and deposition record index";
  H["eh_depo_dL"] = "Returns total travelled distance for the given EventHistory record and deposition record index";
}

double InterfaceToCheckerScript::ev_X(int i)
{
  if (i<0 || i>EnergyVector->size()-1) return std::numeric_limits<double>::quiet_NaN();
  return EnergyVector->at(i)->r[0];
}

double InterfaceToCheckerScript::ev_Y(int i)
{
  if (i<0 || i>EnergyVector->size()-1) return std::numeric_limits<double>::quiet_NaN();
  return EnergyVector->at(i)->r[1];
}

double InterfaceToCheckerScript::ev_Z(int i)
{
  if (i<0 || i>EnergyVector->size()-1) return std::numeric_limits<double>::quiet_NaN();
  return EnergyVector->at(i)->r[2];
}

double InterfaceToCheckerScript::ev_dE(int i)
{
  if (i<0 || i>EnergyVector->size()-1) return std::numeric_limits<double>::quiet_NaN();
  return EnergyVector->at(i)->dE;
}

double InterfaceToCheckerScript::ev_dL(int i)
{
  if (i<0 || i>EnergyVector->size()-1) return std::numeric_limits<double>::quiet_NaN();
  return EnergyVector->at(i)->cellLength;
}

int InterfaceToCheckerScript::ev_pId(int i)
{
  if (i<0 || i>EnergyVector->size()-1) return std::numeric_limits<double>::quiet_NaN();
  return EnergyVector->at(i)->ParticleId;
}

int InterfaceToCheckerScript::ev_mId(int i)
{
  if (i<0 || i>EnergyVector->size()-1) return std::numeric_limits<double>::quiet_NaN();
  return EnergyVector->at(i)->MaterialId;
}

int InterfaceToCheckerScript::ev_pIndex(int i)
{
  if (i<0 || i>EnergyVector->size()-1) return std::numeric_limits<double>::quiet_NaN();
  return EnergyVector->at(i)->index;
}

int InterfaceToCheckerScript::pr_init(int position)
{
  return PR.fill(EnergyVector, position);
}


int InterfaceToCheckerScript::pr_mr_mId(int mIndex)
{
  if (mIndex<0 || mIndex > PR.MaterialRecord.size()-1) return -1;
  return PR.MaterialRecord.at(mIndex).MaterialId;
}

int InterfaceToCheckerScript::pr_mr_firstEV(int mIndex)
{
  if (mIndex<0 || mIndex > PR.MaterialRecord.size()-1) return -1;
  return PR.MaterialRecord.at(mIndex).StartIndexInEV;
}

int InterfaceToCheckerScript::pr_mr_lastEV(int mIndex)
{
  if (mIndex<0 || mIndex > PR.MaterialRecord.size()-1) return -1;
  return PR.MaterialRecord.at(mIndex).StartIndexInEV + PR.MaterialRecord.at(mIndex).LengthInEV - 1;
}

int InterfaceToCheckerScript::pr_mr_lengthEV(int mIndex)
{
  if (mIndex<0 || mIndex > PR.MaterialRecord.size()-1) return -1;
  return PR.MaterialRecord.at(mIndex).LengthInEV;
}

double InterfaceToCheckerScript::pr_mr_x0(int mIndex)
{
  if (mIndex<0 || mIndex > PR.MaterialRecord.size()-1) return -1;
  return PR.MaterialRecord.at(mIndex).StartPosition[0];
}

double InterfaceToCheckerScript::pr_mr_y0(int mIndex)
{
  if (mIndex<0 || mIndex > PR.MaterialRecord.size()-1) return -1;
  return PR.MaterialRecord.at(mIndex).StartPosition[1];
}

double InterfaceToCheckerScript::pr_mr_z0(int mIndex)
{
  if (mIndex<0 || mIndex > PR.MaterialRecord.size()-1) return -1;
  return PR.MaterialRecord.at(mIndex).StartPosition[2];
}

int InterfaceToCheckerScript::eh_pId(int i)
{
  if (i<0 || i>EventHistory->size()-1) return -1;
  return EventHistory->at(i)->ParticleId;
}

int InterfaceToCheckerScript::eh_pIndex(int i)
{
  if (i<0 || i>EventHistory->size()-1) return -1;
  return EventHistory->at(i)->index;
}

int InterfaceToCheckerScript::eh_Secondary(int i)
{
  if (i<0 || i>EventHistory->size()-1) return -1;
  return EventHistory->at(i)->flagSecondary;
}

int InterfaceToCheckerScript::eh_dX(int i)
{
  if (i<0 || i>EventHistory->size()-1) return -1;
  return EventHistory->at(i)->dx;
}

int InterfaceToCheckerScript::eh_dY(int i)
{
  if (i<0 || i>EventHistory->size()-1) return -1;
  return EventHistory->at(i)->dy;
}

int InterfaceToCheckerScript::eh_dZ(int i)
{
  if (i<0 || i>EventHistory->size()-1) return -1;
  return EventHistory->at(i)->dz;
}

int InterfaceToCheckerScript::eh_termination(int i)
{
  if (i<0 || i>EventHistory->size()-1) return -1;
  return EventHistory->at(i)->Termination;
}

QString InterfaceToCheckerScript::eh_terminationStr(int i)
{
  if (i<0 || i>EventHistory->size()-1) return "unknown termination";
  switch (EventHistory->at(i)->Termination)
    {
    case 1:  return "escaped";
    case 2:  return "all energy dissipated";
    case 3:  return "photoeffect";
    case 4:  return "compton";
    case 5:  return "neutron capture";
    case 6:  return "bug detected!";
    case 7:  return "was created outside of the world";
    case 8:  return "tracking forbidden";
    default: return "unknown termination"; // should never get here (VS)
    }
}

int InterfaceToCheckerScript::eh_depo_count(int i)
{
  if (i<0 || i>EventHistory->size()-1) return -1;
  return EventHistory->at(i)->Deposition.size();
}

int InterfaceToCheckerScript::eh_depo_matId(int i, int imat)
{
  if (i<0 || i>EventHistory->size()-1) return -1;
  if (imat>EventHistory->at(i)->Deposition.size()-1) return -1;
  return EventHistory->at(i)->Deposition.at(imat).MaterialId;
}

double InterfaceToCheckerScript::eh_depo_dE(int i, int imat)
{
  if (i<0 || i>EventHistory->size()-1) return std::numeric_limits<double>::quiet_NaN();
  if (imat>EventHistory->at(i)->Deposition.size()-1) return std::numeric_limits<double>::quiet_NaN();
  return EventHistory->at(i)->Deposition.at(imat).DepositedEnergy;
}

double InterfaceToCheckerScript::eh_depo_dL(int i, int imat)
{
  if (i<0 || i>EventHistory->size()-1) return std::numeric_limits<double>::quiet_NaN();
  if (imat>EventHistory->at(i)->Deposition.size()-1) return std::numeric_limits<double>::quiet_NaN();
  return EventHistory->at(i)->Deposition.at(imat).Distance;
}

int ParticleRecordClass::fill(QVector<EnergyDepositionCell *> *EnergyVector, int indexEV)
{  
   //qDebug() << "EV size:"<< EnergyVector->size() << "  request at index:"<<indexEV;
   if (indexEV<0 || indexEV > EnergyVector->size()-1) return -1;
   ParticleId = EnergyVector->at(indexEV)->ParticleId;
   StartInEV = indexEV;
   int pIndex = EnergyVector->at(indexEV)->index;
   //qDebug() << "  particle id:"<<ParticleId;

   // first scan, defining the length and set material record except starting positions
   int pos = StartInEV;
   int prevMat = -1;
   MaterialRecord.clear();
   do
     {
        int mId = EnergyVector->at(pos)->MaterialId;
        //qDebug() << "-pr->"<<pos<<mId;

        if (mId == prevMat) MaterialRecord.last().LengthInEV++;
        else
          {
            MaterialRecord.append(MaterialRecordClass());
            MaterialRecord.last().StartIndexInEV = pos;
            MaterialRecord.last().MaterialId = mId;
            prevMat = mId;
          }
        pos++;
     }
   while (pos<EnergyVector->size() && EnergyVector->at(pos)->index == pIndex);
   //qDebug() << "  length in EV for this particle:"<<pos;

   //extracting direction
   if (pos-StartInEV > 1)
     {
       double* r0 = EnergyVector->at(StartInEV)->r;
       double* r1 = EnergyVector->at(pos-1)->r;
       double dx = r1[0]-r0[0];
       double dy = r1[1]-r0[1];
       double dz = r1[2]-r0[2];
       double norm = sqrt(dx*dx + dy*dy + dz*dz);
       if (norm > 1e-10)
         {
            Direction[0] = dx/norm; Direction[1] = dy/norm; Direction[2] = dz/norm;
         }
       else Direction[0] = Direction[1] = Direction[2] = 0;
     }
   else Direction[0] = Direction[1] = Direction[2] = 0; //undefined for length of one

   //filling start positions
   for (int ir = 0; ir<MaterialRecord.size(); ir++)
     {
       int stIndex = MaterialRecord.at(ir).StartIndexInEV;
       double* r = EnergyVector->at( stIndex )->r;
       double  l = EnergyVector->at( stIndex )->cellLength;
       for (int i=0; i<3; i++)
         MaterialRecord[ir].StartPosition[i] = r[i] - l * Direction[i];
     }

   EVlength = pos - StartInEV;
   return pos;
}
