#ifndef INTERFACETOCHECKERSCRIPT_H
#define INTERFACETOCHECKERSCRIPT_H

#include "classes.h"
#include "scriptinterfaces.h"

#include <QVector>

class MainWindow;
class EventsDataClass;

struct DepoNode
{
  double R[3];                  // x,y,z -coordinates of the cell
  double CellLength;            // cell length for charged particles
  double Energy;                // energy deposited in this cell
};

class MaterialRecord
{
public:
  int MatId;                    // material Id
  double StartPosition[3];      // position at which particle entered this material
  QVector<DepoNode> ByMaterial; // detailed depo info in this material
};

class ParticleRecord
{
public:
  EventHistoryStructure* History;     // migrated from EventHistory
  QVector<MaterialRecord> Deposition; // extracted data per material - from EnergyVector
};

class InterfaceToInteractionScript : public BaseInterfaceClass
{
  Q_OBJECT
public:
  InterfaceToInteractionScript(MainWindow* MW, EventsDataClass* EventsDataHub);
  ~InterfaceToInteractionScript(){clearPR();}

private: 
  MainWindow* MW;
  EventsDataClass* EventsDataHub;
  QVector<ParticleRecord> PR;

public slots:
  void ClearStack();
  void AddParticleToStack(int particleID, double X, double Y, double Z, double dX, double dY, double dZ, double time, double Energy, int numCopies);
  void TrackStack();  //run simulationwith particles in the stack

  void ReleaseMemory();

  int count() {return PR.length();} // total number of records

  int termination(int i);
  //1 - escaped
  //2 - all energy dissipated
  //3 - photoeffect
  //4 - compton
  //5 - capture
  //6 - error - undefined termination
  //7 - created outside defined geometry
  //8 - found untrackable material
  QString terminationStr(int i);
  double dX(int i);
  double dY(int i);
  double dZ(int i);
  int particleId(int i);
  int sernum(int i);
  int isSecondary(int i);

  // MaterialCrossed - info from DepositionHistory
  int MaterialsCrossed_count(int i);
  int MaterialsCrossed_matId(int i, int m);
  int MaterialsCrossed_energy(int i, int m);
  int MaterialsCrossed_distance(int i, int m);

  // Deposition - info from EnergyVector
  int Deposition_countMaterials(int i);
  int Deposition_matId(int i, int m);
  double Deposition_startX(int i, int m);
  double Deposition_startY(int i, int m);
  double Deposition_startZ(int i, int m);
    // +subindex - node index
  int Deposition_countNodes(int i, int m);
  double Deposition_X(int i, int m, int n);
  double Deposition_Y(int i, int m, int n);
  double Deposition_Z(int i, int m, int n);
  double Deposition_dE(int i, int m, int n);
  double Deposition_dL(int i, int m, int n);

private:
  void populateParticleRecords();
  void clearEVandEH();
  void clearPR();
  bool fOwnEventsHistory;
};


//--------old-------
class MaterialRecordClass  // intended for one charged particle!
{
public:
  int MaterialId;          // material Id
  int StartIndexInEV;      // index in EnergyVector
  int LengthInEV;          // length of EnergyVector for this material
  double StartPosition[3]; // position at which particle started in this material

  MaterialRecordClass() : LengthInEV(1) {}
};

class ParticleRecordClass
{
public:
  int ParticleId;           // particle Id
  double Direction[3];      // particle direction
  int StartInEV;            // start in EnergyVector
  int EVlength;             // length in EnergyVector
  QVector< MaterialRecordClass > MaterialRecord; // extracted data per material

  // fills the info for one particle.
  // indexEV is the start position in the EnergyVector from where to start looking for data
  // return the start index in EnergyVector of next particle, or, if its the last record, size of EnergyVector.
  // returns -1 if error encountered
  int fill(QVector<EnergyDepositionCell*> *EnergyVector, int indexEV);
};


class InterfaceToCheckerScript : public BaseInterfaceClass
{
  Q_OBJECT
public:
  InterfaceToCheckerScript(QVector<EnergyDepositionCell*> *EnergyVector, QVector<EventHistoryStructure*> *EventHistory);
  ~InterfaceToCheckerScript(){}

private:
  QVector<EnergyDepositionCell*> *EnergyVector;
  QVector<EventHistoryStructure*> *EventHistory;
  ParticleRecordClass PR;

public slots:
  //EnergyVector
  int ev_count() {return EnergyVector->size();}
  double ev_X(int i);
  double ev_Y(int i);
  double ev_Z(int i);
  double ev_dE(int i);
  double ev_dL(int i);
  int ev_pId(int i);
  int ev_mId(int i);
  int ev_pIndex(int i);

  // advanced EnergyVector processing
  int pr_init(int position); // extract data to ParticleRecord container starting from pos position in EnergyVector
  int pr_pId() {return PR.ParticleId;}
  double pr_dX() {return PR.Direction[0];} //direction
  double pr_dY() {return PR.Direction[1];}
  double pr_dZ() {return PR.Direction[2];}
  int pr_firstEV() {return PR.StartInEV;}
  int pr_lastEV() {return PR.StartInEV+PR.EVlength-1;}
  int pr_lengthEV() {return PR.EVlength;}
    // over material records
  int pr_matCount() {return PR.MaterialRecord.size();}
  int pr_mr_mId(int mIndex);
  int pr_mr_firstEV(int mIndex);
  int pr_mr_lastEV(int mIndex);
  int pr_mr_lengthEV(int mIndex);
  double pr_mr_x0(int mIndex);
  double pr_mr_y0(int mIndex);
  double pr_mr_z0(int mIndex);

  //event history log
  int eh_count() {return EventHistory->size();}
  int eh_pId(int i);
  int eh_pIndex(int i); // this is particle index! - "serial number" of the particle
  int eh_Secondary(int i);
  int eh_dX(int i);
  int eh_dY(int i);
  int eh_dZ(int i);
  int eh_termination(int i);
  QString eh_terminationStr(int i);
  //1 - escaped
  //2 - all energy dissipated
  //3 - photoeffect
  //4 - compton
  //5 - capture
  //6 - error - undefined termination
  //7 - created outside defined geometry
  //8 - found untrackable material
  int eh_depo_count(int i);
  int eh_depo_matId(int i, int imat);
  double eh_depo_dE(int i, int imat);
  double eh_depo_dL(int i, int imat);

};

#endif // INTERFACETOCHECKERSCRIPT_H
