#include "interfacetoglobscript.h"
#include "exampleswindow.h"
#include "reconstructionwindow.h"
#include "simulationmanager.h"
#include "reconstructionwindow.h"
#include "lrfwindow.h"
#include "outputwindow.h"
#include "tmpobjhubclass.h"

#include <QThread>
#include <QPlainTextEdit>
#include <QVBoxLayout>
#include <QDateTime>

#include "TRandom2.h"
#include "TGeoTrack.h"
#include "TColor.h"
#include "TH1D.h"
#include "TH1.h"
#include "TH2D.h"
#include "TH2.h"
#include "TGraph.h"

bool InterfaceToConfig::keyToNameAndIndex(QString Key, QString& Name, int& Index)
{
  //qDebug() << "Array candidate:"<<Key;
  if (Key.contains("["))
    {
      if (!Key.endsWith("]"))
        {
          LastError = "Format error: "+Key;
          qDebug() << LastError;
          return false;
        }

      Name = "";
      QString indexStr = Key;
      for (int i=0; i<Key.size(); i++)
        {
          QChar c = Key[i];
          indexStr = indexStr.mid(1);
          if (c == '[') break;
          else Name += c;
        }
      indexStr.chop(1);
      //qDebug() << "  Index candidate:"<<indexStr;
      bool ok;
      Index = indexStr.toInt(&ok);
      if (!ok)
        {
          LastError = "Index extraction error for "+Key;
          qDebug() << LastError;
          return false;
        }
    }
  else
    { // It is NOT an array
      Name = Key;
      Index = -1;
    }

  //qDebug() << "Extracted Name/Index:"<<Name<<Index;
  return true;
}

bool InterfaceToConfig::modifyJsonValue(QJsonObject& obj, const QString& path, const QJsonValue& newValue)
{
  int indexOfDot = path.indexOf('.');
  QString propertyName = path.left(indexOfDot);
  QString subPath = indexOfDot>0 ? path.mid(indexOfDot+1) : QString();

  QString name;
  int index;
  bool ok = keyToNameAndIndex(propertyName, name, index);
  if (!ok) return false;
  propertyName = name;

  //qDebug() << "Attempting to extract:"<<propertyName<<index;
  QJsonValue subValue = obj[propertyName];
  //qDebug() << "QJsonvalue extraction success?" << (subValue != QJsonValue());
  if (subValue == QJsonValue())
    {
      LastError = "Property not found:"+propertyName;
      qDebug() << LastError;
      return false;
    }

  //updating QJsonObject
  if(subPath.isEmpty())
    {
      subValue = newValue;
    }
  else
    {
      QJsonObject obj;
      if (index == -1)
        {
          obj = subValue.toObject();
          bool ok = modifyJsonValue(obj, subPath, newValue);
          if (!ok) return false;
          subValue = obj;
        }
      else
        {
          QJsonArray arr = subValue.toArray();
          //qDebug() << arr;
          if (index<0 || index>arr.size()-1)
            {
              LastError = "Wrong array index ("+QString::number(index)+") of "+propertyName+": array size is " + QString::number(arr.size());
              qDebug() << LastError;
              return false;
            }
          obj = arr[index].toObject();
          if (obj.isEmpty())
            {
              LastError = "Array element of "+propertyName+"["+QString::number(index)+"] is not QJsonObject!";
              qDebug() << LastError;
              return false;
            }
          bool ok = modifyJsonValue(obj, subPath, newValue);
          if (!ok) return false;
          arr[index] = obj;
          subValue = arr;
        }
    }

  obj[propertyName] = subValue;
  return true;
}

void InterfaceToConfig::find(const QJsonObject &obj, QStringList Keys, QStringList &Found, QString Path)
{
  // script interface replaces ".." or the leading "." with an empty string
  bool fLast = (Keys.size() == 1); //it is the last key in the Keys list
  if (fLast && Keys.last()=="")
    { //bad format - Keys have to end with a concrete key
      LastError = "Bad format - Last object has to be concrete";
      qDebug() << LastError;
      return;
    }

  QString Key = Keys.first();
  QString Name;
  int Index;
  if (Key != "")
    { //looking for a concrete key
      if (!keyToNameAndIndex(Key, Name, Index)) return; //format error

      if (obj.contains(Name))
        { //object does contain the key name! (we do not check here if the index is adequate)
          Path += "." + Key;
          if (fLast)
            { //mission accomplished
              Found.append(Path.mid(1)); //remove the starting "." in the found path
              return;
            }
          //pass to the next key in the chain
          Keys.removeFirst();
          if (Index == -1)
            find(obj[Name].toObject(), Keys, Found, Path);
          else
            find(obj[Name].toArray().at(Index).toObject(), Keys, Found, Path);
        }
      else return; //does not contains the given key
    }
  else
    { // "" Key
      QString Key = Keys.at(1); //this is next key to find
      if (Key == "") return;  //format problem - cannot be "" followed by ""
      //does the object contain the next key?
      if (!keyToNameAndIndex(Key, Name, Index)) return; //format error
      if  (obj.contains(Name))
        { //object does contain the key!
          //we can reuse the function:
          Keys.removeFirst(); //remove ""
          find(obj, Keys, Found, Path);
        }
      else
        { //have to check every sub-object
          foreach(QString oneKey, obj.keys())
            {
              QJsonValue Val = obj[oneKey];
              if (Val.isObject())
                find(obj[oneKey].toObject(), Keys, Found, Path+"."+oneKey);
              else if (Val.isArray())
                {
                  QJsonArray arr = Val.toArray();
                  for (int i=0; i<arr.size(); i++)
                    find(arr[i].toObject(), Keys, Found, Path+"."+oneKey+"["+QString::number(i)+"]");
                }
              //else do nothing for other types
            }
        }
    }
}

InterfaceToGlobScript::InterfaceToGlobScript(MainWindow *MW, DetectorClass *Detector, EventsDataClass *EventsDataHub)
  : MW(MW), Detector(Detector), EventsDataHub(EventsDataHub)
{
  H["Str"] = "Converts double value to string with a given precision";  
}

InterfaceToGlobScript::~InterfaceToGlobScript()
{
  MW->fDemoMode = false;

  for (int i=0; i<MW->TrackInfo.size(); i++)
    delete MW->TrackInfo[i];
  MW->TrackInfo.clear();
}

QString InterfaceToGlobScript::Str(double value, int precision)
{
  return QString::number(value, 'g', precision);
}

bool InterfaceToConfig::Replace(QString Key, QVariant val)
{
  LastError = "";
  //qDebug() << Keys << val << val.typeName();
  QString type = val.typeName();

  QJsonValue jv;
  QString rep;
  if (type == "int")
    {
      jv = QJsonValue(val.toInt());
      rep = QString::number(val.toInt());
    }
  else if (type == "double")
    {
      jv = QJsonValue(val.toDouble());
      rep = QString::number(val.toDouble());
    }
  else if (type == "bool")
    {
      jv = QJsonValue(val.toBool());
      rep = val.toBool() ? "true" : "false";
    }
  else if (type == "QString")
    {
      jv = QJsonValue(val.toString());
      rep = val.toString();
    }
  else
    {
      abort("Wrong argument type ("+type+") in Replace with the key line:\n"+Key);
      return false;
    }

  if (Key.startsWith(".") || Key.contains(".."))
    {
      QStringList keys = Key.split(".");
      //qDebug() << keys;
      QStringList res;
      find(Config, keys, res);

      if (!LastError.isEmpty())
        {
          abort("Replace("+Key+", "+rep+") search error:<br>" + LastError);
          return false;
        }
      if (res.isEmpty())
        {
          abort("Replace("+Key+", "+rep+") search error:<br>" +
                "Not found config objects according to the given pattern: "+Key);
          return false;
        }
      if (res.size()>1)
        {
          abort("Replace("+Key+", "+rep+") search error:<br>" +
                "Not unique match (" + QString::number(res.size())+ ") found for the config search pattern: "+Key);
          return false;
        }

      Key = res.first();
      //qDebug() << "*-----------"<<res;
    }

  bool ok = modifyJsonValue(Config, Key, jv);
  if (ok) return true;

  abort("Replace("+Key+", "+rep+") format error:<br>"+LastError);
  return false;
}

void InterfaceToConfig::SetPassivePM(int pm)
{
  SetPassiveStatus(pm, true);
}

void InterfaceToConfig::SetActivePM(int pm)
{
  SetPassiveStatus(pm, false);
}

void InterfaceToConfig::SetPassiveStatus(int pm, bool flag)
{
  if (pm<0 || pm>MW->Detector->PMs->count()-1)
    {
      abort("Attempt to change passive status of non-existent PM");
      return;
    }

  MW->PMs->SetPmPassiveState(pm, flag);
  QJsonArray passives;
  for (int ipm=0; ipm<MW->PMs->count(); ipm++)
    passives.append( MW->PMs->at(ipm).isStaticPassive() );

  //"ReconstructionConfig.PMrelatedSettings.StaticPassives"
  // *** !!! need a dedicated function like Replace()
  QJsonObject recJson = Config["ReconstructionConfig"].toObject();
  QJsonObject pmJson = recJson["PMrelatedSettings"].toObject();
  pmJson["StaticPassives"] = passives;
  recJson["PMrelatedSettings"] = pmJson;
  Config["ReconstructionConfig"] = recJson;

  UpdateRecConfig();
}

bool InterfaceToConfig::UpdateAllConfig()
{
  bool ok = MW->ELwindow->LoadAllConfig(Config);

  MW->on_pbRebuildDetector_clicked();
  MW->GeometryWindow->readRasterWindowProperties();
  //if (MW->GeometryWindow->isVisible()) MW->ShowGeometry();
  return ok;
}

bool InterfaceToConfig::UpdateSimConfig()
{
  return MW->ELwindow->LoadAllConfig(Config, false, true, false, true);
}

bool InterfaceToConfig::UpdateRecConfig()
{
  return MW->ELwindow->LoadAllConfig(Config, false, false, true, true);
}

InterfaceToConfig::InterfaceToConfig(MainWindow *MW) : MW(MW)
{
  MW->ELwindow->SaveAllConfig(Config);
  MW->GeometryWindow->readRasterWindowProperties();

  H["Load"]  = "Load ANTS2 configuration file (.json).";
  H["Replace"] = "Replace the value of the key in the configuration object with the new one. Config update is required for the change to take effect!";
  H["UpdateAllConfig"] = "Updates detector/sim/reconstruction settings. Used after Replace for changes to take effect.";
}

bool InterfaceToConfig::InitOnRun()
{
  //qDebug() << "InitOnRun triggered for config unit";
  MW->ELwindow->SaveAllConfig(Config);
  //qDebug() << "Config object updated!";
  MW->GeometryWindow->readRasterWindowProperties();
  return true;
}

bool InterfaceToConfig::Load(QString FileName)
{
  bool ok = MW->ELwindow->LoadAllConfig(FileName, true, true, true, &Config);
  if (ok)
    {
      MW->ReconstructDetector(false);
      if (MW->GeometryWindow->isVisible()) MW->ShowGeometry(false);
      return true;
    }

  abort("Failed to load config from file "+FileName);
  Config = QJsonObject();
  return false;
}

void InterfaceToGlobScript::ShowOutputWindow(bool flag, int tab)
{
  if (flag)
    {
      MW->Owindow->showNormal();
      MW->Owindow->raise();
    }
  else MW->Owindow->hide();

  if (tab>-1 && tab<4) MW->Owindow->SetTab(tab);
  qApp->processEvents();
}

//-----------------------------------
static int msgH = 500, msgW = 300, msgX=50, msgY=50;
InterfaceToTexter::InterfaceToTexter()
{
  fEnabled = true;
  Init(false);
}

void InterfaceToTexter::Init(bool fTransparent)
{
  D = new QDialog();

  QVBoxLayout* l = new QVBoxLayout;
  e = new QPlainTextEdit();
  e->setReadOnly(true);
  l->addWidget(e);
  D->setLayout(l);

  X = msgX;
  Y = msgY;
  W = msgW;
  H = msgH;

  D->setGeometry(X, Y, W, H);
  D->setWindowTitle("Script msg");

  if (fTransparent)
    {
      D->setWindowFlags(Qt::FramelessWindowHint);
      D->setAttribute(Qt::WA_TranslucentBackground);

      e->setStyleSheet("background: rgba(0,0,255,0%)");
      e->setFrameStyle(QFrame::NoFrame);
    }
}


InterfaceToTexter::~InterfaceToTexter()
{
  delete D;
}

void InterfaceToTexter::SetTransparent(bool flag)
{
  QString text = e->document()->toPlainText();
  delete D;
  Init(flag);
  e->setPlainText(text);
}

void InterfaceToTexter::Append(QString txt)
{
  e->appendHtml(txt);
}

void InterfaceToTexter::Clear()
{
  e->clear();
}

void InterfaceToTexter::Show(QString txt, int ms)
{
  if (!fEnabled) return;
  e->clear();
  e->appendHtml(txt);

  if (ms == -1)
    {
      D->show();
      D->raise();
      return;
    }

  D->show();
  D->raise();
  QTime t;
  t.restart();
  do qApp->processEvents();
  while (t.elapsed()<ms);
  D->hide();
}

void InterfaceToTexter::Move(double x, double y)
{
  X = msgX = x; Y = msgY = y;
  D->move(X, Y);
}

void InterfaceToTexter::Resize(double w, double h)
{
  W = msgW = w; H = msgH = h;
  D->resize(W, H);
}

void InterfaceToTexter::Show()
{
  if (!fEnabled) return;
  D->show();
  D->raise();
}

void InterfaceToTexter::Hide()
{
  D->hide();
}

void InterfaceToTexter::SetFontSize(int size)
{
  QFont f = e->font();
  f.setPointSize(size);
  e->setFont(f);
}

//----------------------------------
InterfaceToSim::InterfaceToSim(MainWindow *MW, DetectorClass *Detector, EventsDataClass *EventsDataHub)
  : MW(MW), Detector(Detector), EventsDataHub(EventsDataHub)
{
  H["RunPhotonSources"] = "Perform simulation with photon sorces.\nPhoton tracks are not shown!";
  H["RunParticleSources"] = "Perform simulation with particle sorces.\nPhoton tracks are not shown!";
  H["SetSeed"] = "Set random generator seed";
  H["GetSeed"] = "Get random generator seed";
  H["SaveAsTree"] = "Save simulation results as a ROOT tree file";
  H["SaveAsText"] = "Save simulation results as an ASCII file";
}

void InterfaceToSim::ForceStop()
{
  emit MW->StopRequested();
}

bool InterfaceToSim::RunPhotonSources()
{
  MW->fDemoMode = true;
  MW->on_pbSimulate_clicked();

  do
    {
      qApp->processEvents();
      if (MW->SimManager->isFinished() && !MW->SimManager->wasSuccessful())
        {
          qDebug()<<"Sim failed or aborted by user!";
          MW->fDemoMode = false;
          return false;
        }
    }
  while (MW->fSimDataNotSaved == false); //used as the ready flag
  //qDebug() << "Sim completed!";
  MW->fSimDataNotSaved = false; //avoid spam
  MW->fDemoMode = false;
  return true;
}

bool InterfaceToSim::RunParticleSources()
{
  MW->fDemoMode = true;
  MW->on_pbParticleSourcesSimulate_clicked();

  do
    {
      qApp->processEvents();
      if (MW->SimManager->isFinished() && !MW->SimManager->wasSuccessful())
        {
          qDebug()<<"Sim failed or aborted by user!";
          MW->fDemoMode = false;
          return false;
        }
    }
  while (MW->fSimDataNotSaved == false); //used as the ready flag
  //qDebug() << "Sim completed!";
  MW->fSimDataNotSaved = false; //avoid spam
  MW->fDemoMode = false;
  return true;
}

void InterfaceToSim::SetSeed(long seed)
{
  Detector->RandGen->SetSeed(seed);
}

long InterfaceToSim::GetSeed()
{
  return Detector->RandGen->GetSeed();
}

void InterfaceToSim::SaveAsTree(QString fileName)
{
  bool ok = EventsDataHub->saveSimulationAsTree(fileName);
  if (!ok) abort("Failed to save simulation data as tree");
  else MW->fSimDataNotSaved = false;
}

void InterfaceToSim::SaveAsText(QString fileName)
{
  bool ok = EventsDataHub->saveSimulationAsText(fileName);
  if (!ok) abort("Failed to save simulation data as text");
  else MW->fSimDataNotSaved = false;
}

//----------------------------------
InterfaceToData::InterfaceToData(MainWindow *MW, DetectorClass *Detector, EventsDataClass* EventsDataHub)
  : MW(MW), Detector(Detector), EventsDataHub(EventsDataHub)
{

}

double InterfaceToData::GetPMsignal(int ievent, int ipm)
{
  int numEvents = EventsDataHub->Events.size();
  int numPMs = Detector->PMs->count();
  if (ievent<0 || ievent>numEvents-1)
    {
      abort("Wrong event number "+QString::number(ievent)+" Events available: "+QString::number(numEvents));
      return 0;
    }
  if (ipm<0 || ipm>numPMs-1)
    {
      abort("Wrong PM number "+QString::number(ipm)+" PMs defined: "+QString::number(numPMs));
      return 0;
    }

  return EventsDataHub->Events.at(ievent).at(ipm);
}

int InterfaceToData::GetNumPMs()
{
  return Detector->PMs->count();
}

int InterfaceToData::GetNumEvents()
{
  return EventsDataHub->Events.size();
}

void InterfaceToData::LoadEventsTree(QString fileName, bool Append)
{
  if (!Append) EventsDataHub->clear();
  MW->LoadSimulationDataFromTree(fileName);
  MW->Rwindow->CheckAvailableEventData();
}

void InterfaceToData::LoadEventsAscii(QString fileName, bool Append)
{
  if (!Append) EventsDataHub->clear();
  MW->LoadPMsignals(fileName);
  MW->Rwindow->CheckAvailableEventData();
}

//----------------------------------
InterfaceToLRF::InterfaceToLRF(MainWindow *MW)
  : MW(MW)
{
  H["Make"] = "Calculates new LRFs.";
  H["SetGrouping"] = "Set LRF grouping on or off.";
  H["SetSource"] = "Selects source for position info:\n0 - sim/calibration data, 1 - reconstructed positions.";
}

void InterfaceToLRF::Make()
{
  QString res = MW->lrfwindow->MakeLRFs();
  if (res.isEmpty()) return;

  abort("Make LRFs failed:\n"+res);
}

void InterfaceToLRF::SetGrouping(bool flag)
{
  QJsonObject obj;
  MW->lrfwindow->writeToJson(obj);
  obj["UseGroupping"] = flag;
  MW->lrfwindow->loadJSON(obj);
}

void InterfaceToLRF::SetSource(int selector)
{
  QJsonObject obj;
  MW->lrfwindow->writeToJson(obj);
  obj["DataSelector"] = selector;
  MW->lrfwindow->loadJSON(obj);
}

//----------------------------------
InterfaceToGeoWin::InterfaceToGeoWin(MainWindow *MW, DetectorClass *Detector)
 : MW(MW), Detector(Detector)
{
  H["SaveImage"] = "Save image currently shown on the graph window to an image file.\nTip: use .png extension";
}

double InterfaceToGeoWin::GetPhi()
{
  return MW->GeometryWindow->Phi;
}

double InterfaceToGeoWin::GetTheta()
{
  return MW->GeometryWindow->Theta;
}

void InterfaceToGeoWin::SetPhi(double phi)
{
  MW->GeometryWindow->Phi = phi;
}

void InterfaceToGeoWin::SetTheta(double theta)
{
  MW->GeometryWindow->Theta = theta;
}

void InterfaceToGeoWin::Rotate(double Theta, double Phi, int Steps, int msPause)
{
  if (Steps <= 0) return;
  double stepT = Theta/Steps;
  double stepP = Phi/Steps;
  double T0 = GetTheta();
  double P0 = GetPhi();

  QTime time;
  MW->GeometryWindow->fNeedZoom = false;
  MW->GeometryWindow->fRecallWindow = true;
  for (int i=0; i<Steps; i++)
    {
      qApp->processEvents();
      time.restart();
      MW->GeometryWindow->Theta = T0 + stepT*(i+1);
      MW->GeometryWindow->Phi   = P0 + stepP*(i+1);
      MW->GeometryWindow->PostDraw();

      int msPassed = time.elapsed();
      if (msPassed<msPause) QThread::msleep(msPause-msPassed);
    }
}

void InterfaceToGeoWin::UpdateView()
{
  MW->GeometryWindow->fRecallWindow = true;
  MW->GeometryWindow->PostDraw();
  MW->GeometryWindow->UpdateRootCanvas();
}

void InterfaceToGeoWin::SetParallel(bool on)
{
  MW->GeometryWindow->ModePerspective = !on;
}

void InterfaceToGeoWin::Show()
{
  MW->GeometryWindow->showNormal();
  MW->GeometryWindow->raise();
}

void InterfaceToGeoWin::Hide()
{
  MW->GeometryWindow->hide();
}

void InterfaceToGeoWin::BlockUpdates(bool on)
{
  MW->DoNotUpdateGeometry = on;
  MW->GeometryDrawDisabled = on;
}

int InterfaceToGeoWin::GetX()
{
  return MW->GeometryWindow->x();
}

int InterfaceToGeoWin::GetY()
{
  return MW->GeometryWindow->y();
}

int InterfaceToGeoWin::GetW()
{
  return MW->GeometryWindow->width();
}

int InterfaceToGeoWin::GetH()
{
  return MW->GeometryWindow->height();
}

void InterfaceToGeoWin::ShowGeometry()
{
  MW->GeometryWindow->readRasterWindowProperties();
  MW->ShowGeometry(false);
}

void InterfaceToGeoWin::ShowPMnumbers()
{
  MW->GeometryWindow->on_pbShowPMnumbers_clicked();
}

void InterfaceToGeoWin::ShowTracks(int num, int OnlyColor)
{
  Detector->GeoManager->ClearTracks();
  if (MW->TrackInfo.isEmpty()) return;

  for (int iTr=0; iTr<MW->TrackInfo.size() && iTr<num; iTr++)
    {
      TrackHolderClass* th = MW->TrackInfo.at(iTr);
      TGeoTrack* track = new TGeoTrack(1, th->UserIndex);
      track->SetLineColor(th->Color);
      track->SetLineWidth(th->Width);
      for (int iNode=0; iNode<th->Nodes.size(); iNode++)
        track->AddPoint(th->Nodes[iNode].R[0], th->Nodes[iNode].R[1], th->Nodes[iNode].R[2], th->Nodes[iNode].Time);

      if (track->GetNpoints()>1)
        {
          if (OnlyColor == -1  || OnlyColor == th->Color)
            Detector->GeoManager->AddTrack(track);
        }
      else delete track;
    }
  MW->ShowTracks();
}

void InterfaceToGeoWin::ShowSPS_position()
{
  MW->on_pbSingleSourceShow_clicked();
}

void InterfaceToGeoWin::ClearTracks()
{
  MW->GeometryWindow->on_pbClearTracks_clicked();
}

void InterfaceToGeoWin::ClearMarkers()
{
  MW->GeometryWindow->on_pbClearDots_clicked();
}

void InterfaceToGeoWin::SaveImage(QString fileName)
{
  MW->GeometryWindow->SaveAs(fileName);
}

void InterfaceToGeoWin::ShowTracksMovie(int num, int steps, int msDelay, double dTheta, double dPhi, double rotSteps, int color)
{
  int tracks = MW->TrackInfo.size();
  if (tracks == 0) return;
  if (num > tracks) num = tracks;

  int toDo;
  int thisNodes = 2;
  do //finished when all nodes of the longest track are shown
    {
      //this iteration will show track nodes < thisNode
      toDo = num; //every track when finished results in toDo--

      // cycle by indication step with interpolation between the nodes
      for (int iStep=0; iStep<steps; iStep++)
        {
          MW->Detector->GeoManager->ClearTracks();

          //cycle by tracks
          for (int iTr=0; iTr<num; iTr++)
            {
              TrackHolderClass* th = MW->TrackInfo.at(iTr);
              int ThisTrackNodes = th->Nodes.size();
              if (ThisTrackNodes <= thisNodes && iStep == steps-1) toDo--; //last node of this track

              TGeoTrack* track = new TGeoTrack(1, th->UserIndex);
              track->SetLineWidth(th->Width+1);
              if (color == -1) track->SetLineColor(15); // during tracking
              else track->SetLineColor(color); // during tracking
              int lim = std::min(thisNodes, th->Nodes.size());
              for (int iNode=0; iNode<lim; iNode++)
                {
                  double x, y, z;
                  if (iNode == thisNodes-1)
                    {
                      //here track is in interpolation mode
                      x = th->Nodes[iNode-1].R[0] + (th->Nodes[iNode].R[0]-th->Nodes[iNode-1].R[0])*iStep/(steps-1);
                      y = th->Nodes[iNode-1].R[1] + (th->Nodes[iNode].R[1]-th->Nodes[iNode-1].R[1])*iStep/(steps-1);
                      z = th->Nodes[iNode-1].R[2] + (th->Nodes[iNode].R[2]-th->Nodes[iNode-1].R[2])*iStep/(steps-1);
                      if (color == -1) track->SetLineColor(15); //black during tracking
                    }
                  else
                    {
                      x = th->Nodes[iNode].R[0];
                      y = th->Nodes[iNode].R[1];
                      z = th->Nodes[iNode].R[2];
                      if (color == -1) track->SetLineColor( th->Color );
                    }
                  track->AddPoint(x, y, z, th->Nodes[iNode].Time);
                }
              if (color == -1)
                if (ThisTrackNodes <= thisNodes && iStep == steps-1) track->SetLineColor( th->Color );

              //adding track to GeoManager
              if (track->GetNpoints()>1)
                Detector->GeoManager->AddTrack(track);
              else delete track;
            }
          MW->ShowTracks();
          Rotate(dTheta, dPhi, rotSteps);
          QThread::msleep(msDelay);
        }
      thisNodes++;
    }
  while (toDo>0);
}

//----------------------------------
InterfaceToGraphs::InterfaceToGraphs(MainWindow *MW)
  : MW(MW)
{
  H["NewGraph"] = "Creates a new graph (Root TGraph object)";
  H["SetMarkerProperties"] = "Default marker properties are 1, 20, 1";
  H["SetLineProperties"] = "Default line properties are 1, 1, 2";
  H["Draw"] = "Draws the graph (use \"APL\" options if in doubt)";
}

bool InterfaceToGraphs::InitOnRun()
{
  MW->TmpHub->ScriptDrawObjects.clear();
  return true;
}

void InterfaceToGraphs::NewGraph(QString GraphName)
{
  int index = MW->TmpHub->ScriptDrawObjects.findIndexOf(GraphName);
  if (index != -1)
    {
      abort("Bad new graph name! Object "+GraphName+" already exists");
      return;
    }

  TGraph* gr = new TGraph();
  MW->TmpHub->ScriptDrawObjects.append(gr, GraphName, "TGraph");
}

void InterfaceToGraphs::SetMarkerProperties(QString GraphName, int MarkerColor, int MarkerStyle, int MarkerSize)
{
  int index = MW->TmpHub->ScriptDrawObjects.findIndexOf(GraphName);
  if (index == -1)
    {
      abort("Graph "+GraphName+" not found!");
      return;
    }

  RootDrawObj& r = MW->TmpHub->ScriptDrawObjects.List[index];
  r.MarkerColor = MarkerColor;
  r.MarkerStyle = MarkerStyle;
  r.MarkerSize = MarkerSize;
}

void InterfaceToGraphs::SetLineProperties(QString GraphName, int LineColor, int LineStyle, int LineWidth)
{
  int index = MW->TmpHub->ScriptDrawObjects.findIndexOf(GraphName);
  if (index == -1)
    {
      abort("Graph "+GraphName+" not found!");
      return;
    }

  RootDrawObj& r = MW->TmpHub->ScriptDrawObjects.List[index];
  r.LineColor = LineColor;
  r.LineStyle = LineStyle;
  r.LineWidth = LineWidth;
}

void InterfaceToGraphs::SetTitles(QString GraphName, QString X_Title, QString Y_Title)
{
  int index = MW->TmpHub->ScriptDrawObjects.findIndexOf(GraphName);
  if (index == -1)
    {
      abort("Graph "+GraphName+" not found!");
      return;
    }

  RootDrawObj& r = MW->TmpHub->ScriptDrawObjects.List[index];
  r.Xtitle = X_Title;
  r.Ytitle = Y_Title;
}

void InterfaceToGraphs::AddPoint(QString GraphName, double x, double y)
{
  int index = MW->TmpHub->ScriptDrawObjects.findIndexOf(GraphName);
  if (index == -1)
    {
      abort("Graph "+GraphName+" not found!");
      return;
    }

  RootDrawObj& r = MW->TmpHub->ScriptDrawObjects.List[index];
  if (r.type == "TGraph")
    {
      TGraph* gr = static_cast<TGraph*>(r.Obj);
      gr->SetPoint(gr->GetN(), x, y);
    }
  else
    {
      abort("Graph "+GraphName+" not found!");
      return;
    }
}

void InterfaceToGraphs::Draw(QString GraphName, QString options)
{
  int index = MW->TmpHub->ScriptDrawObjects.findIndexOf(GraphName);
  if (index == -1)
    {
      abort("Graph "+GraphName+" not found!");
      return;
    }

  RootDrawObj& r = MW->TmpHub->ScriptDrawObjects.List[index];
  if (r.type == "TGraph")
    {
      TGraph* gr = static_cast<TGraph*>(r.Obj);
      if (gr->GetN() == 0) return;

      gr->SetLineColor(r.LineColor);
      gr->SetLineWidth(r.LineWidth);
      gr->SetLineStyle(r.LineStyle);
      gr->SetMarkerColor(r.MarkerColor);
      gr->SetMarkerSize(r.MarkerSize);
      gr->SetMarkerStyle(r.MarkerStyle);
      gr->SetEditable(false);
      gr->GetYaxis()->SetTitleOffset((Float_t)1.30);
      emit RequestDraw(gr, options, true);

      gr->GetXaxis()->SetTitle(r.Xtitle.toLatin1().data());
      gr->GetYaxis()->SetTitle(r.Ytitle.toLatin1().data());
      emit RequestDraw(0, "", true); //to update canvas so axes titles are visible
    }
  else
    {
      abort("Graph "+GraphName+" not found!");
      return;
    }
}

//----------------------------------
InterfaceToHistD::InterfaceToHistD(MainWindow *MW)
  : MW(MW)
{

}

bool InterfaceToHistD::InitOnRun()
{
  MW->TmpHub->ScriptDrawObjects.clear();
  return true;
}

void InterfaceToHistD::NewHist(QString HistName, int bins, double start, double stop)
{
  int index = MW->TmpHub->ScriptDrawObjects.findIndexOf(HistName);
  if (index != -1)
    {
      abort("Bad new hist name! Object "+HistName+" already exists");
      return;
    }

  TH1D* hist = new TH1D("", HistName.toLatin1().data(), bins, start, stop);
  hist->GetYaxis()->SetTitleOffset((Float_t)1.30);
  MW->TmpHub->ScriptDrawObjects.append(hist, HistName, "TH1D");
}

void InterfaceToHistD::NewHist2D(QString HistName, int binsX, double startX, double stopX, int binsY, double startY, double stopY)
{
  int index = MW->TmpHub->ScriptDrawObjects.findIndexOf(HistName);
  if (index != -1)
    {
      abort("Bad new hist name! Object "+HistName+" already exists");
      return;
    }

  TH2D* hist = new TH2D("", HistName.toLatin1().data(), binsX, startX, stopX, binsY, startY, stopY);
  hist->GetYaxis()->SetTitleOffset((Float_t)1.30);
  MW->TmpHub->ScriptDrawObjects.append(hist, HistName, "TH2D");
}

void InterfaceToHistD::SetTitles(QString HistName, QString X_Title, QString Y_Title, QString Z_Title)
{
  int index = MW->TmpHub->ScriptDrawObjects.findIndexOf(HistName);
  if (index == -1)
    {
      abort("Histogram "+HistName+" not found!");
      return;
    }

  RootDrawObj& r = MW->TmpHub->ScriptDrawObjects.List[index];
  r.Xtitle = X_Title;
  r.Ytitle = Y_Title;
  r.Ztitle = Z_Title;
}

void InterfaceToHistD::SetLineProperties(QString HistName, int LineColor, int LineStyle, int LineWidth)
{
  int index = MW->TmpHub->ScriptDrawObjects.findIndexOf(HistName);
  if (index == -1)
    {
      abort("Histogram "+HistName+" not found!");
      return;
    }

  RootDrawObj& r = MW->TmpHub->ScriptDrawObjects.List[index];
  r.LineColor = LineColor;
  r.LineStyle = LineStyle;
  r.LineWidth = LineWidth;
}

void InterfaceToHistD::Fill(QString HistName, double val, double weight)
{
  int index = MW->TmpHub->ScriptDrawObjects.findIndexOf(HistName);
  if (index == -1)
    {
      abort("Histogram "+HistName+" not found!");
      return;
    }

  RootDrawObj& r = MW->TmpHub->ScriptDrawObjects.List[index];
  if (r.type == "TH1D")
    {
      TH1D* h = static_cast<TH1D*>(r.Obj);
      h->Fill(val, weight);
    }
  else
    {
      abort("TH1D histogram "+HistName+" not found!");
      return;
    }
}

void InterfaceToHistD::Fill2D(QString HistName, double x, double y, double weight)
{
  int index = MW->TmpHub->ScriptDrawObjects.findIndexOf(HistName);
  if (index == -1)
    {
      abort("Histogram "+HistName+" not found!");
      return;
    }

  RootDrawObj& r = MW->TmpHub->ScriptDrawObjects.List[index];
  if (r.type == "TH2D")
    {
      TH2D* h = static_cast<TH2D*>(r.Obj);
      h->Fill(x, y, weight);
    }
  else
    {
      abort("TH2D histogram "+HistName+" not found!");
      return;
    }
}

void InterfaceToHistD::Smooth(QString HistName, int times)
{
  int index = MW->TmpHub->ScriptDrawObjects.findIndexOf(HistName);
  if (index == -1)
    {
      abort("Histogram "+HistName+" not found!");
      return;
    }

  RootDrawObj& r = MW->TmpHub->ScriptDrawObjects.List[index];
  if (r.type == "TH1D")
    {
      TH1D* h = static_cast<TH1D*>(r.Obj);
      h->Smooth(times);
      emit RequestDraw(0, "", true); //to update
    }
  else if (r.type == "TH2D")
    {
      TH2D* h = static_cast<TH2D*>(r.Obj);
      h->Smooth(times);
      emit RequestDraw(0, "", true); //to update
    }
  else
    {
      abort("Object "+HistName+": unknown histogram type!");
      return;
    }
}

void InterfaceToHistD::Draw(QString HistName, QString options)
{
  int index = MW->TmpHub->ScriptDrawObjects.findIndexOf(HistName);
  if (index == -1)
    {
      abort("Histogram "+HistName+" not found!");
      return;
    }

  RootDrawObj& r = MW->TmpHub->ScriptDrawObjects.List[index];
  if (r.type == "TH1D")
    {
      TH1D* h = static_cast<TH1D*>(r.Obj);
      h->SetXTitle(r.Xtitle.toLatin1().data());
      h->SetYTitle(r.Ytitle.toLatin1().data());
      h->SetLineColor(r.LineColor);
      h->SetLineWidth(r.LineWidth);
      h->SetLineStyle(r.LineStyle);
      emit RequestDraw(h, options, true);
    }
  else if (r.type == "TH2D")
    {
      TH2D* h = static_cast<TH2D*>(r.Obj);
      h->SetXTitle(r.Xtitle.toLatin1().data());
      h->SetYTitle(r.Ytitle.toLatin1().data());
      h->SetZTitle(r.Ztitle.toLatin1().data());
      h->SetLineColor(r.LineColor);
      h->SetLineWidth(r.LineWidth);
      h->SetLineStyle(r.LineStyle);
      emit RequestDraw(h, options, true);
    }
  else
    {
      abort("Object "+HistName+": unknown histogram type!");
      return;
    }
}

//----------------------------------
InterfaceToReconstructor::InterfaceToReconstructor(MainWindow *MW, EventsDataClass *EventsDataHub)
 : MW(MW), EventsDataHub(EventsDataHub)
{

}

void InterfaceToReconstructor::ForceStop()
{
  emit MW->Rwindow->StopRequested();
}

void InterfaceToReconstructor::ReconstructEvents(bool fShow)
{
  MW->Rwindow->ReconstructAll(fShow);
}

void InterfaceToReconstructor::UpdateFilters()
{
  MW->Rwindow->on_pbUpdateFilters_clicked();
}

void InterfaceToReconstructor::DoBlur(double sigma)
{
  MW->Rwindow->DoBlur(sigma);
}

void InterfaceToReconstructor::PlotDensityXY()
{
  MW->Rwindow->DoPlotXY(0);
}

void InterfaceToReconstructor::PlotEnergyXY()
{
  MW->Rwindow->DoPlotXY(2);
}

void InterfaceToReconstructor::PlotChi2XY()
{
  MW->Rwindow->DoPlotXY(3);
}

void InterfaceToReconstructor::SaveAsTree(QString fileName)
{
  EventsDataHub->saveReconstructionAsTree(fileName, MW->Detector->PMs);
}

void InterfaceToReconstructor::SaveAsText(QString fileName)
{
  EventsDataHub->saveReconstructionAsText(fileName);
}

void InterfaceToReconstructor::SetShowOnlyFirstEvents(bool fOn, int number)
{
  MW->Rwindow->SetShowFirst(fOn, number);
}

void InterfaceToReconstructor::SetShowTrue(bool fOn, bool fActualOnTop)
{
  MW->Rwindow->SetShowActual(fOn, fActualOnTop);
}

void InterfaceToReconstructor::ShowTruePositions()
{
  MW->Rwindow->DotActualPositions();
  MW->ShowGeometry(false, false);
}

//----------------------------------
InterfaceToGraphWin::InterfaceToGraphWin(MainWindow *MW)
  : MW(MW)
{
  H["SaveImage"] = "Save image currently shown on the graph window to an image file.\nTip: use .png extension";
}

void InterfaceToGraphWin::Show()
{
  MW->GraphWindow->showNormal();
}

void InterfaceToGraphWin::Hide()
{
  MW->GraphWindow->hide();
}

void InterfaceToGraphWin::SetLog(bool Xaxis, bool Yaxis)
{
  MW->GraphWindow->SetLog(Xaxis, Yaxis);
}

void InterfaceToGraphWin::addToBasket(QString Title)
{
  MW->GraphWindow->AddCurrentToBasket(Title);
}

void InterfaceToGraphWin::SaveImage(QString fileName)
{
  MW->GraphWindow->SaveGraph(fileName);
}
