#ifndef INTERFACETOGLOBSCRIPT
#define INTERFACETOGLOBSCRIPT

#include "mainwindow.h"
#include "detectorclass.h"
#include "eventsdataclass.h"
#include "geometrywindowclass.h"
#include "graphwindowclass.h"
#include "globalsettingsclass.h"
#include "scriptinterfaces.h"

#include <QtGui>

class QPlainTextEdit;

class InterfaceToGlobScript : public BaseInterfaceClass
{
  Q_OBJECT

public:
  InterfaceToGlobScript(MainWindow* MW, DetectorClass* Detector, EventsDataClass* EventsDataHub);
  ~InterfaceToGlobScript();

public slots:    

  //output window
  void ShowOutputWindow(bool flag = true, int tab = -1);
  //show text need formatting
  QString Str(double value, int precision);
  //Random gen?

private:
  MainWindow* MW;
  DetectorClass* Detector;
  EventsDataClass* EventsDataHub;
};

//unit that handles RECONSTRUCTION
class InterfaceToReconstructor : public BaseInterfaceClass
{
  Q_OBJECT

public:
  InterfaceToReconstructor(MainWindow* MW, EventsDataClass* EventsDataHub);
  ~InterfaceToReconstructor(){}

  virtual void ForceStop();

public slots:
  void ReconstructEvents(bool fShow = true);
  void UpdateFilters();
  void DoBlur(double sigma);

  void PlotDensityXY();
  void PlotEnergyXY();
  void PlotChi2XY();

  void SaveAsTree(QString fileName);
  void SaveAsText(QString fileName);

  void SetShowOnlyFirstEvents(bool fOn, int number = -1);
  void SetShowTrue(bool fOn, bool fActualOnTop = false);
  void ShowTruePositions();

private:
  MainWindow* MW;
  EventsDataClass* EventsDataHub;
};

//unit to build and draw HISTOGRAMS (TH1D of ROOT)
class InterfaceToHistD : public BaseInterfaceClass
{
  Q_OBJECT

public:
  InterfaceToHistD(MainWindow* MW);
  ~InterfaceToHistD(){}

  virtual bool InitOnRun();

public slots:
  void NewHist(QString HistName, int bins, double start, double stop);
  void NewHist2D(QString HistName, int binsX, double startX, double stopX, int binsY, double startY, double stopY);

  void SetTitles(QString HistName, QString X_Title, QString Y_Title, QString Z_Title = "");
  void SetLineProperties(QString HistName, int LineColor, int LineStyle, int LineWidth);

  void Fill(QString HistName, double val, double weight);
  void Fill2D(QString HistName, double x, double y, double weight);

  void Draw(QString HistName, QString options);

  void Smooth(QString HistName, int times);

signals:
  void RequestDraw(TObject* obj, QString options, bool fFocus);

private:
  MainWindow* MW;
};

//unit to build and draw a GRAPH (TGraph of ROOT)
class InterfaceToGraphs : public BaseInterfaceClass
{
  Q_OBJECT

public:
  InterfaceToGraphs(MainWindow* MW);
  ~InterfaceToGraphs(){}

  virtual bool InitOnRun();

public slots:
  void NewGraph(QString GraphName);
  void SetMarkerProperties(QString GraphName, int MarkerColor, int MarkerStyle, int MarkerSize);
  void SetLineProperties(QString GraphName, int LineColor, int LineStyle, int LineWidth);
  void SetTitles(QString GraphName, QString X_Title, QString Y_Title);

  void AddPoint(QString GraphName, double x, double y);
  void Draw(QString GraphName, QString options);

signals:
  void RequestDraw(TObject* obj, QString options, bool fFocus);

private:
  MainWindow* MW;
};

//unit handling GRAPH WINDOW
class InterfaceToGraphWin : public BaseInterfaceClass
{
  Q_OBJECT

public:
  InterfaceToGraphWin(MainWindow* MW);
  ~InterfaceToGraphWin(){}

public slots:
  void Show();
  void Hide();

  void SetLog(bool Xaxis, bool Yaxis);

  //basket operation
  void addToBasket(QString Title);

  void SaveImage(QString fileName);

private:
  MainWindow* MW;
};

//unit handling GEOMETRY WINDOW
class InterfaceToGeoWin : public BaseInterfaceClass
{
  Q_OBJECT

public:
  InterfaceToGeoWin(MainWindow* MW, DetectorClass* Detector);
  ~InterfaceToGeoWin(){}

public slots:  
  void Show();
  void Hide();

  void BlockUpdates(bool on); //forbids updates

  //orientation of TView3D
  double GetPhi();
  double GetTheta();
  void SetPhi(double phi);
  void SetTheta(double theta);
  void Rotate(double Theta, double Phi, int Steps, int msPause = 50);

  //view manipulation
  void SetParallel(bool on);
  void UpdateView();

  //position and size
  int  GetX();
  int  GetY();
  int  GetW();
  int  GetH();

  //show things
  void ShowGeometry();
  void ShowPMnumbers();
  void ShowTracks(int num, int OnlyColor = -1);
  void ShowSPS_position();
  void ShowTracksMovie(int num, int steps, int msDelay, double dTheta, double dPhi, double rotSteps, int color = -1);

  //clear things
  void ClearTracks();
  void ClearMarkers();

  void SaveImage(QString fileName);

private:
  MainWindow* MW;
  DetectorClass* Detector;
};

//unit that handles LRF calculation / visualization
class InterfaceToLRF : public BaseInterfaceClass
{
  Q_OBJECT

public:
  InterfaceToLRF(MainWindow* MW);
  ~InterfaceToLRF(){}

public slots:
  void Make();

  void SetGrouping(bool flag);
  void SetSource(int selector);

private:
  MainWindow* MW;
};

//unit that handles events DATA
class InterfaceToData : public BaseInterfaceClass
{
  Q_OBJECT

public:
  InterfaceToData(MainWindow* MW, DetectorClass* Detector, EventsDataClass* EventsDataHub);
  ~InterfaceToData(){}

public slots:
  double GetPMsignal(int ievent, int ipm);
  int GetNumPMs();
  int GetNumEvents();

  //load data
  void LoadEventsTree(QString fileName, bool Append = false);
  void LoadEventsAscii(QString fileName, bool Append = false);

private:
  MainWindow* MW;
  DetectorClass* Detector;
  EventsDataClass* EventsDataHub;
};

//unit that handles SIMULATIONS
class InterfaceToSim : public BaseInterfaceClass
{
  Q_OBJECT

public:
  InterfaceToSim(MainWindow* MW, DetectorClass* Detector, EventsDataClass* EventsDataHub);
  ~InterfaceToSim(){}

  virtual void ForceStop();

public slots:
  bool RunPhotonSources();
  bool RunParticleSources();

  void SetSeed(long seed);
  long GetSeed();

  void SaveAsTree(QString fileName);
  void SaveAsText(QString fileName);

private:
  MainWindow* MW;
  DetectorClass* Detector;
  EventsDataClass* EventsDataHub;
};

//unit that handles CONFIGURATION load/safe/modification
class InterfaceToConfig : public BaseInterfaceClass
{
  Q_OBJECT

public:
  InterfaceToConfig(MainWindow* MW);
  ~InterfaceToConfig(){}

  virtual bool InitOnRun();

  QJsonObject Config;
  QString LastError;

public slots:
  bool Load(QString FileName);

  bool Replace(QString Key, QVariant val);

  //shortcuts in config
  void SetPassivePM(int pm);
  void SetActivePM(int pm);

  bool UpdateAllConfig();
  bool UpdateSimConfig();  // does not reconstruct the detector! but also no redraw
  bool UpdateRecConfig();  // does not reconstruct the detector! but also no redraw

private:
  MainWindow* MW;

  //void abort(QString message = "Aborted!") {emit AbortScriptEvaluation(message);}
  bool modifyJsonValue(QJsonObject& obj, const QString& path, const QJsonValue& newValue);
  void find(const QJsonObject &obj, QStringList Keys, QStringList& Found, QString Path = "");
  bool keyToNameAndIndex(QString Key, QString &Name, int &index);

  void SetPassiveStatus(int pm, bool flag);

};

// MESSENGER window
class InterfaceToTexter : public BaseInterfaceClass
{
  Q_OBJECT

public:
  InterfaceToTexter();
  ~InterfaceToTexter();

  QDialog *D;
  double X, Y;
  double W, H;

  QPlainTextEdit* e;
  bool fEnabled;

public slots:
  void Enable(bool flag) {fEnabled = flag;}
  void Append(QString txt);
  void Clear();
  void Show();
  void Hide();
  void Show(QString txt, int ms = -1);
  void SetTransparent(bool flag);

  void Move(double x, double y);
  void Resize(double w, double h);

  void SetFontSize(int size);

private:
  void Init(bool fTransparent);
};

#endif // INTERFACETOGLOBSCRIPT

