#ifndef GRAPHWINDOWCLASS_H
#define GRAPHWINDOWCLASS_H

#include <QMainWindow>
#include "Rtypes.h"
#include "TGraph.h"
#include "TAxis.h"
#include <TVectorD.h>
//#include "TPolyLine.h"

class MainWindow;
class RasterWindowGraphClass;
class TH1;
class TH2;
class TH1D;
//class TGraph;
class TGraph2D;
class TF1;
class TF2;
class TMultiGraph;
class QGraphicsView;
class ToolboxScene;
class QListWidgetItem;

class DrawObjectStructure
{
public:
    void setPointer(TObject* p) {Pointer = p;}
    TObject* getPointer() {return Pointer;}

    void setOptions(QString str) {Options = str;}
    void setOptions(const char* chars) {Options = chars;}
    QString getOptions() {return Options;}

    DrawObjectStructure() {Pointer=0;}
    DrawObjectStructure(TObject* pointer, const char* options) {Pointer = pointer; Options = options;}
    DrawObjectStructure(TObject* pointer, QString options) {Pointer = pointer; Options = options;}

private:
    TObject* Pointer;
    QString Options;
};

class BasketItemClass
{
public:
    QString Name;
    QString Type;
    QVector<DrawObjectStructure> DrawObjects;

    BasketItemClass(QString name, QVector<DrawObjectStructure>* drawObjects);
    BasketItemClass(){}
    ~BasketItemClass();

    void clearObjects();
};

namespace Ui {
class GraphWindowClass;
}

class GraphWindowClass : public QMainWindow
{
    Q_OBJECT

public:
    explicit GraphWindowClass(QWidget *parent, MainWindow *mw);
    ~GraphWindowClass();

    //main usage
public:
    void Draw(TObject* obj, const char* options = "", bool DoUpdate = true);
    void DrawWithoutFocus(TObject* obj, const char* options = "", bool DoUpdate = true);

    //control
    void ShowAndFocus();
    void SetAsActiveRootWindow();
    void ClearRootCanvas();
    void UpdateRootCanvas();
    void SetLog(bool X, bool Y);

    void ClearDrawObjects();

    //Canvas size in actual coordinates of plotted data
    double getCanvasMinX();
    double getCanvasMaxX();
    double getCanvasMinY();
    double getCanvasMaxY();

    //extraction of coordinates from graph
    bool IsExtractionComplete();
    bool IsExtractionCanceled() {return ExtractionCanceled;}

    //commands to start extraction of shapes on canvas
    void ExtractX();  //start extraction of X coordinate from a 1D graph/histogram using mouse
    void Extract2DLine();  //start extraction of ABC line coordinate from a 2D graph/histogram using mouse
    void Extract2DEllipse();  //start extraction of (T)ellipse from a 2D graph/histogram using mouse
    void Extract2DBox();  //start extraction of 2D box (2 opposite corners)
    void Extract2DPolygon();  //start extraction of 2D polygon, extraction ends by right click (or doubleclick?)

    //commands to get results of extraction
    double extractedX();
    double extracted2DLineA();
    double extracted2DLineB();
    double extracted2DLineC();
    double extracted2DEllipseX();
    double extracted2DEllipseY();
    double extracted2DEllipseR1();
    double extracted2DEllipseR2();
    double extracted2DEllipseTheta();
    double extractedX1();
    double extractedY1();
    double extractedX2();
    double extractedY2();
    QList<double> extractedPolygon();

    template <class T_int_double> TGraph* ShowGraph(const QVector<T_int_double> *x, const QVector<T_int_double> *y, Color_t color, const char *XTitle, const char *YTitle, int MarkerStyle=20, int MarkerSize=1, int LineStyle=1, int LineWidth = 2, const char* options = "", bool OnlyBuild = false)
    {
        int numEl = x->size();
        TVectorD xx(numEl);
        TVectorD yy(numEl);
        for (int i=0; i < numEl; i++)
        {
            xx[i] = x->at(i);
            yy[i] = y->at(i);
        }

        TGraph* gr = new TGraph(xx,yy);
        gr->SetTitle(""); gr->GetXaxis()->SetTitle(XTitle); gr->GetYaxis()->SetTitle(YTitle);
        gr->SetMarkerStyle(MarkerStyle); gr->SetMarkerColor(color); gr->SetMarkerSize(MarkerSize);
        gr->SetEditable(false); gr->GetYaxis()->SetTitleOffset((Float_t)1.30);
        gr->SetLineWidth(LineWidth); gr->SetLineColor(color); gr->SetLineStyle(LineStyle);

        //in the case of OnlyBuild return the graph (no drawing)
        if (OnlyBuild) return gr; //gr is not registered here!!!

        //drawing the graph
        QString opts = options;
        if (opts.contains("same",Qt::CaseInsensitive))
        {
            if (LineWidth == 0) GraphWindowClass::Draw(gr, "P"); else GraphWindowClass::Draw(gr, "PL");
        }
        else
        {
            if (LineWidth == 0) GraphWindowClass::Draw(gr, "AP"); else GraphWindowClass::Draw(gr, "APL");

            for (int i=0; i<TmpGraphs.size(); i++) delete TmpGraphs[i];
            TmpGraphs.clear();
        }

        //registering graph
        TmpGraphs.append(gr);
        return 0;
    }

    //void ShowGraph(const QVector<double> *x, const QVector<int> *y, Color_t color, const char *XTitle, const char *YTitle, int MarkerStyle=20, int MarkerSize=1, int LineStyle=1, int LineWidth = 2, const char* options = "");

    void onBusyOn();
    void OnBusyOff();

    void switchOffBasket();

    TObject *GetMainPlottedObject();

    void SaveGraph(QString fileName);

    void ShowProjection(QString type);
    void EnforceOverlayOff();

protected:
    void resizeEvent(QResizeEvent *event);
    void mouseMoveEvent(QMouseEvent *event);
    bool event(QEvent *event);
    void closeEvent(QCloseEvent *event);

public slots:
    // if the first object (no "same" option) was drawn without update, call this function after the manual update
    void UpdateControls(); //updates visualisation of the current master graph parameters
    void DoSaveGraph(QString name);
    void AddCurrentToBasket(QString name);

    void DrawStrOpt(TObject* obj, QString options = "", bool DoUpdate = true);

private slots:
    void Reshape();

    void on_cbToolBox_toggled(bool checked);
    void on_cobToolBox_currentIndexChanged(int index);
    void on_pbToolboxDragMode_clicked();

    void selBoxGeometryChanged();
    void selBoxControlsUpdated();
    void on_pbSelBoxToCenter_clicked();
    void on_pbSelBoxFGColor_clicked();
    void on_pbSelBoxBGColor_clicked();

    void rulerGeometryChanged();
    void rulerControlsP1Updated();
    void rulerControlsP2Updated();
    void rulerControlsLenAngleUpdated();
    void on_ledRulerTicksLength_editingFinished();
    void on_pbRulerFGColor_clicked();
    void on_pbRulerBGColor_clicked();
    void on_pbResetRuler_clicked();

    void on_pbAddToBasket_clicked();

    void on_cbGridX_toggled(bool checked);

    void on_cbGridY_toggled(bool checked);

    void on_cbLogX_toggled(bool checked);

    void on_cbLogY_toggled(bool checked);

    void on_ledXfrom_editingFinished();

    void on_ledXto_editingFinished();

    void on_ledYfrom_editingFinished();

    void on_ledYto_editingFinished();

    void on_ledZfrom_editingFinished();

    void on_ledZto_editingFinished();

    void on_cbShowLegend_toggled(bool checked);

    void on_pbZoom_clicked();

    void on_pbUnzoom_clicked();

    void on_leOptions_editingFinished();

    void on_pbXprojection_clicked();

    void on_pbYprojection_clicked();

    void on_cbShowBasket_toggled(bool checked);

    void on_lwBasket_itemDoubleClicked(QListWidgetItem *item);

    void on_pbBasketBackToLast_clicked();

    void contextMenuBasket(const QPoint &pos);
    void on_actionSave_image_triggered();

    void on_actionExport_data_as_text_triggered();

    void on_actionExport_data_using_bin_start_positions_TH1_triggered();

    void on_actionBasic_ROOT_triggered();

    void on_actionDeep_sea_triggered();

    void on_actionGrey_scale_triggered();

    void on_actionDark_body_radiator_triggered();

    void on_actionTwo_color_hue_triggered();

    void on_actionRainbow_triggered();

    void on_actionInverted_dark_body_triggered();

    void on_pbToolboxDragMode_2_clicked();

    void on_actionSave_root_object_triggered();

    void on_pbSmooth_clicked();

    void on_actionTop_triggered();

    void on_actionSide_triggered();

    void on_actionFront_triggered();

    void on_pbAttributes_clicked();

    void on_actionToggle_toolbar_toggled(bool arg1);

private:
    Ui::GraphWindowClass *ui;
    MainWindow *MW;
    RasterWindowGraphClass *RasterWindow;
    QWidget *QWinContainer;
    bool ExtractionCanceled;

    QVector<DrawObjectStructure> DrawObjects;  //currently drawn
    QVector<DrawObjectStructure> MasterDrawObjects; //last draw made from outse of the graph window
    QVector< BasketItemClass > Basket; //container with user selected "drawings"
    int CurrentBasketItem;
    int BasketMode;

    QVector<TGraph*> TmpGraphs;
    double TG_X0, TG_Y0;

    QGraphicsView* gvOver;
    ToolboxScene* scene;
    TH1D* hProjection;

    void DoDraw(TObject *obj, const char *options, bool DoUpdate); //actual drawing, does not have window focussing - done to avoid refocussing issues leading to bugs

    //flags
    bool TMPignore; //temporarily forbid updates - need for bulk update to avoid cross-modification
    bool BarShown;
    bool ColdStart;
    bool fFirstTime; //signals that "UnZoom" range values (xmin0, etc...) have to be stored

    double xmin, xmax, ymin, ymax, zmin, zmax;
    double xmin0, xmax0, ymin0, ymax0, zmin0, zmax0; //start values - filled on first draw, can be used to reset view with "Unzoom"
    QString old_option;

    void RedrawAll();
    void updateLegendVisibility();
    void startOverlayMode();
    void endOverlayMode();
    void UpdateBasketGUI();
    void ClearBasket();
    void ExportData(bool fUseBinCenters=true);
    void SaveBasket();
    void AppendBasket();
    QVector<DrawObjectStructure> *getCurrentDrawObjects();
};

#endif // GRAPHWINDOWCLASS_H
