#include "lrfwindow.h"
#include "ui_lrfwindow.h"
#include "mainwindow.h"
#include "pms.h"
#include "graphwindowclass.h"
#include "sensorlrfs.h"
#include "lrfaxial.h"
#include "lrfcaxial.h"
#include "lrfxy.h"
#include "lrfcomposite.h"
#include "lrfaxial3d.h"
#include "lrfsliced3d.h"
#include "sensorlocalcache.h"
#include "guiutils.h"
#include "viewer2darrayobject.h"
#include "geometrywindowclass.h"
#include "outputwindow.h"
#include "eventsdataclass.h"
#include "globalsettingsclass.h"

#include "bspline3.h"
#include "spline.h"
#include "windownavigatorclass.h"
#include "reconstructionwindow.h"
#include "jsonparser.h"

//Qt
#include <QtGui>
#include <QMessageBox>
#include <QPair>
#include <QFileDialog>
#include <qmath.h>
#include <QListWidgetItem>
#include <QVBoxLayout>

//std
#include <iostream>
#include <fstream>
#include <algorithm>

//ROOT
#include "TMath.h"
#include "TTree.h"
#include "TFile.h"
#include "TGraph.h"
#include "TMultiGraph.h"
#include "TGraph2D.h"
#include "TF2.h"
#include "TProfile2D.h"
#include "TH2D.h"

LRFwindow::LRFwindow(QWidget *parent, MainWindow *mw, EventsDataClass *eventsDataHub) :
  QMainWindow(parent),
  ui(new Ui::LRFwindow)
{
  MW = mw;
  EventsDataHub = eventsDataHub;

  d2t = 0;
  f2d = 0;
  ff2d = 0;
  d2f = 0;
  f1d = 0;
  ff1d = 0;
  MGraph = 0;
  h2d = 0;

  LRFdialog = 0;
  LRFviewObj = 0;
  gv = 0;

  tmpIgnore = false;
  lastPlotted = "";
  secondIter = -1;
  fOneClick = false;

  ui->setupUi(this);
  this->setFixedSize(this->width(), this->height());

  Qt::WindowFlags windowFlags = (Qt::Window | Qt::CustomizeWindowHint);
  windowFlags |= Qt::WindowCloseButtonHint;
  this->setWindowFlags( windowFlags );

  StopSignal = false;

  ui->pbUpdateGUI->setVisible(false);
  ui->pbUpdateHistory->setVisible(false);
  //ui->fSingleGroup->setEnabled(false);

  ui->lwLRFhistory->setContextMenuPolicy(Qt::CustomContextMenu);
  connect(ui->lwLRFhistory, SIGNAL(customContextMenuRequested(const QPoint &)),
          this,        	    SLOT(contextMenuHistory(const QPoint &)));

  LRFwindow::on_pbUpdateGUI_clicked(); //update GUI to set enable/visible/index status
  LRFwindow::on_pbShrink_clicked();
  ui->fSingleGroup->setEnabled(false);
}

LRFwindow::~LRFwindow()
{
  if (d2t) delete d2t;
  if (f2d) delete f2d;
  if (ff2d) delete ff2d;
  if (d2f) delete d2f;

  if (h2d) delete h2d;
  if (f1d) delete f1d;
  if (ff1d) delete ff1d;
  if (MGraph) delete MGraph; //deletes all the members

  if (LRFdialog)
    {
      LRFdialog->hide(); //does not help
      disconnect(gv, SIGNAL(MouseMovedSignal(QPointF*)), this, SLOT(paintLRFonDialog(QPointF*)));
    }
  if (gv) delete gv;
  gv = 0;
  if (LRFviewObj) delete LRFviewObj;
  LRFviewObj = 0;
  if (LRFdialog) delete LRFdialog;
  LRFdialog = 0;

  delete ui;
}

void LRFwindow::SetProgress(int val)
{
  ui->pbarLRF->setValue(val);
}

void LRFwindow::onBusyOn()
{
  ui->fAll->setEnabled(false);
}

void LRFwindow::onBusyOff()
{
  ui->fAll->setEnabled(true);
}

void LRFwindow::showLRF()
{
  //Watchdogs
  //are LRFs defined?
  if (!MW->SensLRF->isAllLRFsDefined())
    {
      message("Not all LRFs are defined!", this);
      return;
    }
  //if showing data or diff, have to check the data are available
  int datais = ui->cb_data_selector->currentIndex();
  bool plot_data = ui->cb_data->isChecked();
  bool plot_diff = ui->cb_diff->isChecked();
  //bool have_data = !MW->Reconstructor->isEventsEmpty();
  bool have_data = !EventsDataHub->isEmpty();
  //    qDebug() << "Have data connected?" << have_data;
  if (have_data && ( plot_data || plot_diff) )
    {
      //        qDebug() << "User selected to use scan(0)/loaded(1) data:" << datais;
      if (datais == 0)
        {
          //if (MW->Reconstructor->isScanEmpty())
          if (EventsDataHub->isScanEmpty())
            {
              qWarning() << "There is no scan data, attempting to show reconstruction data instead";
              datais = 1; //checked below!
            }
        }
      if (datais == 1)
        {
          //if (EventsDataHub->ReconstructionData.isEmpty() || !MW->Reconstructor->isReconstructionDataReady() )
          if (!EventsDataHub->isReconstructionDataReady() )
            {
              qWarning() << "Reconstruction was not performed!";
              have_data = false;
              plot_data = false;
              plot_diff = false;
            }
        }
    }

  //is there something to show?
  bool plot_lrf = ui->cb_LRF->isChecked();
  if (!plot_lrf && !plot_data && !plot_diff)
    {
      message("Nothing to show!", this);
      return;
    }

  int PMnumber = ui->sbPMnoButons->value();
  int maxPM = MW->PMs->count()-1;
  if (PMnumber>maxPM)  //protection
    {
      ui->sbPMnoButons->setValue(maxPM);
      PMnumber = maxPM;
    }
  bool check_z = ui->cb3D_LRFs->isChecked();
  double z0 = ui->ledZcenter->text().toDouble();
  double dz = ui->ledZrange->text().toDouble();
  double x0 = MW->PMs->X(PMnumber);
  double y0 = MW->PMs->Y(PMnumber);
  bool scale_by_energy = ui->cbEnergyScalling->isChecked();

  //    qDebug() << "PM# " << PMnumber << "x0=" << x0 << " y0=" << y0;

  double x, y, z, r, fdata, fdiff;
  std::vector <double> xx, yy, rr, ffdata, ffdiff;
  int counter = 0;
  double factor = 1.0;

  if (have_data && (plot_data || plot_diff) )
    {
      //int npts = MW->Reconstructor->Events->size();
      int npts = EventsDataHub->Events.size();
      //        qDebug() << "Number of connected events: " << npts;

      if (!datais)
        {
          //if (MW->Reconstructor->Scan->size() != npts)
          if ( EventsDataHub->Scan.size() != npts)
            {
              message("Error: Scan size not equal to Events size!", this);
              return;
            }
        }
      else
        {
          if (EventsDataHub->ReconstructionData.size() != npts)
            {
              message("Error: Reconstruction size not equal to Events size!", this);
              return;
            }
        }

      for (int i=0; i<npts; i++)
        {
          if (!datais)
            {
              //x = (*MW->Reconstructor->Scan)[i]->Points[0].r[0];
              x = EventsDataHub->Scan[i]->Points[0].r[0];
              y = EventsDataHub->Scan[i]->Points[0].r[1];
              z = EventsDataHub->Scan[i]->Points[0].r[2];
            }
          else
            { // mercury
              if (!EventsDataHub->ReconstructionData[i]->GoodEvent) continue;
              x = EventsDataHub->ReconstructionData[i]->Points[0].r[0];
              y = EventsDataHub->ReconstructionData[i]->Points[0].r[1];
              z = EventsDataHub->ReconstructionData[i]->Points[0].r[2];
              factor = EventsDataHub->ReconstructionData[i]->Points[0].energy;
            }

          if (check_z && (z<z0-dz || z>z0+dz)) continue;

          r = sqrt((x-x0)*(x-x0) + (y-y0)*(y-y0));
          //fdata = (*MW->Reconstructor->Events)[i][PMnumber];
          fdata = EventsDataHub->Events[i][PMnumber];
          if (scale_by_energy) fdata /= factor;
          fdiff = fdata - MW->SensLRF->getLRF(PMnumber, x, y, z);

          xx.push_back(x);
          yy.push_back(y);
          rr.push_back(r);
          ffdata.push_back(fdata);
          ffdiff.push_back(fdiff);
          //            qDebug() << "data size:" << ffdata.size();
          counter++;
        }
    }

  double minmax[4];
  if (plot_lrf)
    {      
      const PMsensor *sensor = MW->SensLRF->getIteration()->sensor(PMnumber);
      sensor->getGlobalMinMax(minmax);
      //qDebug() << "global xmin xmax ymin ymax"<<minmax[0]<<minmax[1]<<minmax[2]<<minmax[3];

      if (f2d) delete f2d;
      f2d = new TF2("f2d", MW->SensLRF, &SensorLRFs::getFunction2D,
                    minmax[0], minmax[1], minmax[2], minmax[3], 2,
                    "SensLRF", "getFunction2D");

      f2d->SetParameter(0, PMnumber);
      f2d->SetParameter(1, z0);
      MW->GraphWindow->DrawWithoutFocus(f2d, "surf");

      if (secondIter != -1)
        if (secondIter < MW->SensLRF->countIterations())
          {            
            //have to draw in the same range of coordnates - ROOT's "same" option does not work with "surf"
            if (ff2d) delete ff2d;
            ff2d = new TF2("f2d", MW->SensLRF, &SensorLRFs::getFunction2Dplus,
                          minmax[0], minmax[1], minmax[2], minmax[3], 3,
                          "SensLRF", "getFunction2D");

            double min = f2d->GetMinimum();
            double max = f2d->GetMaximum();
            ff2d->GetZaxis()->SetRangeUser(min, max);

            ff2d->SetParameter(0, PMnumber);
            ff2d->SetParameter(1, z0);
            ff2d->SetParameter(2, secondIter);

            ff2d->SetLineColor(1);
            ff2d->SetLineStyle(9);

            MW->GraphWindow->DrawWithoutFocus(ff2d, "surf same");
            MW->GraphWindow->UpdateRootCanvas();
          }
    }

  if (!have_data) return;

  if (plot_data)
    {
      if (d2t) delete d2t;
      d2t = new TGraph2D(ffdata.size(), &xx[0], &yy[0], &ffdata[0]);
      d2t->SetMarkerStyle(7);
      if (plot_lrf) MW->GraphWindow->DrawWithoutFocus(d2t, "samepcol");
      else          MW->GraphWindow->DrawWithoutFocus(d2t, "pcol");
    }

  if (plot_diff)
    {
      if (d2f) delete d2f;
      d2f = new TGraph2D(ffdiff.size(), &xx[0], &yy[0], &ffdiff[0]);
      d2f->SetMarkerStyle(7);
      if (plot_lrf || plot_data) MW->GraphWindow->DrawWithoutFocus(d2f, "samepcol");
      else                       MW->GraphWindow->DrawWithoutFocus(d2f, "pcol");
    }

}

void LRFwindow::on_pbStopLRFmake_clicked()
{
  StopSignal = true;
  ui->pbStopLRFmake->setText("stopping...");
  MW->Detector->SensLRF->requestStop();
}

bool LRFwindow::event(QEvent *event)
{
  if (!MW->WindowNavigator) return QMainWindow::event(event);

  if (event->type() == QEvent::Hide) MW->WindowNavigator->HideWindowTriggered("lrf");
  if (event->type() == QEvent::Show) MW->WindowNavigator->ShowWindowTriggered("lrf");

  return QMainWindow::event(event);
}

void LRFwindow::on_pbUpdateGUI_clicked()
{
  if (tmpIgnore) return;

  // PM groups
  ui->fPMgroups->setEnabled(ui->cbUseGroupping->isChecked());

  // 2D or 3D
  bool f3D = ui->cb3D_LRFs->isChecked();
  //    qDebug()<<"3D?"<<f3D;
  if (f3D) ui->swLRFtype->setCurrentIndex(1); else ui->swLRFtype->setCurrentIndex(0);

  // Compressing for axial
  bool fCompression = true;
  if (f3D)
    {
      if (ui->cob3Dtype->currentIndex() == 1) fCompression = false; //not for sliced XY
    }
  else
    {
      if (ui->cob2Dtype->currentIndex() == 1) fCompression = false; //for XY cannot use compression
    }
  ui->fCompression->setVisible(fCompression);

  // Nodes
  if (f3D)
    {
      //        qDebug()<<"3D type:"<< ui->cob3Dtype->currentIndex();
      if (ui->cob3Dtype->currentIndex() == 0)
        {
          // 3D axial
          ui->lbl_nodes_X->setText("R:");
          ui->lbl_nodes_Y->setText("Z:");
          ui->sbSplNodesY->setEnabled(true);
        }
      if (ui->cob3Dtype->currentIndex() == 1)
        {
          // 3D sliced
          ui->lbl_nodes_X->setText("X&Y:");
          ui->lbl_nodes_Y->setText("Z:");
          ui->sbSplNodesY->setEnabled(true);
        }
    }
  else
    {
      //       qDebug()<<"2D type:"<< ui->cob2Dtype->currentIndex();
      switch (ui->cob2Dtype->currentIndex())
        {
        case 0: // axial
          ui->lbl_nodes_X->setText("R:");
          ui->lbl_nodes_Y->setText("");
          ui->sbSplNodesY->setEnabled(false);
          break;
        case 1: // XY
          ui->lbl_nodes_X->setText("X:");
          ui->lbl_nodes_Y->setText("Y:");
          ui->sbSplNodesY->setEnabled(true);
          break;
        case 2: // polar
          ui->lbl_nodes_X->setText("R:");
          ui->lbl_nodes_Y->setText("Phi:");
          ui->sbSplNodesY->setEnabled(true);
          break;
        case 3: // composite
          ui->lbl_nodes_X->setText("R:");
          ui->lbl_nodes_Y->setText("X&Y:");
          ui->sbSplNodesY->setEnabled(true);
          break;
        }
    }

  //In show, Z control - better to synchronize with actual LRF ***
  ui->fZ->setEnabled(f3D);
}

void LRFwindow::LRF_ModuleReadySlot(bool ready)
{
  //status label
  QString status;
  if (ready)
    {
      status = "Ready";
    }
  else
    {
      status = "Not Ready";
      if (MW->SensLRF)
        if (ui->lwLRFhistory->count() != 0) LRFwindow::on_pbUpdateHistory_clicked();
    }

  ui->label_ready->setText(status);
  if (ready) ui->label_ready->setStyleSheet("QLabel { color : green; }");
  else ui->label_ready->setStyleSheet("QLabel { color : red; }");

  ui->pbSaveLRFs->setEnabled(ready); //save enabled
  ui->pbSaveIterations->setEnabled(ready); //save enabled
  ui->fPlot->setEnabled(ready); //Plot enabled  
}

void LRFwindow::on_pbSaveLRFs_clicked()
{
  LRFwindow::SaveLRFDialog(this);
}

void LRFwindow::SaveLRFDialog(QWidget* wid)
{
  if (!MW->SensLRF->isAllLRFsDefined())
    {
      message("Cannot save empty or incomplete LRF set!", wid);
      return;
    }

  QFileDialog *fileDialog = new QFileDialog;
  fileDialog->setDefaultSuffix("json");
  QString fileName = fileDialog->getSaveFileName(this, "Save LRF Block", MW->GlobSet->LastOpenDir, "json files(*.json);;all files(*.*)");
  if (fileName.isEmpty()) return;
  MW->GlobSet->LastOpenDir = QFileInfo(fileName).absolutePath();

  int iCur = MW->SensLRF->getCurrentIterIndex();
  if (iCur == -1)
    {
      message("There is no LRF data!", wid);
      return;
    }

  LRFwindow::SaveIteration(iCur, fileName);
}

void LRFwindow::SaveIteration(int iter, QString fileName)
{
  QFileInfo file(fileName);
  if(file.suffix().isEmpty()) fileName += ".json";
  QFile saveFile(fileName);
  if (!saveFile.open(QIODevice::WriteOnly))
    {
      message("Cannot open file "+fileName + " to save LRFs!");
      return;
    }

  QJsonObject json;
  if (!MW->SensLRF->saveIteration(iter, json))
    {
      message(MW->SensLRF->getLastError());
      saveFile.close();
      return;
    }

  QJsonDocument saveDoc(json);
  saveFile.write(saveDoc.toJson());
  saveFile.close();
}

void LRFwindow::on_pbSaveIterations_clicked()
{
  int numIter = MW->SensLRF->countIterations();
  if (numIter == 0)
    {
      message("The iteration history is empty!");
      return;
    }

  QString fileName = QFileDialog::getSaveFileName(this, "Save all Iterations: provided name will be used to create a dir", MW->GlobSet->LastOpenDir, "json files(*.json);;all files(*.*)");
  if (fileName.isEmpty()) return;
  MW->GlobSet->LastOpenDir = QFileInfo(fileName).absolutePath();

  QFileInfo fi(fileName);
  QString dir = fi.baseName();
  if (!QDir(fi.path()+"/"+dir).exists()) QDir().mkdir(fi.path()+"/"+dir);
  else
    {
      message("This directory already exists, please choose another name");
      return;
    }

  for (int iter=0; iter<numIter; iter++)
    {
      QString name = MW->SensLRF->getIterationName(iter);
      name = fi.path()+"/"+dir+"/"+name;
      if (QFileInfo(name+".json").exists())
        {
          int index = 1;
          while (QFileInfo(name+QString::number(index)+".json").exists()) index++;
          name += QString::number(index);
        }
      name += ".json";

      LRFwindow::SaveIteration(iter, name);
    }  
}

void LRFwindow::on_pbLoadLRFs_clicked()
{
  LRFwindow::LoadLRFDialog(this);
}

void LRFwindow::LoadLRFDialog(QWidget* wid)
{
  QStringList fileNames = QFileDialog::getOpenFileNames(this, "Load LRF Block", MW->GlobSet->LastOpenDir, "JSON files (*.json)");
  if (fileNames.isEmpty()) return;
  MW->GlobSet->LastOpenDir = QFileInfo(fileNames.first()).absolutePath();

  for (int ifile=0; ifile<fileNames.size(); ifile++)
    {
      QString fileName = fileNames[ifile];
      QFile loadFile(fileName);
      if (!loadFile.open(QIODevice::ReadOnly))
        {
          message("Cannot open save file", wid);
          continue;
        }

      QByteArray saveData = loadFile.readAll();
      loadFile.close();

      QJsonDocument loadDoc(QJsonDocument::fromJson(saveData));
      QJsonObject json = loadDoc.object();
      if(!MW->SensLRF->loadAll(json))
      {
          message("Failed to load LRFs: "+MW->SensLRF->getLastError());
          continue;
      }

      if (json.contains("LRF_window"))
        {
          QJsonObject a = json["LRF_window"].toObject();
          if (!a.isEmpty()) LRFwindow::loadJSON(a);
          else qDebug() << "LRF window settings were empty!";
        }
    }

  //update GUI  
  LRFwindow::updateIterationIndication(); 
}

void LRFwindow::on_pbShow2D_clicked()
{  
  if (MW->SensLRF->getIteration() == 0) return;
  MW->GraphWindow->ShowAndFocus();
  LRFwindow::showLRF();
  lastPlotted = "scatter";
}

void LRFwindow::on_sbPMnoButons_editingFinished()
{
  if (MW->GraphWindow->isVisible())
    {
      if (lastPlotted == "scatter") LRFwindow::showLRF();
      if (lastPlotted == "distrib") LRFwindow::drawRadial();
    }
}

void LRFwindow::on_sbPMnumberSpinButtons_valueChanged(int arg1)
{
  if (ui->sbPMnoButons->value() != arg1) ui->sbPMnoButons->setValue(arg1);

  if (tmpIgnore) return;
  if (MW->GraphWindow->isVisible())
    {
      if (lastPlotted == "scatter") LRFwindow::showLRF();
      if (lastPlotted == "distrib") LRFwindow::drawRadial();
    }
}

void LRFwindow::on_sbPMnoButons_valueChanged(int arg1)
{
  tmpIgnore = true;
  ui->sbPMnumberSpinButtons->setValue(arg1); //link to respect 0
  tmpIgnore = false;
}

void LRFwindow::writeToJson(QJsonObject &json) const
{
  //Version number
  int versionNumber = ANTS2_VERSION;
  json["ANTS2build"] = versionNumber;

  //General
  json["DataSelector"] = ui->cb_data_selector->currentIndex();
  json["UseGrid"] = ui->cb_use_grid->isChecked();
  json["StoreError"] = ui->cb_store_error->isChecked();
  json["UseEnergy"] = ui->cbEnergyScalling->isChecked();

  //Grouping
  bool fUseGrouping = ui->cbUseGroupping->isChecked();
  json["UseGroupping"] = fUseGrouping;
  if (fUseGrouping)
    {
      json["GrouppingType"] = ui->cb_pmt_group_type->currentIndex();
      json["AdjustGains"] = ui->cb_adj_gains->isChecked();
      json["GroupingOption"] = ui->cb_pmt_group_type->currentText();
    }
  else json["GroupingOption"] = "Individual";

  //LRF type
  bool f3D = ui->cb3D_LRFs->isChecked();
  json["LRF_3D"] = f3D;
  int LRFtype;
  if (f3D) LRFtype = ui->cob3Dtype->currentIndex();
  else LRFtype = ui->cob2Dtype->currentIndex();
  json["LRF_type"] = LRFtype;

  //Compression properties
  bool fCompress = ui->cb_compress_r->isChecked();
  json["LRF_compress"] = fCompress;
  if (fCompress)
  {
    json["Compression_k"] = ui->led_compression_k->text().toDouble();
    json["Compression_r0"] = ui->led_compression_r0->text().toDouble();
    json["Compression_lam"] = ui->led_compression_lam->text().toDouble();
  }

  //Number of nodes
  json["Nodes_x"] = ui->sbSplNodesX->value();
  json["Nodes_y"] = ui->sbSplNodesY->value();

  //Limit group making
  bool fLimitGroup = ui->cbMakePMGroup->isChecked();
  json["LimitGroup"] = fLimitGroup;
  if (fLimitGroup) json["GroupToMake"] = ui->sbMakePMgroup->value();
}

void LRFwindow::loadJSON(QJsonObject &json)
{
  tmpIgnore = true; //starting bulk update

  JsonParser parser(json);
  int version;
  parser.ParseObject("ANTS2 version", version);
  bool bTmp; int iTmp;

  parser.ParseObject("DataSelector", iTmp);
  ui->cb_data_selector->setCurrentIndex(iTmp);
  parser.ParseObject("UseGrid", bTmp);
  ui->cb_use_grid->setChecked(bTmp);
  parser.ParseObject("StoreError", bTmp);
  ui->cb_store_error->setChecked(bTmp);
  parser.ParseObject("UseEnergy", bTmp);
  ui->cbEnergyScalling->setChecked(bTmp);

  parser.ParseObject("UseGroupping", bTmp);
  ui->cbUseGroupping->setChecked(bTmp);
  parser.ParseObject("GrouppingType", iTmp);
  if (iTmp>ui->cb_pmt_group_type->count()-1) iTmp = 0;
  if (bTmp) ui->cb_pmt_group_type->setCurrentIndex(iTmp);
  else ui->cb_pmt_group_type->setCurrentIndex(0);
  parser.ParseObject("AdjustGains", bTmp);
  ui->cb_adj_gains->setChecked(bTmp);

  bool f3D;
  parser.ParseObject("LRF_3D", f3D);
  ui->cb3D_LRFs->setChecked(f3D);
  parser.ParseObject("LRF_type", iTmp);
  if (f3D) ui->cob3Dtype->setCurrentIndex(iTmp);
  else ui->cob2Dtype->setCurrentIndex(iTmp);

  JsonToCheckbox(json, "LRF_compress", ui->cb_compress_r);
  JsonToLineEdit(json, "Compression_k", ui->led_compression_k);
  JsonToLineEdit(json, "Compression_r0", ui->led_compression_r0);
  JsonToLineEdit(json, "Compression_lam", ui->led_compression_lam);

  JsonToSpinBox(json, "Nodes_x", ui->sbSplNodesX);
  JsonToSpinBox(json, "Nodes_y", ui->sbSplNodesY);

  tmpIgnore = false;

  LRFwindow::on_pbUpdateGUI_clicked();
  LRFwindow::on_pbUpdateHistory_clicked();
}

void LRFwindow::on_pbShow1Ddistr_clicked()
{
  if (MW->SensLRF->getIteration() == 0) return;
  MW->GraphWindow->ShowAndFocus();
  LRFwindow::drawRadial();
  lastPlotted = "distrib";
}

void LRFwindow::on_cb_data_2_clicked()
{
  bool data = ui->cb_data_2->isChecked();
  bool dif  = ui->cb_diff_2->isChecked();

  if (data && dif) ui->cb_diff_2->setChecked(false);
  LRFwindow::on_pbShow1Ddistr_clicked();
}

void LRFwindow::on_cb_diff_2_clicked()
{
  bool data = ui->cb_data_2->isChecked();
  bool dif  = ui->cb_diff_2->isChecked();

  if (data && dif) ui->cb_data_2->setChecked(false);
  LRFwindow::on_pbShow1Ddistr_clicked();
}

bool LRFwindow::checkWatchdogs()
{
    const int minDataSize = 100;

    int dataScanRecon = ui->cb_data_selector->currentIndex(); //0 - Scan, 1 - Reconstr data
    QVector < QVector <float> > *events = &EventsDataHub->Events;

    int dataSize = events->size();
    if (dataSize == 0)
      {
        LastError = "Data is empty!";
        return false;
      }
    if (dataSize < minDataSize)
      {
        LastError = "Data set is too small!";
        return false;
      }
    if (EventsDataHub->countGoodEvents() < minDataSize)
      {
        LastError = "Data set is too small!";
        return false;
      }

    if (dataScanRecon == 0)
      {      
        if (EventsDataHub->isScanEmpty())
          {
            LastError = "Scan data are not available!";
            return false;
          }
      }
    else
      {       
        if (!EventsDataHub->isReconstructionDataReady())
          {
            LastError = "Reconstruction was not yet performed!";
            return false;
          }
      }

    bool fLimitGroup = ui->cbMakePMGroup->isChecked();
    int igrToMake = ui->sbMakePMgroup->value();
    if(fLimitGroup)
       {
         if (!MW->SensLRF->getIteration())
           {
             LastError = "Current iteration is empty!";
             return false;
           }
         if (igrToMake >= MW->SensLRF->countGroups())
           {
             LastError = "Group index out of bounds";
             return false;
           }
      }

    int type;
    bool f3D = ui->cb3D_LRFs->isChecked();
    if (f3D) type = ui->cob3Dtype->currentIndex() + 4; // *** absolute number!!!
    else type = ui->cob2Dtype->currentIndex();
    //  qDebug() << "LRF type: " << type;
    if (type == 2)
      {
        LastError = "Polar LRFs are not implemented in this version";
        return false;
      }
    return true;
}

void LRFwindow::on_pbMakeLRF_clicked()
{
  MW->WindowNavigator->ResetAllProgressBars();
  MW->WindowNavigator->BusyOn();
  ui->pbStopLRFmake->setEnabled(true);
  qApp->processEvents();

  QString res = LRFwindow::MakeLRFs();
  if (!res.isEmpty()) message(res, this);

  ui->pbStopLRFmake->setEnabled(false);
  ui->pbStopLRFmake->setText("Stop");
  MW->WindowNavigator->BusyOff(false);
}

QString LRFwindow::MakeLRFs()
{
  //checking that all necessary data is present and valid
  if (!LRFwindow::checkWatchdogs()) return LastError;

  //packing settings
  QJsonObject json;
  writeToJson(json);
  SaveJsonToFile(json, "makeLRF.json");

  bool ok = MW->Detector->SensLRF->makeLRFs(json, EventsDataHub, MW->Detector->PMs);

  LRFwindow::on_pbUpdateHistory_clicked(); //updating iteration history indication

  if (!ok) return MW->Detector->SensLRF->getLastError();
  else return "";
}

void LRFwindow::on_pbAdjustOnlyGains_clicked()
{
    //checking that all necessary data is present and valid
    if (!LRFwindow::checkWatchdogs())
      {
        message(LastError, this);
        return;
      }

    //packing settings
    QJsonObject json;
    writeToJson(json);
    SaveJsonToFile(json, "makeLRF.json");

    MW->WindowNavigator->ResetAllProgressBars();
    MW->WindowNavigator->BusyOn();
    ui->pbStopLRFmake->setEnabled(true);
    qApp->processEvents();

    bool ok = MW->Detector->SensLRF->onlyGains(json, EventsDataHub, MW->Detector->PMs);

    if (!ok) message(MW->Detector->SensLRF->getLastError(), this);
    LRFwindow::on_pbUpdateHistory_clicked(); //updating iteration history indication
    ui->pbStopLRFmake->setEnabled(false);
    ui->pbStopLRFmake->setText("Stop");
    MW->WindowNavigator->BusyOff(false);
}

void LRFwindow::onProgressReportReceived(int progress)
{
  MW->WindowNavigator->setProgress(progress);
  ui->pbarLRF->setValue(progress);
}

void LRFwindow::DoMakeGroupLRF(int igroup)
{
  qDebug()<< "\nScript requested to make group LRF, group #:"<<igroup;
  if (igroup<0 || igroup>= MW->SensLRF->countGroups())
    {
      qDebug() << "Wrong iteration group selected!";
      return;
    }
  ui->cbMakePMGroup->setChecked(true);
  ui->sbMakePMgroup->setValue(igroup);
  LRFwindow::on_pbMakeLRF_clicked();
  ui->cbMakePMGroup->setChecked(false);
}

void LRFwindow::drawRadial()
{
  //Watchdogs
  //are LRFs defined?
  if (!MW->SensLRF->isAllLRFsDefined())
    {
      message("LRFs are not defined!", this);
      return;
    }
  //if showing data or diff, have to check the data are available
  int datais = ui->cb_data_selector->currentIndex(); //0 - sim, 1 - loaded
  bool plot_data = ui->cb_data_2->isChecked();
  bool plot_diff = ui->cb_diff_2->isChecked();
  //bool have_data = !MW->Reconstructor->isEventsEmpty();
  bool have_data = !EventsDataHub->isEmpty();
  //    qDebug() << "Have data connected?" << have_data;
  if (have_data && ( plot_data || plot_diff) )
    {
      //        qDebug() << "User selected to use scan(0)/loaded(1) data:" << datais;
      if (datais == 0)
        {
          //if (MW->Reconstructor->isScanEmpty())
          if (EventsDataHub->isScanEmpty())
            {
              qWarning() << "There is no scan data, attempting to show reconstruction data instead";
              datais = 1; //checked below!
            }
        }
      if (datais == 1)
        {
          //if (EventsDataHub->ReconstructionData.isEmpty() || !MW->Reconstructor->isReconstructionDataReady() )
          if (!EventsDataHub->isReconstructionDataReady() )
            {
              qWarning() << "Reconstruction was not performed!";
              have_data = false;
              plot_data = false;
              plot_diff = false;
            }
        }
    }

  //is there something to show?
  bool plot_lrf = ui->cb_LRF_2->isChecked();
  if (!plot_lrf && !plot_data && !plot_diff)
    {
      message("Nothing to show!", this);
      return;
    }

  int PMnumber = ui->sbPMnoButons->value();
  int maxPM = MW->PMs->count()-1;
  if (PMnumber>maxPM)  //protection
    {
      ui->sbPMnoButons->setValue(maxPM);
      PMnumber = maxPM;
    }
  bool check_z = ui->cb3D_LRFs->isChecked();
  double z0 = ui->ledZcenter->text().toDouble();
  double dz = ui->ledZrange->text().toDouble();
  double x0 = MW->PMs->X(PMnumber);
  double y0 = MW->PMs->Y(PMnumber);
  bool scale_by_energy = ui->cbEnergyScalling->isChecked();

  //    qDebug() << "PM# " << PMnumber << "x0=" << x0 << " y0=" << y0;

  if (have_data && (plot_data || plot_diff) )
    {
      //int npts = MW->Reconstructor->Events->size();
      int npts = EventsDataHub->Events.size();
      //        qDebug() << "Number of connected events: " << npts;

      if (!datais)
        {
          //if (MW->Reconstructor->Scan->size() != npts)
          if (EventsDataHub->Scan.size() != npts)
            {
              message("Error: Scan size not equal to Events size!", this);
              return;
            }
        }
      else
        {
          if (EventsDataHub->ReconstructionData.size() != npts)
            {
              message("Error: Reconstruction size not equal to Events size!", this);
              return;
            }
        }    

      int bins = ui->sbNumBins_2->value();
      if (h2d) delete h2d;
      h2d = new TH2D("tmpTestHist", "", bins, 0, 0, bins, 0, 0);

      double factor = 1.0;
      for (int i=0; i<npts; i++)
        {
          double *pos;
          if (!datais)          
            pos = EventsDataHub->Scan[i]->Points[0].r;
          else
            { // mercury
              if (!EventsDataHub->ReconstructionData[i]->GoodEvent) continue;
              pos = EventsDataHub->ReconstructionData[i]->Points[0].r;
              factor = EventsDataHub->ReconstructionData[i]->Points[0].energy;
            }

          if (check_z && (pos[2]<z0-dz || pos[2]>z0+dz)) continue;

          double r = hypot(pos[0]-x0, pos[1]-y0);
          //double fdata = (*MW->Reconstructor->Events)[i][PMnumber];
          double fdata = EventsDataHub->Events[i][PMnumber];
          if (scale_by_energy) fdata /= factor;
          double fdiff = fdata - MW->SensLRF->getSensor(PMnumber)->eval(pos);

          if (plot_diff) h2d->Fill(r, fdiff);
          else h2d->Fill(r, fdata);
        }

      h2d->GetXaxis()->SetTitle("Radial distance, mm");
      MW->GraphWindow->DrawWithoutFocus(h2d, "colz");
    }

  if (plot_lrf)
    {
      const LRF2 *lrf = (*MW->SensLRF)[PMnumber];
      if (f1d) delete f1d;
      f1d = new TF1("f1d", MW->SensLRF, &SensorLRFs::getRadial, 0, lrf->getRmax()*1.01, 2,
                    "SensorLRFs", "getRadial");
      f1d->SetParameter(0, PMnumber);
      f1d->SetParameter(1, z0);

      if (have_data && (plot_data || plot_diff) ) MW->GraphWindow->DrawWithoutFocus(f1d, "same");
      else
        {
          f1d->GetXaxis()->SetTitle("Radial distance, mm");
          f1d->GetYaxis()->SetTitle("LRF");
          TString str = "LRF of pm#";
          str += PMnumber;
          f1d->SetTitle(str);
          MW->GraphWindow->DrawWithoutFocus(f1d);
        }


      if (secondIter != -1)
        if (secondIter < MW->SensLRF->countIterations())
          {
            //qDebug() << "adding second curve for iter#"<<secondIter;
            //adding second lrf
            const LRF2 *lrf = MW->SensLRF->getIteration(secondIter)->sensor(PMnumber)->GetLRF();
            if (ff1d) delete ff1d;
            ff1d = new TF1("ff1d", MW->SensLRF, &SensorLRFs::getRadialPlus, 0, lrf->getRmax(), 3,
                                "SensorLRFs", "getRadial");
            ff1d->SetParameter(0, PMnumber);
            ff1d->SetParameter(1, z0);
            ff1d->SetParameter(2, secondIter);

            ff1d->SetLineColor(1);
            ff1d->SetLineStyle(9);

            MW->GraphWindow->DrawWithoutFocus(ff1d, "same");
          }
    }

  if (ui->cbShowNodePositions->isChecked())
    {
      const LRF2 *lrfGen = (*MW->SensLRF)[PMnumber];
      QString type = lrfGen->type();
      if (type != "ComprAxial" && type != "Axial")
        {
          message("Unsupported LRF type -cannot show LRF nodes", this);
          return;
        }
      LRFaxial* lrf = (LRFaxial*)lrfGen;
      int nodes = lrf->getNint();
      double rmax = lrf->getRmax();
      //qDebug() <<"type, nodes, rmax:"<< type << nodes << rmax;
      double Rmin = lrf->compress(0);
      double Rmax = lrf->compress(rmax);
      double DX = Rmax-Rmin;
      //qDebug() <<"Rmin, Rmax:" <<Rmin<<Rmax<<"\n";

      QVector <double> GrX, GrY;
      int lastnode = -1;
      for (int ix=0; ix<102; ix++)
        {
          double x = rmax*ix/100.0;
          double X = lrf->compress(x);
          int node = X * nodes/DX;
          if (node>lastnode)
            {
             GrX << x;
             GrY << 0;
             lastnode = node;
             //qDebug() << lastnode << x;
            }
        }
      TGraph* gr = MW->GraphWindow->ShowGraph(&GrX, &GrY, kBlue, "", "", 2, 6, 0, 0, "", true);
      MW->GraphWindow->Draw(gr, "P same");
    }
}

void LRFwindow::on_pbUpdateHistory_clicked()
{
  ui->lwLRFhistory->clear();

  if (!MW->SensLRF->isCurrentIterDefined())
    {
//      qDebug()<< "Current iter is NULL";
      ui->leCurrentLRFs->setText("--");
      emit MW->SensLRF->SensorLRFsReadySignal(false);
      return;
    }

  ui->leCurrentLRFs->setText(MW->SensLRF->getIterationName());

  int curIt = MW->SensLRF->getCurrentIterIndex();
//  qDebug() << "--cur it index:" << curIt;
  if (secondIter == curIt) secondIter = -1; //protection

  //updating the list widget
  int numIters = MW->SensLRF->countIterations();
//  qDebug() << "--it groups:"<<numIters;

  for (int iter = numIters-1; iter>-1; iter--)
    {
      QString name = MW->SensLRF->getIterationName(iter);
//      qDebug() << "---- iter:"<<iter<<"name:"<<name;
      QListWidgetItem* pItem =new QListWidgetItem(name);
      if (iter == secondIter)
        {          
          pItem->setText(name);
          pItem->setBackgroundColor(QColor(55,142,192));
          pItem->setForeground(Qt::white);
        }
      else if (iter == curIt)
        {
          //name = name;
          pItem->setText(name);
          pItem->setForeground(Qt::red);
          QFont font = pItem->font();
          font.setBold(true);
          pItem->setFont(font);
        }
      else pItem->setForeground(Qt::black);


      QString tt = name;
    //  tt += "<br>---------------------------";
      tt += "<br>"+MW->SensLRF->getIteration(iter)->getGrouppingOption();
      const LRF2* lrf = MW->SensLRF->getIteration(iter)->getGroups()->first().getLRF();
      if (lrf)
        {
          tt += "<br>"+QString(lrf->type());
          QJsonObject json = lrf->reportSettings();
          if (json.contains("n1"))
            {
              tt += "<br>"+ QString::number(json["n1"].toInt());
              if (json.contains("n2")) tt += " x "+ QString::number(json["n2"].toInt());
            }
        }
      pItem->setToolTip(tt);

      ui->lwLRFhistory->addItem(pItem);
    }
}

void LRFwindow::on_leCurrentLRFs_editingFinished()
{
  MW->SensLRF->setCurrentIterName(ui->leCurrentLRFs->text());
  LRFwindow::on_pbUpdateHistory_clicked();
}

void LRFwindow::historyClickedTimeout()
{
  //was a single click
  fOneClick = false;

  int row = ui->lwLRFhistory->currentRow();
//  qDebug() << "Single-clicked row:" << row;
  int numIters = MW->SensLRF->countIterations();
  int thisSel = numIters-1 - row;
  if (thisSel == secondIter) secondIter = -1;
  else secondIter = thisSel;

  LRFwindow::on_pbUpdateHistory_clicked();
}

void LRFwindow::on_lwLRFhistory_itemClicked(QListWidgetItem * /*item*/)
{
  if (!fOneClick)
    {
      //giving user chance to click again - detecting double click
      timer.singleShot(300, this, SLOT(historyClickedTimeout()));
      fOneClick = true;
      return;
    }

  //already the second click!
  timer.stop();
  int row = ui->lwLRFhistory->currentRow();
//  qDebug() << "Double clicked row:" << row;
  int numIters = MW->SensLRF->countIterations();
  int iterSelected = numIters-1 - row;
  MW->SensLRF->setCurrentIter(iterSelected);
  if (secondIter == iterSelected) secondIter = -1;

  LRFwindow::updateIterationIndication();
}

void LRFwindow::updateIterationIndication()
{
    if (MW->SensLRF->getIteration())
      {
        //update groupping option
        LRFwindow::updateGroupsInfo();

        //update GUI for LRF-based info
        QJsonObject json = (*MW->SensLRF)[0]->reportSettings();
        LRFwindow::updateLRFinfo(json);
      }

    //update history widget
    LRFwindow::on_pbUpdateHistory_clicked();
}

void LRFwindow::updateGroupsInfo()
{
  tmpIgnore = true;
  IterationGroups *iteration = MW->SensLRF->getIteration();
  if(!iteration)
  {
      qWarning()<<"LRFwindow::updateGroupsInfo() currentIteration is NULL";
      return;
  }
  QString option = iteration->getGrouppingOption();
  //qDebug() << "Groupping option: " << option;
  if (option == "Individual")
    {
      ui->cbUseGroupping->setChecked(false);
      ui->fPMgroups->setEnabled(false);
    }
  else
    {
      ui->cbUseGroupping->setChecked(true);
      ui->fPMgroups->setEnabled(true);
      if (option.startsWith("One LRF")) ui->cb_pmt_group_type->setCurrentIndex(0);
      else if (option.startsWith("Group PMs by radius")) ui->cb_pmt_group_type->setCurrentIndex(1);
      else if (option.startsWith("Assume square")) ui->cb_pmt_group_type->setCurrentIndex(2);
      else if (option.startsWith("Assume hexagonal")) ui->cb_pmt_group_type->setCurrentIndex(3);
      else
        {
          ui->cbUseGroupping->setChecked(false);
          ui->fPMgroups->setEnabled(false);
          //message("Unknown groupping option!", this);
        }
    }

  tmpIgnore = false;
}


void LRFwindow::updateLRFinfo(QJsonObject &json)
{
  //qDebug() << json;
  tmpIgnore = true; //starting bulk update
  JsonParser parser(json);

  //LRF type
  QString type;
  if (!parser.ParseObject("type", type))
    {
      message("Unknown format!", this);
      return;
    }

  //reconstruct Z?
  if (type == "Axial3D" || type == "Radial3D" || type == "Sliced3D")
    {
      ui->cb3D_LRFs->setChecked(true);
      ui->swLRFtype->setCurrentIndex(1);
    }
  else
    {
      ui->cb3D_LRFs->setChecked(false);
      ui->swLRFtype->setCurrentIndex(0);
    }

  //type indicator
  if (type == "Axial" || type == "Radial" || type == "ComprAxial" || type == "ComprRad" ) ui->cob2Dtype->setCurrentIndex(0);
  if (type == "XY" || type == "Freeform") ui->cob2Dtype->setCurrentIndex(1);
  if (type == "Polar") ui->cob2Dtype->setCurrentIndex(2);
  if (type == "Composite") ui->cob2Dtype->setCurrentIndex(3);
  if (type == "Axial3D" || type == "Radial3D") ui->cob3Dtype->setCurrentIndex(0);
  if (type == "Sliced3D") ui->cob3Dtype->setCurrentIndex(1);

  //compress panel?
  if (type == "Axial" || type == "Radial" || type == "ComprAxial" || type == "ComprRadial" || type == "Composite")
    {
      ui->fCompression->setVisible(true);

      if (type == "ComprAxial" || type == "ComprRad")
        {
          ui->cb_compress_r->setChecked(true);
          double tmp;
          parser.ParseObject("r0", tmp);
          ui->led_compression_r0->setText(QString::number(tmp));
          parser.ParseObject("comp_lam", tmp);
          ui->led_compression_lam->setText(QString::number(tmp));
          parser.ParseObject("comp_k", tmp);
          ui->led_compression_k->setText(QString::number(tmp));
        }
      else if (type == "Composite")
        {
          //expecting first LRF in th deck to be Axial or cAxial
          QVector<QJsonObject> LRFdeck;
          QJsonArray deck;
           ui->cb_compress_r->setChecked(false);
          if (parser.ParseObject("deck", deck))
           if (parser.ParseArray(deck, LRFdeck))
            if (LRFdeck.size()>0)
              {
                QJsonObject axLRF = LRFdeck[0];
                QString type = axLRF["type"].toString();
                /////if (1) // (LRFdeck[0]["type"] == "ComprAxial" || LRFdeck[0]["type"] == "ComprRad")
                if (type == "ComprAxial" || type == "ComprRad")
                {
  //                qDebug() << "-----" << LRFdeck[0]["type"];
                  JsonParser parser1(LRFdeck[0]);
                  ui->cb_compress_r->setChecked(true);
                  double tmp;
                  parser1.ParseObject("r0", tmp);
                  ui->led_compression_r0->setText(QString::number(tmp));
                  parser1.ParseObject("comp_lam", tmp);
                  ui->led_compression_lam->setText(QString::number(tmp));
                  parser1.ParseObject("comp_k", tmp);
                  ui->led_compression_k->setText(QString::number(tmp));
                }
              }

        }
      else ui->cb_compress_r->setChecked(false);
    }
  else ui->fCompression->setVisible(false);

  //number of nodes
  int val;
  if (parser.ParseObject("n1", val)) ui->sbSplNodesX->setValue(val);
  if (parser.ParseObject("n2", val)) ui->sbSplNodesY->setValue(val);

  tmpIgnore = false;
  LRFwindow::on_pbUpdateGUI_clicked(); //additional interface update
}

void LRFwindow::on_pushButton_clicked()
{
    double wx = ui->ledWx->text().toDouble();
    double wy = ui->ledWy->text().toDouble();

    double rw[3] = {wx, wy, 0};
    double rl[3];
    MW->SensLRF->getSensor(ui->spinBox->value())->transform(rw, rl);

    ui->ledLx->setText(QString::number(rl[0]));
    ui->ledLy->setText(QString::number(rl[1]));
}

void LRFwindow::on_pushButton_2_clicked()
{
  double lx = ui->ledLx->text().toDouble();
  double ly = ui->ledLy->text().toDouble();

  double rl[3] = {lx, ly, 0};
  double rw[3];
  MW->SensLRF->getSensor(ui->spinBox->value())->transformBack(rl, rw);

  ui->ledWx->setText(QString::number(rw[0]));
  ui->ledWy->setText(QString::number(rw[1]));
}

void LRFwindow::contextMenuHistory(const QPoint &pos)
{
  QMenu historyMenu;
  int row;

  QListWidgetItem* temp = ui->lwLRFhistory->itemAt(pos);
  if(temp != NULL) row = ui->lwLRFhistory->row(temp);
  else return;

  historyMenu.addSeparator();
  QAction* cur = historyMenu.addAction("Select as current");
  historyMenu.addSeparator();
  QAction* sec = historyMenu.addAction("Select as second");
  historyMenu.addSeparator();
  QAction* del =historyMenu.addAction("Delete");
  historyMenu.addSeparator();

  QAction* selectedItem = historyMenu.exec(ui->lwLRFhistory->mapToGlobal(pos));

  int numIters = MW->SensLRF->countIterations();
  int iter = numIters-1 - row;

  if (selectedItem == cur)
    {
      //Setting as current
      MW->SensLRF->setCurrentIter(iter);
      if (secondIter == iter) secondIter = -1;
    }
  else if (selectedItem == sec)
    {
      //Setting as second
      if (iter == secondIter) secondIter = -1;
      else secondIter = iter;
    }
  else if (selectedItem == del)
    {
      //Delete iter
      if (iter == MW->SensLRF->getCurrentIterIndex())
        {
          //deleting actual current
          QMessageBox::StandardButton reply;
          reply = QMessageBox::question(this, "Confirmation needed", "Delete currently selected LRFs?",
                                          QMessageBox::Yes|QMessageBox::Cancel);
          if (reply == QMessageBox::Cancel) return;
        }

      MW->SensLRF->deleteIteration(iter); //it takes care of current/not current and being last iter
      secondIter = -1;

      LRFwindow::updateIterationIndication();
    }

  LRFwindow::on_pbUpdateHistory_clicked();
}

void LRFwindow::on_pbClearHistory_clicked()
{
  QMessageBox::StandardButton reply;
  reply = QMessageBox::question(this, "Confirmation needed", "Delete all history?",
                                  QMessageBox::Yes|QMessageBox::Cancel);
  if (reply == QMessageBox::Cancel) return;
  MW->SensLRF->clearHistory();
  LRFwindow::on_pbUpdateHistory_clicked();
}

void LRFwindow::on_pbClearTmpHistory_clicked()
{
  QMessageBox::StandardButton reply;
  reply = QMessageBox::question(this, "Confirmation needed", "Delete all iterations with --#-- name pattern?",
                                  QMessageBox::Yes|QMessageBox::Cancel);
  if (reply == QMessageBox::Cancel) return;

  MW->SensLRF->clearHistory(true);
  LRFwindow::on_pbUpdateHistory_clicked();
}

void LRFwindow::on_pbShowLRFMouse_clicked()
{ 
  if (gv) delete gv;

  if (LRFviewObj) delete LRFviewObj;
  if (LRFdialog) delete LRFdialog;

  LRFdialog = new QDialog(this);
  LRFdialog->setWindowTitle("LRF viewer");
  LRFdialog->resize(800,800);
  LRFdialog->show();

  QPushButton *okButton = new QPushButton("Close");
  connect(okButton,SIGNAL(clicked()),LRFdialog,SLOT(accept()));
  QVBoxLayout *mainLayout = new QVBoxLayout;
  gv = new myQGraphicsView(this);
  mainLayout->addWidget(gv);
  mainLayout->addWidget(okButton);
  LRFdialog->setLayout(mainLayout);

  LRFviewObj = new Viewer2DarrayObject(gv, MW->PMs);
  LRFviewObj->SetCursorMode(1);
  gv->show();
  okButton->show();

  LRFviewObj->DrawAll();
  LRFviewObj->ResetViewport();
  connect(gv, SIGNAL(MouseMovedSignal(QPointF*)), this, SLOT(paintLRFonDialog(QPointF*)));

  QPointF tmpp(0,0);
  LRFwindow::paintLRFonDialog(&tmpp);

  LRFdialog->exec();

  if (gv) delete gv;
  gv = 0;
  if (LRFviewObj) delete LRFviewObj;
  LRFviewObj = 0;
  if (LRFdialog) delete LRFdialog;
  LRFdialog = 0;
}

void LRFwindow::paintLRFonDialog(QPointF* pos)
{
  if (MW->SensLRF->getIteration() == 0) return;

  double r[3];
  r[0] = pos->x();
  r[1] = -pos->y(); //inverted!
  r[2] = MW->Rwindow->getSuggestedZ();
  //qDebug() << r[0] << r[1] << r[2];
  QString title = QString::number(r[0], 'f', 1) +" , "+ QString::number(r[1], 'f', 1) +" , "+ QString::number(r[2], 'f', 1);
  LRFdialog->setWindowTitle(title);

  int numPMs = MW->PMs->count();
  QVector<double> lrfs;
  lrfs.resize(numPMs);

  double max = -100;
  for (int ipm=0; ipm<numPMs; ipm++)
    {
      lrfs[ipm] = MW->SensLRF->getLRF(ipm, r);
      if (lrfs[ipm]>max) max = lrfs[ipm];
    }

  for (int ipm=0; ipm<numPMs; ipm++)
    {
      int g = 255 - lrfs[ipm]/max*254.0;
      LRFviewObj->SetText(ipm, QString::number(lrfs[ipm], 'f', 1));

      if (lrfs[ipm] <= 0)
        {
          LRFviewObj->SetBrushColor(ipm, Qt::red);
          LRFviewObj->SetTextColor(ipm, Qt::black);
        }
      else
        {
          LRFviewObj->SetBrushColor(ipm, QColor(g,g,g));
          if (g>128) LRFviewObj->SetTextColor(ipm, Qt::black);
          else LRFviewObj->SetTextColor(ipm, Qt::white);
        }
    }
  LRFviewObj->DrawAll();
}

void LRFwindow::on_pbExpand_clicked()
{
   ui->swAdvancedFeatures->setCurrentIndex(1);
   this->setFixedSize(690, this->height());
}

void LRFwindow::on_pbShrink_clicked()
{
   ui->swAdvancedFeatures->setCurrentIndex(0);
   this->setFixedSize(554, this->height());
}

void LRFwindow::on_pbRadialToText_clicked()
{
    //watchdogs
    if (!MW->SensLRF->isAllLRFsDefined())
      {
        message("There is no LRF data!", this);
        return;
      }
    double energy = 1.0;
    if (ui->cbScaleForEnergy->isChecked())
    {
        //if (EventsDataHub->ReconstructionData.isEmpty() || !MW->Reconstructor->isReconstructionDataReady() )
        if (!EventsDataHub->isReconstructionDataReady() )
          {
            message("Reconstruction was not yet performed, energy scaling is not possible!", this);
            return;
          }

        int goodEv = 0;
        energy = 0;
        for (int iev=0; iev<EventsDataHub->ReconstructionData.size(); iev++)
        {
            if (EventsDataHub->ReconstructionData[iev]->GoodEvent)
            {
                goodEv++;
                energy += EventsDataHub->ReconstructionData[iev]->Points[0].energy;
            }
        }
        if (goodEv == 0)
        {
            message("There are no events passing all the filters!", this);
            return;
        }
        energy /= goodEv;
        //qDebug() << "average energy: "<< energy;
    }

    QString str = ui->sbPMnoButons->text();
    QString fileName = QFileDialog::getSaveFileName(this, "Save LRF # " +str+ " vs radius", MW->GlobSet->LastOpenDir,
                                                   "Text files (*.txt);;Data files (*.dat);;All files (*.*)");
    if (fileName.isEmpty()) return;
    MW->GlobSet->LastOpenDir = QFileInfo(fileName).absolutePath();

    QFile outputFile(fileName);
    outputFile.open(QIODevice::WriteOnly);
    if(!outputFile.isOpen())
      {
        message("Unable to open file " +fileName+ " for writing!", this);
        return;
      }
    QTextStream outStream(&outputFile);

    double step = ui->ledStep->text().toDouble();
    double p[2];  //p[0] is pm number, p[1] is z
    ///p[1] = MW->Reconstructor->getSuggestedZ();
    p[1] = MW->Rwindow->getSuggestedZ();
    double r[1];
    int ipm = ui->sbPMnoButons->value();

    PMsensor* sens = MW->SensLRF->getSensor(ipm);
    const LRF2 *lrf = sens->GetLRF();
    double rmax = lrf->getRmax();
    double gain = 1.0;
    if (ui->cbScaleForGain->isChecked()) gain = sens->GetGain();

    p[0] = ipm;

    for (r[0] = 0.0; r[0]<=rmax; r[0] += step)
    {
        double val = MW->SensLRF->getRadial(r, p)*gain*energy;
        outStream << r[0] << " " << val << "\n";
    }

    outputFile.close();
}

void LRFwindow::on_pbShowSensorGroups_clicked()
{
  QVector<QString> tmp(MW->PMs->count());

  for (int igrp=0; igrp<MW->SensLRF->countGroups(); igrp++)
    {
      const std::vector<PMsensor> *sensors = MW->SensLRF->getSensorGroup(igrp)->getPMs();
      for(unsigned int j = 0; j < sensors->size(); j++)
        tmp[(*sensors)[j].GetIndex()].append(QString::number(igrp,'g', 4));
    }

  //MW->DotsTGeo.resize(0);
  MW->clearGeoMarkers();
  MW->GeometryWindow->on_pbTop_clicked();
  MW->ShowGeometry();
  MW->ShowTextOnPMs(tmp, kBlue);
}

void LRFwindow::on_pbShowSensorGains_clicked()
{
    QVector<QString> tmp;

    for (int ipm=0; ipm<MW->PMs->count(); ipm++)
    {
        double gain = MW->SensLRF->getSensor(ipm)->GetGain();
        tmp.append(QString::number(gain, 'g', 4));
    }

    //MW->DotsTGeo.resize(0);
    MW->clearGeoMarkers();
    MW->GeometryWindow->on_pbTop_clicked();
    MW->ShowGeometry();
    MW->ShowTextOnPMs(tmp, kBlue);
}

void LRFwindow::on_pbCompareRadials_clicked()
{
  int bins = ui->sbCompareBins->value();
  double R0 = ui->ledCompareFrom->text().toDouble();
  double R1 = ui->ledCompareTo  ->text().toDouble();

  if (R1<0 || R0<0 || R0>R1)
  {
      message("Input error", this);
      return;
  }

  if (!MW->SensLRF->isAllLRFsDefined())
    {
      message("LRFs are not defined!", this);
      return;
    }

  if (secondIter == -1)
    {
      message("Second iteration is not selected", this);
      return;
    }

    if (secondIter >= MW->SensLRF->countIterations())
    {
       message("Error in second iteration number", this);
       return;
    }

  double p[3];
  p[0] = ui->sbPMnoButons->value(); //ipm
  p[1] = ui->ledZcenter->text().toDouble(); //z0
  p[2] = secondIter;

  int step = (R1-R0)/bins;
  double MeanDelta = 0, RMSdelta = 0, Mean1 = 0, Mean2 = 0;
  for (int i=0; i<=bins; i++)
  {
      double r = R0 + step*i;

      double LRF1 = MW->SensLRF->getRadial(&r, p);
      double LRF2 = MW->SensLRF->getRadialPlus(&r, p);

//      qDebug() << i << "LRF1, LRF2=" << LRF1 << LRF2;
      Mean1 += LRF1;
      Mean2 += LRF2;
      MeanDelta += (LRF1-LRF2);
      RMSdelta += (LRF1-LRF2)*(LRF1-LRF2);
  }

  qDebug() << "\n----------\n";
  qDebug() << "Comparing LRFs for"<< ui->sbPMnoButons->value() << "PM.";
  qDebug() << "range:" << R0 << R1;
  qDebug() << "Mean delta: "<< MeanDelta;
  qDebug() << "RMS of delta: "<<RMSdelta;
  double factor = Mean2/Mean1;
  qDebug() << "Scaling: "<< factor;

  MW->Owindow->show();
  MW->Owindow->raise();
  MW->Owindow->activateWindow();
  MW->Owindow->OutText("\n----------\n");
  MW->Owindow->OutText("Mean delta: "+QString::number(MeanDelta));
  MW->Owindow->OutText("RMS of delta: "+QString::number(RMSdelta));
  MW->Owindow->OutText("Scaling: "+QString::number(factor));
}

void LRFwindow::on_led_compression_k_editingFinished()
{
  double val = ui->led_compression_k->text().toDouble();
  if (val<=1)
    {
      message("Compression factor should be greater than 1.0. Setting to 3.0", this);
      ui->led_compression_k->setText("3.0");
    }
}

void LRFwindow::on_cbUseGroupping_toggled(bool checked)
{
  QString txt;
  if (checked) txt = "Only for group#:";
  else         txt = "Only for PM#:";
  ui->cbMakePMGroup->setText(txt);
  ui->fSingleGroup->setEnabled(checked);
}
