#include "materialinspectorwindow.h"
#include "ui_materialinspectorwindow.h"

#include "mainwindow.h"
#include "graphwindowclass.h"
#include "geometrywindowclass.h"
#include "classes.h"
#include "windownavigatorclass.h"
#include "primaryparticletracker.h"
#include "outputwindow.h"
#include "detectorclass.h"
#include "globalsettingsclass.h"

#include <QtGui>
#include <QLayout>
#include <QFileDialog>
#include <QMessageBox>
#include <QTextEdit>

//Root
#include "TGraph.h"

MaterialInspectorWindow::MaterialInspectorWindow(QWidget* parent, MainWindow *mw, DetectorClass* detector) :
    QMainWindow(parent),
    ui(new Ui::MaterialInspectorWindow)
{
    MW = mw;
    Detector = detector;
    ui->setupUi(this);
    this->move(15,15);
    this->setFixedSize(this->size());

    Qt::WindowFlags windowFlags = (Qt::Window | Qt::CustomizeWindowHint);
    windowFlags |= Qt::WindowCloseButtonHint;
    this->setWindowFlags( windowFlags );

    ui->swNeutral2->setGeometry(237, 0, 121, 65); //in editor it is displaced for convenience of access
    ui->labWasModified->setVisible(false);
    ui->pbWasModified->setVisible(false);
    ui->pbUpdateIndication->setVisible(false);

    ui->pbUpdateTmpMaterial->setVisible(false);
    ui->cobStoppingPowerUnits->setCurrentIndex(1);
    ui->cobTotalInteractionLoadEnergyUnits->setCurrentIndex(2);
    graphOver = 0;

    QDoubleValidator* dv = new QDoubleValidator(this);
    dv->setNotation(QDoubleValidator::ScientificNotation);
    QList<QLineEdit*> list = this->findChildren<QLineEdit *>();
    foreach(QLineEdit *w, list) if (w->objectName().startsWith("led")) w->setValidator(dv);

    DoNotDefocus = false;
    flagDisreguardChange = false;
    LastSelectedParticle = 0;
    DoNotUpdateOnCOBchange = false;
}

MaterialInspectorWindow::~MaterialInspectorWindow()
{
    delete ui;
}

void MaterialInspectorWindow::on_pbAddToLibrary_clicked()
{
  //checkig this material
  int ierror = MW->MaterialCollection->CheckTmpMaterial();
  if (ierror != 0)
    {
      QString ErrStr = MW->MaterialCollection->getErrorString(ierror);
      message(ErrStr, this);
      return;
    }

  QString starter = (MW->GlobSet->LibMaterials.isEmpty()) ? MW->GlobSet->LastOpenDir : MW->GlobSet->LibMaterials;
  int imat = ui->cobActiveMaterials->currentIndex();
  if (imat == -1) starter += "/Material_";
  else starter += "/Material_"+(*Detector->MaterialCollection)[imat]->name;
  QString fileName = QFileDialog::getSaveFileName(this,"Save material", starter, "json files (*.json);;All files (*.*)");
  if (fileName.isEmpty()) return;
  QFileInfo fileInfo(fileName);
  if (fileInfo.suffix().isEmpty()) fileName += ".json";

  QJsonObject json, js;
  Detector->MaterialCollection->writeMaterialToJson(imat, json);
  js["Material"] = json;
  SaveJsonToFile(js, fileName);
}

void MaterialInspectorWindow::on_pbLoadFromLibrary_clicked()
{
  QString starter = (MW->GlobSet->LibMaterials.isEmpty()) ? MW->GlobSet->LastOpenDir : MW->GlobSet->LibMaterials;
  DoNotDefocus = true;
  QString fileName = QFileDialog::getOpenFileName(this, "Load material", starter, "json files (*.json);;All files (*.*)");
  DoNotDefocus = false;
  if (fileName.isEmpty()) return;

  QJsonObject json, js;
  LoadJsonFromFile(json, fileName);
  if (!json.contains("Material"))
    {
      message("File format error: Json with material settings not found");
      return;
    }
  js = json["Material"].toObject();
  Detector->MaterialCollection->tmpMaterial.readFromJson(js, Detector->MaterialCollection, &Detector->ParticleCollection);

  MaterialInspectorWindow::on_pbWasModified_clicked();
  MaterialInspectorWindow::UpdateWaveButtons();
  MW->ListActiveParticles();

  ui->cobActiveMaterials->setCurrentIndex(-1); //to avoid confusion (and update is disabled for -1)
  MaterialInspectorWindow::VisualizeTmpMaterial(); //refresh indication of tmpMaterial
  MaterialInspectorWindow::UpdateWaveButtons(); //refresh button state for Wave-resolved properties
  SetParticleSelection(0);
}

void MaterialInspectorWindow::UpdateActiveMaterials()
{   
    ui->cobActiveMaterials->clear();    
    for (int i=0;i<MW->MaterialCollection->size();i++) ui->cobActiveMaterials->addItem( (*MW->MaterialCollection)[i]->name);
}

void MaterialInspectorWindow::on_pbAddToActive_clicked()
{    
    //checkig this material
    int ierror = MW->MaterialCollection->CheckTmpMaterial();
    if (ierror != 0)
      {
        QString ErrStr = MW->MaterialCollection->getErrorString(ierror);
        message(ErrStr, this);
        return;
      }

    QString name = MW->MaterialCollection->tmpMaterial.name;
//    qDebug() << "Handling material:"<<name;
    int index = MW->MaterialCollection->FindMaterial(name);
    if (index > -1)
      {
//        qDebug()<<"This name already in use! index = "<<index;
        DoNotDefocus = true;
        switch( QMessageBox::information( this, "", "Update properties for material "+name+"?", "Overwrite", "Cancel", 0, 1 ) )
          {
           case 0:  break;  //overwrite
           default: return; //cancel
          }        
      }
    else
      {
        //new material
        index = MW->MaterialCollection->size(); //then it will be appended, so index is = current size
        MW->AddMaterialToCOBs(MW->MaterialCollection->tmpMaterial.name);
      }

    MW->MaterialCollection->CopyTmpToMaterialCollection(); //if absent, new material is created!
    MW->UpdateMaterialListEdit();

    ui->cobActiveMaterials->setCurrentIndex(index);
    ui->labWasModified->setVisible(false);

    MW->ReconstructDetector(true);
}

void MaterialInspectorWindow::on_pbShowTotalInteraction_clicked()
{
    MaterialInspectorWindow::ShowTotalInteraction();
}

void MaterialInspectorWindow::on_pbShowStoppingPower_clicked()
{
    MaterialInspectorWindow::ShowTotalInteraction();
}

void MaterialInspectorWindow::ShowTotalInteraction()
{
    int particleId = ui->cobParticle->currentIndex();
    int elements = MW->MaterialCollection->tmpMaterial.MatParticle[particleId].InteractionDataX.size();
    if (elements<1) return;

    ParticleType& type = Detector->ParticleCollection[particleId]->type;

    if (type == _charged_)
      {
        MW->GraphWindow->ShowGraph(&MW->MaterialCollection->tmpMaterial.MatParticle[particleId].InteractionDataX,
                                   &MW->MaterialCollection->tmpMaterial.MatParticle[particleId].InteractionDataF, kRed,
                                   "Energy, keV", "Stopping power, keV cm2/g");
       }
    else if (type == _gamma_)
      {
        MW->GraphWindow->ShowGraph(&MW->MaterialCollection->tmpMaterial.MatParticle[particleId].InteractionDataX,
                                   &MW->MaterialCollection->tmpMaterial.MatParticle[particleId].InteractionDataF, kRed,
                                   "Energy, keV", "Total interaction coefficient, cm2/g");
      }
    else //this is neutron
    {
        QVector<double> X(MW->MaterialCollection->tmpMaterial.MatParticle[particleId].InteractionDataX);
        QVector<double> Y(MW->MaterialCollection->tmpMaterial.MatParticle[particleId].InteractionDataF);
        for (int i=0; i<X.size(); i++)
          {
            X[i] *= 1.0e6; // keV -> meV
            Y[i] *= 1.0e24; // back to barns
          }

        MW->GraphWindow->ShowGraph(&X, &Y, kRed,
                                   "Energy, meV", "Capture cross section, barns");
    }
}

void MaterialInspectorWindow::on_pbShowThisScenarioCrossSection_clicked()
{
    int particleId = ui->cobParticle->currentIndex();
    int TermScenario = ui->cobTerminationScenarios->currentIndex();
    int elements = MW->MaterialCollection->tmpMaterial.MatParticle[particleId].Terminators[TermScenario].PartialCrossSection.size();
    if (elements<1) return;

    MW->GraphWindow->ShowGraph(&MW->MaterialCollection->tmpMaterial.MatParticle[particleId].Terminators[TermScenario].PartialCrossSectionEnergy,
                           &MW->MaterialCollection->tmpMaterial.MatParticle[particleId].Terminators[TermScenario].PartialCrossSection,
                           kRed, "Energy, keV", "Partial interaction coefficient, cm2/g", 20, 1, 1, 0);
}

void MaterialInspectorWindow::on_pbShowThisScenarioCrossSectionInterpolated_clicked()
{
   MaterialInspectorWindow::on_pbShowThisScenarioCrossSection_clicked();

   MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

   int particleId = ui->cobParticle->currentIndex();
   int TermScenario = ui->cobTerminationScenarios->currentIndex();
   int entries = tmpMaterial.MatParticle[particleId].Terminators[TermScenario].PartialCrossSection.size();

   QVector<double> xx((entries-1)*50 +1);
   QVector<double> yy((entries-1)*50 +1);

   int LogLogInterpolation = Detector->MaterialCollection->fLogLogInterpolation;
   int counter = 0;
   for (int i=1; i<entries; i++)
     for (int j=0; j<50; j++)
         {
           double previousOne = tmpMaterial.MatParticle[particleId].Terminators[TermScenario].PartialCrossSectionEnergy[i-1];
           double thisOne = tmpMaterial.MatParticle[particleId].Terminators[TermScenario].PartialCrossSectionEnergy[i];
           double XX = previousOne + 0.02* j * (thisOne-previousOne);
           xx[counter] = XX;
           double YY;
           if (XX<tmpMaterial.MatParticle[particleId].Terminators[TermScenario].PartialCrossSectionEnergy.last()) YY = InteractionValue(XX, &tmpMaterial.MatParticle[particleId].Terminators[TermScenario].PartialCrossSectionEnergy, &tmpMaterial.MatParticle[particleId].Terminators[TermScenario].PartialCrossSection, LogLogInterpolation);
           else YY = tmpMaterial.MatParticle[particleId].Terminators[TermScenario].PartialCrossSection.last();
           yy[counter] = YY;
           counter++;
         }
   xx[counter] = tmpMaterial.MatParticle[particleId].Terminators[TermScenario].PartialCrossSectionEnergy.last();
   yy[counter] = tmpMaterial.MatParticle[particleId].Terminators[TermScenario].PartialCrossSection.last();

   if (graphOver) delete graphOver;
   graphOver = new TGraph(counter+1, xx.data(), yy.data());
   graphOver->SetLineColor(kBlue);
   graphOver->SetLineWidth(1);
   graphOver->SetTitle("Interpolated");
   MW->GraphWindow->Draw(graphOver, "SAME L");
   MW->GraphWindow->UpdateRootCanvas();
}

void MaterialInspectorWindow::SetMaterialSelection(int i)
{
  if (ui->cobActiveMaterials->currentIndex() != i) ui->cobActiveMaterials->setCurrentIndex(i);
  else on_cobActiveMaterials_currentIndexChanged(i);
}

void MaterialInspectorWindow::setLogLog(bool flag)
{
  ui->cbLogLogInterpolation->setChecked(flag);
}

void MaterialInspectorWindow::on_cobActiveMaterials_currentIndexChanged(int index)
{
    if (index < 0)
      {
        ui->pbShowUsage->setEnabled(false);
        ui->pbRename->setText("");
        return;
      }
    else ui->pbShowUsage->setEnabled(true);
    if (DoNotUpdateOnCOBchange) return;

    ui->leoGammaDiagnosticsCoefficient->setText("n.a.");
    if ( MW->MaterialCollection->size() == 0) return;
    if ( (*MW->MaterialCollection)[index]->MatParticle.size() == 0)
      {
        qWarning()<<"Error: MatParticle is empty for this material!";
        return;
      }
    MW->MaterialCollection->CopyMaterialToTmp(index);
    MaterialInspectorWindow::VisualizeTmpMaterial();

    //resetting particle-related section
    int indexToSet = (LastSelectedParticle < Detector->ParticleCollection.size()) ? LastSelectedParticle : 0;
    if (ui->cobParticle->currentIndex() != indexToSet) ui->cobParticle->setCurrentIndex(indexToSet); //update on change
    else on_cobParticle_currentIndexChanged(indexToSet);

    ui->labWasModified->setVisible(false);
    ui->pbRename->setText("Rename\n"+ui->cobActiveMaterials->currentText());
}

void MaterialInspectorWindow::UpdateWaveButtons()
{   
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

  if (tmpMaterial.PrimarySpectrum_lambda.size()>0)
    {
      ui->pbShowPrimSpectrum->setEnabled(true);
      ui->pbDeletePrimSpectrum->setEnabled(true);
    }
  else
    {
      ui->pbShowPrimSpectrum->setEnabled(false);
      ui->pbDeletePrimSpectrum->setEnabled(false);
    }

  if (tmpMaterial.SecondarySpectrum_lambda.size()>0)
    {
      ui->pbShowSecSpectrum->setEnabled(true);
      ui->pbDeleteSecSpectrum->setEnabled(true);
    }
  else
    {
      ui->pbShowSecSpectrum->setEnabled(false);
      ui->pbDeleteSecSpectrum->setEnabled(false);
    }

  if (tmpMaterial.nWave_lambda.size()>0)
    {
      ui->pbShowNlambda->setEnabled(true);
      ui->pbDeleteNlambda->setEnabled(true);
    }
  else
    {
      ui->pbShowNlambda->setEnabled(false);
      ui->pbDeleteNlambda->setEnabled(false);
    }

  if (tmpMaterial.absWave_lambda.size()>0)
    {
      ui->pbShowABSlambda->setEnabled(true);
      ui->pbDeleteABSlambda->setEnabled(true);
    }
  else
    {
      ui->pbShowABSlambda->setEnabled(false);
      ui->pbDeleteABSlambda->setEnabled(false);
    }
}

void MaterialInspectorWindow::VisualizeTmpMaterial()
{
    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

    QString str;
    ui->leName->setText(tmpMaterial.name);  

    str.setNum(tmpMaterial.density, 'g');
    ui->ledDensity->setText(str);
    str.setNum(tmpMaterial.atomicDensity, 'g');
    if (tmpMaterial.atomicDensity > 0) ui->ledAtomicDensity->setText(str);
    else ui->ledAtomicDensity->setText("");

    str.setNum(tmpMaterial.n, 'g');
    ui->ledN->setText(str);
    str.setNum(tmpMaterial.abs, 'g');
    ui->ledAbs->setText(str);

    if (tmpMaterial.rayleighMFP > 0)  ui->ledRayleigh->setText(QString::number(tmpMaterial.rayleighMFP));
    else ui->ledRayleigh->setText("");
    ui->ledRayleighWave->setText(QString::number(tmpMaterial.rayleighWave));

    str.setNum(tmpMaterial.PriScintDecayTime, 'g');
    ui->ledPriT->setText(str);
    str.setNum(tmpMaterial.W*1000.0, 'g'); //keV->eV
    ui->ledW->setText(str);
    str.setNum(tmpMaterial.SecYield, 'g');
    ui->ledSecYield->setText(str);
    str.setNum(tmpMaterial.SecScintDecayTime, 'g');
    ui->ledSecT->setText(str);    
    str.setNum(tmpMaterial.e_driftVelocity, 'g');
    ui->ledEDriftVelocity->setText(str);
    str.setNum(tmpMaterial.p1, 'g');
    ui->ledP1->setText(str);
    str.setNum(tmpMaterial.p2, 'g');
    ui->ledP2->setText(str);
    str.setNum(tmpMaterial.p3, 'g');
    ui->ledP3->setText(str);

    int tmp = LastSelectedParticle;
    ui->cobParticle->clear();
    ui->cobSecondaryParticleToAdd->clear();
    for (int i=0; i<Detector->ParticleCollection.size(); i++)
      {
        ui->cobParticle->addItem(Detector->ParticleCollection[i]->ParticleName);
        ui->cobSecondaryParticleToAdd->addItem(Detector->ParticleCollection[i]->ParticleName);
      }

    LastSelectedParticle = tmp;
    if (LastSelectedParticle < Detector->ParticleCollection.size()) ui->cobParticle->setCurrentIndex(LastSelectedParticle);
    else ui->cobParticle->setCurrentIndex(0); //will trigger MaterialInspectorWindow::on_cobParticle_currentIndexChanged(int index)

    MaterialInspectorWindow::UpdateWaveButtons();
}

void MaterialInspectorWindow::UpdateSelection()
{
   if (ui->cobActiveMaterials->currentIndex() == -1)
     {
       if (ui->cobActiveMaterials->count() > 0) ui->cobActiveMaterials->setCurrentIndex(0);
       return; //update on change
     }
    MaterialInspectorWindow::on_cobActiveMaterials_currentIndexChanged(ui->cobActiveMaterials->currentIndex());
}

void MaterialInspectorWindow::on_cobParticle_currentIndexChanged(int /*index*/)
{
  if (MW->MaterialCollection->tmpMaterial.MatParticle.size() == 0 ) return; //for init phase
  MaterialInspectorWindow::on_pbUpdateIndication_clicked();
}

void MaterialInspectorWindow::on_pbUpdateIndication_clicked()
{
    int index = ui->cobParticle->currentIndex();
    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

    flagDisreguardChange = true; //to skip auto-update "modified!" sign

    if (index <0 )
    {  //on undefined
        ui->cobTerminationScenarios->setCurrentIndex(-1);
        ui->pbShowTotalInteraction->setEnabled(false);
        flagDisreguardChange = false;
        return;
    }

    LastSelectedParticle = index;
    bool TrackingAllowed = tmpMaterial.MatParticle[index].TrackingAllowed;
    ui->cbTrackingAllowed->setChecked(TrackingAllowed);
    ui->fEnDepProps->setEnabled(TrackingAllowed);

    bool MaterialIsTransparent = tmpMaterial.MatParticle[index].MaterialIsTransparent;
    ui->cbTransparentMaterial->setChecked(MaterialIsTransparent);
    MaterialInspectorWindow::on_cbTransparentMaterial_toggled(MaterialIsTransparent);

    QString str;
    str.setNum(tmpMaterial.MatParticle[index].PhYield, 'g');
    ui->ledPrimaryYield->setText(str);

    ui->pbShowTotalInteraction->setEnabled(true);
    ui->ledPrimaryYield->setEnabled(true);

    if (Detector->ParticleCollection[index]->type != _charged_)
    {
        //neutral particles        
        ui->swMainMatParticle->setCurrentIndex(1);
        if (Detector->ParticleCollection[index]->type == _neutron_)
        {            
           //neutron            
            ui->ledPrimaryYield->setEnabled(false);
            ui->swNeutral->setCurrentIndex(1);
            ui->swNeutral2->setCurrentIndex(1);
            ui->pbShowTotalInteraction->setText("Show total cross section [barns]");          

            if (tmpMaterial.MatParticle[index].Terminators.size() == 0)
              {
                ui->ledBranching->setText("0");
                ui->fNeutron->setEnabled(false);
              }
            else ui->fNeutron->setEnabled(true);

            if (tmpMaterial.MatParticle[index].InteractionDataX.size() == 0)
              {
                ui->fNeutron->setEnabled(false);
                ui->pbShowTotalInteraction->setEnabled(false);
                ui->pbAddNewTerminationScenario->setEnabled(false);
              }
            else
              {
                if (tmpMaterial.MatParticle[index].Terminators.size() > 0) ui->fNeutron->setEnabled(true);
                ui->pbShowTotalInteraction->setEnabled(true);
                ui->pbAddNewTerminationScenario->setEnabled(true);
              }
            MaterialInspectorWindow::on_ledMFPenergy_editingFinished();
        }
        else  if (Detector->ParticleCollection[index]->type == _gamma_)
        {
            //gamma
            ui->swNeutral->setCurrentIndex(0);
            ui->swNeutral2->setCurrentIndex(0);

            ui->pbShowTotalInteraction->setText("Show total interaction coefficient");
            ui->pbAddNewTerminationScenario->setEnabled( tmpMaterial.MatParticle[index].Terminators.size() == 0 );
            ui->fGamma->setEnabled( tmpMaterial.MatParticle[index].Terminators.size() != 0 );

            if (tmpMaterial.MatParticle[index].Terminators.size() == 0)
              {
                //seems no need to disable GUI elements here, its automatic
              }
            else on_cobTerminationScenarios_currentIndexChanged(ui->cobTerminationScenarios->currentIndex());
        }
        else
          {
            message("Critical error: unknown neutral particle", this);
            return;
          }

        QString tmpStr;
        int size = tmpMaterial.MatParticle[index].Terminators.size();
        tmpStr.setNum(size);
        tmpStr = "(Total defined: " + tmpStr + " )";
        //ui->labTotal->setText(tmpStr);

        ui->cobTerminationScenarios->clear();
        flagDisreguardChange = true;
        for (int i=0; i<size; i++)
          {
            tmpStr.setNum(i);
            ui->cobTerminationScenarios->addItem(tmpStr);
            flagDisreguardChange = true;
          }
        //further indication is handled by  MaterialInspectorWindow::on_cobTerminationScenarios_currentIndexChanged(int index)
        if (tmpMaterial.MatParticle[index].InteractionDataX.size()>0) ui->pbShowTotalInteraction->setEnabled(true);
        else ui->pbShowTotalInteraction->setEnabled(false);
    }
    else
    {
       //charged particle
       ui->swMainMatParticle->setCurrentIndex(0);
       if (tmpMaterial.MatParticle[index].InteractionDataX.size()>0) ui->pbShowStoppingPower->setEnabled(true);
       else ui->pbShowStoppingPower->setEnabled(false);
    };

    MaterialInspectorWindow::on_ledMFPenergy_2_editingFinished();

    flagDisreguardChange = false;
}

void MaterialInspectorWindow::on_cobTerminationScenarios_currentIndexChanged(int index)
{
    if (index < 0)
    {
        ui->leThisScenarioType->clear();
        ui->lwGeneratedParticlesEnergies->clear();
        return;
    }

    flagDisreguardChange = true; //"modified!" sign control

    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

    int particleId = ui->cobParticle->currentIndex();        
    if (Detector->ParticleCollection[particleId]->type == _gamma_)
       {
         bool fPCSdefined = tmpMaterial.MatParticle[particleId].Terminators[index].PartialCrossSection.size() != 0;
         ui->pbShowThisScenarioCrossSection->setEnabled(fPCSdefined);
         ui->pbShowThisScenarioCrossSectionInterpolated->setEnabled(fPCSdefined);

         MaterialInspectorWindow::on_ledGammaDiagnosticsEnergy_editingFinished(); //diagnostics for gammas
       }

    int ReactionType = tmpMaterial.MatParticle[particleId].Terminators[index].ReactionType;
    QString text;
    ui->leThisScenarioType->clear();
    if (ReactionType == 0) text = "Photoelectric absorption";
    else if (ReactionType == 1) text = "Compton scattering";
    else if (ReactionType == 2)
    {
        text = "Neutron capture";
        QString str;
        str.setNum(100.0 * tmpMaterial.MatParticle[particleId].Terminators[index].branching);
        ui->ledBranching->setText(str);
        //secondary particle indication     
        ui->lwGeneratedParticlesEnergies->clear();
        int size = tmpMaterial.MatParticle[particleId].Terminators[index].GeneratedParticles.size();
        if (size < 1) return;

        QString tmpStr, tmpStr1;
        for (int i=0; i<size; i++)
        {
            int SecParticle = tmpMaterial.MatParticle[particleId].Terminators[index].GeneratedParticles[i];
            tmpStr = Detector->ParticleCollection[SecParticle]->ParticleName;
            double energy = tmpMaterial.MatParticle[particleId].Terminators[index].GeneratedParticleEnergies[i];

            if (energy > 0) {tmpStr1.setNum(energy); tmpStr += " / " + tmpStr1 + " keV";}           
            ui->lwGeneratedParticlesEnergies->addItem(tmpStr);
        }
    }
    ui->leThisScenarioType->setText(text);
    flagDisreguardChange = false;
}

void MaterialInspectorWindow::on_pbUpdateTmpMaterial_clicked()
{  
    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

    tmpMaterial.name = ui->leName->text();
    tmpMaterial.density = ui->ledDensity->text().toDouble();
    QString str = ui->ledAtomicDensity->text();
    if (!str.compare("")) tmpMaterial.atomicDensity = 0;
    else tmpMaterial.atomicDensity = str.toDouble();
    tmpMaterial.n = ui->ledN->text().toDouble();
    tmpMaterial.abs = ui->ledAbs->text().toDouble();
    tmpMaterial.PriScintDecayTime = ui->ledPriT->text().toDouble();
    tmpMaterial.W = ui->ledW->text().toDouble()*0.001; //eV -> keV
    tmpMaterial.SecYield = ui->ledSecYield->text().toDouble();
    tmpMaterial.SecScintDecayTime = ui->ledSecT->text().toDouble();   
    tmpMaterial.e_driftVelocity = ui->ledEDriftVelocity->text().toDouble();
    tmpMaterial.p1 = ui->ledP1->text().toDouble();
    tmpMaterial.p2 = ui->ledP2->text().toDouble();
    tmpMaterial.p3 = ui->ledP3->text().toDouble();
    MaterialInspectorWindow::on_ledMFPenergy_editingFinished(); //update mean free path
}

void MaterialInspectorWindow::on_pbLoadDeDr_clicked()
{
    QString fileName;
    fileName = QFileDialog::getOpenFileName(this, "Load dE/dr data", MW->GlobSet->LastOpenDir, "Data files (*.dat)");
    qDebug()<<fileName;    

    if (!fileName.isEmpty()) return;
    MW->GlobSet->LastOpenDir = QFileInfo(fileName).absolutePath();

    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

        QFile file(fileName);

        if(!file.open(QIODevice::ReadOnly | QFile::Text))
        {
            QMessageBox::information(0, "error", file.errorString());
            return;
        }

        QTextStream in(&file);

        int particleId = ui->cobParticle->currentIndex();
        tmpMaterial.MatParticle[particleId].InteractionDataX.resize(0);
        tmpMaterial.MatParticle[particleId].InteractionDataF.resize(0);

        double Multiplier;
        switch (ui->cobStoppingPowerUnits->currentIndex())
        {
        case (0): {Multiplier = 0.001; break;}
        case (1): {Multiplier = 1.0; break;}
        case (2): {Multiplier = 1000.0; break;}
        }

        //separators
        QRegExp rx("(\\ |\\,|\\:|\\t)"); //RegEx for ' ' or ',' or ':' or '\t'

        while(!in.atEnd())
        {
            QString line = in.readLine();
            QStringList fields = line.split(rx);

            //*** TO ADD error control
            double x = fields[0].toDouble();
//            qDebug()<<x;
            double f = fields[1].toDouble();
//            qDebug()<<f;

            tmpMaterial.MatParticle[particleId].InteractionDataX.append(x);
            tmpMaterial.MatParticle[particleId].InteractionDataF.append(f*Multiplier);
        }
        file.close();
        ui->pbShowStoppingPower->setEnabled(true);

        MaterialInspectorWindow::on_pbWasModified_clicked();
        MaterialInspectorWindow::on_cobParticle_currentIndexChanged(ui->cobParticle->currentIndex());//refresh indication   
}

void MaterialInspectorWindow::SetParticleSelection(int index)
{
  int old = ui->cobParticle->currentIndex();
  ui->cobParticle->setCurrentIndex(index);
  if (old == index) on_cobParticle_currentIndexChanged(index);
}

void MaterialInspectorWindow::SetMaterial(int index)
{
  int old = ui->cobActiveMaterials->currentIndex();
  ui->cobActiveMaterials->setCurrentIndex(index);
  if (old == index) on_cobActiveMaterials_currentIndexChanged(index);
}

void MaterialInspectorWindow::on_pbLoadThisScenarioCrossSection_clicked()
{
    QString fileName;
    fileName = QFileDialog::getOpenFileName(this, "Load partial interaction coefficient data", MW->GlobSet->LastOpenDir, "Data files (*.dat)");
    qDebug()<<fileName;

    if (fileName.isEmpty()) return;
    MW->GlobSet->LastOpenDir = QFileInfo(fileName).absolutePath();

    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

    int particleId = ui->cobParticle->currentIndex();
    int scenario = ui->cobTerminationScenarios->currentIndex();
    int error = LoadDoubleVectorsFromFile(fileName,
                                           &tmpMaterial.MatParticle[particleId].Terminators[scenario].PartialCrossSectionEnergy,
                                           &tmpMaterial.MatParticle[particleId].Terminators[scenario].PartialCrossSection);

    if (error>0)
      {
        message("error loading file!");
        tmpMaterial.MatParticle[particleId].Terminators[scenario].PartialCrossSectionEnergy.resize(0);
        tmpMaterial.MatParticle[particleId].Terminators[scenario].PartialCrossSection.resize(0);
        ui->pbShowThisScenarioCrossSection->setEnabled(false);
        ui->pbShowTotalInteraction->setEnabled(false);
      }
    else
      {
        ui->pbShowThisScenarioCrossSection->setEnabled(true);
        //checking for misatch
        int otherScenario;
        if (scenario == 0) otherScenario = 1;
        else otherScenario = 0;

        bool ok = true;
        if (!tmpMaterial.MatParticle[particleId].Terminators[otherScenario].PartialCrossSectionEnergy.isEmpty())
          {
            if (tmpMaterial.MatParticle[particleId].Terminators[scenario].PartialCrossSectionEnergy.size() !=  tmpMaterial.MatParticle[particleId].Terminators[otherScenario].PartialCrossSectionEnergy.size()) ok = false;

            if (ok)
              {
                //cycle all nodes
                for (int i=0; i<tmpMaterial.MatParticle[particleId].Terminators[scenario].PartialCrossSectionEnergy.size(); i++)
                  if (tmpMaterial.MatParticle[particleId].Terminators[scenario].PartialCrossSectionEnergy[i] != tmpMaterial.MatParticle[particleId].Terminators[otherScenario].PartialCrossSectionEnergy[i])
                    {
                      ok = false;
                      break;
                    }
              }
          }

        if (ok)
          {
            ui->pbShowTotalInteraction->setEnabled(true);
          }
        else
          {
            //mismatch detected
            message("Found mismatch in energy range between photoelectric and compton! Resetting the other process data!");
            tmpMaterial.MatParticle[particleId].Terminators[otherScenario].PartialCrossSectionEnergy.resize(0);
            tmpMaterial.MatParticle[particleId].Terminators[otherScenario].PartialCrossSection.resize(0);
            ui->pbShowTotalInteraction->setEnabled(false);
          }
      }

    //recalculating total cross-section
    tmpMaterial.MatParticle[particleId].CalculateTotal();

    MaterialInspectorWindow::on_pbWasModified_clicked();
    MaterialInspectorWindow::on_ledGammaDiagnosticsEnergy_editingFinished();
}

void MaterialInspectorWindow::on_pbAddNewTerminationScenario_clicked()
{
    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
    int particleId = ui->cobParticle->currentIndex();
    int scenarios = tmpMaterial.MatParticle[particleId].Terminators.size();
    ParticleType Type = Detector->ParticleCollection[particleId]->type;
    if (Type == _gamma_)
    {
        //gamma
        if (scenarios > 0) return; //already defined
        tmpMaterial.MatParticle[particleId].Terminators.resize(2);
        tmpMaterial.MatParticle[particleId].Terminators[0].ReactionType = 0; //photoeffect
        tmpMaterial.MatParticle[particleId].Terminators[1].ReactionType = 1; //compton
        tmpMaterial.MatParticle[particleId].Terminators[0].GeneratedParticles.resize(0); // *** TO ADD clean up
        tmpMaterial.MatParticle[particleId].Terminators[1].GeneratedParticles.resize(0);
        //adding gamma to secondary particles
        int Id = MW->MaterialCollection->FindCreateParticle("gamma", _gamma_, 0, 0, true);
        if (Id<0)
          {
            message("Error: gamma was not found!");
            Id = 0;
          }
        tmpMaterial.MatParticle[particleId].Terminators[1].GeneratedParticles.append(Id);
        tmpMaterial.MatParticle[particleId].Terminators[1].GeneratedParticleEnergies.resize(0);
        tmpMaterial.MatParticle[particleId].Terminators[1].GeneratedParticleEnergies.append(0);
        scenarios = 2;
        ui->fGamma->setEnabled(true);
        ui->pbAddNewTerminationScenario->setEnabled( false );
    }

    if (Type == _neutron_)
    {
        //neutron
        tmpMaterial.MatParticle[particleId].Terminators.resize(scenarios+1);
        tmpMaterial.MatParticle[particleId].Terminators[scenarios].ReactionType = 2;
        tmpMaterial.MatParticle[particleId].Terminators[scenarios].GeneratedParticles.resize(0);   // *** TO ADD clean up
        tmpMaterial.MatParticle[particleId].Terminators[scenarios].GeneratedParticleEnergies.resize(0);

        if (scenarios == 0)
        {  //adding the first
            tmpMaterial.MatParticle[particleId].Terminators[scenarios].branching = 1.0;
            ui->ledBranching->setText("100");
            ui->fNeutron->setEnabled(true);
        }
        else
        {
            tmpMaterial.MatParticle[particleId].Terminators[scenarios].branching = 0;
            ui->ledBranching->setText("0");
        }
        scenarios++;
    }

    QString tmpStr;
    ui->cobTerminationScenarios->clear();
    for (int i=0; i<scenarios; i++)
    {
        tmpStr.setNum(i);
        ui->cobTerminationScenarios->addItem(tmpStr);
    }

    tmpStr.setNum(scenarios);
    tmpStr = "(Total defined: " + tmpStr + " )";
    //ui->labTotal->setText(tmpStr);

    ui->cobTerminationScenarios->setCurrentIndex(scenarios-1);
    if (Type == _neutron_) MaterialInspectorWindow::on_ledBranching_editingFinished(); //to update cross-sections
}

void MaterialInspectorWindow::on_ledPrimaryYield_editingFinished()
{
     MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
     tmpMaterial.MatParticle[ui->cobParticle->currentIndex()].PhYield = ui->ledPrimaryYield->text().toDouble();
}

void MaterialInspectorWindow::on_ledBranching_editingFinished()
{
    double BValue = ui->ledBranching->text().toDouble();

    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

    int particleId = ui->cobParticle->currentIndex();
    int scenario = ui->cobTerminationScenarios->currentIndex();
    int Scenarios = tmpMaterial.MatParticle[particleId].Terminators.size();

    if (BValue<0 || BValue>100.0)
    {
        ui->ledBranching->setText("0");        
        message("Branching value should be >= 0 and <=100");
        MaterialInspectorWindow::on_ledBranching_editingFinished();
        return;
    }

    if (Scenarios==1 && BValue != 100.0)
    {
        ui->ledBranching->setText("100");
        MaterialInspectorWindow::on_ledBranching_editingFinished();
        return;
    }

    BValue = 0.01*BValue;
    tmpMaterial.MatParticle[particleId].Terminators[scenario].branching = BValue;

    //Rescaling total to 1 - scalling all other branchings
    if (Scenarios != 1) //could be the first just was defined
    {
       double ExclusiveSum =0;
       for (int i=0; i<Scenarios; i++) if (i != scenario) ExclusiveSum += tmpMaterial.MatParticle[particleId].Terminators[i].branching;
       double correction = (1.0 - BValue)/ExclusiveSum;
       for (int i=0; i<Scenarios; i++) if (i != scenario) tmpMaterial.MatParticle[particleId].Terminators[i].branching *= correction;
    }

    MW->MaterialCollection->RecalculateCrossSections(particleId);
}

void MaterialInspectorWindow::on_pbImportStoppingPowerFromTrim_clicked()
{
    QString fileName;
    fileName = QFileDialog::getOpenFileName(this, "Import stopping power data from a Trim file", MW->GlobSet->LastOpenDir, "Data files (*.dat);;Text files (*.txt);;All files (*)");
    //qDebug()<<fileName;
    if (fileName.isEmpty()) return;
    MW->GlobSet->LastOpenDir = QFileInfo(fileName).absolutePath();

    int particleId = ui->cobParticle->currentIndex();
    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

    tmpMaterial.MatParticle[particleId].InteractionDataX.resize(0);    //***ADD cleanUp
    tmpMaterial.MatParticle[particleId].InteractionDataF.resize(0);

    QFile file(fileName);

        if(!file.open(QIODevice::ReadOnly | QFile::Text))
        {
            QMessageBox::information(0, "error", file.errorString());
            return;
        }

        QTextStream in(&file);
        //separators
        QRegExp rx("(\\ |\\,|\\:|\\(|\\)|\\t)"); //RegEx for ' ' or ',' or ':' or '\t'
        QStringList fields;

        // looking for the line with e.g.     Target Density =  2.5000E+00 g/cm3 = 1.3623E+23 atoms/cm3
        bool flagFound = false;
        while(!in.atEnd())
        {
            QString line = in.readLine();
            fields = line.split(rx, QString::SkipEmptyParts);
            if (fields.isEmpty()) continue;
//            qDebug()<<fields[0];
            if (!fields[0].compare("Target")) {flagFound = true; break;}
        }

        if (flagFound) ;//qDebug()<<"found!";
        else
        {
            message("File format not recognized!");
            file.close();
            return;
        }

        //now have the target density
        qDebug()<<"Target density: "<<fields[3]<<fields[4]; //[3] - density  [4] - units
        double TrimDensity = fields[3].toDouble(); //*** assuming its always g/cm3

        message("Setting material density to one used in Trim!");
        tmpMaterial.density = TrimDensity;
        QString str;
        str.setNum(TrimDensity);
        ui->ledDensity->setText(str);

        // looking for the line with e.g.     Stopping Units =  keV / (mg/cm2)
        while(!in.atEnd())
        {
            QString line = in.readLine();
            fields = line.split(rx, QString::SkipEmptyParts);
            if (fields.isEmpty()) continue;
//            qDebug()<<fields[0];
            if (!fields[0].compare("Stopping")) {flagFound = true; break;}
        }

        if (flagFound) ;//qDebug()<<"found!";
        else
        {        
            message("File format not recognized!");
            file.close();
            return;
        }

        //now have the stopping power units
        qDebug()<<"Stopping power units: "<<fields[3]<<fields[5]; //[3] - keV  [5] - mg/cm2
        double StoppingMultiplier;
        if (!fields[3].compare("eV")) StoppingMultiplier = 0.001;
        else if (!fields[3].compare("keV")) StoppingMultiplier = 1.0;
        else if (!fields[3].compare("MeV")) StoppingMultiplier = 1000.0;
        else
        {        
            message("Stopping power units not recognized!");
            file.close();
            return;
        }

        //maybe we have already density-corrected?
        if (!fields[5].compare("mm") || !fields[5].compare("micron") || !fields[5].compare("Angstrom"))
        { //Linear stopping power (dE/dx)
            StoppingMultiplier = StoppingMultiplier / TrimDensity *10.0; // in the body : dE/dx = TabValue * Density  //10.0 to take care of cm->mm (cm3/g -> mm cm2/g)

            if (!fields[5].compare("mm"));
            else if (!fields[5].compare("micron")) StoppingMultiplier *= 1.0e3; //divide by!
            else if (!fields[5].compare("Angstrom")) StoppingMultiplier *= 1.0e7; //divide by!
        }
        else
        { //Mass stopping power (dE / (g/cm2))

            if (!fields[5].compare("g/cm2"));
            else if (!fields[5].compare("mg/cm2")) StoppingMultiplier *= 1.0e3; //divide by!
            else if (!fields[5].compare("ug/cm2")) StoppingMultiplier *= 1.0e6;
            else
            {             
                message("Stopping power units not recognized!");
                file.close();
                return;
            }
        }
        qDebug()<<"StoppingMultiplier="<<StoppingMultiplier;

        //skipping the rest of the header
        flagFound=false;
        while(!in.atEnd())
        {
            QString line = in.readLine();
            fields = line.split(rx, QString::SkipEmptyParts);
            if (fields.isEmpty()) continue;
//            qDebug()<<fields[0];
            if (!fields[0].compare("-----------")) {flagFound = true; break;}
        }
        if (flagFound) qDebug()<<"Found beginning of the data block!";
        else
        {         
            message("File format not recognized!");
            file.close();
            return;
        }

        //reading the data
        bool flagWrongTermination = true;
        while(!in.atEnd())
        {
            QString line = in.readLine();
//            qDebug()<<"line:"<<line;
            fields = line.split(rx, QString::SkipEmptyParts);

            if (fields.isEmpty()) break; //something is wrong

 //           qDebug()<<fields[0];
            if (!fields[0].compare("-----------------------------------------------------------")) {flagWrongTermination = false; break;} //ending OK!

            if (fields.size() < 4) break; //something is wrong

            double Multiplier;
            QString mstr = fields[1];
            if (!mstr.compare("eV")) Multiplier = 0.001;
            else if (!mstr.compare("keV")) Multiplier = 1.0;
            else if (!mstr.compare("MeV")) Multiplier = 1000.0;
            else
            {          
               message("Unknown energy unit!");
               file.close();
               return;
            }
           // qDebug()<<"multiplier: "<<Multiplier;

            double x = fields[0].toDouble()*Multiplier;
            double f = (fields[2].toDouble() + fields[3].toDouble()) * StoppingMultiplier;

//            qDebug()<<x<<f;

            //assigning the data
            tmpMaterial.MatParticle[particleId].InteractionDataX.append(x);
            tmpMaterial.MatParticle[particleId].InteractionDataF.append(f);

            MaterialInspectorWindow::on_pbWasModified_clicked();
        }

        if (flagWrongTermination)
        {        
            message("Unexpected termination!");
            return;
        }

        MW->Owindow->OutText("File "+fileName+ " with Trim data has been loaded.");
        file.close();
        ui->pbShowStoppingPower->setEnabled(true);
        MaterialInspectorWindow::on_cobParticle_currentIndexChanged(ui->cobParticle->currentIndex());//refresh indication
}

void MaterialInspectorWindow::on_pbImportXCOM_clicked()
{
    int particleId = ui->cobParticle->currentIndex();
    if (Detector->ParticleCollection[particleId]->type != _gamma_) return; //only for gamma
    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

    QString fileName;
    fileName = QFileDialog::getOpenFileName(this, "Import partial interaction coefficients from an XCOM file", MW->GlobSet->LastOpenDir, "Data files (*.dat);;Text files (*.txt);;All files (*)");
//    qDebug()<<fileName;
    if (fileName.isEmpty()) return;
    MW->GlobSet->LastOpenDir = QFileInfo(fileName).absolutePath();

    //tmp data vectors
    QVector<double> x, f1, f2;
    x.resize(0);
    f1.resize(0);
    f2.resize(0);

    bool flagError = false;

    QFile file(fileName);
    if(!file.open(QIODevice::ReadOnly | QFile::Text))
        {
            QMessageBox::information(0, "error", file.errorString());
            return;
        }

    QTextStream in(&file);
        //separators
    QRegExp rx("(\\ |\\,|\\:|\\(|\\)|\\t)"); //RegEx for ' ' or ',' or ':' or '\t'

        //reading the data
        //header blocks can be between the data block
    bool flagHeaderBlock = true;

    QStringList fields;
    while(!in.atEnd())
        {
            QString line = in.readLine();

            if (flagHeaderBlock)
            { //if header block
                // looking for the line containing "MeV"
                fields = line.split(rx, QString::SkipEmptyParts);
                if (fields.isEmpty()) continue;
//                qDebug()<<"Assuming its header, first string:"<<fields[0];
                if (fields[0] == "MeV")
                  {
                    flagHeaderBlock = false;
//                    qDebug()<<"-->Header finished!<--";
                    continue; //jump to data section
                }
              //end header block
            }
            else
            { //data block
//                qDebug()<<"->assuming it is data line:"<<line;
                if (line.isEmpty() || line == " ")
                  {
//                    qDebug()<<"Empty line detected, skip"; //could be line between data paragraphs - shells
                    continue;
                  }
                line.remove(0,7); //removing frst characters where shell info can be
                fields = line.split(rx, QString::SkipEmptyParts);
                if (fields.size()<8)
                  {
                    //expecting 8 doubles!
                    //can be another header
//                    qDebug()<<"Short str found - assuming another header starts here";
                    flagHeaderBlock = true;
                    continue;
                  }



                bool ok1, ok2, ok3;
                double tmpX = fields[0].toDouble(&ok1);
                double tmpF1 = fields[2].toDouble(&ok2);
                double tmpF2 = fields[3].toDouble(&ok3);
                if (!ok1 || !ok2 || !ok3)
                  {
                    //wrong format
                    flagError = true;
                    break;
                  }
                x.append(tmpX*1.0e3);  //MeV -> keV
                f1.append(tmpF1);
                f2.append(tmpF2);
//                qDebug()<<tmpX<<"MeV"<<tmpF1<<tmpF2;
            } //end data block
        }
    file.close();

    if (x.isEmpty()) flagError = true;
    if (flagError)
        {
            message("Error reading the file!");
            return;
        }

//     qDebug()<<"-->check: shell lines...";
     //checking for shell lines - leave the first value, all from 2nd to nth - averaging
     //result: two point, first having the value "before line", second having averaged over all lines
     double xPrevious = 0;
     int SameValueTimes = 0;
     QVector<double> xx, ff1, ff2;
     xx.resize(0);
     ff1.resize(0);
     ff2.resize(0);

     double Accu1=0, Accu2=0;
     for (int i=0; i<x.size(); i++)
     {
         if (x[i] == xPrevious)
         {
             Accu1 += f1[i];
             Accu2 += f2[i];
             SameValueTimes++;
             continue; //not copying to the output vectors, waiting for the change!
         }
         else
         {
             if (SameValueTimes == 0)
             {
                //there was no overlap, normal operation
             }
             else
             {
                //there was overlap!
//                 qDebug()<<"->same value appeared times:"<<SameValueTimes;
                 xx.append(xPrevious);
                 ff1.append(Accu1/ (double)SameValueTimes);
                 ff2.append(Accu2/ (double)SameValueTimes);
                 Accu1 = 0;
                 Accu2 = 0;
                 SameValueTimes = 0;
                 //now appending the new value
             }
            xPrevious = x[i];
         }
         //normal
         xx.append(x[i]);
         ff1.append(f1[i]);
         ff2.append(f2[i]);                 
     }

//     qDebug()<<"-->shell lines check completed";
//     for (int i=0; i<xx.size(); i++) qDebug()<<xx[i]<<f1[i]<<f2[i];

//     qDebug()<<"\n-->populating tmpMaterial";
     //assigning data to the tmpMaterial
     tmpMaterial.MatParticle[particleId].InteractionDataF.resize(0);
     tmpMaterial.MatParticle[particleId].InteractionDataX.resize(0);
     tmpMaterial.MatParticle[particleId].Terminators.resize(0);

     QVector<NeutralTerminatorStructure> Terminators;
     NeutralTerminatorStructure terminator;

     //preparing termination scenarios
     Terminators.resize(0);
             //photoelectric
     terminator.ReactionType = 0;
     terminator.GeneratedParticles.resize(0); //none
     terminator.GeneratedParticleEnergies.resize(0); //none

     terminator.PartialCrossSectionEnergy = xx;
     terminator.PartialCrossSection = ff2;
     Terminators.append(terminator);
             //compton
     terminator.ReactionType = 1;
     terminator.GeneratedParticles.resize(0);
     terminator.GeneratedParticles.append(particleId); //adding gamma //OBSOLETE - generation code does not use this index
     terminator.GeneratedParticleEnergies.resize(0);
     terminator.GeneratedParticleEnergies.append(0);

     terminator.PartialCrossSectionEnergy.clear();
     terminator.PartialCrossSectionEnergy = xx;
    // terminator.PartialCrossSection.clear();
     terminator.PartialCrossSection = ff1;
     Terminators.append(terminator);

//     for (int i=0; i<xx.size(); i++) qDebug()<<xx[i]<<ff1[i]<<ff2[i];

     tmpMaterial.MatParticle[particleId].Terminators = Terminators;
      //calculating total interaction cross-section
      bool TotCalc =  tmpMaterial.MatParticle[particleId].CalculateTotal();
      if (!TotCalc) qCritical()<<"ERROR in calculation of the total interaction coefficient!";

//      qDebug()<<"-->tmpMaterial terminators updated and total interaction calculated";
//      qDebug()<<tmpMaterial.MatParticle[particleId].Terminators.size();
//      qDebug()<<tmpMaterial.MatParticle[particleId].InteractionDataX.size();

      MaterialInspectorWindow::on_cobParticle_currentIndexChanged(particleId); //indication
  //    ui->pbShowThisScenarioCrossSection->setEnabled(true);
  //    ui->pbShowTotalInteraction->setEnabled(true);

      MaterialInspectorWindow::on_pbWasModified_clicked();
      MaterialInspectorWindow::on_ledGammaDiagnosticsEnergy_editingFinished();

      qDebug()<<"XCOM data imported from "+ fileName;
      return;
}

void MaterialInspectorWindow::on_pbLoadTotalInteractionCoefficient_clicked()
{
    //available only for neutrons!
    QString fileName;
    fileName = QFileDialog::getOpenFileName(this, "Load capture cross-section vs energy", MW->GlobSet->LastOpenDir, "Data files (*.dat, *.txt);;All files (*)");
    //qDebug()<<fileName;
    if (fileName.isEmpty()) return;
    MW->GlobSet->LastOpenDir = QFileInfo(fileName).absolutePath();
    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

        QFile file(fileName);

        if(!file.open(QIODevice::ReadOnly | QFile::Text))
        {
            QMessageBox::information(0, "error", file.errorString());
            return;
        }

        QTextStream in(&file);

        int particleId = ui->cobParticle->currentIndex();
        //tmpMaterial.MatParticle[particleId].InteractionDataX.resize(0);
        //tmpMaterial.MatParticle[particleId].InteractionDataF.resize(0);

        double Multiplier;
        switch (ui->cobTotalInteractionLoadEnergyUnits->currentIndex())
        {
        case (0): {Multiplier = 1.0e-6; break;} //meV
        case (1): {Multiplier = 1.0e-3; break;} //eV
        case (2): {Multiplier = 1.0; break;}    //keV
        case (3): {Multiplier = 1.0e3; break;}    //MeV
        }

        //separators
        QRegExp rx("(\\ |\\,|\\:|\\t)"); //RegEx for ' ' or ',' or ':' or '\t'

        QVector<double> xx, ff;
        while(!in.atEnd())
        {
            QString line = in.readLine().simplified();
            if (line.isEmpty()) continue; //allow empty lines
            if (line.startsWith("#") || line.startsWith(">")) continue; //comments
            QStringList fields = line.split(rx, QString::SkipEmptyParts);
            if (fields.size() < 2)
              {
                message("Bad file format! Should be Energy and Cross-section columns, comments/headers should start from > or #", this);
                return;
              }
            bool ok1, ok2;
            double x = fields[0].toDouble(&ok1);
            double f = fields[1].toDouble(&ok2);
            if (!ok1 || !ok2)
              {
                message("Bad file format! Should be Energy and Cross-section columns, comments/headers should start from > or #", this);
                return;
              }
            //qDebug()<<x<<f;
            xx.append(x*Multiplier);
            ff.append(f*1.0e-24);
        }
        file.close();
        tmpMaterial.MatParticle[particleId].InteractionDataX = xx;
        tmpMaterial.MatParticle[particleId].InteractionDataF = ff;

        if (tmpMaterial.MatParticle[particleId].Terminators.size() > 0) ui->fNeutron->setEnabled(true);
        ui->pbShowTotalInteraction->setEnabled(true);
        ui->pbAddNewTerminationScenario->setEnabled(true);

        MaterialInspectorWindow::on_pbWasModified_clicked();

        MaterialInspectorWindow::on_cobParticle_currentIndexChanged(ui->cobParticle->currentIndex());
        MaterialInspectorWindow::on_ledBranching_editingFinished(); //to update cross-sections
}

void MaterialInspectorWindow::on_pbNeutronClear_clicked()
{
    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
    int particleId = ui->cobParticle->currentIndex();
    int scenario = ui->cobTerminationScenarios->currentIndex();

    MatParticleStructure& MatParticle = tmpMaterial.MatParticle[particleId];
    if (scenario >= MatParticle.Terminators.size())
    {
        qWarning() << "Wrong scenario index!";
        return;
    }

    MatParticle.Terminators.remove(scenario);
    int Scenarios = MatParticle.Terminators.size();
    double Sum = 0;
    for (int i=0; i<Scenarios; i++) Sum += MatParticle.Terminators[i].branching;
    if (Sum == 0) Sum = 1.0;
    for (int i=0; i<Scenarios; i++) MatParticle.Terminators[i].branching /= Sum;

    MW->MaterialCollection->RecalculateCrossSections(particleId);
    MaterialInspectorWindow::on_pbUpdateIndication_clicked();
}


void MaterialInspectorWindow::on_pbCleanTmpMaterialMatParticles_clicked()
{
  QMessageBox *msgBox = new QMessageBox( this );
  msgBox->setWindowTitle("Confirmation");
  msgBox->setText("Clean interaction data for ALL particles?");
  msgBox->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
  msgBox->setDefaultButton(QMessageBox::Yes);

  if (msgBox->exec() == QMessageBox::No)
    {
      if (msgBox) delete msgBox;
      return;
    }
  if (msgBox) delete msgBox;

  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

    for (int i=0; i<Detector->ParticleCollection.size(); i++)
    {
        tmpMaterial.MatParticle[i].TrackingAllowed = true;
        tmpMaterial.MatParticle[i].MaterialIsTransparent = true;
        tmpMaterial.MatParticle[i].PhYield=0;
        tmpMaterial.MatParticle[i].InteractionDataX.resize(0);
        tmpMaterial.MatParticle[i].InteractionDataF.resize(0);
        tmpMaterial.MatParticle[i].Terminators.resize(0);
    }
    //to refresh indication:
    MaterialInspectorWindow::on_pbUpdateIndication_clicked();
    MaterialInspectorWindow::on_pbWasModified_clicked();
}

void MaterialInspectorWindow::on_pbCleanThisParticleMatParticleData_clicked()
{
  QMessageBox *msgBox = new QMessageBox( this );
  msgBox->setWindowTitle("Confirmation");
  msgBox->setText("Clean interaction data for this particle?");
  msgBox->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
  msgBox->setDefaultButton(QMessageBox::Yes);

  if (msgBox->exec() == QMessageBox::No)
    {
      if (msgBox) delete msgBox;
      return;
    }
  if (msgBox) delete msgBox;

    int i = ui->cobParticle->currentIndex();
    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

    tmpMaterial.MatParticle[i].TrackingAllowed = true;
    tmpMaterial.MatParticle[i].MaterialIsTransparent = true;
    tmpMaterial.MatParticle[i].PhYield=0;
    tmpMaterial.MatParticle[i].InteractionDataX.resize(0);
    tmpMaterial.MatParticle[i].InteractionDataF.resize(0);
    tmpMaterial.MatParticle[i].Terminators.resize(0);

    //to refresh indication:
    MaterialInspectorWindow::on_pbUpdateIndication_clicked();
    MaterialInspectorWindow::on_pbWasModified_clicked();
}

void MaterialInspectorWindow::AddMatToCobs(QString str)
{
   DoNotUpdateOnCOBchange = true;
   ui->cobActiveMaterials->addItem(str);  
   DoNotUpdateOnCOBchange = false;
}

void MaterialInspectorWindow::UpdateMaterialInCobs(int imat)
{
  QString name = (*Detector->MaterialCollection)[imat]->name;
  ui->cobActiveMaterials->setItemText(imat, name);
}

void MaterialInspectorWindow::RemoveMaterial(int imat)
{
   DoNotUpdateOnCOBchange = true;
   ui->cobActiveMaterials->removeItem(imat);
   DoNotUpdateOnCOBchange = false;
   ui->cobActiveMaterials->setCurrentIndex(-1);
}

void MaterialInspectorWindow::on_pbLoadPrimSpectrum_clicked()
{
  QString fileName;
  DoNotDefocus = true;
  fileName = QFileDialog::getOpenFileName(this, "Load primary scintillation spectrum", MW->GlobSet->LastOpenDir, "Data files (*.dat);;All files (*.*)");
  DoNotDefocus = false;
  qDebug()<<fileName;

  if (fileName.isEmpty()) return;
  MW->GlobSet->LastOpenDir = QFileInfo(fileName).absolutePath();
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

  LoadDoubleVectorsFromFile(fileName, &tmpMaterial.PrimarySpectrum_lambda, &tmpMaterial.PrimarySpectrum);  //cleans previous data

  int numPoints = tmpMaterial.PrimarySpectrum_lambda.size();
  QString str;
  str.setNum(numPoints);
  MW->Owindow->OutText(fileName + " - loaded "+str+" data points");

  if (numPoints>0)
    {
      ui->pbShowPrimSpectrum->setEnabled(true);
      ui->pbDeletePrimSpectrum->setEnabled(true);     
    }
  else
    {
      ui->pbShowPrimSpectrum->setEnabled(false);
      ui->pbDeletePrimSpectrum->setEnabled(false);
    }
  MaterialInspectorWindow::on_pbWasModified_clicked();
}

void MaterialInspectorWindow::on_pbShowPrimSpectrum_clicked()
{
  //MW->ShowGraphWindow();
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
  MW->GraphWindow->ShowGraph(&tmpMaterial.PrimarySpectrum_lambda, &tmpMaterial.PrimarySpectrum, kRed, "Wavelength, nm", "Photon flux, a.u.");
}

void MaterialInspectorWindow::on_pbDeletePrimSpectrum_clicked()
{
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
  tmpMaterial.PrimarySpectrum_lambda.resize(0);
  tmpMaterial.PrimarySpectrum.resize(0); 

  ui->pbShowPrimSpectrum->setEnabled(false);
  ui->pbDeletePrimSpectrum->setEnabled(false);
  MaterialInspectorWindow::on_pbWasModified_clicked();
}

void MaterialInspectorWindow::ConvertToStandardWavelengthes(QVector<double>* sp_x, QVector<double>* sp_y, QVector<double>* y)
{
  y->resize(0);

  //qDebug()<<"Data range:"<<sp_x->at(0)<<sp_x->at(sp_x->size()-1);
  double xx, yy;
  int points = MW->WaveNodes;
  for (int i=0; i<points; i++)
    {
      xx = MW->WaveFrom + MW->WaveStep*i;
      if (xx <= sp_x->at(0)) yy = sp_y->at(0);
      else
        {
          if (xx >= sp_x->at(sp_x->size()-1)) yy = sp_y->at(sp_x->size()-1);
          else
            {
              //general case
              yy = InteractionValue(xx, sp_x, sp_y); //reusing interpolation function from functions.h
            }
        }
//      qDebug()<<xx<<yy;
      y->append(yy);
    }
}

void MaterialInspectorWindow::on_pbLoadSecSpectrum_clicked()
{
  QString fileName;
  DoNotDefocus = true;
  fileName = QFileDialog::getOpenFileName(this, "Load secondary scintillation spectrum", MW->GlobSet->LastOpenDir, "Data files (*.dat);;All files (*.*)");
  DoNotDefocus = false;
  qDebug()<<fileName;

  if (fileName.isEmpty()) return;
  MW->GlobSet->LastOpenDir = QFileInfo(fileName).absolutePath();
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

  LoadDoubleVectorsFromFile(fileName, &tmpMaterial.SecondarySpectrum_lambda, &tmpMaterial.SecondarySpectrum);  //cleans previous data

  int numPoints = tmpMaterial.SecondarySpectrum_lambda.size();
  QString str;
  str.setNum(numPoints); 
  MW->Owindow->OutText(fileName + " - loaded "+str+" data points");

  if (numPoints>0)
    {
      ui->pbShowSecSpectrum->setEnabled(true);
      ui->pbDeleteSecSpectrum->setEnabled(true);     
    }
  else
    {
      ui->pbShowSecSpectrum->setEnabled(false);
      ui->pbDeleteSecSpectrum->setEnabled(false);
    }
  MaterialInspectorWindow::on_pbWasModified_clicked();
}

void MaterialInspectorWindow::on_pbShowSecSpectrum_clicked()
{
  //MW->ShowGraphWindow();
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
  MW->GraphWindow->ShowGraph(&tmpMaterial.SecondarySpectrum_lambda, &tmpMaterial.SecondarySpectrum, kRed, "Wavelength, nm", "Photon flux, a.u.");
}

void MaterialInspectorWindow::on_pbDeleteSecSpectrum_clicked()
{
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
  tmpMaterial.SecondarySpectrum_lambda.resize(0);
  tmpMaterial.SecondarySpectrum.resize(0);  

  ui->pbShowSecSpectrum->setEnabled(false);
  ui->pbDeleteSecSpectrum->setEnabled(false);
  MaterialInspectorWindow::on_pbWasModified_clicked();
}

void MaterialInspectorWindow::on_pbLoadNlambda_clicked()
{
  QString fileName;
  DoNotDefocus = true;
  fileName = QFileDialog::getOpenFileName(this, "Load refractive index data", MW->GlobSet->LastOpenDir, "Data files (*.dat);;All files (*.*)");
  DoNotDefocus = false;
//  qDebug()<<fileName;

  if (fileName.isEmpty()) return;
  MW->GlobSet->LastOpenDir = QFileInfo(fileName).absolutePath();
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

  LoadDoubleVectorsFromFile(fileName, &tmpMaterial.nWave_lambda, &tmpMaterial.nWave);  //cleans previous data too

  int numPoints = tmpMaterial.nWave_lambda.size();
  QString str;
  str.setNum(numPoints);
  MW->Owindow->OutText(fileName + " - loaded "+str+" data points");

  if (numPoints>0)
    {
      ui->pbShowNlambda->setEnabled(true);
      ui->pbDeleteNlambda->setEnabled(true);    
    }
  else
    {
      ui->pbShowNlambda->setEnabled(false);
      ui->pbDeleteNlambda->setEnabled(false);
    }
  MaterialInspectorWindow::on_pbWasModified_clicked();
}

void MaterialInspectorWindow::on_pbShowNlambda_clicked()
{
  //MW->ShowGraphWindow();
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
  MW->GraphWindow->ShowGraph(&tmpMaterial.nWave_lambda, &tmpMaterial.nWave, kRed, "Wavelength, nm", "Refractive index");
}

void MaterialInspectorWindow::on_pbDeleteNlambda_clicked()
{
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
  tmpMaterial.nWave_lambda.resize(0);
  tmpMaterial.nWave.resize(0);  

  ui->pbShowNlambda->setEnabled(false);
  ui->pbDeleteNlambda->setEnabled(false);

  MaterialInspectorWindow::on_pbWasModified_clicked();
}

void MaterialInspectorWindow::on_pbLoadABSlambda_clicked()
{
  QString fileName;
  DoNotDefocus = true;
  fileName = QFileDialog::getOpenFileName(this, "Load exponential bulk absorption data", MW->GlobSet->LastOpenDir, "Data files (*.dat);;All files (*.*)");
  DoNotDefocus = false;
  qDebug()<<fileName;

  if (fileName.isEmpty()) return;
  MW->GlobSet->LastOpenDir = QFileInfo(fileName).absolutePath();
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

  LoadDoubleVectorsFromFile(fileName, &tmpMaterial.absWave_lambda, &tmpMaterial.absWave);  //cleans previous data too

  int numPoints = tmpMaterial.absWave_lambda.size();
  QString str;
  str.setNum(numPoints);
  MW->Owindow->OutText(fileName + " - loaded "+str+" data points");

  if (numPoints>0)
    {
      ui->pbShowABSlambda->setEnabled(true);
      ui->pbDeleteABSlambda->setEnabled(true);     
    }
  else
    {
      ui->pbShowABSlambda->setEnabled(false);
      ui->pbDeleteABSlambda->setEnabled(false);
    }
  MaterialInspectorWindow::on_pbWasModified_clicked();
}

void MaterialInspectorWindow::on_pbShowABSlambda_clicked()
{
  //MW->ShowGraphWindow();
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
  MW->GraphWindow->ShowGraph(&tmpMaterial.absWave_lambda, &tmpMaterial.absWave, kRed, "Wavelength, nm", "Attenuation coefficient");
}

void MaterialInspectorWindow::on_pbDeleteABSlambda_clicked()
{
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
  tmpMaterial.absWave_lambda.resize(0);
  tmpMaterial.absWave.resize(0); 

  ui->pbShowABSlambda->setEnabled(false);
  ui->pbDeleteABSlambda->setEnabled(false);
  MaterialInspectorWindow::on_pbWasModified_clicked();
}

void MaterialInspectorWindow::on_pbWasModified_clicked()
{
  if (flagDisreguardChange) return;
  ui->labWasModified->setVisible(true);
}

bool MaterialInspectorWindow::event(QEvent * e)
{
 //  if (!MW->isVisible()) return QMainWindow::event(e); //if we are closing the application

    switch(e->type())
    {
      /*
        case QEvent::WindowActivate :
         {
            //focussed!
//            qDebug()<<"in";
            if (MW->GeometryWindow->isVisible())
               {
                 MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
                 QString name = tmpMaterial.name;
                 int size = MW->MaterialCollection->size();
                 bool flagFound = false;
                 int index;
                 for (index = 0; index < size; index++)
                    if (!name.compare( (*MW->MaterialCollection)[index]->name,Qt::CaseSensitive)) {flagFound = true; break;}

                 if (!flagFound) index = 111111;
                 MW->ColorVolumes(2, index);
                 //MW->ShowGeometry(2);
                 MW->ShowGeometry(false, true, false);
                 //this->raise();
               }
            break ;
          }
        case QEvent::WindowDeactivate :
            // lost focus
            if (DoNotDefocus) break;            
//            qDebug()<<"out!";
            //if (MW->GeometryWindow->isVisible()) MW->ShowGeometry(3);
            if (MW->GeometryWindow->isVisible()) MW->ShowGeometry(false, true, true);
            break;
*/
      case QEvent::Hide :
        if (MW->WindowNavigator) MW->WindowNavigator->HideWindowTriggered("mat");
        break;
      case QEvent::Show :
        if (MW->WindowNavigator) MW->WindowNavigator->ShowWindowTriggered("mat");
        break;
      default:
        break;
    } ;
    return QMainWindow::event(e);
}

void MaterialInspectorWindow::on_pbGammaDiagnosticsPhComptRation_clicked()
{
  int particleId = ui->cobParticle->currentIndex();
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

  if (tmpMaterial.MatParticle[particleId].Terminators.size() < 2) return;

  int elements0 = tmpMaterial.MatParticle[particleId].Terminators[0].PartialCrossSection.size();
  int elements1 = tmpMaterial.MatParticle[particleId].Terminators[1].PartialCrossSection.size();
  if (elements0 == 0 && elements1 == 0) return;

  if (elements0 !=elements1)
    {
      message("Error: mismatch in number of entries!");
      return;
    }

  if (elements0 == 0) return;
  double nan;
  QString str = "nan";
  nan = str.toDouble();

  QVector<double> *X;
  X = &tmpMaterial.MatParticle[particleId].Terminators[0].PartialCrossSectionEnergy;
  QVector<double> Y;
  for (int i=0; i<elements0; i++)
    {
      double sum = tmpMaterial.MatParticle[particleId].Terminators[0].PartialCrossSection[i] + tmpMaterial.MatParticle[particleId].Terminators[1].PartialCrossSection[i];
      if (sum == 0) Y.append(nan);
      else Y.append(tmpMaterial.MatParticle[particleId].Terminators[0].PartialCrossSection[i] / sum);
    }

  //MW->ShowGraphWindow();
  MW->GraphWindow->ShowGraph(X, &Y, kRed, "Energy, keV", "Fraction of photoelectrics");
}

void MaterialInspectorWindow::on_leName_editingFinished()
{
  QString name = ui->leName->text();
  name.replace("+","_");
  name.replace(".","_");
  ui->leName->setText(name);

  if ( (name.at(0) >= QChar('a') && name.at(0) <= QChar('z')) || (name.at(0) >= QChar('A') && name.at(0) <= QChar('Z'))) ;//ok
  else
    {
      name.remove(0,1);
      name.insert(0, "A");
      ui->leName->setText(name);
      DoNotDefocus = true;
      message("Name should start with an alphabetic character!");
      DoNotDefocus = false;
      return;
    }
  MaterialInspectorWindow::on_pbUpdateTmpMaterial_clicked();
}

void MaterialInspectorWindow::on_cbTrackingAllowed_toggled(bool checked)
{  
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
  ui->fEnDepProps->setEnabled(checked);
  tmpMaterial.MatParticle[ui->cobParticle->currentIndex()].TrackingAllowed = checked;
  //ui->fTransparent->setEnabled(checked);

  QFont font = ui->cbTrackingAllowed->font();
  font.setBold(checked);
  ui->cbTrackingAllowed->setFont(font);

  //invisible control
  ui->fClear->setVisible(checked);
  ui->lineClear->setVisible(checked);
  ui->fEnDepProps->setVisible(checked);
  ui->cbLogLogInterpolation->setVisible(checked);
}

void MaterialInspectorWindow::on_leName_textChanged(const QString &)
{
    //if tmpMaterial holds binned properties (invisible in mat inspector) and aquired from material, they has to be disconnected
    //on text change - on chage this is a signal that it will be another material. These properties are recalculated anyway on
    //accepting changes/new material
    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

    tmpMaterial.absWaveBinned.resize(0);
    tmpMaterial.nWaveBinned.resize(0);
    tmpMaterial.GeoMat = 0;  //no delete! the original material has to have them
    tmpMaterial.GeoMed = 0;
    tmpMaterial.PrimarySpectrumHist = 0;
    tmpMaterial.SecondarySpectrumHist = 0;
}

void MaterialInspectorWindow::on_ledGammaDiagnosticsEnergy_editingFinished()
{
  ui->leoGammaDiagnosticsCoefficient->setText("n.a.");
  ui->leoMFPgamma->setText("n.a.");

  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

  int particleId = ui->cobParticle->currentIndex();
 // int TermScenario = ui->cobTerminationScenarios->currentIndex();
 // int elements = tmpMaterial.MatParticle[particleId].Terminators[TermScenario].PartialCrossSection.size();
 // if (elements<1) return;

  bool ok;
  double energy = ui->ledGammaDiagnosticsEnergy->text().toDouble(&ok);
  if (!ok) return;

  int DataPoints = tmpMaterial.MatParticle[particleId].InteractionDataX.size();  
  if (energy<tmpMaterial.MatParticle[particleId].InteractionDataX[0] || energy>tmpMaterial.MatParticle[particleId].InteractionDataX[DataPoints-1])  return;

  QString str, str1;
  //double coefficient = InteractionValue(energy, &tmpMaterial.MatParticle[particleId].Terminators[TermScenario].PartialCrossSectionEnergy, &tmpMaterial.MatParticle[particleId].Terminators[TermScenario].PartialCrossSection, LogLogInterpolation);
  //str.setNum(coefficient, 'g', 4);

  double Density = tmpMaterial.density;
  int LogLogInterpolation = Detector->MaterialCollection->fLogLogInterpolation;
  double InteractionCoefficient = InteractionValue(energy, &tmpMaterial.MatParticle[particleId].InteractionDataX, &tmpMaterial.MatParticle[particleId].InteractionDataF, LogLogInterpolation);
  //qDebug()<<energy<<InteractionCoefficient;
  str.setNum(InteractionCoefficient, 'g', 4);

  double MeanFreePath = 10.0/InteractionCoefficient/Density;  //1/(cm2/g)/(g/cm3) - need in mm (so that 10.)
  str1.setNum(MeanFreePath, 'g', 4);

  ui->leoGammaDiagnosticsCoefficient->setText(str);
  ui->leoMFPgamma->setText(str1);
}

void MaterialInspectorWindow::on_cbTransparentMaterial_toggled(bool checked)
{
   MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
   int particleId = ui->cobParticle->currentIndex();
   tmpMaterial.MatParticle[particleId].MaterialIsTransparent = checked;

   ui->frameMatParticleData->setEnabled(!checked);

   QFont font = ui->cbTransparentMaterial->font();
   font.setBold(checked);
   ui->cbTransparentMaterial->setFont(font);
}

void MaterialInspectorWindow::on_pbComments_clicked()
{
  QDialog* dialog = new QDialog(this);

  QString str = ui->leName->text();
  dialog->setWindowTitle(str);

  QPushButton *okButton = new QPushButton("Close");
  connect(okButton,SIGNAL(clicked()),dialog,SLOT(accept()));

  QTextEdit *text = new QTextEdit();
  text->append(MW->MaterialCollection->tmpMaterial.Comments);

  QVBoxLayout *mainLayout = new QVBoxLayout;
  mainLayout->addWidget(text);
  mainLayout->addWidget(okButton);

  dialog->setLayout(mainLayout);
  dialog->exec();

  MW->MaterialCollection->tmpMaterial.Comments = text->document()->toPlainText();
  ui->labWasModified->setVisible(true);
  delete dialog;
}

void MaterialInspectorWindow::on_ledRayleighWave_editingFinished()
{
    double wave = ui->ledRayleighWave->text().toDouble();
    if (wave <= 0)
      {
        message("Wavelength should be a positive number!", this);
        wave = 500.0;
        ui->ledRayleighWave->setText("500");
      }

    MW->MaterialCollection->tmpMaterial.rayleighWave = wave;
}

void MaterialInspectorWindow::on_ledRayleigh_textChanged(const QString &arg1)
{
    if (arg1 == "") ui->ledRayleighWave->setEnabled(false);
    else ui->ledRayleighWave->setEnabled(true);
}

void MaterialInspectorWindow::on_ledRayleigh_editingFinished()
{
  double ray;
  if (ui->ledRayleigh->text() == "") ray = 0;
  else ray = ui->ledRayleigh->text().toDouble();
  MW->MaterialCollection->tmpMaterial.rayleighMFP = ray;
}

void MaterialInspectorWindow::on_pbRemoveRayleigh_clicked()
{
  ui->ledRayleigh->setText("");
  MW->MaterialCollection->tmpMaterial.rayleighMFP = 0;
  MaterialInspectorWindow::on_pbWasModified_clicked();
}

void MaterialInspectorWindow::on_pbShowUsage_clicked()
{
  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
  QString name = tmpMaterial.name;
  int index = ui->cobActiveMaterials->currentIndex();

  bool flagFound = false;
  TObjArray* list = Detector->GeoManager->GetListOfVolumes();
  int size = list->GetEntries();
  for (int i=0; i<size; i++)
    {
      TGeoVolume* vol = (TGeoVolume*)list->At(i);
      if (!vol) break;
      if (index == vol->GetMaterial()->GetIndex())
        {
          flagFound = true;
          break;
        }
    }

  if (flagFound)
    {
      Detector->colorVolumes(2, index);
      MW->ShowGeometry(true, true, false);
    }
  else
    {
      MW->ShowGeometry(false);
      message("Current detector configuration does not have objects referring to material "+name, this);
    }

}

void MaterialInspectorWindow::on_ledMFPenergy_editingFinished()
{
    if (ui->ledMFPenergy->text().isEmpty())
      {
        ui->leMFP->setText("");
        return;
      }

    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
    if (tmpMaterial.atomicDensity <= 0) return;

    double energy = ui->ledMFPenergy->text().toDouble() *0.001; //energy in keV
    int particleId = ui->cobParticle->currentIndex();

    if ( tmpMaterial.MatParticle[particleId].InteractionDataX.size() < 2)
      {
        ui->leMFP->setText("no data");
        return;
      }
    if (energy<tmpMaterial.MatParticle[particleId].InteractionDataX.first() || energy > tmpMaterial.MatParticle[particleId].InteractionDataX.last())
      {
        ui->leMFP->setText("out range");
        return;
      }

    int LogLogInterpolation = Detector->MaterialCollection->fLogLogInterpolation;
    double CrossSection = InteractionValue(energy, &tmpMaterial.MatParticle[particleId].InteractionDataX, &tmpMaterial.MatParticle[particleId].InteractionDataF, LogLogInterpolation);
    //qDebug()<<CrossSection;
    double AtomicDensity = tmpMaterial.atomicDensity;
    double MeanFreePath = 10.0/CrossSection/AtomicDensity;  //1/(cm2)/(1/cm3) - need in mm (so that 10.)
    QString str;
    str.setNum(MeanFreePath, 'g', 4);
    ui->leMFP->setText(str);
}

void MaterialInspectorWindow::on_pbNistPage_clicked()
{
    QDesktopServices::openUrl(QUrl("http://physics.nist.gov/PhysRefData/Xcom/html/xcom1-t.html", QUrl::TolerantMode));
}

void MaterialInspectorWindow::on_pbAddNewSecondary_clicked()
{
    int particleId = ui->cobParticle->currentIndex();
    int scenario = ui->cobTerminationScenarios->currentIndex();
    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

    tmpMaterial.MatParticle[particleId].Terminators[scenario].GeneratedParticles.append(ui->cobSecondaryParticleToAdd->currentIndex());
    tmpMaterial.MatParticle[particleId].Terminators[scenario].GeneratedParticleEnergies.append(ui->ledInitialEnergy->text().toDouble());

    MaterialInspectorWindow::on_cobTerminationScenarios_currentIndexChanged(scenario);
    ui->labWasModified->setVisible(true);
}

void MaterialInspectorWindow::on_pbRemoveSecondary_clicked()
{
    int iSecondary = ui->lwGeneratedParticlesEnergies->currentRow();
    qDebug() << "Selected:" << iSecondary;
    if (iSecondary < 0)
    {
        message("First select a particle to remove!", this);
        return;
    }

    int scenario = ui->cobTerminationScenarios->currentIndex();
    if (scenario < 0) return;
    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
    int ParticleId = ui->cobParticle->currentIndex();
    if (tmpMaterial.MatParticle[ParticleId].Terminators[scenario].GeneratedParticles.size() > iSecondary)
    {
        tmpMaterial.MatParticle[ParticleId].Terminators[scenario].GeneratedParticles.remove(iSecondary);
        tmpMaterial.MatParticle[ParticleId].Terminators[scenario].GeneratedParticleEnergies.remove(iSecondary);
    }

    MaterialInspectorWindow::on_cobTerminationScenarios_currentIndexChanged(scenario);
    ui->labWasModified->setVisible(true);
}

void MaterialInspectorWindow::on_lwGeneratedParticlesEnergies_currentRowChanged(int currentRow)
{
    //qDebug() << "Row changed!"<< currentRow;
    if (currentRow<0) return;

    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
    int ParticleId = ui->cobParticle->currentIndex();
    int scenario = ui->cobTerminationScenarios->currentIndex();

    int secP = tmpMaterial.MatParticle[ParticleId].Terminators[scenario].GeneratedParticles[currentRow];
    double energy = tmpMaterial.MatParticle[ParticleId].Terminators[scenario].GeneratedParticleEnergies[currentRow];
    ui->cobSecondaryParticleToAdd->setCurrentIndex(secP);
    ui->ledInitialEnergy->setText(QString::number(energy));
}

void MaterialInspectorWindow::on_ledInitialEnergy_editingFinished()
{
    int row = ui->lwGeneratedParticlesEnergies->currentRow();
    if (row<0) return;

    MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
    int ParticleId = ui->cobParticle->currentIndex();
    int scenario = ui->cobTerminationScenarios->currentIndex();

    double energy = ui->ledInitialEnergy->text().toDouble();
    tmpMaterial.MatParticle[ParticleId].Terminators[scenario].GeneratedParticleEnergies[row] = energy;
    //on_pbUpdateIndication_clicked();
    MaterialInspectorWindow::on_cobTerminationScenarios_currentIndexChanged(ui->cobTerminationScenarios->currentIndex()); //refresh indication
    ui->labWasModified->setVisible(true);
}

void MaterialInspectorWindow::on_cobSecondaryParticleToAdd_activated(int index)
{
  int row = ui->lwGeneratedParticlesEnergies->currentRow();
  if (row<0) return;

  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;
  int ParticleId = ui->cobParticle->currentIndex();
  int scenario = ui->cobTerminationScenarios->currentIndex();

  tmpMaterial.MatParticle[ParticleId].Terminators[scenario].GeneratedParticles[row] = index;
  on_pbUpdateIndication_clicked();
  ui->labWasModified->setVisible(true);
}

void MaterialInspectorWindow::on_pbRename_clicked()
{
  QString name = ui->leName->text();
  int aMat = ui->cobActiveMaterials->currentIndex();
  if (aMat<0) return;
  QString aName = (*Detector->MaterialCollection)[aMat]->name;
  if (name == aName) return; //not changed

  for (int i=0; i<Detector->MaterialCollection->size(); i++)
    if (i!=aMat && aName==(*Detector->MaterialCollection)[i]->name)
      {
        DoNotDefocus = true;
        message("There is already a material with name "+name);
        DoNotDefocus = false;
        return;
      }
  (*Detector->MaterialCollection)[aMat]->name = name;
  DoNotDefocus = true;
  MW->UpdateMaterialListEdit();
  MW->UpdateMaterialNameInCobs(aMat);
  DoNotDefocus = false;
  ui->pbRename->setText("Rename\n"+name);

  MW->ReconstructDetector(true);
}

void MaterialInspectorWindow::on_cbLogLogInterpolation_clicked(bool checked)
{
   Detector->MaterialCollection->fLogLogInterpolation = checked;
   MW->ReconstructDetector(true);
}

void MaterialInspectorWindow::on_ledMFPenergy_2_editingFinished()
{
  if (ui->ledMFPenergy_2->text().isEmpty())
    {
      ui->leMFP_2->setText("");
      return;
    }

  MaterialStructure& tmpMaterial = MW->MaterialCollection->tmpMaterial;

  double energy0, energy;
  energy0 = energy = ui->ledMFPenergy_2->text().toDouble();
  int particleId = ui->cobParticle->currentIndex();

  if ( tmpMaterial.MatParticle[particleId].InteractionDataX.size() < 2)
    {
      ui->leMFP->setText("no data");
      return;
    }
  if (energy<tmpMaterial.MatParticle[particleId].InteractionDataX.first() || energy > tmpMaterial.MatParticle[particleId].InteractionDataX.last())
    {
      ui->leMFP->setText("out range");
      return;
    }

  int LogLogInterpolation = Detector->MaterialCollection->fLogLogInterpolation;
  //double InteractionCoefficient = InteractionValue(energy, &tmpMaterial.MatParticle[particleId].InteractionDataX, &tmpMaterial.MatParticle[particleId].InteractionDataF, LogLogInterpolation);
  //qDebug()<<InteractionCoefficient<<tmpMaterial.density;
  //double MeanFreePath = 10.0/InteractionCoefficient/tmpMaterial.density;  // 1/(cm2/g)/(g/cm3) - need in mm (so that 10.)

  double range = 0;
  double Step;
  int counter = 1000; //1000 steps!
  do
    {
      if (energy < tmpMaterial.MatParticle[particleId].InteractionDataX.first()) break;

      double InteractionCoefficient = InteractionValue(energy, &tmpMaterial.MatParticle[particleId].InteractionDataX, &tmpMaterial.MatParticle[particleId].InteractionDataF, LogLogInterpolation);
      //qDebug()<<InteractionCoefficient<<tmpMaterial.density;
      //dE/dx [keV/mm] = Density[g/cm3] * [cm2/g*keV] * 0.1  //0.1 since local units are mm, not cm
      double dEdX = 0.1 * tmpMaterial.density * InteractionCoefficient;
      Step = 0.03 * energy/dEdX;
      double dE = dEdX * Step;
      range += Step;

      energy -= dE;
      counter--;
    }
  while (counter>0);


  QString str;
  str.setNum(range, 'g', 4);
  ui->leMFP_2->setText(str);
}
