#include "detectorclass.h"

#include "TGeoManager.h"
#include "pms.h"
#include "sensorlrfs.h"
#include "materialcollectionclass.h"
#include "functions.h"
//#include "detectorsandwichclass.h"

#include <QtGui>

#include "TRandom2.h"

void DetectorClass::clearParticleCollection()
{
  for (int i=0; i<ParticleCollection.size(); i++)
    delete ParticleCollection[i];
  ParticleCollection.clear();
}

DetectorClass::DetectorClass() : Sandwich(this)
{  
  //qDebug() << "--> Detector constructor";
  //RandGen = new TRandom2(123456789);
  RandGen = new TRandom2(QDateTime::currentMSecsSinceEpoch());
  MaterialCollection = new MaterialCollectionClass(&ParticleCollection);
  //qDebug() << "    Material collection created";
  //GeoManager = new TGeoManager();
  //GeoManager->AddNavigator(); /// *** !!!
  GeoManager = 0;  // without it we have a crash in Linux on attempt to remove empty GeoManager: so we create GeoManager first time detector is build
  //qDebug() << "    GeoManager created";
  PMs = new pms(MaterialCollection);
  PMs->clearPMtypes();
  //qDebug() << "    Photomultipliers manager created";
  SensLRF = new SensorLRFs(0);
  //qDebug() << "    SensLRF module created";

  GDML = "";
  LastConfig = QJsonObject();
  PMarrays.resize(2);
  fAdditionalGeoObjectsEnabled = true;
  fPrScintCont = false;
  fSecScintPresent = false;
}

DetectorClass::~DetectorClass()
{
  if (SensLRF) delete SensLRF;
  //qDebug() << "  --SensLRF module deleted";

  if (PMs) delete PMs;
  //qDebug() << "  --Photomultiplier manager deleted";

  if (GeoManager) delete GeoManager;
  //qDebug()<<"  --GeoManager deleted";

  for (int i=0; i<ParticleCollection.size(); i++) delete ParticleCollection[i];
  //qDebug()<<"  --Particle collection cleaned";

  if (MaterialCollection) delete MaterialCollection;
  //qDebug() << "  --Material collection deleted";

  delete RandGen;
}

bool DetectorClass::makeDetector(QJsonObject &json)
{
  LastConfig = json;
  bool fOK = makeSandwichDetector(json);
  if (!fOK) return false;

  //handling GDML
  QJsonObject js = json["DetectorConfig"].toObject();
  if (!js.contains("GDML")) return true; //No GDML in json
  QString gdml = js["GDML"].toString();
  fOK = importGDML(gdml);
  return fOK;
}

bool DetectorClass::remakeLastDetector()
{
  if (LastConfig.isEmpty()) return false;
  //qDebug() << "Remaking last detector!";
  return makeDetector(LastConfig);
}

QJsonArray PosAngTypeStructure::writeToJArrayAutoZ()
{
  QJsonArray el;
  el << type << x << y << psi;
  return el;
}

void PosAngTypeStructure::readFromJArrayAutoZ(QJsonArray &arr)
{
  type = arr[0].toInt();
  x =    arr[1].toDouble();
  y =    arr[2].toDouble();
  z = 0;
  phi =   0;
  theta = 0;

  if (arr.size()<4) psi = 0;
  else psi = arr[3].toDouble();
}

QJsonArray PosAngTypeStructure::writeToJArrayFull()
{
  QJsonArray el;
  el << type << x << y << z << phi << theta << psi;
  return el;
}

void PosAngTypeStructure::readFromJArrayFull(QJsonArray &arr)
{
  type =  arr[0].toInt();
  x =     arr[1].toDouble();
  y =     arr[2].toDouble();
  z =     arr[3].toDouble();
  phi =   arr[4].toDouble();
  theta = arr[5].toDouble();
  psi =   arr[6].toDouble();
}

void PMarrayDataClass::writeToJson(QJsonObject &json)
{
  QJsonObject js;
  js["Activated"] = fActive;
  //if (fActive)
    {
      js["Regularity"] = Regularity;   //0-regular, 1-auto_z, 2-custom
      switch (Regularity)
        {
         case 0:
          js["PMtype"] = PMtype;
          js["Packing"] = Packing;
          js["CenterToCenter"] = CenterToCenter;
          js["UseRings"] = fUseRings;
          if (fUseRings) js["NumRings"] = NumRings;
          else
            {
              js["NumX"] = NumX;
              js["NumY"] = NumY;
            }
          break;
        case 1:
          {
            js["PMtype"] = PMtype;
            QJsonArray Arr;
            for (int ipm=0; ipm<PositionsAnglesTypes.size(); ipm++)
              Arr.append(PositionsAnglesTypes[ipm].writeToJArrayAutoZ());
            js["PositionsTypes"] = Arr;
            break;
          }
        case 2:
          {
            QJsonArray Arr;
            for (int ipm=0; ipm<PositionsAnglesTypes.size(); ipm++)
              Arr.append(PositionsAnglesTypes[ipm].writeToJArrayFull());
            js["PositionsAnglesTypes"] = Arr;
            break;
          }
        default:
          qWarning()<<"Unknown array regularity type!";
        }
      if (!PositioningScript.isEmpty()) js["PositioningScript"] = PositioningScript;
    }
  json["PMarrayData"] = js;
}

bool PMarrayDataClass::readFromJson(QJsonObject &json)
{
  if (!json.contains("PMarrayData")) return false;

  QJsonObject js = json["PMarrayData"].toObject();
  parseJson(js, "Activated", fActive);
//  if (!fActive)
//    {
//      //qDebug() << "--- Array not active";
//      return true;
//    }
  parseJson(js, "Regularity", Regularity);
  parseJson(js, "PMtype", PMtype);
  parseJson(js, "Packing", Packing);
  parseJson(js, "CenterToCenter", CenterToCenter);
  parseJson(js, "UseRings", fUseRings);
  parseJson(js, "NumRings", NumRings);
  parseJson(js, "NumX", NumX);
  parseJson(js, "NumY", NumY);
  if (Regularity == 1)
    {
      if (!js.contains("PositionsTypes"))
        {
          qWarning() << "Missing positions xy data!";
          return false;
        }
      QJsonArray Parr = QJsonArray();
      parseJson(js, "PositionsTypes", Parr);
      PositionsAnglesTypes.resize(0);
      for (int i=0; i<Parr.size(); i++)
        {
          QJsonArray el = Parr[i].toArray();
          PositionsAnglesTypes.append( PosAngTypeStructure() );
          PositionsAnglesTypes.last().readFromJArrayAutoZ(el);
        }
    }
  if (Regularity==2)
    {
      if (!js.contains("PositionsAnglesTypes"))
        {
          qWarning() << "Missing PM position/angle/type data!";
          return false;
        }
      QJsonArray Parr = QJsonArray();
      parseJson(js, "PositionsAnglesTypes", Parr);
      PositionsAnglesTypes.resize(0);
      for (int i=0; i<Parr.size(); i++)
        {
          QJsonArray el = Parr[i].toArray();
          PositionsAnglesTypes.append( PosAngTypeStructure() );
          PositionsAnglesTypes.last().readFromJArrayFull(el);
        }
    }
  parseJson(js, "PositioningScript", PositioningScript);
  //qDebug() << "--> Loaded PM array with regularity"<<Regularity;
  return true;
}

void DetectorClass::writePMarraysToJson(QJsonObject &json)
{
  QJsonArray arr;
  for (int i=0; i<2;i++)
    {
      QJsonObject js;
      PMarrays[i].writeToJson(js);
      arr.append(js);
    }
  json["PMarrays"] = arr;
}

bool DetectorClass::readPMarraysFromJson(QJsonObject &json)
{
  if (!json.contains("PMarrays"))
    {
      qWarning()<<"Json file does not contain PM arrays data";
      return false;
    }
  PMarrays.clear();
  PMarrays.resize(2); //protection

  QJsonArray arr = json["PMarrays"].toArray();
  if (arr.size() != 2)
    {
      qWarning()<< "Wrong size of PMarrays in Json file";
      return false;
    }
  for (int i=0; i<2;i++)
    {
      QJsonObject js = arr[i].toObject();
      bool fOK = PMarrays[i].readFromJson(js);
      if (!fOK)
        {
          qWarning() << "Error reading PM array #"<<i;
          return false;
        }
    }
  return true;
}

void PMdummyStructure::writeToJson(QJsonObject &json)
{
  QJsonObject js;
  js["Type"] = PMtype;
  js["Array"] = UpperLower;
  QJsonArray arrR;
  arrR.append(r[0]);
  arrR.append(r[1]);
  arrR.append(r[2]);
  js["R"] = arrR;
  QJsonArray arrA;
  arrA.append(Angle[0]);
  arrA.append(Angle[1]);
  arrA.append(Angle[2]);
  js["Angle"] = arrA;
  json["DummyPM"] = js;
}

bool PMdummyStructure::readFromJson(QJsonObject &json)
{
  if (!json.contains("DummyPM"))
    {
      qWarning() << "Not a dummy pm json!";
      return false;
    }
  QJsonObject js = json["DummyPM"].toObject();
  parseJson(js, "Type", PMtype);
  parseJson(js, "Array", UpperLower);
  QJsonArray arrR = js["R"].toArray();
  r[0] = arrR[0].toDouble();
  r[1] = arrR[1].toDouble();
  r[2] = arrR[2].toDouble();
  QJsonArray arrA = js["Angle"].toArray();
  Angle[0] = arrA[0].toDouble();
  Angle[1] = arrA[1].toDouble();
  Angle[2] = arrA[2].toDouble();
  return true;
}

void DetectorClass::writeDummyPMsToJson(QJsonObject &json)
{
  QJsonArray arr;
  for (int i=0; i<PMdummies.size(); i++)
    {
      QJsonObject js;
      PMdummies[i].writeToJson(js);
      arr.append(js);
    }
  json["DummyPMs"] = arr;
}

void DetectorClass::writeLightguidesToJson(QJsonObject &json)
{
  QJsonArray arr;
  for (int i=0; i<2; i++)
    {
      QJsonObject js;
      Lightguides[i].writeToJson(js);
      arr.append(js);
    }
  json["Lightguides"] = arr;
}

void DetectorClass::writeAddGeoObjectsToJson(QJsonObject &json)
{
  QJsonObject js;
  js["MasterActivate"] = fAdditionalGeoObjectsEnabled;
  if (AdditionalGeoObjects.size()>0)
    {
      QJsonArray ar;
      for (int i=0; i<AdditionalGeoObjects.size(); i++)
        {
          QJsonObject jj;
          AdditionalGeoObjects[i].writeToJson(jj);
          ar.append(jj);
        }
      js["Objects"] = ar;
    }
  js["Script"] = AddObjPositioningScript;

  json["AddGeoObjects"] = js;
}

void DetectorClass::writeGridsToJson(QJsonObject &json)
{
   if (Grids.size()>0)
     {
       QJsonArray ar;
       for (int i=0; i<Grids.size(); i++)
         {
           QJsonObject jj;
           Grids[i].writeToJson(jj);
           ar.append(jj);
         }
       json["OpticalGrids"] = ar;
     }
}

void DetectorClass::writeParticleCollectionToJson(QJsonObject &json)
{
  QJsonArray ar;
  for (int i=0; i<ParticleCollection.size(); i++)
    {
      QJsonObject js;
      ParticleCollection[i]->writeToJson(js);
      ar.append(js);
    }
  json["ParticleCollection"] = ar;
}

void DetectorClass::writeGDMLtoJson(QJsonObject &json)
{
  json["GDML"] = GDML;
}

bool DetectorClass::readParticleCollectionFromJson(QJsonObject &json)
{
  if (!json.contains("ParticleCollection"))
    {
      qCritical() << "Json does not contain particle collections!";
      exit(-1);
    }
  //ParticleCollection.clear();
  clearParticleCollection();
  QJsonArray ar = json["ParticleCollection"].toArray();
  if (ar.isEmpty())
    {
      qCritical() << "Particle collections in json is empty!";
      exit(-1);
    }
  for (int i=0; i<ar.size(); i++)
    {
      QJsonObject js = ar[i].toObject();
      ParticleCollection.append(new ParticleDescriptor());
      ParticleCollection.last()->readFromJson(js);
    }
  //qDebug() << "--> Loaded particle collection with"<<ParticleCollection.size()<<"particles";
  return true;
}

bool DetectorClass::readDummyPMsFromJson(QJsonObject &json)
{
  if (!json.contains("DummyPMs"))
    {
      //qDebug() << "--- Json does not contain dummy PMs data";
      return false;
    }
  QJsonArray arr = json["DummyPMs"].toArray();
  PMdummies.resize(arr.size());
  for (int i=0; i<arr.size(); i++)
    {
      QJsonObject js = arr[i].toObject();
      PMdummies[i].readFromJson(js);
    }
  //qDebug() << "--> Loaded"<<PMdummies.size()<<"dummy PMs";
  return true;
}

bool DetectorClass::readLightguidesFromJson(QJsonObject &json)
{
  if (!json.contains("Lightguides"))
    {
      //qDebug() << "--- Json does not contain lightguides data";
      return false;
    }
  QJsonArray arr = json["Lightguides"].toArray();
  Lightguides.resize(2); //upper:0 and lower:1
  for (int i=0; i<2; i++)
    {
      QJsonObject js = arr[i].toObject();
      bool fOK = Lightguides[i].readFromJson(js);
      if (!fOK)
        {
          qWarning() << "Error reading lightguides data";
          return false;
        }      
    }
  //qDebug() << "--> Lightguides loaded";
  return true;
}

bool DetectorClass::readAddGeoObjectsFromJson(QJsonObject &json)
{
  if (!json.contains("AddGeoObjects"))
    {
      //qDebug() << "--- Json does not contain AddGeoObjects";
      return false;
    }
  QJsonObject js = json["AddGeoObjects"].toObject();
  parseJson(js, "MasterActivate", fAdditionalGeoObjectsEnabled);
  AdditionalGeoObjects.clear();
  if (js.contains("Objects"))
    {
      QJsonArray ar;
      parseJson(js, "Objects", ar);
      for (int i=0; i<ar.size(); i++)
        {
          AdditionalGeoObjects.append(AdditionalGeoObjectClass());
          QJsonObject jj = ar[i].toObject();
          bool fOK = AdditionalGeoObjects.last().readFromJson(jj);
          if (!fOK)
            {
              qWarning() << "Error loading AddGeoObject!";
              continue;
            }
        }
    }
  AddObjPositioningScript = js["Script"].toString();

  //qDebug() << "--> Loaded"<<AdditionalGeoObjects.size()<<"add geo objects";
  return true;
}

bool DetectorClass::readGridsFromJson(QJsonObject &json)
{
  if (!json.contains("OpticalGrids"))
    {
      //qDebug() << "--- Json does not contain optical grids data";
      return false;
    }
  QJsonArray ar = json["OpticalGrids"].toArray();
  Grids.clear();
  for (int i=0; i<ar.size(); i++)
    {
      QJsonObject jj = ar[i].toObject();
      Grids.append(OpticalGridStructure());
      bool fOK = Grids.last().readFromJson(jj);
      if (!fOK) qWarning() << "Error reading optical grid";
    }
  //qDebug() << "--> Loaded"<<Grids.size()<<"optical grids";
  return true;
}

bool DetectorClass::makeSandwichDetector(QJsonObject &json)
{
  //qDebug() << "--->MakeSandwichDetector triggered";
  if (!json.contains("DetectorConfig"))
    {
      qWarning() << "Json does NOT contain detector config!";
      return false;
    }
  QJsonObject js = json["DetectorConfig"].toObject();

  readParticleCollectionFromJson(js);
  MaterialCollection->readFromJson(js);
  PMs->readPMtypesFromJson(js);
  readPMarraysFromJson(js);
  PMdummies.clear();
  readDummyPMsFromJson(js);
  Lightguides.clear();
  Lightguides.resize(2); //to reset
  readLightguidesFromJson(js);
  AdditionalGeoObjects.clear();
  fAdditionalGeoObjectsEnabled = true;
  readAddGeoObjectsFromJson(js); 
  //Mask.fActive = false;
  Masks.readFromJson(js);
//  PrScintCont.fActive = false;
//  PrScintCont.readFromJson(js);
  fPrScintCont = false;
  parseJson(js, "PrScintCont", fPrScintCont);
  Grids.clear();
  readGridsFromJson(js);
  Sandwich.readFromJson(js);

  //making detector
  Sandwich.constructDetector();
  //post-construction load
  PMs->readInividualOverridesFromJson(js);
  PMs->readElectronicsFromJson(js);

  top->SetVisContainers(true); //making contaners visible
  return true;
}

bool DetectorClass::importGDML(QString gdml)
{
  GDML = gdml;

  bool fOK = processGDML();
  if (fOK)
    {
      qDebug() << "--> GDML successfully registered";
    }
  else
    {
      qDebug() << "||| GDML import failed: "+ErrorString;
      GeoManager = new TGeoManager();
      //GeoManager->AddNavigator(); /// *** !!!
      GDML = "";
      makeSandwichDetector(LastConfig);
    }
  return fOK;
}

bool DetectorClass::processGDML()   //if failed, caller have to do: GeoManager = new TGeoManager();GDML = ""; then reconstruct Sandwich Detector
{
  ErrorString = "";
  //have to delete old GeoManager
  if (GeoManager)
    {
      delete GeoManager;
      GeoManager = 0;
    }
  //making tmp file to be used by root
  QFile file("tmpGDML.gdml");
  file.open(QIODevice::WriteOnly);
  if(!file.isOpen())
    {
      ErrorString = "Cannot open file for writing";
      return false;
    }
  QTextStream out(&file);
  out << GDML;
  file.close();
  //importing file
  GeoManager = TGeoManager::Import("tmpGDML.gdml");
  //checking for the result
  if (!GeoManager)
    {
      ErrorString = "Failed to import GDML file!";
      return false;
    }
  qDebug() << "--> GeoManager loaded from GDML file";
  //GeoManager->AddNavigator(); /// *** !!!

  //check validity
  top = GeoManager->GetTopVolume();
    //Correcting NodeNumbers so the PM hodes are properly indexed:
    //after gdml file is loaded, NodeNumber field is set to the number of the volume (0th will be always Top!)
  int totNodes = top->GetNdaughters();
  // qDebug()<<totNodes;
  bool flagFound = false;
  int offsetNumber;
  for (int i=0; i<totNodes; i++)
    {
      TGeoNode* thisNode = (TGeoNode*)top->GetNodes()->At(i);
      QString nodename = top->GetNodes()->At(i)->GetName();
      if (nodename.startsWith("PM"))
        {
          flagFound = true;
          offsetNumber =  thisNode->GetNumber();
          break;
        }
    }
  if (flagFound)
    {
      //        qDebug()<<"first declared PMs has number of: "<<offsetNumber;
      int counter = 0;
      for (int i=0; i<totNodes; i++)
        {
          TGeoNode* thisNode = (TGeoNode*)top->GetNodes()->At(i);
          QString nodename = top->GetNodes()->At(i)->GetName();
          //            qDebug()<<"-------------loaded:"<<nodename;
          thisNode->GetVolume()->SetTransparency(0);
          thisNode->GetVolume()->SetLineColor(kGray);
          if (nodename.startsWith("PM"))
            {
              thisNode->SetNumber(counter);
              //updating cordinates for this PM
              double Local[3] = {0,0,0};
              double Master[3];
              thisNode->LocalToMaster(Local, Master);
              //qDebug()<<counter<<Master[0]<<Master[1]<<Master[2];
              if (counter > PMs->count()-1)
                {
                  ErrorString = "Too many PMs in the GDML file";
                  return false;
                }
              PMs->at(counter++).setCoords(Master);
              thisNode->GetVolume()->SetLineColor(kGreen);
            }
        }
      //qDebug()<<"In this GDML found "<<counter<<" PMs";
      if (counter != PMs->count())
        {
          ErrorString = "Too few PMs in the GDML file";
          return false;
        }
    }
  else
    {
      ErrorString = "No PMs were found in the geometry file!";
      return false;
    }

  //checking materials
  TObjArray* list = GeoManager->GetListOfVolumes();
  int size = list->GetEntries();
  for (int i=0; i<size; i++)
    {
      TGeoVolume* vol = (TGeoVolume*)list->At(i);
      if (!vol) break;
      QString Vname = vol->GetName();
      QString MatName = vol->GetMaterial()->GetName();
      //        qDebug()<<"Volume: "<<Vname<<" Material name:"<<MatName;
      if (Vname.startsWith("PM")) vol->SetLineColor(kGreen);
      else if (Vname.startsWith("Sp")) vol->SetLineColor(kYellow);
      else if (Vname.startsWith("PrScint")) vol->SetLineColor(kRed);
      else if (Vname.startsWith("SecScint")) vol->SetLineColor(kMagenta);
      else if (Vname.startsWith("UpWin")) vol->SetLineColor(kBlue);
      else if (Vname.startsWith("LowWin")) vol->SetLineColor(kBlue);
      else if (Vname.startsWith("Mask")) vol->SetLineColor(40);
      else if (Vname.startsWith("MaskHole")) vol->SetLineColor(40);
      else if (Vname.startsWith("GuideUp")) vol->SetLineColor(38);
      else if (Vname.startsWith("GuideLow")) vol->SetLineColor(38);
      else if (Vname.startsWith("UpperOpening")) vol->SetLineColor(38);
      else if (Vname.startsWith("LowerOpening")) vol->SetLineColor(38);
      else vol->SetLineColor(kGray);
      vol->SetTransparency(0);

      bool foundMaterial = false;
      for (int imat =0; imat<MaterialCollection->size(); imat++)
        {
          QString name = (*MaterialCollection)[imat]->name;
          if (name == MatName)
            {
              foundMaterial = true;
              vol->GetMaterial()->SetIndex(imat);
              //                qDebug()<<"Found! mat id="<<imat;
              break;
            }
        }
      if (!foundMaterial)
        {
          ErrorString = "Found node ("+ Vname+") with unknown material: "+MatName;
          return false;
        }
    }
  //all is fine!
  checkSecScintPresent();
  return true;
}

void DetectorClass::checkSecScintPresent()
{
  fSecScintPresent = false;
  TObjArray* list = GeoManager->GetListOfVolumes();
  int size = list->GetEntries();
  for (int i=0; i<size; i++)
    {
      TGeoVolume* vol = (TGeoVolume*)list->At(i);
      if (!vol) break;
      QString name = vol->GetName();
      if (name == "SecScint")
        {
          fSecScintPresent = true;
          break;
        }
    }
}

void DetectorClass::colorVolumes(int scheme, int id)
{
  //scheme = 0 - default, according to the name
  //scheme = 1 - by material
  //scheme = 2 - medium with id will be red, the rest - black
  TObjArray* list = GeoManager->GetListOfVolumes();
  int size = list->GetEntries();
  for (int i=0; i<size; i++)
    {
      TGeoVolume* vol = (TGeoVolume*)list->At(i);
      if (!vol) break;
      QString name = vol->GetName();
      //      qDebug()<<"name: "<<name<<" medium name: "<<vol->GetMedium()->GetName()<<" med index: "<<vol->GetMedium()->GetId();
      switch (scheme)
        { //default color volumes using name
        case 0:
          vol->SetLineColor(kGray);
          if (name.startsWith("Top")) vol->SetLineColor(kGreen);
          if (name.startsWith("PM")) vol->SetLineColor(kGreen);
          if (name.startsWith("dPM")) vol->SetLineColor(30);
          if (name.startsWith("Sp")) vol->SetLineColor(28);
          if (name.startsWith("PrScint")) vol->SetLineColor(kRed);
          if (name.startsWith("SecScint")) vol->SetLineColor(kMagenta);
          if (name.startsWith("UpWin")) vol->SetLineColor(kBlue);
          if (name.startsWith("LowWin")) vol->SetLineColor(kBlue);
          if (name.startsWith("Mask")) vol->SetLineColor(40);
          if (name.startsWith("MaskHole")) vol->SetLineColor(40);
          if (name.startsWith("LGu")) vol->SetLineColor(38);
          if (name.startsWith("LGl")) vol->SetLineColor(38);
          if (name.startsWith("UOp")) vol->SetLineColor(38);
          if (name.startsWith("LOp")) vol->SetLineColor(38);
          break;
        case 1:
          { //color by material
            int MatId = vol->GetMaterial()->GetIndex();
            vol->SetLineColor(MatId+1);
            break;
          }
        case 2:  //highlight a given material
          if (vol->GetMaterial()->GetIndex() == id) vol->SetLineColor(kRed);
          else vol->SetLineColor(kBlack);
        }
  }
}

int DetectorClass::pmCount() const
{
  return PMs->count();
}

void DetectorClass::findPM(int ipm, int &ul, int &index)
{
  if (ipm<PMarrays[0].PositionsAnglesTypes.size())
    {
      ul=0;
      index = ipm;
      return;
    }
  index = ipm - PMarrays[0].PositionsAnglesTypes.size();
  if (index<PMarrays[1].PositionsAnglesTypes.size())
    {
      ul=1;
      return;
    }
  index = -1; //bad ipm
  return;
}

void LightguidePropertiesStructure::writeToJson(QJsonObject &json)
{
  QJsonObject js;
  js["Active"] = fActive;
  if (fActive)
    {
      js["Autoshape"] = fAutoshape;
      if (!fAutoshape)
        {
          js["Shape"] = Shape;
          if (Shape==2) js["Sides"] = Sides;
          js["SizeX"] = SizeX;
          if (Shape==0) js["SizeY"] = SizeY;
          if (Shape!=1) js["Phi"] = Phi;
        }
      js["Thickness"] = Thickness;
      js["OpeningsShape"] = OpeningsShape;
      js["OpeningsSize"] = OpeningsSize;
      js["Cone"] = fCone;
      if (fCone) js["ConeFactor"] = ConeFactor;
      js["BulkMaterial"] = BulkMaterial;
      js["OpeningsMaterial"] = OpeningsMaterial;
      js["Angle"] = Angle;
    }
  json["Lightguide"] = js;
}

bool LightguidePropertiesStructure::readFromJson(QJsonObject &json)
{
  if (!json.contains("Lightguide"))
    {
      qWarning() << "Lightguide data are not found in json!";
      return false;
    }
  QJsonObject js = json["Lightguide"].toObject();
  parseJson(js,"Active", fActive);
  parseJson(js,"Autoshape", fAutoshape);
  parseJson(js,"Shape", Shape);
  parseJson(js,"Sides", Sides);
  parseJson(js,"SizeX", SizeX);
  parseJson(js,"SizeY", SizeY);
  parseJson(js,"Phi", Phi);
  parseJson(js,"Thickness", Thickness);
  parseJson(js,"OpeningsShape", OpeningsShape);
  parseJson(js,"OpeningsSize", OpeningsSize);
  parseJson(js,"Cone", fCone);
  parseJson(js,"ConeFactor", ConeFactor);
  parseJson(js,"BulkMaterial", BulkMaterial);
  parseJson(js,"OpeningsMaterial", OpeningsMaterial);
  parseJson(js,"Angle", Angle);
  return true;
}

void AdditionalGeoObjectClass::init(QString name)
{
  Name = name;
  Container = "Top";
  Material = 0;
  Shape = 0;
  Size[0]=10.0; Size[1]=10.0; Size[2]=10.0;
  R[0]=0; R[1]=0; R[2]=0;
  Angles[0]=0; Angles[1]=0; Angles[2]=0;
  fActivated = false;

  Arb8height = 10;
  Vertices[0][0] = 0; Vertices[0][1] = 0;
  Vertices[1][0] = 0; Vertices[1][1] = 20;
  Vertices[2][0] = 10; Vertices[2][1] = 20;
  Vertices[3][0] = 10; Vertices[3][1] = 0;

  Vertices[4][0] = 0; Vertices[4][1] = 0;
  Vertices[5][0] = 0; Vertices[5][1] = 10;
  Vertices[6][0] = 10; Vertices[6][1] = 10;
  Vertices[7][0] = 10; Vertices[7][1] = 0;
}

bool AdditionalGeoObjectClass::isOverlapping(const AdditionalGeoObjectClass *o)
{
  if (!fActivated || !o->fActivated) return false;
  if (Container != o->Container) return false;
  if (R[0]!=o->R[0] || R[1]!=o->R[1] || R[2]!=o->R[2]) return false;
  if (Angles[0]!=o->Angles[0] || Angles[1]!=o->Angles[1] || Angles[2]!=o->Angles[2]) return false;
  if (Shape != o->Shape) return false;
  if (Shape == 4)
    {
      if (Arb8height != o->Arb8height) return false;
      for (int i8=0; i8<8; i8++)
       for (int j=0; j<2; j++)
        {
          if (Vertices[i8][j] != o->Vertices[i8][j]) return false;
        }
    }
  else
    {
      if (Size[0]!=o->Size[0] || Size[1]!=o->Size[1] || Size[2]!=o->Size[2]) return false;
    }

  return true;
}

void AdditionalGeoObjectClass::writeToJson(QJsonObject &json)
{
  QJsonObject js;
  js["Name"] = Name;
  js["Activated"] = fActivated;
  js["Container"] = Container;
  js["Material"] = Material;
  QJsonArray arR;
  arR.append(R[0]);
  arR.append(R[1]);
  arR.append(R[2]);
  js["Position"] = arR;
  QJsonArray arA;
  arA.append(Angles[0]);
  arA.append(Angles[1]);
  arA.append(Angles[2]);
  js["Orientation"] = arA;

  js["Shape"] = Shape;
  if (Shape == 4)
    { //4-Arb8
      js["Arb8height"] = Arb8height;
      QJsonArray ar8;
      for (int i8=0; i8<8; i8++)
        {
         QJsonArray ar2;
         ar2.append(Vertices[i8][0]);
         ar2.append(Vertices[i8][1]);
         ar8.append(ar2);
        }
      js["Vertices"] = ar8;
    }
  else
    { //0-box, 1-cylinder, 2-cone, 3-sphere
      QJsonArray ar;
      ar.append(Size[0]);
      ar.append(Size[1]);
      ar.append(Size[2]);
      js["Size"] = ar;
    }
  if (Shape == 5) js["Sides"] = Sides;
  json["AddGeoObj"] = js;
}

bool AdditionalGeoObjectClass::readFromJson(QJsonObject &json)
{
  if (!json.contains("AddGeoObj"))
    {
      qWarning()<<"json does not contain data for AddGeoObject!";
      return false;
    }
  QJsonObject js = json["AddGeoObj"].toObject();
  parseJson(js, "Name", Name);
  parseJson(js, "Activated", fActivated);
  parseJson(js, "Container", Container);
  parseJson(js, "Material", Material);
  QJsonArray arR;
  parseJson(js, "Position", arR);
  R[0] = arR[0].toDouble();
  R[1] = arR[1].toDouble();
  R[2] = arR[2].toDouble();
  QJsonArray arA;
  parseJson(js, "Orientation", arA);
  Angles[0] = arA[0].toDouble();
  Angles[1] = arA[1].toDouble();
  Angles[2] = arA[2].toDouble();
  parseJson(js, "Shape", Shape);
  Sides = 3;
  parseJson(js, "Sides", Sides);
  parseJson(js, "Arb8height", Arb8height);
  if (js.contains("Vertices"))
    {
      QJsonArray ar8 = js["Vertices"].toArray();
      if (ar8.size()<8)
        {
          qWarning()<<"Error in arb8 record in json";
          return false;
        }
      for (int i8=0; i8<8; i8++)
        {
         QJsonArray ar2 = ar8[i8].toArray();
         Vertices[i8][0] = ar2[0].toDouble();
         Vertices[i8][1] = ar2[1].toDouble();
        }
    }
  if (js.contains("Size"))
    {
      QJsonArray ar;
      parseJson(js, "Size", ar);
      Size[0] = ar[0].toDouble();
      Size[1] = ar[1].toDouble();
      Size[2] = ar[2].toDouble();
    }
  return true;
}

MaskClass::MaskClass()
{
  fActive = true;
  Shape = 0;
  Size = 60.0;
  Thickness = 3.0;
  Container = "Top";
  OriginX = OriginY = OriginZ = 0;
  Phi = Theta = Psi = 0;
  MaterialBulk = MaterialHoles = 0;
  MaskScript = "";
}

void MaskElementStructure::writeToJson(QJsonObject &json)
{
  QJsonObject js;
  js["DX"] = DX;
  js["DY"] = DY;
  js["Type"] = Type;
  if (Type == 0) //0-box, 1-tube
    {
      js["SizeX"] = SizeX;
      js["SizeY"] = SizeY;
      js["Angle"] = Angle;
    }
  else js["Radius"] = Radius;
  json["MaskElement"] = js;
}

bool MaskElementStructure::readFromJson(QJsonObject &json)
{
  if (!json.contains("MaskElement"))
    {
      qWarning() << "It is not a mask element json!";
      return false;
    }
  QJsonObject js = json["MaskElement"].toObject();
  parseJson(js, "DX", DX);
  parseJson(js, "DY", DY);
  parseJson(js, "Type", Type);
  parseJson(js, "SizeX", SizeX);
  parseJson(js, "SizeY", SizeY);
  parseJson(js, "Angle", Angle);
  parseJson(js, "Radius", Radius);
  return true;
}

void MaskClass::writeToJson(QJsonObject &json)
{
  QJsonObject js;
  js["Active"] = fActive;
  js["Shape"] = Shape;
  js["Size"] = Size;
  js["Thickness"] = Thickness;
  js["Container"] = Container;
  js["OriginX"] = OriginX;
  js["OriginY"] = OriginY;
  js["OriginZ"] = OriginZ;
  js["Phi"] = Phi;
  js["Theta"] = Theta;
  js["Psi"] = Psi;
  js["MaterialBulk"] = MaterialBulk;
  js["MaterialHoles"] = MaterialHoles;
  js["Script"] = MaskScript;
  QJsonArray ar;
  for (int i=0; i<MaskElements.size(); i++)
    {
      QJsonObject jj;
      MaskElements[i].writeToJson(jj);
      ar.append(jj);
    }
  js["Elements"] = ar;

  json["Mask"] = js;
}

void MaskCollectionClass::append(MaskClass &Mask)
{
  Masks.append(Mask);
}

void MaskCollectionClass::remove(int index)
{
  if (index<0 || index>Masks.size()-1) return;
  Masks.remove(index);
}

void MaskCollectionClass::writeToJson(QJsonObject &json)
{
    QJsonArray ar;
    for (int i=0; i<Masks.size(); i++)
      {
        QJsonObject js;
        Masks[i].writeToJson(js);
        ar << js;
      }
    json["Mask"] = ar;
}

bool MaskCollectionClass::readFromJson(QJsonObject &json)
{
   if (!json.contains("Mask"))
     {
       Masks.clear();
       return true;
     }

   if (json["Mask"].isArray())
     {
       QJsonArray ar = json["Mask"].toArray();
//       qDebug() << "New system. Array size:"<<ar.size();
       Masks.resize(ar.size());
       for (int i=0; i<ar.size(); i++)
         {
           QJsonObject js = ar[i].toObject();
           Masks[i].readFromJson(js);
         }
     }
   else if (json["Mask"].isObject())
     {
//       qDebug() << "Old mask system";
       Masks.resize(1);
       Masks.last().readFromJson(json);
     }
   else return false;

   return true;
}

bool MaskClass::readFromJson(QJsonObject &json)
{
  if (!json.contains("Mask"))
    {
      //qDebug() << "--- json does not contain mask (phantom) data";
      fActive = false;
      return false;
    }
  QJsonObject js = json["Mask"].toObject();
  parseJson(js, "Active", fActive);
  parseJson(js, "Shape", Shape);
  parseJson(js, "Size", Size);
  Container = "Top"; //compatibility
  parseJson(js, "Container", Container);
  parseJson(js, "Thickness", Thickness);
  parseJson(js, "OriginX", OriginX);
  parseJson(js, "OriginY", OriginY);
  parseJson(js, "OriginZ", OriginZ);
  parseJson(js, "Phi", Phi);
  parseJson(js, "Theta", Theta);
  parseJson(js, "Psi", Psi);
  parseJson(js, "MaterialBulk", MaterialBulk);
  parseJson(js, "MaterialHoles", MaterialHoles);
  parseJson(js, "Script", MaskScript);
  MaskElements.clear();
  QJsonArray ar = js["Elements"].toArray();
  for (int i=0; i<ar.size(); i++)
    {
      MaskElements.append(MaskElementStructure());
      QJsonObject jj = ar[i].toObject();
      bool fOK = MaskElements.last().readFromJson(jj);
      if (!fOK) qWarning() << "Error while reading mask element from json!";
    }
  //qDebug() << "--> Loaded a mask (phantom) with"<<MaskElements.size()<<"elements";
  return true;
}

void OpticalGridStructure::writeToJson(QJsonObject &json)
{
  QJsonObject js;
  js["Active"] = fActive;
  if (fActive)
    {
      js["Diameter"] = Diameter;
      js["Pitch"] = Pitch;
      js["Type"] = Type;
      js["WireAngle"] = WireAngle;
      js["Abs"] = Abs;
      js["Container"] = Container;
      js["Z"] = Z;
      js["Shape"] = Shape; // 0 - rectangular, 1 - round, 3 - polygon
      js["SizeX"] = SizeX;
      if (Shape == 0) js["SizeY"] = SizeY;
      if (Shape != 1) js["Phi"] = Phi;
      if (Shape == 2) js["Sides"] = Sides;
    }
  json["OpticalGrid"] = js;
}

bool OpticalGridStructure::readFromJson(QJsonObject &json)
{
  if (!json.contains("OpticalGrid"))
    {
      qWarning() << "It is not an optical grid json!";
      return false;
    }
  QJsonObject js = json["OpticalGrid"].toObject();
  parseJson(js, "Active", fActive);
  parseJson(js, "Diameter", Diameter);
  parseJson(js, "Pitch", Pitch);
  parseJson(js, "Type", Type);
  WireAngle = 0;                //compatibility
  parseJson(js, "WireAngle", WireAngle);
  parseJson(js, "Abs", Abs);

  parseJson(js, "Container", Container);  
  parseJson(js, "Z", Z);
  Shape = 0;// compatibility
  parseJson(js, "Shape", Shape);
  SizeX = 100;                  // compatibility
  parseJson(js, "SizeX", SizeX);
  SizeY = 100;                  // compatibility
  parseJson(js, "SizeY", SizeY);
  Phi = 0;                      // compatibility
  parseJson(js, "Phi", Phi);
  Sides = 6;
  parseJson(js, "Sides", Sides);

  return true;
}
