#include "detectorsandwichclass.h"
#include "classes.h"
#include "functions.h"
#include "detectorclass.h"
#include "materialcollectionclass.h"
#include "pms.h"

#include <QtGui>

#include "TGeoManager.h"
#include "TGeoPcon.h"
#include "TGeoPgon.h"

DetectorSandwichClass::DetectorSandwichClass(DetectorClass *detector)
{
  Detector = detector;

  widthPMs = widthDummyPMs = 0.01;
  Sides = 6;
  DefaultSizeX = 100.0;
  DefaultSizeY = 100.0;

  Layers.resize(9);
  ZeroLayer = 5;
  Layers[0].Name = "Sp1";
  Layers[1].Name = "UpWin";
  Layers[2].Name = "Sp2";
  Layers[3].Name = "SecScint";
  Layers[4].Name = "Sp3";
  Layers[5].Name = "PrScint";
  Layers[6].Name = "Sp4";
  Layers[7].Name = "LowWin";
  Layers[8].Name = "Sp5";
}

DetectorSandwichClass::~DetectorSandwichClass()
{
  //Mediums are destroyed NOT here
}

void DetectorSandwichClass::readFromJson(QJsonObject &json)
{
  QJsonObject jws = json["FixedWorldSizes"].toObject();
  fWorldSizeFixed = false;
  parseJson(jws, "WorldSizeFixed", fWorldSizeFixed);
  if (fWorldSizeFixed)
    {
      parseJson(jws, "XY", WorldSizeXY);
      parseJson(jws, "Z", WorldSizeZ);
    }

  QJsonObject sj = json["SandwichSettings"].toObject();

  parseJson(sj, "UseDefaultSizes", fUseDefaultSizes);
  parseJson(sj, "DefaultSize1", DefaultSizeX);
  parseJson(sj, "DefaultSize2", DefaultSizeY);
  parseJson(sj, "Shape", Shape);
  Sides = 6;
  parseJson(sj, "Sides", Sides);
  Angle = 0;
  parseJson(sj, "Angle", Angle);
  parseJson(sj, "TopMaterial", TopMaterial);

  QJsonArray lar = sj["Layers"].toArray();
  if (lar.size() != Layers.size())
    {
      ErrorString = "Mismatch in number of layers!";
      qWarning() << ErrorString;
    }
  for (int i=0; i<lar.size(); i++)
    {
      QJsonObject jo = lar[i].toObject();
      bool fOn = jo["Active"].toBool();
      Layers[i].fActive = fOn;
      if (fOn)
        {
          parseJson(jo, "Height", Layers[i].Height);
          parseJson(jo, "Material", Layers[i].Material);
          parseJson(jo, "Size1", Layers[i].Size1);
          parseJson(jo, "Size2", Layers[i].Size2);
          parseJson(jo, "Shape", Layers[i].Shape);
          parseJson(jo, "Sides", Layers[i].Sides);
          parseJson(jo, "Angle", Layers[i].Angle);
        }
    }
}

void DetectorSandwichClass::constructDetector()
{
  //qDebug() << "--->Creating detector sandwich";
  //qDebug() << "--> Deleting old GeoManager and creating new";
  if (Detector->GeoManager)
    {
      //gGeoManager = 0;

      //   if (Detector->GeoManager->GetListOfNavigators()) qDebug() << "-------------Navigators listed in GeoManager"<< Detector->GeoManager->GetListOfNavigators()->GetEntries();
      //   else qDebug() << "------------- GeoManager did not have navigators";

      delete Detector->GeoManager;
     }
  Detector->GeoManager = new TGeoManager();
  //Detector->GeoManager->AddNavigator(); /// *** !!!
  Detector->GeoManager->SetVerboseLevel(0);

  //qDebug() << "--> Creating materials and media";
  for (int i=0; i<Detector->MaterialCollection->size(); i++)
    {
      QString tmpStr = (*Detector->MaterialCollection)[i]->name;
      QByteArray ba = tmpStr.toLocal8Bit();
      char *cname = ba.data();
      (*Detector->MaterialCollection)[i]->GeoMat = new TGeoMaterial(cname, (*Detector->MaterialCollection)[i]->p1, (*Detector->MaterialCollection)[i]->p2, (*Detector->MaterialCollection)[i]->p3);
      (*Detector->MaterialCollection)[i]->GeoMed = new TGeoMedium (cname, i, (*Detector->MaterialCollection)[i]->GeoMat);
    }
  //qDebug() << "--> Calculating Z positions of layers";
  calculateLayersZpositions();
  //qDebug() << "--> Populating PMs module with individual PMs";
  populatePMs();
  //for (int ul=0; ul<2; ul++)
  //  for (int i=0; i<Detector->PMarrays[ul].PositionsAnglesTypes.size(); i++)
  //    qDebug() << ul<<i<<"xyz:"<<Detector->PMarrays[ul].PositionsAnglesTypes[i].x<<Detector->PMarrays[ul].PositionsAnglesTypes[i].y<<Detector->PMarrays[ul].PositionsAnglesTypes[i].z;
  //for (int i=0; i<Detector->PMs->count(); i++)
  //  qDebug() << i<<Detector->PMs->at(i).upperLower<<"xyz:"<<Detector->PMs->at(i).x<<Detector->PMs->at(i).y<<Detector->PMs->at(i).z;

  if (fWorldSizeFixed)
    {
      //qDebug() << "--> Using user-defined world size";
    }
  else
    {
      //qDebug() << "--> Calculating world size";
      WorldSizeXY = 0;
      WorldSizeZ = 0;
      updateWorldSize(WorldSizeXY, WorldSizeZ);
      WorldSizeXY *= 1.05; WorldSizeZ *= 1.05; //adding margings
    }
  //qDebug() << "    World size XYmax:"<<WorldSizeXY<<"Zmax:"<<WorldSizeZ;

  //qDebug() << "--> Creating top volume";
  Detector->top = Detector->GeoManager->MakeBox("Top", (*Detector->MaterialCollection)[TopMaterial]->GeoMed, WorldSizeXY, WorldSizeXY, WorldSizeZ);
  //Detector->top->SetLineColor(kGreen);
  Detector->GeoManager->SetTopVolume(Detector->top);
  Detector->GeoManager->SetTopVisible(true);
  //qDebug() << "--> Positioning sandwich layers";
//  TGeoVolume* PrScintContainer = 0;
  for (int i=0; i<Layers.size(); i++)
    {
      if (Layers[i].fActive)
        {
          TString name = Layers[i].Name;
          if (name == "PrScint" && Detector->fPrScintCont) name = "PrScCont";
          TGeoVolume* tmp = generateVolume(name, (*Detector->MaterialCollection)[Layers[i].Material]->GeoMed,
                                           Layers[i].Shape, 0.5*Layers[i].Size1, 0.5*Layers[i].Size2, 0.5*Layers[i].Height, Layers[i].Sides);
          //Detector->top->AddNode(tmp, 0, new TGeoTranslation(0, 0, Layers[i].Z));
//          if (name == "PrScCont") PrScintContainer = tmp;
          TGeoRotation *lRot = new TGeoRotation("lRot", 0, 0, Layers[i].Angle);
          lRot->RegisterYourself();
          TGeoCombiTrans *lTrans = new TGeoCombiTrans("lTrans", 0, 0, Layers[i].Z, lRot);
          Detector->top->AddNode(tmp, 0, lTrans);
        }
    }

//  //PrScint in PrScintCont
//  if (Detector->PrScintCont.fActive)
//    {
//      TGeoVolume* PrScint = generateVolume( "PrScint", (*Detector->MaterialCollection)[Detector->PrScintCont.Material]->GeoMed,
//                                              Detector->PrScintCont.Shape, 0.5*Detector->PrScintCont.SizeX, 0.5*Detector->PrScintCont.SizeY, 0.5*Detector->PrScintCont.SizeZ, Detector->PrScintCont.Sides);
//      TGeoRotation *psRot = new TGeoRotation("psRot", Detector->PrScintCont.Phi, Detector->PrScintCont.Theta, Detector->PrScintCont.Psi);
//      psRot->RegisterYourself(); //***
//      TGeoCombiTrans *psTrans = new TGeoCombiTrans("psTrans", Detector->PrScintCont.OriginX, Detector->PrScintCont.OriginY, Detector->PrScintCont.OriginZ, psRot);
//      PrScintContainer->AddNode(PrScint, 0, psTrans);
//    }

  //qDebug() << "--> Positioning PMs";
  positionPMs();
  if (Detector->PMdummies.size()>0)
    {
      //qDebug() << "--> Positioning dummy PMs";
      positionDummies();
    }
  if (Detector->Lightguides[0].fActive || Detector->Lightguides[1].fActive)
    {
      //qDebug() << "--> Positioning lightguides";
      positionLightGuides();
    }
  if (Detector->fAdditionalGeoObjectsEnabled && Detector->AdditionalGeoObjects.size()>0)
    {
      //qDebug() << "--> Positioning additional objects";
      positionAddObjects();
    }
  if (!Detector->Masks.isEmpty())
    {
      //qDebug() << "--> Positioning mask";
      positionMasks();
    }
  if (Detector->Grids.size()>0)
    {
      //qDebug() << "--> Positioning grids";
      positionGrids();
    }
  //qDebug() << "--> Closing geometry";
  Detector->GeoManager->CloseGeometry();
  //qDebug() << "===> All done!";
}

void DetectorSandwichClass::calculateLayersZpositions()
{
  Layers[ZeroLayer].Z = 0;
  UpperEdge = 0.5*Layers[ZeroLayer].Height;
  for (int i=ZeroLayer-1; i>-1; i--)
    {
      if (Layers[i].fActive)
        {
          Layers[i].Z = UpperEdge + 0.5*Layers[i].Height;
          //qDebug() << i << Layers[i].Z;
          UpperEdge += Layers[i].Height;
        }
    }
  LowerEdge = -0.5*Layers[ZeroLayer].Height;
  for (int i=ZeroLayer+1; i<Layers.size(); i++)
    {
      if (Layers[i].fActive)
        {
          Layers[i].Z = LowerEdge - 0.5*Layers[i].Height;
          LowerEdge -= Layers[i].Height;
        }
    }
    //special layers - lightguides
  if (Detector->Lightguides[0].fActive)
    {
      LightguidePropertiesStructure& lg = Detector->Lightguides[0];
      if (lg.fAutoshape)
        {
          for (int i=Layers.size()-1; i>-1; i--)
            if (Layers[i].fActive)
              {  //assuming for now it is the lightguide neighbor
                lg.Shape = Layers[i].Shape;
                lg.Sides = Layers[i].Sides;
                lg.SizeX = 0.5*Layers[i].Size1;
                lg.SizeY = 0.5*Layers[i].Size2;
                lg.Phi   = Layers[i].Angle;
              }
        }
      lg.Z = UpperEdge + 0.5*lg.Thickness;
      UpperEdge += lg.Thickness;
    }
  if (Detector->Lightguides[1].fActive)
    {
      LightguidePropertiesStructure& lg = Detector->Lightguides[1];
      if (lg.fAutoshape)
        {
          for (int i=0; i<Layers.size(); i++)
            if (Layers[i].fActive)
              {  //assuming for now it is the lightguide neighbor
                lg.Shape = Layers[i].Shape;
                lg.Sides = Layers[i].Sides;
                lg.SizeX = 0.5*Layers[i].Size1;
                lg.SizeY = 0.5*Layers[i].Size2;
                lg.Phi   = Layers[i].Angle;
                //qDebug() << i<<">"<<lg.SizeX;
              }
        }
      lg.Z = LowerEdge - 0.5*lg.Thickness;
      LowerEdge -= lg.Thickness;
    }
}

TGeoVolume *DetectorSandwichClass::generateVolume(const char *Name, TGeoMedium *Medium, int Shape, Double_t SizeX, Double_t SizeY, Double_t SizeZ, int Sides)
{
  //Shape 0 -  box, 1 - cylinder, 2 - polygon
  switch (Shape)
    {
    case 0: return Detector->GeoManager->MakeBox (Name, Medium, SizeX, SizeY, SizeZ); //box
    case 1: return Detector->GeoManager->MakeTube(Name, Medium, 0, SizeX, SizeZ);    //tube
    case 2:
      { //polygon
        TGeoVolume* tgv = Detector->GeoManager->MakePgon(Name, Medium, 0, 360.0, Sides, 2);
        ((TGeoPcon*)tgv->GetShape())->DefineSection(0, -SizeZ, 0, SizeX);
        ((TGeoPcon*)tgv->GetShape())->DefineSection(1, +SizeZ, 0, SizeX);
        return tgv;
      }
    default:
      ErrorString = "Error: unrecognized volume type!";
      qWarning() << ErrorString;
      return 0;
    }
}

void DetectorSandwichClass::populatePMs()
{
  Detector->PMs->clear();
  for (int ul=0; ul<2; ul++)
    {
      if (Detector->PMarrays[ul].fActive)
        {
          // calculate XY positions for regular array and create the record with positions (otherwise already there)
          if (Detector->PMarrays[ul].Regularity == 0) calculatePmsXY(ul);
          // calculate Z position for regular and auto-z array
          if (Detector->PMarrays[ul].Regularity < 2)
            {
              double Z = (ul == 0) ? UpperEdge+widthPMs : LowerEdge-widthPMs;
              for (int ipm=0; ipm<Detector->PMarrays[ul].PositionsAnglesTypes.size(); ipm++)
                Detector->PMarrays[ul].PositionsAnglesTypes[ipm].z = Z;
            }
          //registering PMTs in PMs module
          for (int ipm=0; ipm<Detector->PMarrays[ul].PositionsAnglesTypes.size(); ipm++)
              Detector->PMs->add(ul, &Detector->PMarrays[ul].PositionsAnglesTypes[ipm]);
        }
    }
}

void DetectorSandwichClass::positionPMs()
{
  //creating volumes for each type of PMs
  int numPMtypes = Detector->PMs->countPMtypes();
  //qDebug()<<"Defined PM types: "<<PMtypes;
  QVector<TGeoVolume*> pmTypes;
  pmTypes.resize(numPMtypes);
  for (int itype=0; itype<numPMtypes; itype++)
    {
      QString str = "PM" + QString::number(itype);
      QByteArray ba = str.toLocal8Bit();
      char *name = ba.data();
      PMtypeClass *tp = Detector->PMs->getType(itype);
      pmTypes[itype] = generateVolume(name, (*Detector->MaterialCollection)[tp->MaterialIndex]->GeoMed,
          tp->Shape, 0.5*tp->SizeX, 0.5*tp->SizeY, widthPMs, 6);
      pmTypes[itype]->SetLineColor(kGreen);
    }

  int index = 0;
  for (int ul=0; ul<2; ul++)
    {
      if (Detector->PMarrays[ul].fActive)
        {
          //adding PMTs to GeoManager            
          for (int i=0; i<Detector->PMarrays[ul].PositionsAnglesTypes.size(); i++)
            {
              const PosAngTypeStructure *PM = &Detector->PMarrays[ul].PositionsAnglesTypes.at(i);
              bool hex = (Detector->PMs->getType(PM->type)->Shape == 2);
              TGeoRotation *pmRot = new TGeoRotation("pmRot", PM->phi, PM->theta, PM->psi+(hex ? 90.0 : 0.0));
              pmRot->RegisterYourself(); //experiment ***
              TGeoCombiTrans *pmTrans = new TGeoCombiTrans("pmTrans", PM->x, PM->y, PM->z, pmRot);
              Detector->top->AddNode(pmTypes[PM->type], index, pmTrans);
              index++; //need unique index (lower continues after upper)
            }
        }
    }  
}

void DetectorSandwichClass::calculatePmsXY(int ul)
{
  Detector->PMarrays[ul].PositionsAnglesTypes.clear(); //since its regular, there is no data there

  bool hexagonal = (Detector->PMarrays[ul].Packing == 0) ? false : true;
  int iX = Detector->PMarrays[ul].NumX;
  int iY = Detector->PMarrays[ul].NumY;
  double CtC = Detector->PMarrays[ul].CenterToCenter;
  double CtCbis = CtC*cos(30.*3.14159/180.);
  int rings = Detector->PMarrays[ul].NumRings;
  bool XbyYarray = !Detector->PMarrays[ul].fUseRings;

  double x, y;
  QVector<PosAngTypeStructure> *pat = &Detector->PMarrays[ul].PositionsAnglesTypes;

  if (XbyYarray)
    { //X by Y type of array
      if (hexagonal)
        {
          for (int j=iY-1; j>-1; j--)
            for (int i=0; i<iX; i++)
              {
                x = CtC*( -0.5*(iX-1) +i  + 0.5*(j%2));
                y = CtCbis*(-0.5*(iY-1) +j);
                pat->append( PosAngTypeStructure(x,y,0, 0,0,0, Detector->PMarrays[ul].PMtype) );                
              }
        }
      else
        {
          for (int j=iY-1; j>-1; j--)
            for (int i=0; i<iX; i++)
              {
                x = CtC*(-0.5*(iX-1) +i);
                y = CtC*(-0.5*(iY-1) +j);
                pat->append( PosAngTypeStructure(x,y,0, 0,0,0, Detector->PMarrays[ul].PMtype) );                
              }
        }
    }
  else
    { //rings array
      if (hexagonal)
        {
          pat->append( PosAngTypeStructure(0,0,0, 0,0,0, Detector->PMarrays[ul].PMtype) );

          for (int i=1; i<rings+1; i++) //every new ring adds 6i PMs
            {
              x = i*CtC;
              y = 0;
              pat->append( PosAngTypeStructure(x,y,0, 0,0,0, Detector->PMarrays[ul].PMtype) );
              for (int j=1; j<i+1; j++)
                {  //   //
                  x = pat->last().x - 0.5*CtC;
                  y = pat->last().y - CtCbis;
                  Detector->PMarrays[ul].PositionsAnglesTypes.append( PosAngTypeStructure(x,y,0, 0,0,0, Detector->PMarrays[ul].PMtype) );
                }
              for (int j=1; j<i+1; j++)
                {   // --
                  x = pat->last().x - CtC;
                  y = pat->last().y;
                  Detector->PMarrays[ul].PositionsAnglesTypes.append( PosAngTypeStructure(x,y,0, 0,0,0, Detector->PMarrays[ul].PMtype) );
                }
              for (int j=1; j<i+1; j++)
                {  // \\ //
                  x = pat->last().x - 0.5*CtC;
                  y = pat->last().y + CtCbis;
                  //PMs->insert(ipos, ul,x,y,z,Psi,typ);
                  Detector->PMarrays[ul].PositionsAnglesTypes.append( PosAngTypeStructure(x,y,0, 0,0,0, Detector->PMarrays[ul].PMtype) );
                }
              for (int j=1; j<i+1; j++)
                {  // //
                  x = pat->last().x + 0.5*CtC;
                  y = pat->last().y + CtCbis;
                  Detector->PMarrays[ul].PositionsAnglesTypes.append( PosAngTypeStructure(x,y,0, 0,0,0, Detector->PMarrays[ul].PMtype) );
                }
              for (int j=1; j<i+1; j++)
                {   // --
                  x = pat->last().x + CtC;
                  y = pat->last().y;
                  Detector->PMarrays[ul].PositionsAnglesTypes.append( PosAngTypeStructure(x,y,0, 0,0,0, Detector->PMarrays[ul].PMtype) );
                }
              for (int j=1; j<i; j++)
                {  // \\       //dont do the last step - already positioned PM
                  x = pat->last().x + 0.5*CtC;
                  y = pat->last().y - CtCbis;
                  Detector->PMarrays[ul].PositionsAnglesTypes.append( PosAngTypeStructure(x,y,0, 0,0,0, Detector->PMarrays[ul].PMtype) );
                }
            }
        }
      else {
          //using the same algorithm as X * Y
          iX = rings*2 + 1;
          iY = rings*2 + 1;
          for (int j=0; j<iY; j++)
            for (int i=0; i<iX; i++) {
                x = CtC*(-0.5*(iX-1) +i);
                y = CtC*(-0.5*(iY-1) +j);
                Detector->PMarrays[ul].PositionsAnglesTypes.append( PosAngTypeStructure(x,y,0, 0,0,0, Detector->PMarrays[ul].PMtype) );
              }
        }
    }
}

void DetectorSandwichClass::positionDummies()
{
  //creating all types - possible improvement - create only in use types of dummy pms
  int PMtypes = Detector->PMs->countPMtypes();
  QVector<TGeoVolume*> pmtDummy;
  pmtDummy.resize(PMtypes);
  for (int i=0; i<PMtypes; i++)
    {
      QString str = "dPM" + QString::number(i);
      QByteArray ba = str.toLocal8Bit();
      char *name = ba.data();
      PMtypeClass *tp = Detector->PMs->getType(i);
      pmtDummy[i] = generateVolume(name, (*Detector->MaterialCollection)[tp->MaterialIndex]->GeoMed,
                                          tp->Shape, 0.5*tp->SizeX, 0.5*tp->SizeY, widthDummyPMs, 6);
      pmtDummy[i]->SetLineColor(30);
    }
  //positioning dummies
  for (int idum = 0; idum<Detector->PMdummies.size(); idum++)
    {
      const PMdummyStructure &dum = Detector->PMdummies.at(idum);
      bool hex = (Detector->PMs->getType(dum.PMtype)->Shape == 2);
      TGeoRotation *dummyPmRot = new TGeoRotation("dummypmRot", dum.Angle[0], dum.Angle[1], dum.Angle[2]+(hex ? 90.0 : 0.0));
      dummyPmRot->RegisterYourself(); //***
      TGeoCombiTrans *dummyPmTrans = new TGeoCombiTrans("dummypmTrans", dum.r[0], dum.r[1], dum.r[2], dummyPmRot);
      Detector->top->AddNode(pmtDummy[dum.PMtype], idum, dummyPmTrans);
    }
}

void DetectorSandwichClass::positionLightGuides()
{
  for (int ul=0; ul<2; ul++)
    if (Detector->Lightguides[ul].fActive)
      {
        const LightguidePropertiesStructure &lg = Detector->Lightguides[ul];

        //creating and positioning the bulk container
        TGeoVolume* LightGuide = generateVolume( (ul==0)?"LGu":"LGl", (*Detector->MaterialCollection)[lg.BulkMaterial]->GeoMed,
                                                lg.Shape, lg.SizeX, lg.SizeY, 0.5*lg.Thickness, lg.Sides);
        TGeoRotation *lgRot = new TGeoRotation("lgRot", 0, 0, lg.Phi);
        lgRot->RegisterYourself(); //***
        TGeoCombiTrans *lgTrans = new TGeoCombiTrans("lgTrans", 0, 0, lg.Z, lgRot);  //used below!
        Detector->top->AddNode(LightGuide, 0, lgTrans);

        //creating openings volumes
        TGeoVolume* Opening;        
        if (lg.fCone)
          { //cone: dif sizes
            if (lg.OpeningsShape == 1)
              {
                Opening = Detector->GeoManager->MakePcon((ul==0)?"UOp":"LOp", (*Detector->MaterialCollection)[lg.OpeningsMaterial]->GeoMed, 0, 360.0, 2);
                ((TGeoPcon*)Opening->GetShape())->DefineSection(0, -0.5*lg.Thickness, 0, 0.5*lg.OpeningsSize * ((ul==0)?lg.ConeFactor:1.0) );
                ((TGeoPcon*)Opening->GetShape())->DefineSection(1, +0.5*lg.Thickness, 0, 0.5*lg.OpeningsSize * ((ul==1)?lg.ConeFactor:1.0) );
              }
            else
              {
                Opening = Detector->GeoManager->MakePgon((ul==0)?"UOp":"LOp", (*Detector->MaterialCollection)[lg.OpeningsMaterial]->GeoMed, 0, 360.0, 4, 2);
                ((TGeoPcon*)Opening->GetShape())->DefineSection(0, -0.5*lg.Thickness, 0, 0.5*lg.OpeningsSize * ((ul==0)?lg.ConeFactor:1.0) );
                ((TGeoPcon*)Opening->GetShape())->DefineSection(1, +0.5*lg.Thickness, 0, 0.5*lg.OpeningsSize * ((ul==1)?lg.ConeFactor:1.0) );
              }
          }
        else
          { //same sizes
            if (lg.OpeningsShape == 1)
              Opening = Detector->GeoManager->MakeTube((ul==0)?"UOp":"LOp", (*Detector->MaterialCollection)[lg.OpeningsMaterial]->GeoMed,
                                                                             0, 0.5*lg.OpeningsSize, 0.5*lg.Thickness);
            else
              Opening = Detector->GeoManager->MakeBox((ul==0)?"UOp":"LOp", (*Detector->MaterialCollection)[lg.OpeningsMaterial]->GeoMed,
                                                                             0.5*lg.OpeningsSize, 0.5*lg.OpeningsSize, 0.5*lg.Thickness);
          }

        //adding nodes for openings
        for (int ipm=0; ipm<Detector->PMs->count(); ipm++)
          if (Detector->PMs->at(ipm).upperLower == ul)
            {              
              Double_t master[3];
              master[0] = Detector->PMs->X(ipm);
              master[1] = Detector->PMs->Y(ipm);
              master[2] = 0;
              Double_t local[3];
              lgTrans->MasterToLocal(master, local);
              //TGeoCombiTrans *Trans = new TGeoCombiTrans("Trans", Detector->PMs->X(ipm), Detector->PMs->Y(ipm), 0, Rot);
              TGeoRotation *Rot = new TGeoRotation("Rot", lg.Angle, 0, 0);
              Rot->RegisterYourself();//***
              TGeoCombiTrans *Trans = new TGeoCombiTrans("Trans", local[0], local[1], 0, Rot);
              LightGuide->AddNode(Opening, ipm, Trans);
            }
        //adding nodes for oppenings in front of dummyPMs
        if (Detector->PMdummies.size()>0)
          {
            for (int idpm=0; idpm<Detector->PMdummies.size(); idpm++)
              if (Detector->PMdummies[idpm].UpperLower == ul)
                {
                  Double_t master[3];
                  master[0] = Detector->PMdummies[idpm].r[0];
                  master[1] = Detector->PMdummies[idpm].r[1];
                  master[2] = 0;
                  Double_t local[3];
                  lgTrans->MasterToLocal(master, local);
                  TGeoRotation *Rot = new TGeoRotation("Rot", lg.Angle, 0, 0);
                  Rot->RegisterYourself();//***
                  //TGeoCombiTrans *Trans = new TGeoCombiTrans("Trans", Detector->PMdummies[idpm].r[0], Detector->PMdummies[idpm].r[1], 0, Rot);
                  TGeoCombiTrans *Trans = new TGeoCombiTrans("Trans", local[0], local[1], 0, Rot);
                  LightGuide->AddNode(Opening, idpm, Trans);
                }
          }
      }
}

void DetectorSandwichClass::positionAddObjects()
{
  if (Detector->fAdditionalGeoObjectsEnabled)
   for (int iobj = 0; iobj<Detector->AdditionalGeoObjects.size();iobj++)
    if (Detector->AdditionalGeoObjects[iobj].fActivated)
      {
        //making volume
        AdditionalGeoObjectClass* ob = &Detector->AdditionalGeoObjects[iobj];

        QByteArray ba = ob->Name.toLocal8Bit();
        char *name = ba.data();
        TGeoMedium* medium = (*Detector->MaterialCollection)[ob->Material]->GeoMed;
        TGeoVolume *NewObj;
        switch (ob->Shape)
          {
            case 0: //box
               NewObj = Detector->GeoManager->MakeBox (name, medium, 0.5*ob->Size[0], 0.5*ob->Size[1], 0.5*ob->Size[2]);
               break;
            case 1: //cylinder
               NewObj = Detector->GeoManager->MakeTube(name, medium, 0, 0.5*ob->Size[0], 0.5*ob->Size[2]);
               break;
            case 2: //cone
               NewObj = Detector->GeoManager->MakeCone(name, medium, 0.5*ob->Size[2], 0, 0.5*ob->Size[1], 0, 0.5*ob->Size[0]);
               break;
            case 3: //sphere
               NewObj = Detector->GeoManager->MakeSphere(name, medium, 0, 0.5*ob->Size[0]);
               break;
            case 4: //Arb8
               NewObj = Detector->GeoManager->MakeArb8(name, medium, 0.5*ob->Arb8height, (double*)ob->Vertices);
               break;
            case 5: //TGeoPgone
               {
                 NewObj = Detector->GeoManager->MakePgon(name, medium, 0, 360, ob->Sides, 2);
                 TGeoPgon* obj = dynamic_cast<TGeoPgon*>(NewObj->GetShape());
                 obj->DefineSection(0, -0.5*ob->Size[2], 0, 0.5*ob->Size[1]);
                 obj->DefineSection(1, +0.5*ob->Size[2], 0, 0.5*ob->Size[0]);
               }
               break;
            default:
              qWarning() << "Unknown AddGeoObj shape!";
              continue;
          }
          // qDebug()<<"Addon object"<<iobj<<"name"<<ob->name;

        //container?
        ba = ob->Container.toLocal8Bit();
        char *contName = ba.data();
          // qDebug()<<"Container name:"<<contName;
        TGeoVolume *Container = Detector->GeoManager->GetVolume(contName);
        if (Container)
          {
             // qDebug()<<"Container found!"<<Container->GetName();
            TGeoRotation Rot = TGeoRotation("Rot", ob->Angles[0], ob->Angles[1], ob->Angles[2]);
            TGeoTranslation Translat = TGeoTranslation("Tra", ob->R[0], ob->R[1], ob->R[2]);
            TGeoCombiTrans *Trans = new TGeoCombiTrans(Translat, Rot);
            Container->AddNode(NewObj,0,Trans);            
          }
        else qWarning()<<"Container volume"<<contName<<"not found! Requested by:"<<name;
      }
}

void DetectorSandwichClass::positionMasks()
{
  for (int im=0; im<Detector->Masks.size(); im++)
  if (Detector->Masks[im].fActive)
    {
      MaskClass& m = Detector->Masks[im];
      double MaskSize = 0.5*m.Size;
      double MaskThickness = 0.5*m.Thickness;
      double maskOriginX = m.OriginX;
      double maskOriginY = m.OriginY;
      double maskOriginZ = m.OriginZ;
      double Phi = m.Phi;
      double Theta = m.Theta;
      double Psi = m.Psi;

      TGeoVolume *Mask = generateVolume("Mask", (*Detector->MaterialCollection)[m.MaterialBulk]->GeoMed,
                                                    m.Shape, MaskSize, MaskSize, MaskThickness, Sides);
      QVector<TGeoVolume*> MaskHoles;
      MaskHoles.resize(m.MaskElements.size());
      for (int i=0; i<m.MaskElements.size(); i++)
          {
                //qDebug() << "Mask hole #" << i;
                QString str;
                str.setNum(i);
                str = "MaskHole"+str; //name is "Hole" + index of the type
                QByteArray ba = str.toLocal8Bit();
                char *name = ba.data();
                if (m.MaskElements[i].Type == 1)
                     MaskHoles[i] = Detector->GeoManager->MakeTube(name, (*Detector->MaterialCollection)[m.MaterialHoles]->GeoMed,
                                                                   0, m.MaskElements[i].Radius, MaskThickness);
                else MaskHoles[i] = Detector->GeoManager->MakeBox( name,  (*Detector->MaterialCollection)[m.MaterialHoles]->GeoMed,
                                                                   0.5*m.MaskElements[i].SizeX, 0.5*m.MaskElements[i].SizeY, MaskThickness);
                MaskHoles[i]->SetLineColor(kGray);
                  //positioning elements
                if (m.MaskElements[i].Type == 1)
                   Mask->AddNode(MaskHoles[i], i, new TGeoTranslation(m.MaskElements[i].DX, m.MaskElements[i].DY, 0));
                else
                  {
                    TGeoRotation *rot = new TGeoRotation("holeRot", 0, 0, m.MaskElements[i].Angle);
                    rot->RegisterYourself();//***
                    TGeoCombiTrans *trans = new TGeoCombiTrans("totTrans", m.MaskElements[i].DX, m.MaskElements[i].DY, 0, rot);
                    Mask->AddNode(MaskHoles[i], i, trans);
                  }
          }
      //container
      QByteArray ba = m.Container.toLocal8Bit();
      char *contName = ba.data();
        // qDebug()<<"Container name:"<<contName;
      TGeoVolume *Container = Detector->GeoManager->GetVolume(contName);
      if (Container)
        {
           // qDebug()<<"Container found!"<<contName;
          TGeoRotation *maskRot = new TGeoRotation("maskRot", Phi, Theta, Psi);
          maskRot->RegisterYourself();
          TGeoCombiTrans *maskTrans = new TGeoCombiTrans("maskTrans", maskOriginX, maskOriginY, maskOriginZ, maskRot);
          Container->AddNode(Mask, 0, maskTrans);
        }
      else qWarning()<<"Mask container volume"<<contName<<"not found!";
      //Detector->top->AddNode(Mask, 0, maskTrans);

      Mask->SetLineColor(40);
      Mask->SetVisibility(true);
    }
}

void DetectorSandwichClass::positionGrids()
{
  for (int igrid = 0; igrid<Detector->Grids.size(); igrid++)
    if (Detector->Grids[igrid].fActive)
      {
        //qDebug()<<"Grid-> #"<<igrid;
        OpticalGridStructure* gr = &Detector->Grids[igrid];
          //finding the container volume
        QByteArray ba = gr->Container.toLocal8Bit();
        char *contName = ba.data();
        //qDebug()<<"Grid-> Container name:"<<contName;
        TGeoVolume *Container = Detector->GeoManager->GetVolume(contName);
        if (Container)
          {
            //qDebug()<<"Container found!"<<Container->GetName();
              //assigning name
            QString str;
            str.setNum(igrid);
            str = "GR"+str; //name is "GR" + index
            ba = str.toLocal8Bit();
            char *name = ba.data();
            TGeoVolume *grid = generateVolume(name, Container->GetMedium(), gr->Shape,
                                              0.5*gr->SizeX, 0.5*gr->SizeY, 0.5*gr->Diameter, gr->Sides);

            TGeoRotation *grRot = new TGeoRotation("grRot", gr->Phi, 0, 0);
            grRot->RegisterYourself();
            TGeoCombiTrans *grTrans = new TGeoCombiTrans("grTrans", 0, 0, gr->Z, grRot);
            Container->AddNode(grid, igrid, grTrans);
            //Container->AddNode(grid, igrid, new TGeoTranslation(0, 0, gr->Z));
          }
        else qWarning()<<"Grid: Container volume"<<contName<<"not found!";
      }
}

void SandwichLayerClass::updateWorldSize(double &XYm) //Z is updated later - use Edges
{
  updateMax(XYm, Size1);
  if (Shape == 0) updateMax(XYm, Size2);
}

void DetectorSandwichClass::updateWorldSize(double &XYm, double &Zm)
{
  //sandwich layers + lightguides
  for (int il=0; il<Layers.size(); il++)
    if (Layers[il].fActive) Layers[il].updateWorldSize(XYm);
  updateMax(Zm, fabs(UpperEdge));
  updateMax(Zm, fabs(LowerEdge));

  //PMs
  for (int ipm=0; ipm<Detector->PMs->count(); ipm++)
    {
      double msize = 0.5*widthPMs;
      updateMax(msize, 0.5*Detector->PMs->SizeX(ipm));
      updateMax(msize, 0.5*Detector->PMs->SizeY(ipm));

      updateMax(XYm, fabs(Detector->PMs->X(ipm)) + msize);
      updateMax(XYm, fabs(Detector->PMs->Y(ipm)) + msize);
      updateMax(Zm,  fabs(Detector->PMs->Z(ipm)) + msize);
    }

  //DummyPMs
  for (int i=0; i<Detector->PMdummies.size(); i++)
    {
      double msize = 0.5*widthDummyPMs;
      updateMax(msize, 0.5*Detector->PMs->getType(Detector->PMdummies[i].PMtype)->SizeX);
      updateMax(msize, 0.5*Detector->PMs->getType(Detector->PMdummies[i].PMtype)->SizeY);

      updateMax(XYm, fabs(Detector->PMdummies[i].r[0]) + msize);
      updateMax(XYm, fabs(Detector->PMdummies[i].r[1]) + msize);
      updateMax(Zm,  fabs(Detector->PMdummies[i].r[2]) + msize);
    }

  //AddGeoObj
  if (Detector->AdditionalGeoObjects.size()>0 && Detector->fAdditionalGeoObjectsEnabled)
    for (int iob = 0; iob<Detector->AdditionalGeoObjects.size(); iob++)
      if (Detector->AdditionalGeoObjects[iob].fActivated)
        if (Detector->AdditionalGeoObjects[iob].Container == "Top")
          {
            AdditionalGeoObjectClass* ob = &Detector->AdditionalGeoObjects[iob];
            double msize = 0.5*ob->Size[0];
            updateMax(msize, 0.5*ob->Size[1]);
            updateMax(msize, 0.5*ob->Size[2]);

            updateMax(XYm, fabs(ob->R[0])+msize);
            updateMax(XYm, fabs(ob->R[1])+msize);
            updateMax(Zm,  fabs(ob->R[2])+msize);
          }

  //Mask
  if (!Detector->Masks.isEmpty())
    for (int im=0; im<Detector->Masks.size(); im++)
      {
        MaskClass& m = Detector->Masks[im];
        double msize = 0.5*m.Size;
        updateMax(msize, 0.5*m.Thickness);

        updateMax(XYm, fabs(m.OriginX) + msize);
        updateMax(XYm, fabs(m.OriginY) + msize);
        updateMax(Zm,  fabs(m.OriginZ) + msize);
      }
}
