#include "dynamicpassiveshandler.h"
#include "pms.h"
#include "eventsdataclass.h"
#include "reconstructionsettings.h"
#include "functions.h"

DynamicPassivesHandler::DynamicPassivesHandler(pms* Pms, EventsDataClass *eventsDataHub)
{
  EventsDataHub = eventsDataHub;
  PMs = Pms;  
}

void DynamicPassivesHandler::init(ReconstructionSettings *RecSet)
{
  numPMs = PMs->count();
  StatePMs.resize(numPMs);
  for (int ipm=0; ipm<numPMs; ipm++)
    StatePMs[ipm] = (PMs->isStaticPassive(ipm)) ? 1 : 0;

  if (RecSet == 0) return;
  switch (RecSet->ReconstructionAlgorithm)
    {
    case 0:      
      fLocalData = true;
      //PassiveType = 0;
      fByThreshold = RecSet->fCoGIgnoreLow;
      fByDistance =  RecSet->fCoGIgnoreFar;
      Threshold = RecSet->CoGIgnoreThreshold;
      MaxDistance2 = RecSet->CoGIgnoreDistance2;
      break;
    case 1:
      fLocalData = false;      
      //PassiveType =  RecSet->DynamicPassiveType;
      fByThreshold = (RecSet->DynamicPassiveType == 0);
      fByDistance =  (RecSet->DynamicPassiveType == 1);
      CenterOption = RecSet->CGstartOption;
      break;
    case 2:
      fLocalData = false;
      //PassiveType =  RecSet->DynamicPassiveType;
      fByThreshold = (RecSet->DynamicPassiveType == 0);
      fByDistance =  (RecSet->DynamicPassiveType == 1);
      CenterOption = RecSet->RMstartOption;
      break;
    case 3:
      //dynamic passives not applicable
      break;
    case 4:
      {
        fLocalData = true;
        QJsonObject json= RecSet->CGonCUDAsettings;
        int PassiveType;
        parseJson(json, "PassiveOption", PassiveType);
        fByThreshold = (PassiveType == 0);
        fByDistance =  (PassiveType == 1);
        parseJson(json, "Threshold", Threshold);
        double MaxDistance;
        parseJson(json, "MaxDistance", MaxDistance);
        MaxDistance2 = MaxDistance*MaxDistance;
        //qDebug() << "Cuda dyn passive setup: on, Popt, Thresh,MaxDist2"<<RecSet->fUseDynamicPassives<<PassiveType<<Threshold<<MaxDistance2;
        break;
      }
    case 5:
      //dynamic passives not applicable
      break;
    default:
      qWarning() << "Unknown reconstruction algorithm!";
      break;
    }
}

void DynamicPassivesHandler::calculateDynamicPassives(int ievent, const ReconstructionStruct* rec)
{
  if (ievent > EventsDataHub->Events.size())
    {
      clearDynamicPassives();
      return;
    }
  //qDebug()<<"Dyn pass type:" << PassiveType <<"Center opt:"<<CenterOption;

  double X0=0, Y0=0;
  if (fByDistance)
    {
      //some inits
      if (CenterOption == 0)
        {
          //qDebug()<<"using cog";
          X0 = rec->xCoG;
          Y0 = rec->yCoG;
        }
      else
        {
          //qDebug()<<"using max sig pm";
          X0 = PMs->X(rec->iPMwithMaxSignal);
          Y0 = PMs->Y(rec->iPMwithMaxSignal);
        }
    }

  for (int ipm=0; ipm < numPMs; ipm++)
    {
      bool fPassive = false;

      if (fByThreshold)
        {
          double threshold = fLocalData ? Threshold*PMs->getGain(ipm) : PMs->getPassiveThresholdForPM(ipm);       //takes gains into account!!!
          fPassive = (EventsDataHub->Events.at(ievent).at(ipm) < threshold);
        }
      if (fByDistance)
        {
          double dX =  (PMs->X(ipm) - X0);
          double dY =  (PMs->Y(ipm) - Y0);
          double distance2 = dX*dX + dY*dY;
          double maxDist2 = (fLocalData) ? MaxDistance2 : PMs->getPassiveRange2ForPM(ipm);
          fPassive = fPassive || (distance2 > maxDist2);
          //fPassive = fPassive || (distance2 < 1.0); /// TMP!!!***
        }

      if (fPassive) setDynamicPassive(ipm);
      else          clearDynamicPassive(ipm);
    }
//  if (PassiveType == 0)
//    { //use threshold
//      for (int ipm=0; ipm < numPMs; ipm++)
//        {
//          double threshold = fLocalData ? Threshold*PMs->getGain(ipm) : PMs->getPassiveThresholdForPM(ipm);       //takes gains into account!!!
//          if (EventsDataHub->Events.at(ievent).at(ipm) < threshold) setDynamicPassive(ipm);
//          else clearDynamicPassive(ipm);
//        }
//    }
//  else
//    {  //use distance
//      double X0=0, Y0=0;
//      if (CenterOption == 0)
//        {
//          //qDebug()<<"using cog";
//          X0 = rec->xCoG;
//          Y0 = rec->yCoG;
//        }
//      else
//        {
//          //qDebug()<<"using max sig pm";
//          X0 = PMs->X(rec->iPMwithMaxSignal);
//          Y0 = PMs->Y(rec->iPMwithMaxSignal);
//        }


//      for (int ipm=0; ipm<numPMs; ipm++)
//        {
//          double dX =  (PMs->X(ipm) - X0);
//          double dY =  (PMs->Y(ipm) - Y0);
//          double distance2 = dX*dX + dY*dY;
//          double maxDist2 = (fLocalData) ? MaxDistance2 : PMs->getPassiveRange2ForPM(ipm);
//          if (distance2 < maxDist2) clearDynamicPassive(ipm);
//          else setDynamicPassive(ipm);
//        }
//    }
}

int DynamicPassivesHandler::countActives() const
{
   int actives = 0;
   for (int ipm=0; ipm<numPMs; ipm++)
     if (isActive(ipm)) actives++;
   return actives;
}

void DynamicPassivesHandler::clearDynamicPassives()
{
  for (int ipm=0; ipm<numPMs; ipm++) clearDynamicPassive(ipm);
}
