#include "eventsdataclass.h"
#include "functions.h"

//Root
#include "TTree.h"
#include "TFile.h"
#include "pms.h"
#include "TParameter.h"

EventsDataClass::EventsDataClass(const TString nameID) //nameaddon to make unique hist names in multithread
 : QObject()
{
  DetStat = new DetectionStatisticsClass(nameID);
  DetStatNumBins = 100;
  DetStat->initialize(DetStatNumBins);
  fReconstructionDataReady = false;
  ReconstructionTree = 0;
  ResolutionTree = 0;
  fLoadedEventsHaveEnergyInfo = false;  
}

EventsDataClass::~EventsDataClass()
{
  clear();
  delete DetStat;  
}

void EventsDataClass::clearReconstructionTree()
{
  if (ReconstructionTree)
    {
      delete ReconstructionTree;
      ReconstructionTree = 0;
      qDebug() << "  Reconstruction tree deleted";
    }
}

void EventsDataClass::clearResolutionTree()
{
  if (ResolutionTree)
    {
      delete ResolutionTree;
      ResolutionTree = 0;
      qDebug() << "  Resolution tree deleted";
    }
}

void EventsDataClass::clear()
{
  Events.clear();
  TimedEvents.clear();
  PreprocessingSettings.clear();
  fLoadedEventsHaveEnergyInfo = false;

  clearScan();

  for (int i=0; i<EventHistory.size(); i++) delete EventHistory[i];
  EventHistory.clear();
  GeneratedPhotonsHistory.clear();
  DetStat->initialize();

  clearReconstruction();

  clearManifest();
}

void EventsDataClass::clearReconstruction()
{
  for (int i=0; i<ReconstructionData.size(); i++)
      delete ReconstructionData[i];
  ReconstructionData.resize(0);
  fReconstructionDataReady = false;
  clearReconstructionTree();
  clearResolutionTree();
}

void EventsDataClass::clearScan()
{
  for (int i=0; i<Scan.size(); i++) delete Scan[i];
  Scan.clear();
  ScanNumberOfRuns = 1;  //on MW was 0
  fSimulatedData = false;
}

void EventsDataClass::clearManifest()
{
  for (int i=0; i<Manifest.size(); i++) delete Manifest[i];
  Manifest.clear();
}

void EventsDataClass::setDetStatNumBins(int numBins)
{
  DetStatNumBins = numBins;
  DetStat->initialize(DetStatNumBins);
}

void EventsDataClass::squeeze()
{
  Events.squeeze();
  TimedEvents.squeeze();
  Scan.squeeze();
  ReconstructionData.squeeze();
  EventHistory.squeeze();
  GeneratedPhotonsHistory.squeeze();
}

void EventsDataClass::createDefaultReconstructionData()
{
  clearReconstruction();
  ReconstructionData.reserve(Events.size());
  for (int i=0; i<Events.size(); i++)
    {
      ReconstructionData.append(new ReconstructionStruct);
      ReconstructionData.last()->EventId = i;
      ReconstructionData.last()->chi2 = 0;
      ReconstructionData.last()->ReconstructionOK = false;

      ReconstructionData.last()->FilterEventNumber=0;
      ReconstructionData.last()->FilterCutOff=0;
      ReconstructionData.last()->FilterSumCutOff=0;
      ReconstructionData.last()->FilterEnergy=0;
      ReconstructionData.last()->FilterLoadedEnergy=0;
      ReconstructionData.last()->FilterChi2=0;
      ReconstructionData.last()->FilterPosition=0;
      ReconstructionData.last()->FilterCorrelation=0;
      ReconstructionData.last()->FilterkNN=0;
    }
}

void EventsDataClass::resetReconstructionData()
{
  if (ReconstructionData.size() != Events.size())
    EventsDataClass::createDefaultReconstructionData();
  else
    {
      fReconstructionDataReady = false;
      clearReconstructionTree();
      for (int i=0; i<Events.size(); i++)
        {
          ReconstructionData[i]->Points.Reinitialize(1);  //in case it was reconstructed before as a double event
          ReconstructionData[i]->chi2 = 0; //obsolete?
        }
    }
}

int EventsDataClass::getTimeBins()
{
  if (TimedEvents.isEmpty()) return 0;
  return TimedEvents[0].size();
}

const QVector<float> *EventsDataClass::getEvent(int iev)
{
  return &Events.at(iev);
}

const QVector<QVector<float> > *EventsDataClass::getTimedEvent(int iev)
{
  return &TimedEvents.at(iev);
}

int EventsDataClass::countGoodEvents() //counts events passing all filters
{
   int count = 0;
   for (int i=0; i<ReconstructionData.size(); i++)
     {
       if (ReconstructionData[i]->GoodEvent) count++;
       //else ReconstructionData[i]->report();
     }
   return count;
}

void EventsDataClass::prepareStatisticsForEvents(const bool isAllLRFsDefined, int &GoodEvents, double &AvChi2, double &AvDeviation)
{  /// *** not handling multiple events propelrly!
  //   qDebug()<<"   Preparing statistics of EventsDataHub triggered";
  GoodEvents = 0;
  AvChi2 = 0;
  AvDeviation = 0;
  bool DoDeviation = (isScanEmpty() || !isReconstructionDataReady()) ?  false : true;
  bool fDoChi2 = (!isReconstructionDataReady() || RecSettings.ReconstructionAlgorithm==0 || !isAllLRFsDefined) ? false : true;

  for (int iev=0; iev<ReconstructionData.size(); iev++)
    if (ReconstructionData[iev]->GoodEvent)
      {
        if (ReconstructionData[iev]->chi2 != ReconstructionData[iev]->chi2)
          qWarning() << "nan Chi2 detected for event"<<iev<<"RecOK?"<<ReconstructionData[iev]->ReconstructionOK;
        GoodEvents++;
        if (fDoChi2) AvChi2 += ReconstructionData[iev]->chi2;
        if (DoDeviation)
          {
            double r2 = 0;
            for (int i=0; i<2; i++) r2 += (ReconstructionData[iev]->Points[0].r[i] - Scan[iev]->Points[0].r[i]) *
                (ReconstructionData[iev]->Points[0].r[i] - Scan[iev]->Points[0].r[i]);
            AvDeviation += sqrt(r2);
          }
      }

  if (fDoChi2 && GoodEvents>0) AvChi2 /= GoodEvents;
  else AvChi2 = -1;

  if (DoDeviation && GoodEvents>0) AvDeviation /= GoodEvents;
  else AvDeviation = -1;
}

void EventsDataClass::createReconstructionTree(pms* PMs)
{
  qDebug()<<"--> Building reconstruction tree";
  if (ReconstructionTree) delete ReconstructionTree;
  ReconstructionTree = new TTree("T","Reconstruction data");
  int size = ReconstructionData.size();

  //double x, y, z, energy;
  std::vector <double> x; //Can be multiple point reconstruction!
  std::vector <double> y;
  std::vector <double> z;
  std::vector <double> energy;
  double chi2;
  double ssum;
  int ievent; //event number
  int good, recOK, fcut, fsumcut, fen, fchi, fsp, fcorr, fknn;

  ReconstructionTree->Branch("i",&ievent, "i/I");

  ReconstructionTree->Branch("x", &x);
  ReconstructionTree->Branch("y", &y);
  ReconstructionTree->Branch("z", &z);
  ReconstructionTree->Branch("energy", &energy);

  ReconstructionTree->Branch("chi2",&chi2, "chi2/D");
  ReconstructionTree->Branch("ssum",&ssum, "ssum/D");
  //filters
  ReconstructionTree->Branch("good", &good, "good/I");
  ReconstructionTree->Branch("recOK", &recOK, "recOK/I");
  ReconstructionTree->Branch("fcut", &fcut, "fcut/I");
  ReconstructionTree->Branch("fsumcut", &fsumcut, "fsumcut/I");
  ReconstructionTree->Branch("fen", &fen, "fen/I");
  ReconstructionTree->Branch("fchi", &fchi, "fchi/I");
  ReconstructionTree->Branch("fsp", &fsp, "fsp/I");
  ReconstructionTree->Branch("fcorr", &fcorr, "fsp/I");
  ReconstructionTree->Branch("fknn", &fknn, "fsp/I");

  //Pm signals
  int numPMs = PMs->count();
  float* signal = new float[numPMs];
  float* rho = new float[numPMs];
  char buf[32];
  sprintf(buf, "signal[%d]/F", numPMs);
  ReconstructionTree->Branch("signal", signal, buf);
  sprintf(buf, "rho[%d]/F", numPMs);
  ReconstructionTree->Branch("rho", rho, buf);

  //tmp
  int abA;
  ReconstructionTree->Branch("abA",&abA, "abA/I");
  double frac1;
  ReconstructionTree->Branch("frac1",&frac1, "frac1/D");

  //scan data
  std::vector <double> xScan;
  std::vector <double> yScan;
  std::vector <double> zScan;
  std::vector <int> numPhotons;
  double zStop;
  int ScintType, GoodEvent, EventType;

  if (!isScanEmpty())
    {
      ReconstructionTree->Branch("xScan", &xScan);
      ReconstructionTree->Branch("yScan", &yScan);
      ReconstructionTree->Branch("zScan", &zScan);
      ReconstructionTree->Branch("numPhotons", &numPhotons);

      ReconstructionTree->Branch("zStop",&zStop, "zStop/D");
      ReconstructionTree->Branch("ScintType",&ScintType, "ScintType/I");
      ReconstructionTree->Branch("GoodEvent",&GoodEvent, "GoodEvent/I");
      ReconstructionTree->Branch("EventType",&EventType, "EventType/I");
    }

  for (int iev=0; iev<size; iev++)
    {
      int Points = ReconstructionData[iev]->Points.size();
      x.resize(Points);
      y.resize(Points);
      z.resize(Points);
      energy.resize(Points);
      for (int iP=0; iP<Points; iP++)
        {
          x[iP] = ReconstructionData[iev]->Points[iP].r[0];
          y[iP] = ReconstructionData[iev]->Points[iP].r[1];
          z[iP] = ReconstructionData[iev]->Points[iP].r[2];
          energy[iP] = ReconstructionData[iev]->Points[iP].energy;
        }
      ievent = ReconstructionData[iev]->EventId;
      chi2 = ReconstructionData[iev]->chi2;
      ssum = 0;
      double maxSig = 0;
      for (int ipm=0; ipm<numPMs; ipm++) {
          signal[ipm] = Events[iev][ipm];
          double X = PMs->X(ipm);
          double Y = PMs->Y(ipm);
          rho[ipm] = sqrt((x[0]-X)*(x[0]-X)+(y[0]-Y)*(y[0]-Y));
          ssum += signal[ipm] * PMs->at(ipm).relGain;

          double corrSig = signal[ipm] * PMs->at(ipm).relGain;
          if (corrSig > maxSig) maxSig = corrSig;
      }

      //tmp
      abA = 0;
      double avSignal = ssum/numPMs;
      double threshold = (maxSig - avSignal)*0.15 + avSignal;
      for (int ipm=0; ipm<numPMs; ipm++)
          if ( Events[iev][ipm]* PMs->at(ipm).relGain > threshold) abA++;
      frac1 = maxSig/ssum;

      good = ReconstructionData[iev]->GoodEvent;
      recOK = ReconstructionData[iev]->ReconstructionOK;
      fcut = ReconstructionData[iev]->FilterCutOff;
      fsumcut = ReconstructionData[iev]->FilterSumCutOff;
      fen = ReconstructionData[iev]->FilterEnergy;
      fchi = ReconstructionData[iev]->FilterChi2;
      fsp = ReconstructionData[iev]->FilterPosition;
      fcorr = ReconstructionData[iev]->FilterCorrelation;
      fknn = ReconstructionData[iev]->FilterkNN;

      if (!isScanEmpty())
        {
          int Points = Scan[iev]->Points.size();
          xScan.resize(Points);
          yScan.resize(Points);
          zScan.resize(Points);
          numPhotons.resize(Points);
          for (int iP=0; iP<Points; iP++)
            {
              xScan[iP] = Scan[iev]->Points[iP].r[0];
              yScan[iP] = Scan[iev]->Points[iP].r[1];
              zScan[iP] = Scan[iev]->Points[iP].r[2];
              numPhotons[iP] = Scan[iev]->Points[iP].energy;
            }
          zStop = Scan[iev]->zStop;
          ScintType = Scan[iev]->ScintType;
          GoodEvent = Scan[iev]->GoodEvent;
          EventType = Scan[iev]->EventType;
        }
      ReconstructionTree->Fill();
    }

  ReconstructionTree->ResetBranchAddresses();
  delete [] signal;
  delete [] rho;
  qDebug()<<"   Tree created with "<<ReconstructionTree->GetEntries()<<" entries";
}

bool EventsDataClass::createResolutionTree()
{
  //  qDebug()<<"Creating a tree with sigma/distortion data";
  if (ResolutionTree)
    {
      delete ResolutionTree;
      ResolutionTree = 0;
    }
  if (ReconstructionData.size() != Scan.size())
    {
      qWarning() << "||| Cannot build scan tree: reconstruction and simulated data length mismatch";
      return false;
    }

  int events = Scan.size();
  int nodes = events / ScanNumberOfRuns;

  if (nodes*ScanNumberOfRuns != events)
    {
      qWarning() << "||| Cannot build resolution tree - data size is not consistent with the number of runs";
      return false;
    }

  ResolutionTree = new TTree("ResolutionTree","ScanTree");
  float x, y, z;  //actual position
  float rx, ry, rz;  //mean reconstructed position
  float sigmax, sigmay, sigmaz;
  int node;

  ResolutionTree->Branch("node",&node, "node/I");
  ResolutionTree->Branch("x",&x, "x/F");
  ResolutionTree->Branch("y",&y, "y/F");
  ResolutionTree->Branch("z",&z, "z/F");
  ResolutionTree->Branch("rx",&rx, "rx/F");
  ResolutionTree->Branch("ry",&ry, "ry/F");
  ResolutionTree->Branch("rz",&rz, "rz/F");
  ResolutionTree->Branch("sigmax",&sigmax, "sigmax/F");
  ResolutionTree->Branch("sigmay",&sigmay, "sigmay/F");
  ResolutionTree->Branch("sigmaz",&sigmaz, "sigmaz/F");

  for ( node = 0; node<nodes; node++ )
    {
      int i0this = node * ScanNumberOfRuns; //set has the same coordinates for ScanNumberOfRuns events, using 0th
      x = Scan[i0this]->Points[0].r[0];
      y = Scan[i0this]->Points[0].r[1];
      z = Scan[i0this]->Points[0].r[2];
      //qDebug()<<"x, y, z:"<<x<<y<<z;
      //averages
      double sum[3] = {0.0, 0.0, 0.0};
      int goodEvents = 0;
      for (int ievRelative = 0; ievRelative < ScanNumberOfRuns; ievRelative++)
        {
          int iev = ievRelative + node*ScanNumberOfRuns;
          if (iev > ReconstructionData.size()-1) break;
          if (!ReconstructionData[iev]->GoodEvent) continue;
          goodEvents++;
          for (int j=0;j<3; j++) sum[j] += ReconstructionData[iev]->Points[0].r[j];
        }
      float factor = (goodEvents == 0) ? 0 : 1.0/goodEvents;
      for (int j=0;j<3; j++) sum[j] *= factor;
      rx = sum[0];
      ry = sum[1];
      rz = sum[2];
      //qDebug()<<"rx, ry, rz:"<<rx<<ry<<rz;

      //deltas
      float asigma[3] = {0.0, 0.0, 0.0};
      for (int ievRelative = 0; ievRelative < ScanNumberOfRuns; ievRelative++)
        {
          int iev = ievRelative + node*ScanNumberOfRuns;
          if (!ReconstructionData[iev]->GoodEvent) continue;
          for (int j=0;j<3; j++)
            asigma[j] += (ReconstructionData[iev]->Points[0].r[j] - sum[j]) * (ReconstructionData[iev]->Points[0].r[j] - sum[j]);
        }
      factor = (goodEvents < 2) ? 0 : 1.0/(goodEvents-1);
      for (int j=0;j<3; j++) asigma[j] = sqrt(asigma[j]*factor);
      sigmax = asigma[0];
      sigmay = asigma[1];
      sigmaz = asigma[2];
      //qDebug()<<"sigx, sigy, sigz, sig:"<<sigmax<<sigmay<<sigmaz;

      ResolutionTree->Fill();
    }
  //qDebug()<<"Tree is ready";
  return true;
}

bool EventsDataClass::saveReconstructionAsTree(QString fileName, pms *PMs)
{
  clearReconstructionTree();
  QByteArray ba = fileName.toLocal8Bit();
  const char *c_str = ba.data();

  TFile* f = new TFile(c_str,"recreate");
  createReconstructionTree(PMs);
  int result = f->Write();
  f->Close();
  delete f;
  ReconstructionTree = 0; //Tree does not exist after save!
  return (result == 0) ? false : true;
}

bool EventsDataClass::saveReconstructionAsText(QString fileName)
{
  QFile outputFile(fileName);
  outputFile.open(QIODevice::WriteOnly);
  if(!outputFile.isOpen())
    {
      qWarning() << "Unable to open file " +fileName+ " for writing!";
      return false;
    }
  int size = ReconstructionData.size();
  QTextStream outStream(&outputFile);

  for (int iev=0; iev<size; iev++)
    if (ReconstructionData[iev]->GoodEvent)
      {
        int Points = ReconstructionData[iev]->Points.size();
        for (int iP = 0; iP<Points; iP++)
          {            
            outStream<<ReconstructionData[iev]->Points[iP].r[0]<<" ";
            outStream<<ReconstructionData[iev]->Points[iP].r[1]<<" ";
            outStream<<ReconstructionData[iev]->Points[iP].r[2]<<"  "; //2 spaces
            outStream<<ReconstructionData[iev]->Points[iP].energy<<"   "; //3 spaces            
          }
        outStream<<"  "<<ReconstructionData[iev]->chi2<<" "; //5 spaces including trailing
        outStream<<"    "<<ReconstructionData[iev]->EventId<<" "; //event id
        outStream<<"\r\n";
      }
  outputFile.close();
  return true;
}

bool EventsDataClass::saveSimulationAsTree(QString fileName)
{
  QByteArray ba = fileName.toLocal8Bit();
  const char *c_str = ba.data();
  TFile* f = new TFile(c_str,"recreate");
  TTree *tree = new TTree("SimulationTree","Simulation data");

  int numPMs = (Events.size()>0) ? Events[0].size() : 0;
  int tbins = 1;
  if (isTimed())
    if (Events.size()>0)
      tbins = TimedEvents[0].size();
  //qDebug()<<"time bins:"<<tbins;
  std::vector <double> x;
  std::vector <double> y;
  std::vector <double> z;
  double zStop;
  std::vector <int> numPhotons; //int numPhots;
  int ScintType, GoodEvent, EventType;
  int iev;

  tree->Branch("i",&iev, "i/I");
  if (!Scan.isEmpty())
    {
      tree->Branch("x", &x);
      tree->Branch("y", &y);
      tree->Branch("z", &z);
      tree->Branch("zStop",&zStop, "zStop/D");
      tree->Branch("numPhotons", &numPhotons);
      tree->Branch("ScintType",&ScintType, "ScintType/I");
      tree->Branch("GoodEvent",&GoodEvent, "GoodEvent/I");
      tree->Branch("EventType",&EventType, "EventType/I");
    }
  float* signal;
  signal = new float[numPMs];
  char buf[32];
  sprintf(buf, "signal[%d]/F", numPMs);
  tree->Branch("signal", signal, buf);
  std::vector< std::vector <float> > signalTimed;  //  [timebin][ipm]
  signalTimed.resize(tbins);
  for (int itime=0; itime<tbins; itime++) signalTimed[itime].resize(numPMs);
  if (isTimed()) tree->Branch("signalTimed", &signalTimed);

  for (iev=0; iev<Events.size(); iev++)
    {
      for (int ipm=0; ipm<numPMs; ipm++) signal[ipm] = Events[iev][ipm];

      if (isTimed())
        for (int ipm=0; ipm<numPMs; ipm++)
          for (int itime=0; itime<tbins; itime++)
            signalTimed[itime][ipm] = TimedEvents[iev][itime][ipm];

      if (!Scan.isEmpty())
        {
          int Points = Scan[iev]->Points.size();
          x.resize(Points);
          y.resize(Points);
          z.resize(Points);
          numPhotons.resize(Points);
          for (int iP=0; iP<Points; iP++)
            {
              x[iP] = Scan[iev]->Points[iP].r[0];
              y[iP] = Scan[iev]->Points[iP].r[1];
              z[iP] = Scan[iev]->Points[iP].r[2];
              numPhotons[iP] = Scan[iev]->Points[iP].energy;
            }
          zStop = Scan[iev]->zStop;
          ScintType = Scan[iev]->ScintType;
          GoodEvent = Scan[iev]->GoodEvent;
          EventType = Scan[iev]->EventType;
        }
      tree->Fill();
    }

  TParameter<int> *numRuns = new TParameter<int>("numRuns",ScanNumberOfRuns);
  tree->GetUserInfo()->Add(numRuns);

  int result = f->Write() ;
  f->Close();
  delete f;
  tree = 0;
  delete [] signal;
  return (result == 0) ? false : true;
}

bool EventsDataClass::saveSimulationAsText(QString fileName)
{
  QFile outputFile(fileName);
  outputFile.open(QIODevice::WriteOnly);
  if(!outputFile.isOpen())
    {
      qWarning()<<"Unable to open file"<<fileName<<"for writing!";
      return false;
    }
  int size = Events.size();
  bool fTimed = isTimed();
  int numPMs = (size>0) ? Events[0].size() : 0;
  int tbins = (fTimed) ? TimedEvents[0].size() : 1;

  QTextStream outStream(&outputFile);
  if (fTimed) outStream<<"Time bins "<<tbins<<"\r\n";  //header!

  for (int iev=0; iev<size; iev++)
    for (int itime=0; itime<tbins; itime++)
      {
        if (fTimed)
          for (int ipm=0; ipm<numPMs; ipm++)
            outStream << TimedEvents[iev][itime][ipm]<<" ";
        else
          for (int ipm=0; ipm<numPMs; ipm++)
            outStream << Events[iev][ipm]<<" ";

        if (!Scan.isEmpty())
          {
            outStream << "    "; //5 spaces including trailing one
            for (int ip=0; ip<Scan[iev]->Points.size(); ip++)
              {
                outStream << (int)Scan[iev]->Points[ip].energy << " ";
                outStream << Scan[iev]->Points[ip].r[0] << " "
                          << Scan[iev]->Points[ip].r[1] << " "
                          << Scan[iev]->Points[ip].r[2] << "   "; //3 spaces
              }
          }
        outStream<<"\r\n";
      }
  outputFile.close();
  return true;
}

int EventsDataClass::loadEventsFromTxtFile(QString fileName, QJsonObject &jsonPreprocessSettings, pms *PMs)
{
  ErrorString = "";
  QFile file(fileName);
  if(!file.open(QIODevice::ReadOnly | QFile::Text))
    {
      ErrorString = "Could not open: "+fileName;
      qWarning() << ErrorString;
      return -1;
    }

  int result = PreprocessingSettings.readFromJson(jsonPreprocessSettings, PMs, QFileInfo(fileName).fileName());  
  if (result == 1) qDebug() << "Do not perform any preprocessing: Preprocessing settings are absent in the json file";
  else if (result == 2)
    {
      ErrorString = "Preprocessing json read error";
      qWarning() << ErrorString;
      return -1;
    }
  else qDebug() << "Preprocessing settings configured";
  if (PreprocessingSettings.fManifest && PreprocessingSettings.fHavePosition)
    {
      ErrorString = "Preprocessing cannot contain both manifest and true positions!";
      qWarning() << ErrorString;
      return -1;
    }

  QTextStream in(&file);
  QRegExp rx("(\\ |\\,|\\:|\\t)"); //separators: ' ' or ',' or ':' or '\t'
  QString line = in.readLine();
  //  qDebug()<<"Looking for the code line 'Time bins' in the first line of the file";
  // qDebug()<<"Read line:"<<line;
  QStringList fields = line.split(rx, QString::SkipEmptyParts);

  int tBins=1; //if tBins == 1 - no timed data will be kept!
  bool ok;
  if (fields.count() > 2)
     if (!fields[0].compare("Time"))
        if (!fields[1].compare("bins"))
            tBins = fields[2].toInt(&ok);

  //First loaded file defined Time-related strategy:
  //if no time - all timed data will be ignored in later files
  //if timed, all files with different number of time bits will be discarded
  int Forced_tBins;
  if (Events.isEmpty())
    { //this is the first file
      Forced_tBins = tBins;
      //qDebug()<<"tBins fored for all following files to "<<Forced_tBins;
    }
  else
    {
      if (TimedEvents.isEmpty()) Forced_tBins = 1;
      else Forced_tBins = TimedEvents[0].size();
      //qDebug()<<"Forced tBins is "<<Forced_tBins;
      if (Forced_tBins == 1) qDebug()<<"Ignoring timed info!";
      else if (Forced_tBins != tBins)
        {
          ErrorString = "Mismatch in number of time bins detected, the file " + fileName + " is discarded!";
          qWarning() << ErrorString;
          file.close();
          return -1;
        }
    }

  //  qDebug()<<"Bins per PM: "<<tBins;
  if (tBins == 1) in.seek(0); //return to file start

  bool LoadEnergy = PreprocessingSettings.fActive && PreprocessingSettings.fHaveLoadedEnergy;
  int EnergyChannel = PreprocessingSettings.EnergyChannel;
  if (LoadEnergy && EnergyChannel < PMs->count())
    {
      ErrorString = "Error! Energy channel number should be >= the number of PMs";
      qWarning() << ErrorString;
      file.close();
      return -1;
    }
  bool LoadPosition = PreprocessingSettings.fActive && PreprocessingSettings.fHavePosition;
  int PositionXChannel = PreprocessingSettings.PositionXChannel;
  int PositionYChannel = PositionXChannel + 1;
  if (LoadPosition)
    {
      if (PositionXChannel < PMs->count() )
        {
          ErrorString = "Error! X position channel number should be >= the number of PMs";
          qWarning() << ErrorString;
          file.close();
          return -1;
        }
      if (LoadEnergy && (EnergyChannel==PositionXChannel || EnergyChannel==PositionYChannel) )
        {
          ErrorString = "Error! XY position channels overlaps energy channel";
          qWarning() << ErrorString;
          file.close();
          return -1;
        }
    }
  bool LoadZPosition = PreprocessingSettings.fActive && PreprocessingSettings.fHaveZPosition;
  int PositionZChannel = PreprocessingSettings.PositionZChannel;
  if (LoadZPosition)
    {
      if (PositionZChannel < PMs->count() )
        {
          ErrorString = "Error! Z position channel number should be >= the number of PMs";
          qWarning() << ErrorString;
          file.close();
          return -1;
        }
      if (LoadEnergy && EnergyChannel==PositionZChannel )
        {
          ErrorString = "Error! Z position channel overlaps energy channel";
          qWarning() << ErrorString;
          file.close();
          return -1;
        }
      if (LoadPosition && (PositionZChannel==PositionXChannel || PositionZChannel==PositionYChannel) )
        {
          ErrorString = "Error! Z position channel overlaps XY position channels";
          qWarning() << ErrorString;
          file.close();
          return -1;
        }
    }


  int UpperBound = PMs->count();
  if (LoadEnergy) UpperBound = EnergyChannel+1;
  if (LoadPosition) UpperBound = max(UpperBound, PositionYChannel+1);
  if (LoadZPosition) UpperBound = max(UpperBound, PositionZChannel+1);
  int DataSize = PMs->count();
  if (LoadEnergy) DataSize++;
  //position data are NOT added to tmp[][], so they do not influence DataSize

  int eventNumber = 0;

  int AppendedFrom = Events.size(); //data from this file will be appended from this position
  int AppendedTimedFrom = TimedEvents.size(); //will be mismatch with AppendedFrom if tBins == 0
  qint64 fileSize = file.size();
  qint64 alreadyRead = 0;
  double tmpEnergy;
  double positionX = 0, positionY = 0, positionZ = 0;
  while(!in.atEnd())
     {   // reading and storing data to LoadedTimedEvents
        bool error = false;
        //creating and shaping tmp storage   [timeBin][PM]
        QVector< QVector <float> > tmp;
        tmp.resize(tBins);
        tmp.squeeze();
        for (int t=0; t<tBins; t++)
          {
            tmp[t].resize(DataSize);
            tmp[t].squeeze();
          }

        tmpEnergy = 0; //accumulating loaded energy over all time bins
        for (int t=0; t<tBins; t++)
         {
          do  //this way we allow empty lines!
              {
                line = in.readLine();
                alreadyRead += line.size();
                fields = line.split(rx, QString::SkipEmptyParts);
                if (in.atEnd()) break;
              }
          while (fields.isEmpty());

          if (fields.count() < UpperBound)
            {
              if (in.atEnd())
                {
                  error=true;
                  break;
                }
              QString str;
              str.setNum(fields.count());
              if (LoadEnergy || LoadPosition || LoadZPosition) ErrorString = fileName+" - found a line with number of channels ( "+str+" ) less than required to extract the energy or position information!";
              else ErrorString = fileName+" - found a line with number of PMs ( "+str+" ) which is less than the number of PMs defined in the current geometry!";
              qWarning() << ErrorString;
              //clear data loaded from this file
              TimedEvents.resize(AppendedTimedFrom);
              if (LoadEnergy || LoadPosition || LoadZPosition)
                {
                  for (int i=AppendedTimedFrom; i<Scan.size(); i++) delete Scan[i];
                  Scan.resize(AppendedTimedFrom);
                }
              file.close();
              return -1;
            }

          for (int i=0; i<UpperBound; i++)
            {
              bool ok;
              float val = fields[i].toFloat(&ok);
              //  qDebug()<<i<<ok<<val;
              if (!ok)
                {
                  ErrorString = fileName+" - wrong format - expecting float datatype";
                  qWarning() << ErrorString;
                  //clear data loaded from this file
                  TimedEvents.resize(AppendedTimedFrom);
                  if (LoadEnergy || LoadPosition || LoadZPosition)
                    {
                      for (int i=AppendedTimedFrom; i<Scan.size(); i++) delete Scan[i];
                      Scan.resize(AppendedTimedFrom);
                    }
                  file.close();
                  return -1;
                }

              if (i >= PMs->count())
                {
                  //can be only energy or position info here
                  if (LoadEnergy && i==EnergyChannel)
                    {
                      //energy channel
                      double energy;
                      if (PreprocessingSettings.fActive) energy = (val + PreprocessingSettings.LoadEnAdd) * PreprocessingSettings.LoadEnMulti;
                      else  energy = val;
                      tmp[t][PMs->count()] = energy;
                      tmpEnergy += energy;
                    }
                  else if (LoadZPosition && i==PositionZChannel) positionZ = val;
                  else if (LoadPosition)
                    {
                      if (i == PositionXChannel) positionX = val;
                      else if (i == PositionYChannel) positionY = val;
                    }
                  //otherwise do nothing
                }
              else
                {
                  //PM channels
                  if (PreprocessingSettings.fActive)
                    { //saturation control
                      if (PreprocessingSettings.fIgnoreThresholds)
                        {
                          if (val < PreprocessingSettings.ThresholdMin ) error = true;
                          if (val > PreprocessingSettings.ThresholdMax ) error = true;
                        }
                      tmp[t][i] = (val + PMs->at(i).PreprocessingAdd) * PMs->at(i).PreprocessingMultiply;

                      if (PreprocessingSettings.fToPhotoelectrons)
                        { //backtracking electronics
                          if (PMs->isDoADC()) tmp[t][i] *= PMs->getADCstep(i); //backtracking digitalization
                          if (PMs->isDoPHS())
                            { //correcting for average signal - output is photoelectrons (effective)!
                              double AvSigPerPhEl = PMs->getAverageSignalPerPhotoelectron(i);
                              if (AvSigPerPhEl != 0) tmp[t][i] /= AvSigPerPhEl; //ok to test plainly to  0 - it is given by user, not calculated
                            }
                        }
                    }
                  else tmp[t][i] = val;
                }

            } //end cycle by channels
         } // end cycle by time bins

        if (!error)
          {
            eventNumber++;
            //checking event number selecton
            if (PreprocessingSettings.fActive)
              {
                if (PreprocessingSettings.fLimitNumber && eventNumber > PreprocessingSettings.LimitMax) break;
                if (PreprocessingSettings.fManifest && PreprocessingSettings.ManifestItem->LimitEvents != -1)
                  if (eventNumber > PreprocessingSettings.ManifestItem->LimitEvents) break;
              }

            TimedEvents.append(tmp); //adding event

            if (LoadEnergy || LoadPosition || LoadZPosition)
              {
                // *** !!! for timed events: the last time_bin's XYZ is copied to Scan!
                ScanStructure* sc = new ScanStructure();
                sc->Points[0].r[0] = positionX;
                sc->Points[0].r[1] = positionY;
                sc->Points[0].r[2] = positionZ;
                sc->Points[0].energy = tmpEnergy;
                sc->ScintType = 1;
                Scan.append(sc);
              }

            //indication / app update
            int numAdded = TimedEvents.size()-AppendedTimedFrom;
            if (numAdded % 5000 == 0)
              {
                //qDebug() << numAdded;
                emit loaded(numAdded, 100.0/fileSize*alreadyRead);
                qApp->processEvents();
              }
          }
     }

   int numEvents = TimedEvents.size() - AppendedTimedFrom;
   //qDebug()<<"Done! Events found: "<<numEvents;
   file.close();

   //calculating total, deleting LoadedEventsTimed if tBins = 1
   Events.resize(AppendedFrom + numEvents);
   //   qDebug()<<"      numEvents, tBins, DataSize:"<<numEvents<<tBins<<DataSize;
   for (int iev=0; iev<numEvents; iev++)
     {
       Events[AppendedFrom + iev].resize(DataSize);
       Events[AppendedFrom + iev].squeeze();
       for (int ipm = 0; ipm<DataSize; ipm++)
         {
            Events[AppendedFrom + iev][ipm]=0;
            for (int t=0; t<tBins; t++) Events[AppendedFrom + iev][ipm] += TimedEvents[AppendedTimedFrom + iev][t][ipm];
         }

       if (PreprocessingSettings.fManifest) // this mode can not be together with LoadPositions
         {
           ScanStructure* sc = new ScanStructure();
           sc->Points[0].r[2] = 0;
           sc->Points[0].energy = 1.0;
           sc->GoodEvent = true;
           sc->ScintType = 0;

           if (PreprocessingSettings.ManifestItem->getType() == "hole")
             {
               sc->Points[0].r[0] = PreprocessingSettings.ManifestItem->X;
               sc->Points[0].r[1] = PreprocessingSettings.ManifestItem->Y;                          
             }
           else if (PreprocessingSettings.ManifestItem->getType() == "slit")
             {
               double &Angle = dynamic_cast<ManifestItemSlitClass*>(PreprocessingSettings.ManifestItem)->Angle; //angle with X axis in degrees
               if ( Angle == 0)
                 {
                   sc->Points[0].r[0] = 1e10; //undefined
                   sc->Points[0].r[1] = PreprocessingSettings.ManifestItem->Y;
                 }
               else if (Angle == 90)
                 {
                   sc->Points[0].r[0] = PreprocessingSettings.ManifestItem->X;
                   sc->Points[0].r[1] = 1e10; //undefined
                 }
               else
                 {
                   sc->Points[0].r[0] = 1e10; //undefined
                   sc->Points[0].r[1] = 1e10; //undefined
                 }
             }
           else
             {
               sc->Points[0].r[0] = 1e10; //undefined
               sc->Points[0].r[1] = 1e10; //undefined
             }

           Scan.append(sc);
         }
     }
   if (Forced_tBins == 1)
     {
       // qDebug()<<"  Deleting unused time-resolved data...";
       TimedEvents.resize(0);
     }
   fLoadedEventsHaveEnergyInfo = LoadEnergy;
   if (PreprocessingSettings.fManifest)
     {
       //transferring the ManifestItem to the EventsDataHub storage
       Manifest.append(PreprocessingSettings.ManifestItem);
       PreprocessingSettings.ManifestItem = NULL;
     }
   fSimulatedData = false;
   //qDebug() << "File size:"<<fileSize << "read:"<<alreadyRead;
   return numEvents;
}

bool EventsDataClass::overlayAsciiFile(QString fileName, bool fAddMulti, pms* PMs)
{
  qDebug() << fAddMulti<< PMs->at(0).PreprocessingAdd << PMs->at(0).PreprocessingMultiply;
  if (Events.isEmpty())
    {
      ErrorString = "Cannot overlay - events are empty";
      qWarning() << ErrorString;
      return false;
    }

  QFile file(fileName);
  if(!file.open(QIODevice::ReadOnly | QFile::Text))
    {
      ErrorString = "Could not open file "+fileName;
      qWarning() << ErrorString;
      return false;
    }

  QTextStream in(&file);
  QRegExp rx("(\\ |\\,|\\:|\\t)"); //separators: ' ' or ',' or ':' or '\t'
  QString line = in.readLine();
  //  qDebug()<<"Looking for the code line 'Time bins' in the first line of the file";
  // qDebug()<<"Read line:"<<line;
  QStringList fields = line.split(rx, QString::SkipEmptyParts);

  if (fields.count() > 2)
     if (fields[0] == "Time" && fields[1] == "bins")
       {
         ErrorString = "Time resolved files are not currently supported";
         qWarning() << ErrorString;
         return false;
       }
  in.seek(0); //return to file start

  //check that it is enough channels and events
  int numPMs = PMs->count();
  int iEvent = 0;
  while(!in.atEnd())
    {
      line = in.readLine();
      fields = line.split(rx, QString::SkipEmptyParts);
      if (fields.isEmpty()) continue; //allow empty lines
      if (fields.count() < numPMs)
        {
          ErrorString = "Found event line with data channels less than PMs";
          qWarning() << ErrorString;
          return false;
        }
      bool ok;
      for (int i=0; i<fields.count(); i++)
        {
          fields[i].toFloat(&ok);
          if (!ok)
            {
              ErrorString = "Failed to convert text to float";
              qWarning() << ErrorString;
              return false;
            }
        }
      iEvent++;
      if (iEvent == Events.size()) break;
    }
  if (iEvent != Events.size())
    {
      ErrorString = "Not enough data lines in file to cover all events";
      qWarning() << ErrorString;
      return false;
    }

  //all is fine, can read and overlay
  in.seek(0); //return to file start
  iEvent = 0;
  while(!in.atEnd())
    {
      line = in.readLine();
      fields = line.split(rx, QString::SkipEmptyParts);
      if (fields.isEmpty()) continue; //allow empty lines

      for (int i=0; i<fields.count(); i++)
        {
          float delta = fields[i].toFloat();
          if (fAddMulti) delta = (delta + PMs->at(i).PreprocessingAdd) * PMs->at(i).PreprocessingMultiply;
          Events[iEvent][i] += delta;
        }

      iEvent++;
      if (iEvent == Events.size()) break;
    }

  ErrorString = "";
  return true;
}

int EventsDataClass::loadSimulatedEventsFromTree(QString fileName, pms *PMs)
{
  ErrorString = "";
  QByteArray ba = fileName.toLocal8Bit();
  const char *c_str = ba.data();
  TFile* f = new TFile(c_str);
  TTree* T = 0;
  T = (TTree*)f->Get("SimulationTree");
  if (!T)
    {
      ErrorString = "Tree with simulated data not found!";
      qWarning() << ErrorString;
      f->Close();
      return -1;
    }

  bool FirstDataSet = Events.isEmpty();
  //if it is the first data set:
  //if there are no scan data - in any later files scan data is ignored
  //if scan data present - any later files without scan data will NOT be loaded

  int numPMs = PMs->count();
  int numEv = T->GetEntries();
  qDebug() << "Number of events found: "+QString::number(numEv);

  bool ScanDataPresent = false; //Scan data exists?
  TBranch* xBranch = 0;
  xBranch = T->FindBranch("x");
  if (xBranch)
    {
      qDebug()<<"ScanDataFound!";
      ScanDataPresent = true;
      if (!FirstDataSet && Scan.isEmpty())
        {
          qWarning()<<"Ignoring scan data in this file!";
          ScanDataPresent = false;
        }
    }
  else
    {
      qDebug()<<"No scan data in the tree!";
      if (!FirstDataSet && !Scan.isEmpty())
        {
          ErrorString = "Sim tree file cannot be loaded - no required scan data present!";
          qWarning() << ErrorString;
          f->Close();
          return -1;
        }
    }

  //num of PMs in the Tree data?
  TBranch* signalBranch = T->FindBranch("signal");
  TString title = signalBranch->GetTitle();  //  "signal[%d]/D"    or "signal[%d]/F"
  bool fDoubleSignals = (title.Contains("D")) ? true : false;  //(title.Contains("F"))

  title.ReplaceAll("signal[","");
  title.ReplaceAll("]/D","");
  title.ReplaceAll("]/F","");

  int numPMsInTree = ((QString)title).toInt();
  qDebug()<<"PMs in tree file="<<numPMsInTree;
  if (numPMsInTree != numPMs)
    {
      ErrorString = "Error: File contains less PMs than in this detector configuration!";
      qWarning() << ErrorString;
      f->Close();
      return -1;
    }

  //timed data?
  bool TimedData = false;
  TBranch* timeBranch = 0;
  timeBranch = T->FindBranch("signalTimed");
  if (timeBranch)
    {
      qDebug() << "Data is time-resolved!";
      TimedData = true;
    }

  double zStop;
  std::vector <double> *x = 0, *y = 0, *z = 0;
  std::vector <int> *numPhotons = 0;
  int ScintType, GoodEvent, EventType;
  TBranch *bx = 0;

  if (ScanDataPresent)
    {
      T->SetBranchAddress("x", &x, &bx);
      T->SetBranchAddress("y", &y);
      T->SetBranchAddress("z", &z);
      T->SetBranchAddress("numPhotons", &numPhotons);

      T->SetBranchAddress("zStop", &zStop);
      T->SetBranchAddress("ScintType", &ScintType);
      T->SetBranchAddress("GoodEvent", &GoodEvent);
      T->SetBranchAddress("EventType", &EventType);
    }

  float *signalF = new float[numPMs];
  std::vector< std::vector <float> > *signalTimedF = 0;  //  [timebin][ipm]
  double *signalD = new double[numPMs];
  std::vector< std::vector <double> > *signalTimedD = 0;  //  [timebin][ipm]

  if (fDoubleSignals)
    {
      T->SetBranchAddress("signal", signalD);
      if (TimedData) T->SetBranchAddress("signalTimed", &signalTimedD);
    }
  else
    {
      T->SetBranchAddress("signal", signalF);
      if (TimedData) T->SetBranchAddress("signalTimed", &signalTimedF);
    }

  for (int iev=0; iev<numEv; iev++)
    {
      Long64_t tentry = T->LoadTree(iev);
      T->GetEntry(tentry);

      if (ScanDataPresent)
        {
          ScanStructure* scs = new ScanStructure();
          int Points = x->size();
          //                        qDebug()<<"Points this event:"<<Points;
          if (Points != 1) scs->Points.Reinitialize(Points);
          for (int iP = 0; iP<Points; iP++)
            {
              scs->Points[iP].r[0] = x->at(iP);
              scs->Points[iP].r[1] = y->at(iP);
              scs->Points[iP].r[2] = z->at(iP);
              scs->Points[iP].energy = numPhotons->at(iP);
            }
          scs->zStop = zStop;
          scs->GoodEvent = GoodEvent;
          scs->EventType = EventType;
          scs->ScintType = ScintType;

          Scan.append(scs);          
        }

      QVector<float> PMsignals(numPMs);
      if (fDoubleSignals)
        {
         for (int ipm=0; ipm<numPMs; ipm++) PMsignals[ipm] = signalD[ipm];
         Events.append(PMsignals);

         if (TimedData)
           {
             int timeBins = signalTimedD->size();
             QVector< QVector <float> > PMsignalsTimed(timeBins);
             for (int itime=0; itime<timeBins; itime++)
               {
                 PMsignalsTimed[itime].resize(numPMs);

                 for (int ipm=0; ipm<numPMs; ipm++)
                   PMsignalsTimed[itime][ipm] = (*signalTimedD)[itime][ipm];
               }
             TimedEvents.append(PMsignalsTimed);
             delete signalTimedD;
           }
        }
      else
        {
         for (int ipm=0; ipm<numPMs; ipm++) PMsignals[ipm] = signalF[ipm];
         Events.append(PMsignals);

         if (TimedData)
           {
             int timeBins = signalTimedF->size();
             QVector< QVector <float> > PMsignalsTimed(timeBins);
             for (int itime=0; itime<timeBins; itime++)
               {
                 PMsignalsTimed[itime].resize(numPMs);

                 for (int ipm=0; ipm<numPMs; ipm++)
                   PMsignalsTimed[itime][ipm] = (*signalTimedF)[itime][ipm];
               }
             TimedEvents.append(PMsignalsTimed);
             delete signalTimedF;
           }
        }
    }

  TObject* tmp = T->GetUserInfo()->FindObject("numRuns");
  TParameter<int> *numRuns = dynamic_cast<TParameter<int>*>(tmp);
  if (numRuns)
    {
      int NumRuns = numRuns->GetVal();
      qDebug() << "Found numRuns info:"<< NumRuns;
      ScanNumberOfRuns = NumRuns;
    }
  fSimulatedData = true;

  T->ResetBranchAddresses();
  // qDebug()<<"SimulatedScan Object contains now"<<SimulatedScan.size()<<"entries";
  //qDebug()<<"Events object contains now"<<Events.size()<<"entries";
  // qDebug()<<"TimedEvents object contains now"<<TimedEvents.size()<<"entries";
  f->Close();
  delete[] signalD;
  delete[] signalF;
  return numEv;
}

int PreprocessingSettingsClass::readFromJson(QJsonObject &json, pms* PMs, QString FileName)
// 0 - OK
// 1 - Preprocessing data not found in json
// 2 - Error
{
  clear();

  QJsonObject js1 = json["DetectorConfig"].toObject();
  if (!js1.contains("LoadExpDataConfig")) return 1;

  QJsonObject js = js1["LoadExpDataConfig"].toObject();
  if (js.isEmpty()) return 1;

  parseJson(js, "Preprocessing", fActive);
  if (js.contains("AddMulti"))
    {
      QJsonArray ar = js["AddMulti"].toArray();
      if (ar.size() != PMs->count())
        {
          qWarning() << "Mismatch in number of PMs!";
          return 2;
        }
      for (int i=0; i<PMs->count(); i++)
        {
          QJsonArray el = ar[i].toArray();
          PMs->at(i).PreprocessingAdd = el[0].toDouble();
          PMs->at(i).PreprocessingMultiply = el[1].toDouble();
        }
    }
  else
    {
      for (int i=0; i<PMs->count(); i++)
        {
          PMs->at(i).PreprocessingAdd = 0;
          PMs->at(i).PreprocessingMultiply = 1.0;
          LoadEnAdd = 0;
          LoadEnMulti = 1.0;
        }
    }

  fHaveLoadedEnergy = false;
  QJsonObject enj = js["LoadedEnergy"].toObject();
  parseJson(enj, "Activated", fHaveLoadedEnergy);
  parseJson(enj, "Channel", EnergyChannel);
  parseJson(enj, "Add", LoadEnAdd);
  parseJson(enj, "Multi", LoadEnMulti);

  fHavePosition = false;
  QJsonObject posj = js["LoadedPosition"].toObject();
  parseJson(posj, "Activated", fHavePosition);
  parseJson(posj, "Channel", PositionXChannel);

  fHaveZPosition = false;
  QJsonObject zposj = js["LoadedZPosition"].toObject();
  parseJson(zposj, "Activated", fHaveZPosition);
  parseJson(zposj, "Channel", PositionZChannel);

  fIgnoreThresholds = false;
  QJsonObject ij = js["IgnoreThresholds"].toObject();
  parseJson(ij, "Activated", fIgnoreThresholds);
  parseJson(ij, "Min", ThresholdMin);
  parseJson(ij, "Max", ThresholdMax);

  fToPhotoelectrons = false;
  parseJson(js, "ConvertToPhotoelectrons", fToPhotoelectrons);
  if (fToPhotoelectrons)
    {
      bool ok = PMs->readElectronicsFromJson(js1); //to get el settings
      if (!ok) qWarning() << "||| There were no electronics settings in json file";
    }

  fLimitNumber = false;
  int lim = -1;
  parseJson(js, "LoadFirst", lim);
  fLimitNumber = ( lim != -1 );
  if (fLimitNumber) LimitMax = lim;

  fManifest = false;  
  if (js.contains("ManifestFile"))
    {
      //qDebug() << "Manifest file processing";
      fManifest = true;
      ManifestFile = js["ManifestFile"].toString();
      if (FileName != "")
        { //FileName == "" when loading preprocessing settings to GUI -> then skip processing
          ManifestProcessorClass mp(ManifestFile);
          ManifestItem = mp.makeManifestItem(FileName);
          if (!ManifestItem)
            {
              fManifest = false;
              return 2;
            }
        }
    }

  return 0;
}

void PreprocessingSettingsClass::clear()
{
  fActive = false;
  LoadEnAdd = 0;
  LoadEnMulti = 1.0;
  fHaveLoadedEnergy = false;
  fHavePosition = false;
  fHaveZPosition = false;
  fIgnoreThresholds = false;
  fToPhotoelectrons = false;
  fLimitNumber = false;
  if (ManifestItem)
    {
      delete ManifestItem;
      ManifestItem = NULL;
    }
}
