#ifndef EVENTSDATACLASS_H
#define EVENTSDATACLASS_H

#include "classes.h"
#include "detectionstatisticsclass.h"
#include "generalsimsettings.h"
#include "reconstructionsettings.h"
#include "manifesthandling.h"

#include <QVector>
#include <QObject>

class TTree;
class pms;

class PreprocessingSettingsClass
{
public:
  bool fActive;

  bool fHaveLoadedEnergy;
  int EnergyChannel;
  float LoadEnAdd, LoadEnMulti;

  bool fHavePosition;
  int PositionXChannel; //PositionYChannel = PositionXChannel + 1

  bool fHaveZPosition;
  int PositionZChannel;

  bool fIgnoreThresholds;
  float ThresholdMin, ThresholdMax;

  bool fToPhotoelectrons;

  bool fLimitNumber;
  int LimitMax;

  bool fManifest;
  QString ManifestFile;
  ManifestItemBaseClass* ManifestItem;

  int readFromJson(QJsonObject &json, pms *PMs, QString FileName);
  void clear();

  PreprocessingSettingsClass() { ManifestItem = NULL; clear();}
  ~PreprocessingSettingsClass() { clear(); }
};

class EventsDataClass : public QObject
{
  Q_OBJECT
public:
    EventsDataClass(const TString nameID = ""); //nameaddon to make unique hist names in multithread
    ~EventsDataClass();

    // True position
    QVector<ScanStructure*> Scan;
    int ScanNumberOfRuns; //number of runs performed at each position (node) - see simulation module
    // PM signal data
    QVector< QVector <float> > Events; //[event][pm]  - remember, if events energy is loaded, one MORE CHANNEL IS ADDED: last channel is numPMs+1
    QVector< QVector < QVector <float> > > TimedEvents; //event timebin pm
    // Logs
    QVector<EventHistoryStructure*> EventHistory;
    QVector<GeneratedPhotonsHistoryStructure> GeneratedPhotonsHistory;
    //Detection statistics
    DetectionStatisticsClass* DetStat;
    //reconstruction
    QVector<ReconstructionStruct*> ReconstructionData;

    // simulation setting used to produce these events
    GeneralSimSettings LastSimSet;   
    bool fSimulatedData; // true if events data were simulated (false if loaded data)
    ReconstructionSettings RecSettings;
    bool fReconstructionDataReady; // true if reconstruction was already performed

    //Trees
    TTree *ReconstructionTree; //tree with reconstruction data
    void clearReconstructionTree();
    void createReconstructionTree(pms* PMs);
    TTree* ResolutionTree; //tree with resolution data
    void clearResolutionTree();
    bool createResolutionTree();

    //Data Save
    bool saveReconstructionAsTree(QString fileName, pms *PMs);
    bool saveReconstructionAsText(QString fileName);
    bool saveSimulationAsTree(QString fileName);
    bool saveSimulationAsText(QString fileName);

    //Data Load
    bool fLoadedEventsHaveEnergyInfo;
    PreprocessingSettingsClass PreprocessingSettings;
    int loadEventsFromTxtFile(QString fileName, QJsonObject &jsonPreprocessSettings, pms *PMs); //returns -1 if failed, otherwise number of events added
    QVector<ManifestItemBaseClass*> Manifest; //load data can have manifest file with holes/slits
    //sim tree
    int loadSimulatedEventsFromTree(QString fileName, pms *PMs); //returns -1 if failed, otherwise number of events added
    bool overlayAsciiFile(QString fileName, bool fAddMulti, pms *PMs); //true = success, if not, see ErrorString

    //Misc
    QString ErrorString;

    void clear();
    void clearReconstruction();
    void clearScan();
    void clearManifest();
    void setDetStatNumBins(int numBins);
    int getStatDataNumBins() {return DetStatNumBins;}

    void squeeze();

    void createDefaultReconstructionData(); //clean up data completely
    void resetReconstructionData();         //cleanup, but keep eventId intact

    bool isEmpty() {return Events.isEmpty();}
    bool isStatEmpty() {return DetStat->isEmpty();}
    bool isScanEmpty() {return Scan.isEmpty();}
    bool isTimed() {return !TimedEvents.isEmpty();}
    int getTimeBins();
    const QVector<float> *getEvent(int iev);
    const QVector<QVector<float> > *getTimedEvent(int iev);
    bool isReconstructionDataReady() {return (!ReconstructionData.isEmpty() && fReconstructionDataReady);}
    int countGoodEvents();
    void prepareStatisticsForEvents(const bool isAllLRFsDefined, int &GoodEvents, double &AvChi2, double &AvDeviation);

private:
    int DetStatNumBins;

signals:
    void loaded(int events, int progress);
};

#endif // EVENTSDATACLASS_H
