#include "materialcollectionclass.h"
#include "functions.h"
#include "jsonparser.h"
#include "generalsimsettings.h"
#include "opticaloverrideclass.h"

#include <QtDebug>
#include <QJsonObject>
#include <QJsonDocument>

#include "TROOT.h"
#include "TGeoMedium.h"
#include "TGeoMaterial.h"

MaterialCollectionClass::MaterialCollectionClass()
{
  ParticleCollection = 0;
  WaveFrom = 200;
  WaveTo = 800;
  WaveStep = 5;
  WaveNodes = 121;
  WavelengthResolved = false;
  fLogLogInterpolation = false;
}

MaterialCollectionClass::MaterialCollectionClass(QVector<ParticleDescriptor *> *particleCollection)
{
  ParticleCollection = particleCollection;
  WaveFrom = 200;
  WaveTo = 800;
  WaveStep = 5;
  WaveNodes = 121;
  WavelengthResolved = false;
  fLogLogInterpolation = false;
}

MaterialCollectionClass::~MaterialCollectionClass()
{
  MaterialCollectionClass::Clear();
}

void MaterialCollectionClass::SetParticleCollection(QVector<ParticleDescriptor *> *particleCollection)
{
  ParticleCollection = particleCollection;
}

void MaterialCollectionClass::SetWave(bool wavelengthResolved, double waveFrom, double waveTo, double waveStep, int waveNodes)
{
  WavelengthResolved = wavelengthResolved;
  WaveFrom = waveFrom;
  WaveTo = waveTo;
  WaveStep = waveStep;
  WaveNodes = waveNodes;
}

void MaterialCollectionClass::updateWaveProperties(GeneralSimSettings *SimSet)
{
  MaterialCollectionClass::SetWave(SimSet->fWaveResolved, SimSet->WaveFrom, SimSet->WaveTo, SimSet->WaveStep, SimSet->WaveNodes);
  for (int i=0; i<MaterialCollectionData.size(); i++)
    UpdateWaveResolvedProperties(i);
}

void MaterialCollectionClass::getFirstOverridenMaterial(int &ifrom, int &ito)
{
  for (ifrom=0; ifrom<MaterialCollectionData.size(); ifrom++)
    for (ito=0; ito<MaterialCollectionData.size(); ito++)
      //if (MaterialCollectionData[ifrom]->LossOverride[ito] > 0 ||
      //    MaterialCollectionData[ifrom]->ReflectionOverride[ito]>0 ||
      //    MaterialCollectionData[ifrom]->ScatterOverride[ito]>0) return;
      if (MaterialCollectionData[ifrom]->OpticalOverrides[ito]) return;
  ifrom = 0;
  ito = 0;
}

void MaterialCollectionClass::Clear()
{   
    for (int i=0; i<MaterialCollectionData.size(); i++)
      {
        if (MaterialCollectionData[i]->PrimarySpectrumHist)
          {
            delete MaterialCollectionData[i]->PrimarySpectrumHist;
            MaterialCollectionData[i]->PrimarySpectrumHist = 0;
          }
        if (MaterialCollectionData[i]->SecondarySpectrumHist)
          {
            delete MaterialCollectionData[i]->SecondarySpectrumHist;
            MaterialCollectionData[i]->SecondarySpectrumHist = 0;
          }

        for (int im=0; im<MaterialCollectionData[i]->OpticalOverrides.size(); im++)
          if (MaterialCollectionData[i]->OpticalOverrides[im])
              delete MaterialCollectionData[i]->OpticalOverrides[im];
        MaterialCollectionData[i]->OpticalOverrides.clear();

        delete MaterialCollectionData[i];
      }
    MaterialCollectionData.clear();

    MaterialCollectionClass::ClearTmpMaterial();  
}

void MaterialCollectionClass::AddNewMaterial()
{
  MaterialStructure *m = new MaterialStructure;

  int thisMat = MaterialCollectionData.size(); //index of this material (after it is added)
  int numMats = thisMat+1; //collection size after it is added

  //initialize empty optical overrides for all materials
  //m->LossOverride.resize(numMats);
  //m->ScatterOverride.resize(numMats);
  //m->ScatterModel.resize(numMats);
  //m->ReflectionOverride.resize(numMats);
  m->OpticalOverrides.resize(numMats);
  for (int i=0; i<numMats; i++)
    {
      //m->LossOverride[i]=0;
      //m->ScatterOverride[i]=0;
      //m->ScatterModel[i]=1;  //2Pi
      //m->ReflectionOverride[i]=0;
      m->OpticalOverrides[i] = NULL;
    }
  //all other materials have to be resized and set to 0 overrides on this material border:
  for (int i=0; i<numMats-1; i++) //over old materials
    {
      //MaterialCollectionData[i]->LossOverride.resize(numMats);
      //MaterialCollectionData[i]->LossOverride[thisMat]=0;
      //MaterialCollectionData[i]->ScatterOverride.resize(numMats);
      //MaterialCollectionData[i]->ScatterOverride[thisMat]=0;
      //MaterialCollectionData[i]->ScatterModel.resize(numMats);
      //MaterialCollectionData[i]->ScatterModel[thisMat]=1;  //2Pi
      //MaterialCollectionData[i]->ReflectionOverride.resize(numMats);
      //MaterialCollectionData[i]->ReflectionOverride[thisMat]=0;
      MaterialCollectionData[i]->OpticalOverrides.resize(numMats);
      MaterialCollectionData[i]->OpticalOverrides[thisMat] = NULL;
    }

  //inicialize empty MatParticle vector for all defined particles
  int numParticles = ParticleCollection->size();
  m->MatParticle.resize(numParticles);
  for (int i=0; i<numParticles; i++)
    {
      m->MatParticle[i].TrackingAllowed = false;
      m->MatParticle[i].PhYield=0;
      m->MatParticle[i].InteractionDataX.resize(0);
      m->MatParticle[i].InteractionDataF.resize(0);
      m->MatParticle[i].Terminators.resize(0);
    }

  //initialize wavelength-resolved data
  m->PrimarySpectrum_lambda.resize(0);
  m->PrimarySpectrum.resize(0);
  m->PrimarySpectrumHist = 0;
  m->SecondarySpectrum_lambda.resize(0);
  m->SecondarySpectrum.resize(0);
  m->SecondarySpectrumHist = 0;
  m->nWave_lambda.resize(0);
  m->nWave.resize(0);
  m->nWaveBinned.resize(0); //regular step (WaveStep step, WaveNodes bins)
  m->absWave_lambda.resize(0);
  m->absWave.resize(0);
  m->absWaveBinned.resize(0);//regular step (WaveStep step, WaveNodes bins)

  //appending to the material collection
  MaterialCollectionData.append(m);

  //tmpMaterial object has to be updated too!
     //tmpMaterial "knew" one less material, have to set overrides to itself to nothing
  //tmpMaterial.LossOverride.resize(numMats);
  //tmpMaterial.ScatterOverride.resize(numMats);
  //tmpMaterial.ScatterModel.resize(numMats);
  //tmpMaterial.ReflectionOverride.resize(numMats);
  tmpMaterial.OpticalOverrides.resize(numMats);
  for (int i=0; i<numMats; i++)
    {
      //tmpMaterial.LossOverride[thisMat]=0;
      //tmpMaterial.ScatterOverride[thisMat]=0;
      //tmpMaterial.ScatterModel[thisMat]=1;  //2Pi
      //tmpMaterial.ReflectionOverride[thisMat]=0;
      tmpMaterial.OpticalOverrides[thisMat] = NULL;
    }
}

void MaterialCollectionClass::UpdateMaterial(int index, QString name, double density, double atomicDensity, double n, double abs, double PriScintDecayTime,
                                double W, double SecYield, double SecScintDecayTime, double e_driftVelocity,
                                double p1, double p2, double p3)
{
  if (index <0 || index > MaterialCollectionData.size()-1)
    {
      qWarning()<<"Attempt to update non-existing material in UpdateMaterial!";
      return;
    }

  MaterialCollectionData[index]->name = name;

  MaterialCollectionData[index]->density = density;
  MaterialCollectionData[index]->atomicDensity = atomicDensity;
  MaterialCollectionData[index]->n = n;
  MaterialCollectionData[index]->abs = abs;

  MaterialCollectionData[index]->PriScintDecayTime = PriScintDecayTime;

  MaterialCollectionData[index]->W = W;
  MaterialCollectionData[index]->SecYield = SecYield;
  MaterialCollectionData[index]->SecScintDecayTime = SecScintDecayTime;
  MaterialCollectionData[index]->e_driftVelocity = e_driftVelocity;

  MaterialCollectionData[index]->p1 = p1;
  MaterialCollectionData[index]->p2 = p2;
  MaterialCollectionData[index]->p3 = p3;
}

void MaterialCollectionClass::ClearTmpMaterial()
{
  tmpMaterial.name = "";
  tmpMaterial.density = 0;
  tmpMaterial.atomicDensity = -1;
  tmpMaterial.p1 = 0;
  tmpMaterial.p2 = 0;
  tmpMaterial.p3 = 0;
  tmpMaterial.n = 1;
  tmpMaterial.abs = 0;
  tmpMaterial.e_driftVelocity = 0;
  tmpMaterial.W = 0;
  tmpMaterial.SecYield = 0;
  tmpMaterial.PriScintDecayTime = 0;
  tmpMaterial.SecScintDecayTime = 0;
  tmpMaterial.Comments = "";

  int particles = ParticleCollection->size();
  tmpMaterial.MatParticle.resize(particles);
  for (int i=0; i<particles; i++)
    {
      tmpMaterial.MatParticle[i].InteractionDataF.resize(0);
      tmpMaterial.MatParticle[i].InteractionDataX.resize(0);
      tmpMaterial.MatParticle[i].Terminators.resize(0);
      tmpMaterial.MatParticle[i].PhYield = 0;
      tmpMaterial.MatParticle[i].TrackingAllowed = false;
    }

  int materials = MaterialCollectionData.size();
  //tmpMaterial.ReflectionOverride.resize(materials);
  //tmpMaterial.LossOverride.resize(materials);
  //tmpMaterial.ScatterOverride.resize(materials);
  //tmpMaterial.ScatterModel.resize(materials);
  tmpMaterial.OpticalOverrides.resize(materials);
  for (int i=0; i<materials; i++)
    {
      //tmpMaterial.ReflectionOverride[i] = 0;
      //tmpMaterial.LossOverride[i] = 0;
      //tmpMaterial.ScatterOverride[i] = 0;
      //tmpMaterial.ScatterModel[i] = 1;
      if (tmpMaterial.OpticalOverrides[i])
        {
          delete tmpMaterial.OpticalOverrides[i];
          tmpMaterial.OpticalOverrides[i] = NULL;
        }
    }

  tmpMaterial.nWave_lambda.resize(0);
  tmpMaterial.nWave.resize(0);
  tmpMaterial.nWaveBinned.resize(0);
  tmpMaterial.absWave_lambda.resize(0);
  tmpMaterial.absWave.resize(0);
  tmpMaterial.absWaveBinned.resize(0);

  tmpMaterial.PrimarySpectrum_lambda.resize(0);
  tmpMaterial.PrimarySpectrum.resize(0);
  tmpMaterial.SecondarySpectrum_lambda.resize(0);
  tmpMaterial.SecondarySpectrum.resize(0);

  //---POINTERS---
  if (tmpMaterial.PrimarySpectrumHist)
    {
      delete tmpMaterial.PrimarySpectrumHist;
      tmpMaterial.PrimarySpectrumHist = 0;
    }
  if (tmpMaterial.SecondarySpectrumHist)
    {
      delete tmpMaterial.SecondarySpectrumHist;
      tmpMaterial.SecondarySpectrumHist = 0;
    }
  if (tmpMaterial.GeoMat)
    {
      delete tmpMaterial.GeoMat;
      tmpMaterial.GeoMat = 0;
    }
  if (tmpMaterial.GeoMed)
    {
      delete tmpMaterial.GeoMed;
      tmpMaterial.GeoMed = 0;
    }
}

void MaterialCollectionClass::CopyTmpToMaterialCollection()
{
  QString name = tmpMaterial.name;
  int index = MaterialCollectionClass::FindMaterial(name);

  if (index == -1)
    {
//      qDebug()<<"MaterialCollection--> New material: "<<name;
      MaterialCollectionClass::AddNewMaterial();
      index = MaterialCollectionData.size()-1;
    }
  else
    {
//      qDebug()<<"MaterialCollection--> Material "+name+" already defined; index = "<<index;
    }
  //*MaterialCollectionData[index] = tmpMaterial; //updating material properties
  QJsonObject js;
  tmpMaterial.writeToJson(js, ParticleCollection);
  SaveJsonToFile(js,"TMPjson.json");
  MaterialCollectionData[index]->readFromJson(js, this, ParticleCollection);

  //now update pointers!
  MaterialCollectionClass::UpdateWaveResolvedProperties(index); //updating effective properties (hists, Binned), remaking hist objects (Pointers are safe - they objects are recreated on each copy)
  return;
}

int MaterialCollectionClass::FindMaterial(QString name)
{
  int size = MaterialCollectionData.size();

  for (int index = 0; index < size; index++)
      if (name == MaterialCollectionData[index]->name) return index;

  return -1;
}

void MaterialCollectionClass::UpdateWaveResolvedProperties(int imat)
{
  //qDebug()<<"Wavelength-resolved?"<<WavelengthResolved;
  //qDebug()<<"--updating wavelength-resolved properties for material index"<<imat;
  if (WavelengthResolved)
    {
      //calculating histograms and "-Binned" for effective data
      MaterialCollectionData[imat]->nWaveBinned.resize(0);
      if (MaterialCollectionData[imat]->nWave_lambda.size() > 0)
           ConvertToStandardWavelengthes(&MaterialCollectionData[imat]->nWave_lambda, &MaterialCollectionData[imat]->nWave, &MaterialCollectionData[imat]->nWaveBinned);

      MaterialCollectionData[imat]->absWaveBinned.resize(0);
      if (MaterialCollectionData[imat]->absWave_lambda.size() > 0)
           ConvertToStandardWavelengthes(&MaterialCollectionData[imat]->absWave_lambda, &MaterialCollectionData[imat]->absWave, &MaterialCollectionData[imat]->absWaveBinned);

      if (MaterialCollectionData[imat]->rayleighMFP != 0)
        {
          MaterialCollectionData[imat]->rayleighBinned.resize(0);
          double baseWave4 = MaterialCollectionData[imat]->rayleighWave * MaterialCollectionData[imat]->rayleighWave * MaterialCollectionData[imat]->rayleighWave * MaterialCollectionData[imat]->rayleighWave;
          double base = MaterialCollectionData[imat]->rayleighMFP / baseWave4;
          for (int i=0; i<WaveNodes; i++)
            {
              double wave = WaveFrom + WaveStep*i;
              double wave4 = wave*wave*wave*wave;
              MaterialCollectionData[imat]->rayleighBinned.append(base * wave4);
            }
        }

      if (MaterialCollectionData[imat]->PrimarySpectrumHist)
        {
          delete MaterialCollectionData[imat]->PrimarySpectrumHist;
          MaterialCollectionData[imat]->PrimarySpectrumHist = 0;
        }
      if (MaterialCollectionData[imat]->PrimarySpectrum_lambda.size() > 0)
        {
          QVector<double> y;
          ConvertToStandardWavelengthes(&MaterialCollectionData[imat]->PrimarySpectrum_lambda, &MaterialCollectionData[imat]->PrimarySpectrum, &y);
          TString name = "PrimScSp";
          name += imat;
          MaterialCollectionData[imat]->PrimarySpectrumHist = new TH1D(name,"Primary scintillation", WaveNodes, WaveFrom, WaveTo);
          for (int j = 1; j<WaveNodes+1; j++)  MaterialCollectionData[imat]->PrimarySpectrumHist->SetBinContent(j, y[j-1]);
          MaterialCollectionData[imat]->PrimarySpectrumHist->GetIntegral(); //to make thread safe
        }

      if (MaterialCollectionData[imat]->SecondarySpectrumHist)
        {
          delete MaterialCollectionData[imat]->SecondarySpectrumHist;
          MaterialCollectionData[imat]->SecondarySpectrumHist = 0;
        }
      if (MaterialCollectionData[imat]->SecondarySpectrum_lambda.size() > 0)
        {
          QVector<double> y;
          ConvertToStandardWavelengthes(&MaterialCollectionData[imat]->SecondarySpectrum_lambda, &MaterialCollectionData[imat]->SecondarySpectrum, &y);
          TString name = "SecScSp";
          name += imat;
          MaterialCollectionData[imat]->SecondarySpectrumHist = new TH1D(name,"Secondary scintillation", WaveNodes, WaveFrom, WaveTo);
          for (int j = 1; j<WaveNodes+1; j++)  MaterialCollectionData[imat]->SecondarySpectrumHist->SetBinContent(j, y[j-1]);
          MaterialCollectionData[imat]->SecondarySpectrumHist->GetIntegral(); //to make thread safe
        }
    }
  else
    {
      //making empty histograms and "-Binned" for effective data
      MaterialCollectionData[imat]->nWaveBinned.resize(0);
      MaterialCollectionData[imat]->absWaveBinned.resize(0);
      MaterialCollectionData[imat]->rayleighBinned.resize(0);
      if (MaterialCollectionData[imat]->PrimarySpectrumHist)
        {
          delete MaterialCollectionData[imat]->PrimarySpectrumHist;
          MaterialCollectionData[imat]->PrimarySpectrumHist = 0;
        }
      if (MaterialCollectionData[imat]->SecondarySpectrumHist)
        {
          delete MaterialCollectionData[imat]->SecondarySpectrumHist;
          MaterialCollectionData[imat]->SecondarySpectrumHist = 0;
        }
    }
}

void MaterialCollectionClass::CopyMaterialToTmp(int imat)
{
  if (imat<0 || imat>MaterialCollectionData.size()-1)
    {
      qWarning()<<"Error: attempting to copy non-existent material #"<<imat<< " to tmpMaterial!";
      return;
    }
  //tmpMaterial = *MaterialCollectionData[imat];
  QJsonObject js;
  MaterialCollectionData[imat]->writeToJson(js, ParticleCollection);
  tmpMaterial.readFromJson(js, this, ParticleCollection);
}

void MaterialCollectionClass::ConvertToStandardWavelengthes(QVector<double>* sp_x, QVector<double>* sp_y, QVector<double>* y)
{
  y->resize(0);

  //qDebug()<<"Data range:"<<sp_x->at(0)<<sp_x->at(sp_x->size()-1);
  double xx, yy;
  for (int i=0; i<WaveNodes; i++)
    {
      xx = WaveFrom + WaveStep*i;
      if (xx <= sp_x->at(0)) yy = sp_y->at(0);
      else
        {
          if (xx >= sp_x->at(sp_x->size()-1)) yy = sp_y->at(sp_x->size()-1);
          else
            {
              //general case
              yy = InteractionValue(xx, sp_x, sp_y); //reusing interpolation function from functions.h
              if (yy<0) yy = 0; //!!! protection against negative
            }
        }
//      qDebug()<<xx<<yy;
      y->append(yy);
    }
}

int MaterialCollectionClass::SaveMaterialData(MaterialStructure* tmpMaterial, QString outputFilename)
{
     QFile outputFile(outputFilename);
     outputFile.open(QIODevice::WriteOnly);

     if(!outputFile.isOpen()){
             qCritical() << "- Error, unable to open" << outputFilename << "for output";
             return 1;
        }

     QString MatName = tmpMaterial->name;
     QTextStream outStream(&outputFile);

     //writing general data
     outStream << "Material name:" << MatName<<"\n";
     outStream << "Density:" << tmpMaterial->density<<"\n";
     outStream << "AtomicDensity:" << tmpMaterial->atomicDensity<<"\n";
     outStream << "Refractive index:"<< tmpMaterial->n<<"\n";
     outStream << "Bulk absorption:"<< tmpMaterial->abs<<"\n";
     outStream << "Primary DecayTime:"<< tmpMaterial->PriScintDecayTime<<"\n";
     outStream << "Default W:"<< tmpMaterial->W <<"\n";
     outStream << "Secondary Photon Yield:"<< tmpMaterial->SecYield<<"\n";
     outStream << "Secondary DecayTime:"<< tmpMaterial->SecScintDecayTime<<"\n";
     outStream << "Electron drift velocity:"<< tmpMaterial->e_driftVelocity<<"\n";
     outStream << "TGeoMaterial parameter1:"<< tmpMaterial->p1<<"\n";
     outStream << "TGeoMaterial parameter2:"<< tmpMaterial->p2<<"\n";
     outStream << "TGeoMaterial parameter3:"<< tmpMaterial->p3<<"\n";

     for (int i=0; i<ParticleCollection->size(); i++)
       if (tmpMaterial->MatParticle[i].TrackingAllowed)
          //outStream << "Tracking allowed for:"<<MW->ParticleCollection[i]->ParticleName<<":"<<MW->ParticleCollection[i]->charged<<":"<<MW->ParticleCollection[i]->mass<<"\n";
          outStream << "Tracking allowed for:"<<(*ParticleCollection)[i]->ParticleName<<":"<<(*ParticleCollection)[i]->type<<":"<<(*ParticleCollection)[i]->charge<<":"<<(*ParticleCollection)[i]->mass<<"\n";

     //saving wavelength-resolved data
     if (tmpMaterial->nWave_lambda.size() >0)
       {
          QFileInfo QF(outputFilename);
          if (!QDir( QF.path() + "/WaveResolvedProperties").exists()) QDir().mkdir( QF.path() + "/WaveResolvedProperties");
          QString WfileName = QF.path() + "/WaveResolvedProperties/n.dat";
          outStream << "File refractive index:" << "WaveResolvedProperties/n.dat" <<"\n";
          SaveDoubleVectorsToFile(WfileName, &tmpMaterial->nWave_lambda, &tmpMaterial->nWave);
       }
     if (tmpMaterial->absWave_lambda.size() >0)
       {
          QFileInfo QF(outputFilename);
          if (!QDir( QF.path() + "/WaveResolvedProperties").exists()) QDir().mkdir( QF.path() + "/WaveResolvedProperties");
          QString WfileName = QF.path() + "/WaveResolvedProperties/abs.dat";
          outStream << "File attenuation coefficient:" << "WaveResolvedProperties/abs.dat" <<"\n";
          SaveDoubleVectorsToFile(WfileName, &tmpMaterial->absWave_lambda, &tmpMaterial->absWave);
       }
     if (tmpMaterial->PrimarySpectrum_lambda.size() >0)
       {
          QFileInfo QF(outputFilename);
          if (!QDir( QF.path() + "/WaveResolvedProperties").exists()) QDir().mkdir( QF.path() + "/WaveResolvedProperties");
          QString WfileName = QF.path() + "/WaveResolvedProperties/PrimaryScintSpectrum.dat";
          outStream << "File primary scintillation spectrum:" << "WaveResolvedProperties/PrimaryScintSpectrum.dat" <<"\n";
          SaveDoubleVectorsToFile(WfileName, &tmpMaterial->PrimarySpectrum_lambda, &tmpMaterial->PrimarySpectrum);
       }
     if (tmpMaterial->SecondarySpectrum_lambda.size() >0)
       {
          QFileInfo QF(outputFilename);
          if (!QDir( QF.path() + "/WaveResolvedProperties").exists()) QDir().mkdir( QF.path() + "/WaveResolvedProperties");
          QString WfileName = QF.path() + "/WaveResolvedProperties/SecondaryScintSpectrum.dat";
          outStream << "File secondary scintillation spectrum:" << "WaveResolvedProperties/SecondaryScintSpectrum.dat" <<"\n";
          SaveDoubleVectorsToFile(WfileName, &tmpMaterial->SecondarySpectrum_lambda, &tmpMaterial->SecondarySpectrum);
       }

          int numParticles = ParticleCollection->size();

          //counting how many particles have non-zero data
          int counter = 0;
          for (int i=0; i<numParticles; i++) if (tmpMaterial->MatParticle[i].InteractionDataX.size() > 0) counter++;
          outStream<< "Defined particles entries:" << counter <<"\n";

          for (int iParticle = 0; iParticle<numParticles; iParticle++)
            {
              if (tmpMaterial->MatParticle[iParticle].InteractionDataX.size() == 0 ) continue; //no interaction data - do not save

              QString ParticleName = (*ParticleCollection)[iParticle]->ParticleName;
              QFileInfo QF(outputFilename);
              QString ParticleDir = ParticleName;
              if (!QDir( QF.path() + "/" +ParticleDir).exists()) QDir().mkdir( QF.path() + "/" +ParticleDir);

              QString outPFileName = ParticleDir+"/" +ParticleName+".pa";
              outStream << "Interaction File:"<< outPFileName<<"\n";

              QFile outPFile( QF.path() + "/" +outPFileName);
              outPFile.open(QIODevice::WriteOnly);

              if(!outPFile.isOpen())
                {
                  qCritical() << "- Error, unable to open" << outPFileName << "for output";
                  outputFile.close();
                  return 2;
                }
              QTextStream outStream1(&outPFile);
              outStream1 << "Name:"<<ParticleName<<"\n";
              outStream1 << "Particle_type:"<<(*ParticleCollection)[iParticle]->type<<"\n";
              //outStream1 << "Charged:"<<MW->ParticleCollection[iParticle]->charged<<"\n";
              outStream1 << "Charge:"<<(*ParticleCollection)[iParticle]->charge<<"\n";
              outStream1 << "Mass:"<<(*ParticleCollection)[iParticle]->mass<<"\n";
              outStream1 << "----------\n";
              outStream1 << "Tracking allowed:"<<tmpMaterial->MatParticle[iParticle].TrackingAllowed<<"\n";
              outStream1 << "Material is transparent:"<<tmpMaterial->MatParticle[iParticle].MaterialIsTransparent<<"\n";
              outStream1 << "PhotonYield:"<<tmpMaterial->MatParticle[iParticle].PhYield<<"\n";

              //saving Total Interaction data
              //QString outIFileName ="MaterialLibrary/"+MatName+"/"+ParticleName+"/TotalInteraction.dat";
              QString outIFileName = ParticleDir+"/TotalInteraction.dat";
              QFile outIFile( QF.path() + "/" +outIFileName);
              outIFile.open(QIODevice::WriteOnly);
              if(!outIFile.isOpen())
                {
                  qCritical() << "- Error, unable to open" << outIFileName << "for output";
                  outPFile.close();
                  outputFile.close();
                  return 3;
                }
              QTextStream outStream2(&outIFile);
              for (int i=0; i<tmpMaterial->MatParticle[iParticle].InteractionDataX.size(); i++)
                outStream2 <<tmpMaterial->MatParticle[iParticle].InteractionDataX[i]<<" "<<tmpMaterial->MatParticle[iParticle].InteractionDataF[i]<<"\n";
              outIFile.close();

              //if gamma or neutron, saving terminators
              if ( (*ParticleCollection)[iParticle]->type == _gamma_) // if (iParticle == 0)  //*** absolute!!!
                {
                  int iTerminators = tmpMaterial->MatParticle[iParticle].Terminators.size();
                  outStream1 << "Terminators:"<<iTerminators<<"\n";

                  for (int iTerm=0; iTerm<iTerminators; iTerm++)
                    {
                      int ReactionType = tmpMaterial->MatParticle[iParticle].Terminators[iTerm].ReactionType;
                      outStream1 << "Type:"<<ReactionType<<"\n";

                      //saving partial interaction coefficient
                      QString str;
                      str.setNum(iTerm);
                   //  QString outIFileName1 ="MaterialLibrary/"+MatName+"/"+ParticleName+"/PartialInteraction"+str+".dat";
                      QString outIFileName1 =ParticleDir + "/PartialInteraction"+str+".dat";
                      QFile outIFile1( QF.path() + "/" +outIFileName1);
                      outIFile1.open(QIODevice::WriteOnly);
                      if(!outIFile1.isOpen())
                        {
                          qCritical() << "- Error, unable to open" << outIFileName1 << "for output";
                          outPFile.close();
                          outputFile.close();
                          return 4;
                        }
                      QTextStream outStream3(&outIFile1);
                      for (int i=0; i<tmpMaterial->MatParticle[iParticle].Terminators[iTerm].PartialCrossSectionEnergy.size(); i++)
                        outStream3 <<tmpMaterial->MatParticle[iParticle].Terminators[iTerm].PartialCrossSectionEnergy[i]<<" "<<tmpMaterial->MatParticle[iParticle].Terminators[iTerm].PartialCrossSection[i]<<"\n";
                      outIFile1.close();
                    }
                }

              if ( (*ParticleCollection)[iParticle]->type == _neutron_) //if (iParticle == 2)  //*** absolute!!!
                {
                  int iTerminators = tmpMaterial->MatParticle[iParticle].Terminators.size();
                  outStream1 << "Terminators:"<<iTerminators<<"\n";

                  for (int iTerm=0; iTerm<iTerminators; iTerm++)
                    {
                      //always "capture"
                      outStream1 <<"-----\n";
                      outStream1 <<"Branching:"<<tmpMaterial->MatParticle[iParticle].Terminators[iTerm].branching<<"\n";
                      int secParticles= tmpMaterial->MatParticle[iParticle].Terminators[iTerm].GeneratedParticles.size();
                      outStream1 <<"Secondary particles:"<<secParticles<<"\n";
                      for (int i=0; i<secParticles; i++)
                        {
                          int oneParticle = tmpMaterial->MatParticle[iParticle].Terminators[iTerm].GeneratedParticles[i];
                          //outStream1 <<"Particle:"<<MW->ParticleCollection[oneParticle]->ParticleName<<":"<<MW->ParticleCollection[oneParticle]->charged<<":"<<MW->ParticleCollection[oneParticle]->mass<<"\n";
                          outStream1 <<"Particle:"<< (*ParticleCollection)[oneParticle]->ParticleName<<":"<< (*ParticleCollection)[oneParticle]->type<<":"<< (*ParticleCollection)[oneParticle]->charge<<":"<< (*ParticleCollection)[oneParticle]->mass<<"\n";
                          outStream1 <<"Energy:"<<tmpMaterial->MatParticle[iParticle].Terminators[iTerm].GeneratedParticleEnergies[i]<<"\n";
                        }
                    }
                }

              outPFile.close();
             }

    outputFile.close();
    return 0;
}

int MaterialCollectionClass::LoadMaterialData(QString fileName)
{
    QFile file(fileName);
    if(!file.open(QIODevice::ReadOnly | QFile::Text))
          {
            qCritical()<<"Failed to open file" + fileName + "!";
            return 1;
          }
    //qDebug()<<"clear tmp material";
    MaterialCollectionClass::ClearTmpMaterial();

    QTextStream in(&file);
    QFileInfo QF(fileName);

    tmpMaterial.rayleighMFP = 0;//compatibility

    while(!in.atEnd()) {
            QString line = in.readLine();
            QStringList fields = line.split(":");

            if (!fields[0].compare("Material name",Qt::CaseSensitive)) tmpMaterial.name = fields[1];
            if (!fields[0].compare("Density",Qt::CaseSensitive)) tmpMaterial.density = fields[1].toDouble();
            if (!fields[0].compare("AtomicDensity",Qt::CaseSensitive)) tmpMaterial.atomicDensity = fields[1].toDouble();
            if (!fields[0].compare("Refractive index",Qt::CaseSensitive)) tmpMaterial.n = fields[1].toDouble();
            if (!fields[0].compare("Bulk absorption",Qt::CaseSensitive)) tmpMaterial.abs =fields[1].toDouble();
            if (!fields[0].compare("Primary DecayTime",Qt::CaseSensitive)) tmpMaterial.PriScintDecayTime = fields[1].toDouble();
            if (!fields[0].compare("Default W",Qt::CaseSensitive)) tmpMaterial.W = fields[1].toDouble();
            if (!fields[0].compare("Secondary Photon Yield",Qt::CaseSensitive)) tmpMaterial.SecYield = fields[1].toDouble();
            if (!fields[0].compare("Secondary DecayTime",Qt::CaseSensitive)) tmpMaterial.SecScintDecayTime = fields[1].toDouble();
            if (!fields[0].compare("Electron drift velocity",Qt::CaseSensitive)) tmpMaterial.e_driftVelocity = fields[1].toDouble();
            if (!fields[0].compare("TGeoMaterial parameter1",Qt::CaseSensitive)) tmpMaterial.p1 = fields[1].toDouble();
            if (!fields[0].compare("TGeoMaterial parameter2",Qt::CaseSensitive)) tmpMaterial.p2 = fields[1].toDouble();
            if (!fields[0].compare("TGeoMaterial parameter3",Qt::CaseSensitive)) tmpMaterial.p3 = fields[1].toDouble();

            if (fields[0] == "Tracking allowed for")
              {
//                qDebug()<<"====checking tracking allowed";
                //if found in the active list, set as tracking allowed, otherwise ignore: if interaction data are also given later, the particle will be added
                //added since there are cases when tracking is allowed for a material, but there is no interaction data (transparent material)
                //this is only relevant for load of materials in frame of LoadDetConfig, where the list of particles is force-loaded on the beginning anyway
                QString name;
                ParticleType type;
                double mass;
                int charge;
                if (fields.count() == 4)
                  {
                    //old save
                    name = fields[1];
                    bool chargedFlag = fields[2].toInt();
                    mass = fields[3].toDouble();
                    charge = 0; //now info in old save
                    if (chargedFlag)
                      {
                        type = _charged_;
                        if (name == "alpha") charge = 2;
                        if (name == "electron") charge = -1;
                        if (name == "proton") charge = 1;
                        if (name == "triton") charge = 1;
                      }
                    else
                      {
                        if (name == "neutron") type = _neutron_;
                        else if (name == "gamma") type = _gamma_;
                        else
                          {
                            message("Error: Unknown particle type!");
                            type = _charged_;
                          }
                      }
                  }
                else
                  {
                    //new system
                    name = fields[1];
                    type = static_cast<ParticleType>(fields[2].toInt());
                    charge = fields[3].toInt();
                    mass = fields[4].toDouble();
                  }
                //is this particle already in the list of defined ons?
//                qDebug()<<"Tracking allowed for:";
                int ParticleId = MaterialCollectionClass::FindCreateParticle(name, type, charge, mass);
                if (ParticleId > -1) tmpMaterial.MatParticle[ParticleId].TrackingAllowed = true;                
              }

            //Wave-resolved properties
            if (!fields[0].compare("File refractive index",Qt::CaseSensitive))
              {
                QString fileName = QF.path()+"/"+fields[1];
                LoadDoubleVectorsFromFile(fileName, &tmpMaterial.nWave_lambda, &tmpMaterial.nWave);

              }
            if (!fields[0].compare("File attenuation coefficient",Qt::CaseSensitive))
              {
                QString fileName = QF.path()+"/"+fields[1];
                LoadDoubleVectorsFromFile(fileName, &tmpMaterial.absWave_lambda, &tmpMaterial.absWave);
              }
            if (!fields[0].compare("File primary scintillation spectrum",Qt::CaseSensitive))
              {
                QString fileName = QF.path()+"/"+fields[1];
                LoadDoubleVectorsFromFile(fileName, &tmpMaterial.PrimarySpectrum_lambda, &tmpMaterial.PrimarySpectrum);
              }
            if (!fields[0].compare("File secondary scintillation spectrum",Qt::CaseSensitive))
              {
                QString fileName = QF.path()+"/"+fields[1];
                LoadDoubleVectorsFromFile(fileName, &tmpMaterial.SecondarySpectrum_lambda, &tmpMaterial.SecondarySpectrum);
              }

            //Interaction file
            if (!fields[0].compare("Interaction File",Qt::CaseSensitive))
              {
                QString str = "Interaction File:";
                QString ParticleFileName = line.remove(0, str.size());
                ParticleFileName = QF.path()+"/"+ParticleFileName;
 //               qDebug()<<"Reading particle information: "<<ParticleFileName;
                if (!ParticleFileName.isEmpty())
                  {
                    QFile Pfile(ParticleFileName);

                    if(!Pfile.open(QIODevice::ReadOnly | QFile::Text)) {
                        qCritical()<<"Failed to open file "+ParticleFileName;
                        file.close();
                        return 2;
                    }
                    QTextStream inP(&Pfile);
                    QString Name;
                  //  bool Charged; //old
                    ParticleType Type;
                    double Mass, PhYield;
                    int Charge;
                    int scenarios=0;
                    QVector<int> GammaProcesses;
                    GammaProcesses.resize(0);
                    bool trackingAllowed = true; //since in old versions can be missing in save
                    int materialIsTransparent = -1; //IsTransparent property is absent in old versions;

                    QVector<double> Branchings;
                    Branchings.resize(0);
                    QVector<int> NumSecondaries;
                    NumSecondaries.resize(0);
                    QVector<ParticleDescriptor> LoadedParticles;
                    LoadedParticles.resize(0);
                    QVector<double> Energies;
                    Energies.resize(0);

                    while(!inP.atEnd())
                      {
                        line = inP.readLine();
                        fields = line.split(":");
                        if (!fields[0].compare("Name",Qt::CaseSensitive)) Name = fields[1];
                        if (!fields[0].compare("Particle_type",Qt::CaseSensitive)) Type = static_cast<ParticleType>(fields[1].toInt());
                        if (!fields[0].compare("Charged",Qt::CaseSensitive))
                          {
                            bool chargedFlag = fields[1].toInt();
 //                           qDebug()<<"here, name:"<<Name<<"charged:"<<chargedFlag;
                            //old system

                            Charge = 0; //no info in old save
                            if (chargedFlag)
                              {
                                Type = _charged_;
                                if (Name == "alpha") Charge = 2;
                                if (Name == "electron") Charge = -1;
                                if (Name == "proton") Charge = 1;
                                if (Name == "triton") Charge = 1;
                              }
                            else
                              {
                                if (Name == "gamma") Type = _gamma_;
                                else Type = _neutron_;
                              }

                          }
                        if (!fields[0].compare("Charge",Qt::CaseSensitive)) Charge = fields[1].toDouble();
                        if (!fields[0].compare("Mass",Qt::CaseSensitive)) Mass = fields[1].toDouble();

                        if (!fields[0].compare("Tracking allowed",Qt::CaseSensitive)) trackingAllowed = fields[1].toInt();
                        if (!fields[0].compare("Material is transparent",Qt::CaseSensitive)) materialIsTransparent = fields[1].toInt();

                        if (!fields[0].compare("PhotonYield",Qt::CaseSensitive)) PhYield = fields[1].toDouble();
                        if (!fields[0].compare("Terminators",Qt::CaseSensitive)) scenarios = fields[1].toInt();

                        if (!fields[0].compare("Type",Qt::CaseSensitive)) GammaProcesses.append(fields[1].toInt());

                        if (!fields[0].compare("Branching",Qt::CaseSensitive)) Branchings.append(fields[1].toDouble());
                        if (!fields[0].compare("Secondary particles",Qt::CaseSensitive)) NumSecondaries.append(fields[1].toInt());
                        if (!fields[0].compare("Energy",Qt::CaseSensitive)) Energies.append(fields[1].toDouble());
                        if (!fields[0].compare("Particle",Qt::CaseSensitive)) //secondary particles
                          {
                            ParticleDescriptor tmpP;
                            if (fields.count() == 4)
                              {
                                //old save
                                QString name = fields[1];
                                bool chargedFlag = fields[2].toInt();
                                double mass = fields[3].toDouble();
                                ParticleType type;
                                int charge = 0;
                                if (chargedFlag)
                                  {
                                    type = _charged_;
                                    if (name == "alpha") charge = 2;
                                    if (name == "electron") charge = -1;
                                    if (name == "proton") charge = 1;
                                    if (name == "triton") charge = 1;
                                  }
                                else
                                  {
                                    if (name == "neutron") type = _neutron_;
                                    else if (name == "gamma") type = _gamma_;
                                    else
                                      {
                                        message("Error: Unknown particle type!");
                                        type = _charged_;
                                      }
                                  }
                                tmpP = ParticleDescriptor(name, type, charge, mass);
                              }
                            else
                              {
                                //new system
                                tmpP = ParticleDescriptor(fields[1], static_cast<ParticleType>(fields[2].toInt()), fields[3].toInt(), fields[4].toDouble());
                              }
                            LoadedParticles.append(tmpP);
                          }
                      }
                   // qDebug()<<Name<<Charged<<Mass;
                    int ParticleId = MaterialCollectionClass::FindCreateParticle(Name, Type, Charge, Mass); //if exists - gives Id, otherwise creates new one

                    //setting the MatParticle properties for this particle
                    tmpMaterial.MatParticle[ParticleId].TrackingAllowed = trackingAllowed;
                    tmpMaterial.MatParticle[ParticleId].PhYield = PhYield;

                    //LoadingTotalInteraction data
                    tmpMaterial.MatParticle[ParticleId].InteractionDataX.resize(0);
                    tmpMaterial.MatParticle[ParticleId].InteractionDataF.resize(0);
                    QString IFileName = ParticleFileName;
                    IFileName.truncate(ParticleFileName.size() -Name.size()-3); //-3 is ".pa"
                    IFileName += "TotalInteraction.dat";
 //                   qDebug()<<"loading total interaction data: "<<IFileName;
                    int errorIO = LoadDoubleVectorsFromFile(IFileName,
                                                             &tmpMaterial.MatParticle[ParticleId].InteractionDataX,
                                                             &tmpMaterial.MatParticle[ParticleId].InteractionDataF);

                    if (errorIO > 0)
                      {
                        qCritical()<<"Failed to open file "<<IFileName;
                        Pfile.close();
                        file.close();
                        return 3;
                      }

                    //final decision on transparency of material for this particle
                    if (materialIsTransparent == -1)
                      {
                        if (tmpMaterial.MatParticle[ParticleId].InteractionDataX.size()>0) tmpMaterial.MatParticle[ParticleId].MaterialIsTransparent = false;
                        else tmpMaterial.MatParticle[ParticleId].MaterialIsTransparent = true;
                      }
                    else tmpMaterial.MatParticle[ParticleId].MaterialIsTransparent =  materialIsTransparent;
 //                   qDebug()<<"++++++++"<<tmpMaterial.name<<tmpMaterial.MatParticle[ParticleId].MaterialIsTransparent;

                    //next gamma and netron -related data
                    //if (ParticleId == 0)
                    if ( (*ParticleCollection)[ParticleId]->type == _gamma_)
                      {
                        //gamma
                        tmpMaterial.MatParticle[ParticleId].Terminators.resize(scenarios);
                        for (int i=0; i<scenarios; i++)
                          {
                            tmpMaterial.MatParticle[ParticleId].Terminators[i].ReactionType = GammaProcesses[i];
                            if (GammaProcesses[i] == 1) //compton
                              {
                                //adding gamma to secondary particles
                                int Id = MaterialCollectionClass::FindCreateParticle("gamma", _gamma_, 0, 0, true);
                                if (Id<0)
                                  {
                                    message("Error: gamma was not found!");
                                    Id = 0;
                                  }
                                tmpMaterial.MatParticle[ParticleId].Terminators[i].GeneratedParticles.clear();
                                tmpMaterial.MatParticle[ParticleId].Terminators[i].GeneratedParticles.append(Id);
                              }
                            QString FileName = ParticleFileName;
                            FileName.truncate(ParticleFileName.size() -Name.size()-3); //-3 is ".pa"
                            QString str;
                            str.setNum(i);
                            FileName += "PartialInteraction" + str + ".dat";
//                            qDebug()<<"loading partial interaction data: "<<FileName;
                            int errorIO = LoadDoubleVectorsFromFile(FileName,
                                                      &tmpMaterial.MatParticle[ParticleId].Terminators[i].PartialCrossSectionEnergy,
                                                      &tmpMaterial.MatParticle[ParticleId].Terminators[i].PartialCrossSection);  //cleans previous data
                            if (errorIO > 0)
                              {
                                qCritical()<<"Failed to open file "<< FileName;
                                Pfile.close();
                                file.close();
                                return 4;
                              }
                          }
                      }

                    //if (ParticleId == 2)
                    if ( (*ParticleCollection)[ParticleId]->type == _neutron_)
                      {
                        //neutron
                        tmpMaterial.MatParticle[ParticleId].Terminators.resize(scenarios);
                        for (int s=0; s<scenarios; s++)
                          {
                            tmpMaterial.MatParticle[ParticleId].Terminators[s].branching = Branchings[s] ;
                            tmpMaterial.MatParticle[ParticleId].Terminators[s].ReactionType = 2;
                            tmpMaterial.MatParticle[ParticleId].Terminators[s].GeneratedParticles.resize(0);
                            tmpMaterial.MatParticle[ParticleId].Terminators[s].GeneratedParticleEnergies.resize(0);
                            for (int i=0; i<NumSecondaries[s]; i++)
                              {
                                //int SecId = FindCreateParticle(LoadedParticles[0].ParticleName, LoadedParticles[0].charged, LoadedParticles[0].mass);
                                int SecId = MaterialCollectionClass::FindCreateParticle(LoadedParticles[0].ParticleName, LoadedParticles[0].type, LoadedParticles[0].charge, LoadedParticles[0].mass);
                                LoadedParticles.remove(0);
                                tmpMaterial.MatParticle[ParticleId].Terminators[s].GeneratedParticles.append(SecId);
                                tmpMaterial.MatParticle[ParticleId].Terminators[s].GeneratedParticleEnergies.append(Energies[0]);
                                Energies.remove(0);
                              }
                          }
                        //have to calculate partial interaction data (cross-sections) using total interaction data and branchings
                        int dataPoints = tmpMaterial.MatParticle[ParticleId].InteractionDataX.size();
                        for (int s=0; s<scenarios; s++)
                        {
                            double branching = tmpMaterial.MatParticle[ParticleId].Terminators[s].branching;
                            tmpMaterial.MatParticle[ParticleId].Terminators[s].PartialCrossSectionEnergy.resize(dataPoints); //*** TO ADD clean up
                            tmpMaterial.MatParticle[ParticleId].Terminators[s].PartialCrossSection.resize(dataPoints);
                            for (int i=0; i<dataPoints; i++)
                            {
                                tmpMaterial.MatParticle[ParticleId].Terminators[s].PartialCrossSection[i] = branching * tmpMaterial.MatParticle[ParticleId].InteractionDataF[i];
                                tmpMaterial.MatParticle[ParticleId].Terminators[s].PartialCrossSectionEnergy[i] = tmpMaterial.MatParticle[ParticleId].InteractionDataX[i];
                            }
                        }

                      }
                    Pfile.close();
                  }
              }
        }
        file.close();
        fLogLogInterpolation = false;
        return 0;
}

int MaterialCollectionClass::FindCreateParticle(QString Name, ParticleType Type, int Charge, double Mass, bool fOnlyFind)
{
  //  qDebug()<<"-----Looking for:"<<Name<<Type<<Charge<<Mass;
  int ParticleId;
  bool flagFound = false;
  //is this particle already in the list of defined ons?
  for (int Id=0; Id<ParticleCollection->size(); Id++)
    if (Name == (*ParticleCollection)[Id]->ParticleName && Type==(*ParticleCollection)[Id]->type && Charge==(*ParticleCollection)[Id]->charge && Mass==(*ParticleCollection)[Id]->mass)
      {
        ParticleId = Id;
        flagFound= true;
        //            qDebug()<<"found:"<<ParticleId;
        break;
      }
  //  qDebug()<<"------Found?"<<flagFound<<ParticleId;
  if (fOnlyFind)
    {
      if (flagFound) return ParticleId;
      else return -1;
    }
  if (!flagFound)
    {
      //have to define a new particle
      ParticleDescriptor* tmpP = new ParticleDescriptor(Name, Type, Charge, Mass);
      ParticleId = ParticleCollection->size();
      ParticleCollection->append(tmpP);
      MaterialCollectionClass::RegisterNewParticle(); //resize particle-indexed properties of all materials
      //      qDebug()<<"-------created new particle: "<<ParticleId;
    }
  return ParticleId;
}

int MaterialCollectionClass::findOrCreateParticle(QJsonObject &json)
{
  ParticleDescriptor p;
  p.readFromJson(json);
  return FindCreateParticle(p.ParticleName, p.type, p.charge, p.mass, false);
}

int MaterialCollectionClass::CheckMaterial(MaterialStructure* mat, int iPart)
{
  //returns:
  //0 - success
  //1 - wrong material or particle index
    //2 - total interaction is empty or contains only 1 point
    //3 - mismatch in total interaction (X and F) length
    //4 - total interaction: energy binning is not non-decreasing
    //5x - error in terminators for gamma
    //6 - isotope density is not defined while neutron interaction is defined
    //7 - error in terminators for neutron    
    //71 - error with generated particles for neutron
    //72 - error with generated particles energies for neutron
    //73 - mismatch between container size of genereated particle and their energy
    //74 - non-existent particle declared as secondary particle
    //75 - neutron: decay channels following capture are not defined
    //76 - initial energy of generated particle is outside of interaction data range

    if (iPart<0 || iPart>ParticleCollection->size()-1) return 1;

    //qDebug()<<">"<<mat->name<<" - "<<ParticleCollection->at(iPart)->ParticleName;

    MatParticleStructure* mp = &mat->MatParticle[iPart];
      //shortcuts - if particle tracking is disabled or material is transparent - no checks
    if ( ! mp->TrackingAllowed) return 0;
    if ( mp->MaterialIsTransparent) return 0;

      //common block - check TotalInteraction
    int interSize = mp->InteractionDataX.size();
    if (interSize == 0 || mp->InteractionDataF.size() == 0) return 2;
    if (interSize<2 || mp->InteractionDataF.size()<2) return 21;
    if (interSize != mp->InteractionDataF.size()) return 3;
    for (int i=1; i<interSize; i++) //protected against length = 0 or 1
        if (mp->InteractionDataX[i-1] > mp->InteractionDataX[i]) return 4;

      //getting particle type, and type-specific checks
    ParticleType pt = ParticleCollection->at(iPart)->type;

    if (pt == _charged_) return 0;

    if (pt == _gamma_)
      {
        if (mp->Terminators.isEmpty()) return 50;
        if (mp->Terminators.size() != 2) return 50;
        if ( ! (mp->Terminators[0].ReactionType == 0 && mp->Terminators[1].ReactionType == 1 || mp->Terminators[0].ReactionType == 1 && mp->Terminators[1].ReactionType == 0 ) ) return 51;

        //checking partial crossection for terminators
        for (int iTerm=0; iTerm<2; iTerm++)
          {
            QVector<double>*  e = &mp->Terminators[iTerm].PartialCrossSectionEnergy;
            QVector<double>* cs = &mp->Terminators[iTerm].PartialCrossSection;
            if (e->size() != interSize || cs->size() != interSize) return 52;
            for (int i=0; i<interSize; i++) if (mp->InteractionDataX[i] != e->at(i)) return 53;
          }

        // confirminf that TotalInteraction is indeed the sum of two process partial cross-sections
        for (int i=0; i<interSize; i++)
          {
            double delta = mp->Terminators[0].PartialCrossSection[i] + mp->Terminators[1].PartialCrossSection[i] - mp->InteractionDataF[i];
            if ( fabs(delta) > 0.0001*mp->InteractionDataF[i] ) return 54;
          }

        //properties for gamma are properly defined!
        return 0;
      }

    if (pt == _neutron_)
      {        
        if (mat->atomicDensity == 0) return 6;

        if (mp->Terminators.isEmpty()) return 75;

        int numTerm = mp->Terminators.size();
        //qDebug()<<"Terms:"<<numTerm;
        if (numTerm == 0) return 75;

        //confirming all terminators type is "capture"
        for (int i=0; i<numTerm; i++)
          {
            if (mp->Terminators[i].ReactionType != 2) return 7;
            //qDebug() << mp->Terminators[i].ReactionType;
          }

        //checking all terminator one by one
        for (int iTerm=0; iTerm<numTerm; iTerm++)
          {
            //check partial cross-section
            QVector<double>* e  = &mp->Terminators[iTerm].PartialCrossSectionEnergy;
            QVector<double>* cs = &mp->Terminators[iTerm].PartialCrossSection;
//            qDebug() << "---"<<interSize<<e->size()<<cs->size();
            if (e->size() != interSize || cs->size() != interSize) return 7;
            for (int i=0; i<interSize; i++) if (mp->InteractionDataX[i] != e->at(i)) return 7;

            // confirminf that partial cross section is indeed Branching fractin of TotalInteraction
            double branch = mp->Terminators[iTerm].branching;
            for (int i=0; i<interSize; i++)
              {
                double delta = mp->Terminators[iTerm].PartialCrossSection[i] - branch*mp->InteractionDataF[i];
                if ( fabs(delta) > 0.0001*mp->InteractionDataF[i] ) return 7;
              }

            //checking generated particles and their initial energies
            QVector<int>* gp = &mp->Terminators[iTerm].GeneratedParticles;
            if (gp->isEmpty()) return 71;
            QVector<double>* gpE = &mp->Terminators[iTerm].GeneratedParticleEnergies;
            if (gp->isEmpty()) return 72;
            int numPart = gp->size();
            if (numPart != gpE->size()) return 73;
              //checking particle/energy pairs one by one
            for (int iGP=0; iGP<numPart; iGP++)
              {
                int part = gp->at(iGP);
                if (part<0 || part>ParticleCollection->size()-1) return 74;  //unknown particle

                //test that all materials where this particle can be tracked (not-trasparent),
                 //have initial energy within the total interaction energy range
                int energy = gpE->at(iGP);

                ConflictingMaterialIndex = MaterialCollectionClass::CheckParticleEnergyInRange(part, energy);
                if (ConflictingMaterialIndex != -1)
                  {
                    ConflictingParticleIndex = part;
                    return 76;
                  }
              }
          }
      }

    return 0; //passed all tests
}

int MaterialCollectionClass::CheckMaterial(int iMat, int iPart)
{
  if (iMat<0 || iMat>MaterialCollectionData.size()-1) return 1;
  return MaterialCollectionClass::CheckMaterial(MaterialCollectionData[iMat], iPart);
}

int MaterialCollectionClass::CheckMaterial(int iMat)
{
  for (int iPart=0; iPart<ParticleCollection->size(); iPart++)
    {
      int ierr = MaterialCollectionClass::CheckMaterial(iMat, iPart);
      if (ierr>0) return ierr;
    }
  return 0;
}

int MaterialCollectionClass::CheckTmpMaterial()
{
  for (int iPart=0; iPart<ParticleCollection->size(); iPart++)
    {
      int ierr = MaterialCollectionClass::CheckMaterial(&tmpMaterial, iPart);
      if (ierr>0) return ierr;
    }
  return 0;
}

QString MaterialCollectionClass::getErrorString(int iError)
{
   switch (iError)
     {
     case 0: return "No errors";
     case 1: return "Wrong material or particle index";
     case 2: return "Total interaction data is empty";
     case 21: return "Total interaction data size is less than 2 points";
     case 3: return "Mismatch in total interaction (X and F) length";
     case 4: return "Total interaction: energy binning is not non-decreasing";
     case 5:
     case 51:
     case 52:
     case 53:
     case 54: return "Error in terminators for gamma";
     case 6: return "Isotope density is not defined while neutron interaction is enabled";
     case 7: return "Error in terminators for neutron";     
     case 71: return "Error in generated particles for neutron capture";
     case 72: return "Error with generated particles energies for neutron capture";
     case 73: return "Neutron capture: Mismatch between container size of genereated particle and their energy";
     case 74: return "Non-existent particle declared as secondary particle";
     case 75: return "Neutron: decay channels following capture are not defined";
     case 76:
       {
         QString tmp = "Initial energy of generated ";
         tmp += ParticleCollection->at(ConflictingParticleIndex)->ParticleName;
         tmp += "is outse of interaction range defined for ";
         tmp += MaterialCollectionData[ConflictingMaterialIndex]->name;
         return tmp;
       }
     default:;
     }

   return "Unknown error";
}

void MaterialCollectionClass::RegisterNewParticle()
{
  if (MaterialCollectionData.size()>0)
    if (MaterialCollectionData[0]->MatParticle.size() !=  ParticleCollection->size()-1 )
      qWarning()<<"MaterialCollection--> RegisterNewParticle reports mismatch!";

  int index = ParticleCollection->size()-1;
  //define empty MatParticle properties for the new particle
  for (int imat=0; imat<MaterialCollectionData.size();imat++)
    {
      MaterialCollectionData[imat]->MatParticle.resize(index+1);
      MaterialCollectionData[imat]->MatParticle[index].InteractionDataF.resize(0);
      MaterialCollectionData[imat]->MatParticle[index].InteractionDataX.resize(0);
      MaterialCollectionData[imat]->MatParticle[index].PhYield = 0 ;
      MaterialCollectionData[imat]->MatParticle[index].Terminators.resize(0);
    }

  tmpMaterial.MatParticle.resize(index+1);
  tmpMaterial.MatParticle[index].InteractionDataF.resize(0);
  tmpMaterial.MatParticle[index].InteractionDataX.resize(0);
  tmpMaterial.MatParticle[index].PhYield = 0 ;
  tmpMaterial.MatParticle[index].Terminators.resize(0);
}

bool MaterialCollectionClass::DeleteMaterial(int imat)
{
   int size = MaterialCollectionData.size();
   if (imat<0 || imat >= size)
     {
       qWarning()<<"Attempt to remove material with invalid index!";
       return false;
     }

   //clear dynamic properties of this material
   if (MaterialCollectionData[imat]->PrimarySpectrumHist)
       delete MaterialCollectionData[imat]->PrimarySpectrumHist;
   if (MaterialCollectionData[imat]->SecondarySpectrumHist)
       delete MaterialCollectionData[imat]->SecondarySpectrumHist;   
   for (int iOther=0; iOther<size; iOther++)  //overrides from this materials to other materials
     {
       if ( MaterialCollectionData[imat]->OpticalOverrides[iOther] )
         {
           delete MaterialCollectionData[imat]->OpticalOverrides[iOther];
           MaterialCollectionData[imat]->OpticalOverrides[iOther] = 0;
         }
     }

   //clear overrides from other materials to this one
   for (int iOther=0; iOther<size; iOther++)
     {
       if ( MaterialCollectionData[iOther]->OpticalOverrides[imat] ) delete MaterialCollectionData[iOther]->OpticalOverrides[imat];
       MaterialCollectionData[iOther]->OpticalOverrides.remove(imat);
     }

   //delete this material
   delete MaterialCollectionData[imat];
   MaterialCollectionData.remove(imat);

   //update indices of override materials
   for (int i=0; i<size-1; i++)
     for (int j=0; j<size-1; j++)
       if (MaterialCollectionData[i]->OpticalOverrides[j])
         MaterialCollectionData[i]->OpticalOverrides[j]->updateMatIndices(i, j);

   //tmpmaterial - does not receive overrides!
   return true;
}

void MaterialCollectionClass::RecalculateCrossSections(int particleId)
{
    int Scenarios = tmpMaterial.MatParticle[particleId].Terminators.size();
    int dataPoints = tmpMaterial.MatParticle[particleId].InteractionDataF.size();

 //   qDebug()<<"particle"<<particleId<<"tot scenarios"<<Scenarios<<"data points"<<dataPoints;

    for (int s=0; s<Scenarios; s++)
    {
        double branching = tmpMaterial.MatParticle[particleId].Terminators[s].branching;
        tmpMaterial.MatParticle[particleId].Terminators[s].PartialCrossSectionEnergy.resize(dataPoints);
        tmpMaterial.MatParticle[particleId].Terminators[s].PartialCrossSection.resize(dataPoints);

        for (int i=0; i<dataPoints; i++)
          {
            tmpMaterial.MatParticle[particleId].Terminators[s].PartialCrossSection[i] = branching * tmpMaterial.MatParticle[particleId].InteractionDataF[i];
            tmpMaterial.MatParticle[particleId].Terminators[s].PartialCrossSectionEnergy[i] = tmpMaterial.MatParticle[particleId].InteractionDataX[i];
          }
      }
}

void MaterialCollectionClass::writeToJson(QJsonObject &json)
{
  QJsonObject js;

  QJsonArray ar;
  for (int i=0; i<MaterialCollectionData.size(); i++)
    {
      QJsonObject jj;
      MaterialCollectionData[i]->writeToJson(jj, ParticleCollection);
      ar.append(jj);
    }
  js["Materials"] = ar;

  QJsonArray oar;
  for (int iFrom=0; iFrom<MaterialCollectionData.size(); iFrom++)
    for (int iTo=0; iTo<MaterialCollectionData.size(); iTo++)
      {
//        double loss = MaterialCollectionData.at(iFrom)->LossOverride.at(iTo);
//        double ref  = MaterialCollectionData.at(iFrom)->ReflectionOverride.at(iTo);
//        double scat = MaterialCollectionData.at(iFrom)->ScatterOverride.at(iTo);
//        if (loss>0 || ref >0 || scat >0) //saving only non-empty ones
//          {
//            QJsonObject js;
//            js["MatFrom"] = iFrom;
//            js["MatTo"] = iTo;
//            js["Loss"] = loss;
//            js["Ref"] = ref;
//            js["Scat"] = scat;
//            js["ScatModel"] = MaterialCollectionData.at(iFrom)->ScatterModel.at(iTo);
//            oar.append(js);
//          }
         //qDebug() << "From to override:"<<iFrom<<iTo<<MaterialCollectionData.at(iFrom)->OpticalOverrides.at(iTo);

         if ( !MaterialCollectionData.at(iFrom)->OpticalOverrides.at(iTo) ) continue;
         QJsonObject js;
         MaterialCollectionData.at(iFrom)->OpticalOverrides[iTo]->writeToJson(js);
         oar.append(js);
      }
  if (!oar.isEmpty()) js["Overrides"] = oar;
  js["LogLog"] = fLogLogInterpolation;

  json["MaterialCollection"] = js;
}

void MaterialCollectionClass::writeMaterialToJson(int imat, QJsonObject &json)
{
  if (imat<0 || imat>MaterialCollectionData.size()-1)
    {
      qWarning() << "Attempt to save non-existent material!";
      return;
    }
  MaterialCollectionData[imat]->writeToJson(json, ParticleCollection);
}

bool MaterialCollectionClass::readFromJson(QJsonObject &json)
{
  if (!json.contains("MaterialCollection"))
    {
      qCritical() << "Material collection not found in json";
      exit(-2);
    }
  QJsonObject js = json["MaterialCollection"].toObject();

  //reading materials
  QJsonArray ar = js["Materials"].toArray();
  if (ar.isEmpty())
    {
      qCritical() << "No materials in json";
      exit(-2);
    }
  MaterialCollectionClass::Clear();
  for (int i=0; i<ar.size(); i++)
    {
      QJsonObject jj = ar[i].toObject();
      MaterialCollectionClass::AddNewMaterial(); //also initialize overrides
      MaterialCollectionData.last()->readFromJson(jj, this, ParticleCollection);
    }
  int numMats = size();
  //qDebug() << "--> Loaded material collection with"<<numMats<<"materials";

  //reading overrides if present
  QJsonArray oar = js["Overrides"].toArray();
  for (int i=0; i<oar.size(); i++)
    {
      QJsonObject jj = oar[i].toObject();
      if (jj.contains("Model"))
        {
          //new format
          QString model = jj["Model"].toString();
          int MatFrom = jj["MatFrom"].toInt();
          int MatTo = jj["MatTo"].toInt();
          if (MatFrom>numMats-1 || MatTo>numMats-1)
            {
              qWarning()<<"Attempt to override for non-existent material skipped";
              continue;
            }
          OpticalOverrideClass* ov = OpticalOverrideFactory(model, this, MatFrom, MatTo);
          if (!ov || !ov->readFromJson(jj)) message(MaterialCollectionData[MatFrom]->name+": optical override load failed!");
          else MaterialCollectionData[MatFrom]->OpticalOverrides[MatTo] = ov;
        }
      //compatibility with old format:
      else if (jj.contains("MatFrom") && jj.contains("ScatModel"))
        {
          int MatFrom, MatTo;
          parseJson(jj, "MatFrom", MatFrom);
          parseJson(jj, "MatTo", MatTo);
          if (MatFrom>numMats-1 || MatTo>numMats-1)
            {
              qWarning()<<"Attempt to override for non-existent material skipped";
              continue;
            }
          double Abs, Scat, Spec;
          int ScatMode;
          parseJson(jj, "Loss", Abs);
          parseJson(jj, "Ref", Spec);
          parseJson(jj, "Scat", Scat);
          parseJson(jj, "ScatModel", ScatMode);
          BasicOpticalOverride* ov = new BasicOpticalOverride(this, MatFrom, MatTo, Abs, Spec, Scat, ScatMode );
          if (!ov) message(MaterialCollectionData[MatFrom]->name+": optical override load failed!");
          else MaterialCollectionData[MatFrom]->OpticalOverrides[MatTo] = ov;
        }
    }
  fLogLogInterpolation = false;
  parseJson(js, "LogLog", fLogLogInterpolation);
  //qDebug() << "--> Loaded material border overrides from array with"<<oar.size()<<"entries";
  return true;
}

void MaterialCollectionClass::AddNewMaterial(QJsonObject &json) //have to be sure json is indeed material properties!
{
  AddNewMaterial();
  MaterialStructure* mat = MaterialCollectionData.last();
  mat->readFromJson(json, this, ParticleCollection);
  QString name = mat->name;
  bool fFound;
  do
    {
      fFound = false;
      for (int i=0; i<MaterialCollectionData.size()-2; i++)  //-1 -1 - excluding itself
        if (MaterialCollectionData[i]->name == name)
          {
            fFound = true;
            name += "*";
            break;
          }
    }
  while (fFound);
  mat->name = name;
}

bool MaterialCollectionClass::SaveMaterialJSON(MaterialStructure* mat, QString fileName)
{
  QJsonObject json;

  QFileInfo QF(fileName);
  QString MatDataDir = QF.baseName()+"_data";
  if (!QDir(QF.path()+"/"+MatDataDir).exists()) QDir().mkdir(QF.path()+"/"+MatDataDir);

  //general data
  json["MaterialName"] = mat->name;
  json["Density"] = mat->density;
  json["IsotopeDensity"] = mat->atomicDensity;
  json["RefractiveIndex"] = mat->n;
  json["BulkAbsorption"] = mat->abs;  
  json["RayleighMFP"] = mat->rayleighMFP;
  json["RayleighWave"] = mat->rayleighWave;
  json["PrimScint_Tau"] = mat->PriScintDecayTime;
  json["W"] = mat->W;
  json["SecScint_PhYield"] = mat->SecYield;
  json["SecScint_Tau"] = mat->SecScintDecayTime;
  json["ElDriftVelo"] = mat->e_driftVelocity;
  json["TGeoP1"] = mat->p1;
  json["TGeoP2"] = mat->p2;
  json["TGeoP3"] = mat->p3;
  json["Comments"] = mat->Comments;

  //wavelength-resolved data
  if (mat->nWave_lambda.size() >0 || mat->absWave_lambda.size() >0 || mat->PrimarySpectrum_lambda.size() >0 || mat->SecondarySpectrum_lambda.size() >0)
    if      (!QDir( QF.path() + "/" + MatDataDir + "/WaveResolvedProperties").exists())
      QDir().mkdir( QF.path() + "/" + MatDataDir + "/WaveResolvedProperties");

  if (mat->nWave_lambda.size() > 0)
    {
       QString WfileName = QF.path() + "/" + MatDataDir + "/WaveResolvedProperties/n.dat";
       json["RefractiveIndexFile"] = MatDataDir + "/WaveResolvedProperties/n.dat";
       SaveDoubleVectorsToFile(WfileName, &mat->nWave_lambda, &mat->nWave);
    }
  if (mat->absWave_lambda.size() >0)
    {
       QString WfileName = QF.path() + "/" + MatDataDir + "/WaveResolvedProperties/abs.dat";
       json["BulkAbsorptionFile"] = MatDataDir + "/WaveResolvedProperties/abs.dat";
       SaveDoubleVectorsToFile(WfileName, &mat->absWave_lambda, &mat->absWave);
    }
  if (mat->PrimarySpectrum_lambda.size() >0)
    {
       QString WfileName = QF.path() + "/" + MatDataDir + "/WaveResolvedProperties/PrimaryScintSpectrum.dat";
       json["PrimScintSpectrumFile"] = MatDataDir + "/WaveResolvedProperties/PrimaryScintSpectrum.dat";
       SaveDoubleVectorsToFile(WfileName, &mat->PrimarySpectrum_lambda, &mat->PrimarySpectrum);
    }
  if (mat->SecondarySpectrum_lambda.size() >0)
    {
       QString WfileName = QF.path() + "/" + MatDataDir + "/WaveResolvedProperties/SecondaryScintSpectrum.dat";
       json["SecScintSpectrumFile"] = MatDataDir + "/WaveResolvedProperties/SecondaryScintSpectrum.dat";
       SaveDoubleVectorsToFile(WfileName, &mat->SecondarySpectrum_lambda, &mat->SecondarySpectrum);
    }

  //MatParticle properties
  //will be saved only for particles with defined InteractionData
  QJsonArray jParticleEntries;
  for (int ip=0; ip<ParticleCollection->size(); ip++)    
      {
         QJsonObject jMatParticle;

         QJsonObject jparticle;
         jparticle["name"] = (*ParticleCollection)[ip]->ParticleName;
         jparticle["type"] = (*ParticleCollection)[ip]->type;
         jparticle["charge"] = (*ParticleCollection)[ip]->charge;
         jparticle["mass"] = (*ParticleCollection)[ip]->mass;
           //adding to top level
         jMatParticle["Particle"] = jparticle;

         jMatParticle["TrackingAllowed"] = mat->MatParticle[ip].TrackingAllowed;
         jMatParticle["MatIsTransparent"] = mat->MatParticle[ip].MaterialIsTransparent;
         jMatParticle["PrimScintPhYield"] = mat->MatParticle[ip].PhYield;

         if (mat->MatParticle[ip].InteractionDataF.size() > 0)
           {
             //saving total interaction data
             QString ParticleName = (*ParticleCollection)[ip]->ParticleName;
             if      (!QDir( QF.path() + "/" + MatDataDir + "/" + ParticleName).exists())
               QDir().mkdir( QF.path() + "/" + MatDataDir + "/" + ParticleName);
             jMatParticle["TotalInteractionFile"] = MatDataDir + "/" + ParticleName + "/TotalInteraction.dat";
             SaveDoubleVectorsToFile(QF.path() + "/" + MatDataDir + "/" + ParticleName + "/TotalInteraction.dat",
                                     &mat->MatParticle[ip].InteractionDataX, &mat->MatParticle[ip].InteractionDataF);


             //gamma-specific data
             if ((*ParticleCollection)[ip]->type == _gamma_)
               {
                 QJsonArray jgamma;

                 int iTerminators = mat->MatParticle[ip].Terminators.size();
                 for (int iTerm=0; iTerm<iTerminators; iTerm++)
                   {
                     QJsonObject jterm;

                     jterm["ReactionType"] = mat->MatParticle[ip].Terminators[iTerm].ReactionType;
                     jterm["InteractionFile"] = MatDataDir + "/" + ParticleName + "/Interaction" + QString::number(iTerm) + ".dat";
                     SaveDoubleVectorsToFile(QF.path() + "/" + MatDataDir + "/" + ParticleName + "/Interaction" + QString::number(iTerm) + ".dat",
                                             &mat->MatParticle[ip].Terminators[iTerm].PartialCrossSectionEnergy,
                                             &mat->MatParticle[ip].Terminators[iTerm].PartialCrossSection);

                     jgamma.append(jterm);
                   }
                 jMatParticle["GammaTerminators"] = jgamma;
               }

             //neutron-specific data
             if ((*ParticleCollection)[ip]->type == _neutron_)
               {
                 QJsonArray jneutron;

                 int iTerminators = mat->MatParticle[ip].Terminators.size();
                 for (int iTerm=0; iTerm<iTerminators; iTerm++)
                   {
                     QJsonObject jterm;

                     jterm["Branching"] = mat->MatParticle[ip].Terminators[iTerm].branching;
                     //going through secondary particles
                     QJsonArray jsecondaries;
                     for (int is=0; is<mat->MatParticle[ip].Terminators[iTerm].GeneratedParticles.size();is++ )
                       {
                         QJsonObject jsecpart;
                         int pa = mat->MatParticle[ip].Terminators[iTerm].GeneratedParticles[is];
                         jsecpart["name"] = (*ParticleCollection)[pa]->ParticleName;
                         jsecpart["type"] = (*ParticleCollection)[pa]->type;
                         jsecpart["charge"] = (*ParticleCollection)[pa]->charge;
                         jsecpart["mass"] = (*ParticleCollection)[pa]->mass;

                         jsecpart["energy"] = mat->MatParticle[ip].Terminators[iTerm].GeneratedParticleEnergies[is];

                         jsecondaries.append(jsecpart);
                       }

                     jterm["Secondaries"] = jsecondaries;

                     jneutron.append(jterm);
                   }
                 jMatParticle["NeutronTerminators"] = jneutron;
               }

           }


         //appending this particle entry to the json array
         jParticleEntries.append(jMatParticle);
      }

  //if not empty, appending array with Matparticle properties to main json object
  if (!jParticleEntries.isEmpty()) json["MatParticles"] = jParticleEntries;


  //saving json file
  QFile saveFile(fileName);
  if (!saveFile.open(QIODevice::WriteOnly))
    {
          qWarning()<<"Cannot save to file "+fileName;
          return false;
    }
  QJsonDocument saveDoc(json);
  saveFile.write(saveDoc.toJson());

  return true;
}

bool MaterialCollectionClass::LoadMaterialJSON(QString fileName)
{
  QFile file(fileName);
  if(!file.open(QIODevice::ReadOnly | QFile::Text))
        {
          qCritical()<<"Failed to open file" + fileName + "!";
          return false;
        }

  MaterialCollectionClass::ClearTmpMaterial();

  QFileInfo QF(fileName);
  QByteArray saveData = file.readAll();

  QJsonParseError error;
  QJsonDocument loadDoc(QJsonDocument::fromJson(saveData, &error));
  QJsonObject json = loadDoc.object();

  if (error.error != 0)
    {
      qWarning()<<"Load failed: JSON parser reports error "<<error.errorString();
      return false;
    }
  //parsing successful

  //reading general info
  JsonParser parser(json);
  bool ok = true;
  ok &= parser.ParseObject("MaterialName", tmpMaterial.name);
  ok &= parser.ParseObject("Density", tmpMaterial.density);
  if (parser.ParseObject("IsotopeDensity", tmpMaterial.atomicDensity ));
  else tmpMaterial.atomicDensity = 0;
  ok &= parser.ParseObject("RefractiveIndex", tmpMaterial.n );
  ok &= parser.ParseObject("BulkAbsorption", tmpMaterial.abs );

  tmpMaterial.rayleighMFP = 0;//compatibility
  parser.ParseObject("RayleighMFP", tmpMaterial.rayleighMFP );
  parser.ParseObject("RayleighWave", tmpMaterial.rayleighWave );

  ok &= parser.ParseObject("PrimScint_Tau", tmpMaterial.PriScintDecayTime );
  ok &= parser.ParseObject("W", tmpMaterial.W );
  ok &= parser.ParseObject("SecScint_PhYield", tmpMaterial.SecYield );
  ok &= parser.ParseObject("SecScint_Tau", tmpMaterial.SecScintDecayTime );
  ok &= parser.ParseObject("ElDriftVelo", tmpMaterial.e_driftVelocity );
  ok &= parser.ParseObject("TGeoP1", tmpMaterial.p1 );
  ok &= parser.ParseObject("TGeoP2", tmpMaterial.p2 );
  ok &= parser.ParseObject("TGeoP3", tmpMaterial.p3 );
  if (parser.ParseObject("Comments", tmpMaterial.Comments ));
  else tmpMaterial.Comments = "";

  //wavelength-resolved properties
  QString WfileName;
  if (parser.ParseObject("RefractiveIndexFile", WfileName ))
    {
      int res = LoadDoubleVectorsFromFile(QF.path() + "/" + WfileName, &tmpMaterial.nWave_lambda, &tmpMaterial.nWave);
      if (res != 0)
        {
          qWarning()<<"Error reading file "<<WfileName;
          return false;
        }
    }
  if (parser.ParseObject("BulkAbsorptionFile", WfileName ))
    {
      int res = LoadDoubleVectorsFromFile(QF.path() + "/" + WfileName, &tmpMaterial.absWave_lambda, &tmpMaterial.absWave);
      if (res != 0)
        {
          qWarning()<<"Error reading file "<<WfileName;
          return false;
        }
    }
  if (parser.ParseObject("PrimScintSpectrumFile", WfileName ))
    {
      int res = LoadDoubleVectorsFromFile(QF.path() + "/" + WfileName, &tmpMaterial.PrimarySpectrum_lambda, &tmpMaterial.PrimarySpectrum);
      if (res != 0)
        {
          qWarning()<<"Error reading file "<<WfileName;
          return false;
        }
    }
  if (parser.ParseObject("SecScintSpectrumFile", WfileName ))
    {
      int res = LoadDoubleVectorsFromFile(QF.path() + "/" + WfileName, &tmpMaterial.SecondarySpectrum_lambda, &tmpMaterial.SecondarySpectrum);
      if (res != 0)
        {
          qWarning()<<"Error reading file "<<WfileName;
          return false;
        }
    }

  //MatParticles
  QJsonArray jParticleEntries;
  if (parser.ParseObject("MatParticles", jParticleEntries))
    {
      //there are saved MatParticles entries!
      int numMP = jParticleEntries.size();
//      qDebug()<<"Entries in matparticles:"<<numMP;
      for (int ip = 0; ip<numMP; ip++)
        {
          QJsonObject jMatParticle = jParticleEntries[ip].toObject();

          //extracting the particle
          QJsonObject jparticle;
          JsonParser MatParticleParser(jMatParticle);
          bool ok = MatParticleParser.ParseObject("Particle", jparticle);
          if (!ok)
            {
              qWarning()<<"Cannot extract particle in the material file";
              return false;
            }
          QString name;
          int type;
          int charge;
          double mass;
          JsonParser ParticleParser(jparticle);
          ok &= ParticleParser.ParseObject("name", name );
          ok &= ParticleParser.ParseObject("type", type );
          ok &= ParticleParser.ParseObject("charge", charge );
          ok &= ParticleParser.ParseObject("mass", mass );
          if (!ok)
            {
              qWarning()<<"Cannot extract particle properties in the material file";
              return false;
            }
            //looking for this particle in the collection and create if necessary
          ParticleType Type = static_cast<ParticleType>(type);
          int ParticleId = MaterialCollectionClass::FindCreateParticle(name, Type, charge, mass);
            //MatParticle size of tmpMaterial is already appropriately resized!

            //going through matparticle general properties
          ok &= MatParticleParser.ParseObject("TrackingAllowed",  tmpMaterial.MatParticle[ParticleId].TrackingAllowed );
          ok &= MatParticleParser.ParseObject("MatIsTransparent",  tmpMaterial.MatParticle[ParticleId].MaterialIsTransparent );
          ok &= MatParticleParser.ParseObject("PrimScintPhYield",  tmpMaterial.MatParticle[ParticleId].PhYield );

          QString IfileName;
          if ( MatParticleParser.ParseObject("TotalInteractionFile",  IfileName )  )
            {
              int res = LoadDoubleVectorsFromFile(QF.path() + "/" + IfileName,
                                                  &tmpMaterial.MatParticle[ParticleId].InteractionDataX, &tmpMaterial.MatParticle[ParticleId].InteractionDataF);
              if (res != 0)
                {
                  qWarning()<<"Error reading file "<<WfileName;
                  return false;
                }

              //if file found, can look for additional properties for gamma and neutron

              //gamma-specific data
              if (Type == _gamma_)
                {
                  QJsonArray jgamma;
                  bool ok_g = MatParticleParser.ParseObject("GammaTerminators",  jgamma );
                  if (!ok_g)
                    {
                      qWarning()<<"Missing gamma-specific data declared in material file"<<fileName;
                      return false;
                    }

                  //jgamma - array with gamma terminators
                  tmpMaterial.MatParticle[ParticleId].Terminators.resize(jgamma.size());
                  for (int iTerm = 0; iTerm<jgamma.size(); iTerm++)
                    {
                      QJsonObject jterm = jgamma[iTerm].toObject();
                      JsonParser TermParser(jterm);

                      bool ok = true;
                      ok &= TermParser.ParseObject("ReactionType",  tmpMaterial.MatParticle[ParticleId].Terminators[iTerm].ReactionType );
                      QString DfileName;
                      ok &= TermParser.ParseObject("InteractionFile",  DfileName );
                      int res = LoadDoubleVectorsFromFile(QF.path() + "/" + DfileName,
                                                          &tmpMaterial.MatParticle[ParticleId].Terminators[iTerm].PartialCrossSectionEnergy,
                                                          &tmpMaterial.MatParticle[ParticleId].Terminators[iTerm].PartialCrossSection);
                      if (res != 0)
                        {
                          qWarning()<<"Error reading file "<<DfileName;
                          return false;
                        }
                    }
                }

              //neutron-specific data
              if (Type == _neutron_)
                {
                  QJsonArray jneutron;
                  bool ok_n = MatParticleParser.ParseObject("NeutronTerminators",  jneutron );
                  if (!ok_n)
                    {
                      qWarning()<<"Missing neutron-specific data declared in material file"<<fileName;
                      return false;
                    }

                  //jneutron - array with neutron terminators
                  tmpMaterial.MatParticle[ParticleId].Terminators.resize(jneutron.size());
                  for (int iTerm = 0; iTerm<jneutron.size(); iTerm++)
                    {
                      QJsonObject jterm = jneutron[iTerm].toObject();
                      JsonParser TermParser(jterm);

                      tmpMaterial.MatParticle[ParticleId].Terminators[iTerm].ReactionType = 2;
                      double Branching;
                      bool ok = TermParser.ParseObject("Branching",  Branching );
                      tmpMaterial.MatParticle[ParticleId].Terminators[iTerm].branching = Branching;
                      //going through secondary particles
                      QJsonArray jsecondaries;
                      ok &= TermParser.ParseObject("Secondaries", jsecondaries );
                      if (!ok)
                        {
                          qWarning()<<"neutron information is incomplete in"<<fileName;
                          return false;
                        }
                      int NumSec = jsecondaries.size();
 //                     qDebug()<<"neutron: secondaries: "<<NumSec;
                      tmpMaterial.MatParticle[ParticleId].Terminators[iTerm].GeneratedParticles.resize(NumSec);
                      tmpMaterial.MatParticle[ParticleId].Terminators[iTerm].GeneratedParticleEnergies.resize(NumSec);
                      for (int isec=0; isec<NumSec; isec++)
                        {
                          QJsonObject jsec = jsecondaries[isec].toObject();
                          JsonParser sParser(jsec);

                          bool ok = sParser.ParseObject("energy", tmpMaterial.MatParticle[ParticleId].Terminators[iTerm].GeneratedParticleEnergies[isec]);
                          QString name;
                          int type;
                          int charge;
                          double mass;
                          ok &= sParser.ParseObject("name", name );
                          ok &= sParser.ParseObject("type", type );
                          ok &= sParser.ParseObject("charge", charge );
                          ok &= sParser.ParseObject("mass", mass );
                          if (!ok)
                            {
                              qWarning()<<"Cannot extract neutron-specific information from file "<<fileName;
                              return false;
                            }
                            //looking for this particle in the collection and create if necessary
                          ParticleType Type = static_cast<ParticleType>(type);
                          int SecParticleId = MaterialCollectionClass::FindCreateParticle(name, Type, charge, mass);
                          tmpMaterial.MatParticle[ParticleId].Terminators[iTerm].GeneratedParticles[isec] = SecParticleId;                          
                        }
                      //using branchings to calculate partical cross sections (can be changed in the future!)
                      tmpMaterial.MatParticle[ParticleId].Terminators[iTerm].PartialCrossSectionEnergy = tmpMaterial.MatParticle[ParticleId].InteractionDataX;
                      tmpMaterial.MatParticle[ParticleId].Terminators[iTerm].PartialCrossSection.resize(0);
                      for (int i=0; i<tmpMaterial.MatParticle[ParticleId].InteractionDataF.size(); i++)
                        tmpMaterial.MatParticle[ParticleId].Terminators[iTerm].PartialCrossSection.append(Branching * tmpMaterial.MatParticle[ParticleId].InteractionDataF[i]);
                    }
                }

            }
        }
    }


  //checking for errors
  if (parser.GetErrType() != 0)
    {
      qWarning()<<"Error in material file parsing: type mismatch has been found";
      return false;
    }
  if (!ok)
    {
      qWarning()<<"Error in material file parsing - some of the fields are missings";
      return false;
    }

  //no errors!
  fLogLogInterpolation = false;
  return true;
}

int MaterialCollectionClass::CheckParticleEnergyInRange(int iPart, double Energy)
{
   if (iPart<0 || iPart>ParticleCollection->size()-1) return -1; //not-existent particle

   for (int iMat=0; iMat<MaterialCollectionData.size(); iMat++)
     {
       MatParticleStructure* mp = &MaterialCollectionData[iMat]->MatParticle[iPart];

       if (!mp->TrackingAllowed) continue;
       if (mp->MaterialIsTransparent) continue;

       //we skip empty energy data because this type of check is done in CheckMaterial()
       if(mp->InteractionDataX.isEmpty()) continue;

       //checking is energy in range
       double min = mp->InteractionDataX.first();
       double max = mp->InteractionDataX.last();
       if (Energy<min) return iMat;
       if (Energy>max) return iMat;
     }

   return -1; //no errors
}

void MaterialStructure::clear()
{
  name = "Undefined";
  density = atomicDensity = p1 = p2 = p3 = n = abs = rayleighMFP = 0;
  e_driftVelocity = W = SecYield = PriScintDecayTime = SecScintDecayTime = 0;
  rayleighWave = 500.0;
  Comments = "";

  rayleighBinned.clear();
  nWave_lambda.clear();
  nWave.clear();
  nWaveBinned.clear();
  absWave_lambda.clear();
  absWave.clear();
  absWaveBinned.clear();
  PrimarySpectrum_lambda.clear();
  PrimarySpectrum.clear();
  SecondarySpectrum_lambda.clear();
  SecondarySpectrum.clear();

  if(PrimarySpectrumHist)
    {
      delete PrimarySpectrumHist;
      PrimarySpectrumHist=0;
    }
  if (SecondarySpectrumHist)
    {
      delete SecondarySpectrumHist;
      SecondarySpectrumHist = 0;
    }

  MatParticle.clear();

  GeoMat = 0; //if created, deleted by TGeomanager
  GeoMed = 0; //if created, deleted by TGeomanager

  //Do not touch overrides - handled by loaded (want to keep overrides intact when handling inspector window)
}

void MaterialStructure::writeToJson(QJsonObject &json, QVector<ParticleDescriptor *> *ParticleCollection)
{
  //general data
  json["*MaterialName"] = name;
  json["Density"] = density;
  if (atomicDensity>0) json["IsotopeDensity"] = atomicDensity;
  json["RefractiveIndex"] = n;
  json["BulkAbsorption"] = abs;
  if (rayleighMFP > 0)
    {
      json["RayleighMFP"] = rayleighMFP;
      json["RayleighWave"] = rayleighWave;
    }
  json["PrimScint_Tau"] = PriScintDecayTime;
  json["W"] = W;
  json["SecScint_PhYield"] = SecYield;
  json["SecScint_Tau"] = SecScintDecayTime;
  json["ElDriftVelo"] = e_driftVelocity;
  if (p1!=0 || p2!=0 || p3!=0)
    {
      json["TGeoP1"] = p1;
      json["TGeoP2"] = p2;
      json["TGeoP3"] = p3;
    }
  if (!Comments.isEmpty()) json["Comments"] = Comments;

  //wavelength-resolved data
  if (nWave_lambda.size() > 0)
    {
      QJsonArray ar;
      writeTwoQVectorsToJArray(nWave_lambda, nWave, ar);
      json["RefractiveIndexWave"] = ar;
    }
  if (absWave_lambda.size() >0)
    {
      QJsonArray ar;
      writeTwoQVectorsToJArray(absWave_lambda, absWave, ar);
      json["BulkAbsorptionWave"] = ar;
    }
  if (PrimarySpectrum_lambda.size() >0)
    {
      QJsonArray ar;
      writeTwoQVectorsToJArray(PrimarySpectrum_lambda, PrimarySpectrum, ar);
      json["PrimScintSpectrum"] = ar;
    }
  if (SecondarySpectrum_lambda.size() >0)
    {
      QJsonArray ar;
      writeTwoQVectorsToJArray(SecondarySpectrum_lambda, SecondarySpectrum, ar);
      json["SecScintSpectrum"] = ar;
    }

  //MatParticle properties
  //if a particle has default configuration (TrackingAllowed and MatIsTransparent), skip its record
  QJsonArray jParticleEntries;
  for (int ip=0; ip<ParticleCollection->size(); ip++)
    {
      //default properties?
      if (MatParticle[ip].TrackingAllowed && MatParticle[ip].MaterialIsTransparent && MatParticle[ip].InteractionDataF.isEmpty()) continue; //skip
      //writing info on this particle
      QJsonObject jMatParticle;

      QJsonObject jparticle;
      ParticleCollection->at(ip)->writeToJson(jparticle);
      jMatParticle["*Particle"] = jparticle;
      jMatParticle["TrackingAllowed"] = MatParticle[ip].TrackingAllowed;
      //if (MatParticle[ip].TrackingAllowed) //if not, no need for anything else
      {
        jMatParticle["MatIsTransparent"] = MatParticle[ip].MaterialIsTransparent;
        jMatParticle["PrimScintPhYield"] = MatParticle[ip].PhYield;
        if (MatParticle[ip].InteractionDataF.size() > 0)
          {
            QJsonArray iar;
            writeTwoQVectorsToJArray(MatParticle[ip].InteractionDataX, MatParticle[ip].InteractionDataF, iar);
            jMatParticle["TotalInteraction"] = iar;
            //gamma-specific data
            if ((*ParticleCollection)[ip]->type == _gamma_)
              {
                QJsonArray jgamma;
                int iTerminators = MatParticle[ip].Terminators.size();
                for (int iTerm=0; iTerm<iTerminators; iTerm++)
                  {
                    QJsonObject jterm;
                    jterm["ReactionType"] = MatParticle[ip].Terminators[iTerm].ReactionType;
                    QJsonArray ar;
                    writeTwoQVectorsToJArray(MatParticle[ip].Terminators[iTerm].PartialCrossSectionEnergy, MatParticle[ip].Terminators[iTerm].PartialCrossSection, ar);
                    jterm["InteractionData"] = ar;
                    jgamma.append(jterm);
                  }
                jMatParticle["GammaTerminators"] = jgamma;
              }
            //neutron-specific data
            if ((*ParticleCollection)[ip]->type == _neutron_)
              {
                QJsonArray jneutron;
                int iTerminators = MatParticle[ip].Terminators.size();
                for (int iTerm=0; iTerm<iTerminators; iTerm++)
                  {
                    QJsonObject jterm;
                    jterm["Branching"] = MatParticle[ip].Terminators[iTerm].branching;
                    jterm["ReactionType"] = 2;  //can be only capture
                    //going through secondary particles
                    QJsonArray jsecondaries;
                    for (int is=0; is<MatParticle[ip].Terminators[iTerm].GeneratedParticles.size();is++ )
                      {
                        QJsonObject jsecpart;
                        int pa = MatParticle[ip].Terminators[iTerm].GeneratedParticles[is];
                        QJsonObject jj;
                        (*ParticleCollection)[pa]->writeToJson(jj);
                        jsecpart["SecParticle"] = jj;
                        jsecpart["energy"] = MatParticle[ip].Terminators[iTerm].GeneratedParticleEnergies[is];
                        jsecondaries.append(jsecpart);
                      }
                    jterm["Secondaries"] = jsecondaries;
                    jneutron.append(jterm);
                  }
                jMatParticle["NeutronTerminators"] = jneutron;
              }
          }
        else jMatParticle["MatIsTransparent"] = true; //if tracking allowed, but total interaction is NOT defined, make material transparent
      }
      //appending this particle entry to the json array
      jParticleEntries.append(jMatParticle);
    }
  //if not empty, appending array with Matparticle properties to main json object
  if (!jParticleEntries.isEmpty()) json["MatParticles"] = jParticleEntries;
}

bool MaterialStructure::readFromJson(QJsonObject &json, MaterialCollectionClass *MaterialCollection, QVector<ParticleDescriptor *> *ParticleCollection)
{
  clear(); //clear all settings and set default properties

  //general data
  parseJson(json, "*MaterialName", name);
  parseJson(json, "Density", density);
  parseJson(json, "IsotopeDensity", atomicDensity);
  parseJson(json, "RefractiveIndex", n);
  parseJson(json, "BulkAbsorption", abs);
  parseJson(json, "RayleighMFP", rayleighMFP);
  parseJson(json, "RayleighWave", rayleighWave);
  parseJson(json, "PrimScint_Tau", PriScintDecayTime);
  parseJson(json, "W", W);
  parseJson(json, "SecScint_PhYield", SecYield);
  parseJson(json, "SecScint_Tau", SecScintDecayTime);
  parseJson(json, "ElDriftVelo", e_driftVelocity);
  parseJson(json, "TGeoP1", p1);
  parseJson(json, "TGeoP2", p2);
  parseJson(json, "TGeoP3", p3);
  parseJson(json, "Comments", Comments);

  //wavelength-resolved data
  if (json.contains("RefractiveIndexWave"))
    {
      QJsonArray ar = json["RefractiveIndexWave"].toArray();
      readTwoQVectorsFromJArray(ar, nWave_lambda, nWave);
    }
  if (json.contains("BulkAbsorptionWave"))
    {
      QJsonArray ar = json["BulkAbsorptionWave"].toArray();
      readTwoQVectorsFromJArray(ar, absWave_lambda, absWave);
    }
  if (json.contains("PrimScintSpectrum"))
    {
      QJsonArray ar = json["PrimScintSpectrum"].toArray();
      readTwoQVectorsFromJArray(ar, PrimarySpectrum_lambda, PrimarySpectrum);
    }
  if (json.contains("SecScintSpectrum"))
    {
      QJsonArray ar = json["SecScintSpectrum"].toArray();
      readTwoQVectorsFromJArray(ar, SecondarySpectrum_lambda, SecondarySpectrum);
    }

  //MatParticle properties
    //filling default properties for all particles  
  MatParticle.resize(ParticleCollection->size());
  for (int ip=0; ip<ParticleCollection->size(); ip++)
    {
      MatParticle[ip].TrackingAllowed = true;
      MatParticle[ip].MaterialIsTransparent = true;
    }
    // reading defined properties
  QJsonArray jParticleEntries = json["MatParticles"].toArray();
  for (int index=0; index<jParticleEntries.size(); index++)
    {
      QJsonObject jMatParticle = jParticleEntries[index].toObject();

      QJsonObject jparticle = jMatParticle["*Particle"].toObject();
      int ip = MaterialCollection->findOrCreateParticle(jparticle);

      parseJson(jMatParticle, "TrackingAllowed", MatParticle[ip].TrackingAllowed);
      parseJson(jMatParticle, "MatIsTransparent", MatParticle[ip].MaterialIsTransparent);
      parseJson(jMatParticle, "PrimScintPhYield", MatParticle[ip].PhYield);

      if (jMatParticle.contains("TotalInteraction"))
        {
          QJsonArray iar = jMatParticle["TotalInteraction"].toArray();
          readTwoQVectorsFromJArray(iar, MatParticle[ip].InteractionDataX, MatParticle[ip].InteractionDataF);
        }
      if (jMatParticle.contains("GammaTerminators"))
        { //gamma-specific data
          QJsonArray jgamma = jMatParticle["GammaTerminators"].toArray();
          int numTerms = jgamma.size();
          MatParticle[ip].Terminators.resize(numTerms);
          for (int iTerm=0; iTerm<numTerms; iTerm++)
            {
              QJsonObject jterm = jgamma[iTerm].toObject();
              parseJson(jterm, "ReactionType", MatParticle[ip].Terminators[iTerm].ReactionType);
              QJsonArray ar = jterm["InteractionData"].toArray();
              readTwoQVectorsFromJArray(ar, MatParticle[ip].Terminators[iTerm].PartialCrossSectionEnergy, MatParticle[ip].Terminators[iTerm].PartialCrossSection);
            }
        }
       if (jMatParticle.contains("NeutronTerminators"))
         { //neutron-specific data
           QJsonArray jneutron = jMatParticle["NeutronTerminators"].toArray();
           int numTerms = jneutron.size();
           MatParticle[ip].Terminators.resize(numTerms);
           //qDebug() << "Terminators:"<<numTerms;
           for (int iTerm=0; iTerm<numTerms; iTerm++)
             {
               QJsonObject jterm = jneutron[iTerm].toObject();
               parseJson(jterm, "Branching", MatParticle[ip].Terminators[iTerm].branching);
               MatParticle[ip].Terminators[iTerm].ReactionType = 2; //can be only capture
               //going through secondary particles
               QJsonArray jsecondaries = jterm["Secondaries"].toArray();
               int numSecondaries = jsecondaries.size();
               //qDebug() << "Secondaries"<<numSecondaries;
               MatParticle[ip].Terminators[iTerm].GeneratedParticles.resize(numSecondaries);
               MatParticle[ip].Terminators[iTerm].GeneratedParticleEnergies.resize(numSecondaries);
               for (int is=0; is<numSecondaries;is++ )
                 {
                   QJsonObject jsecpart = jsecondaries[is].toObject();

                   QJsonObject jsp = jsecpart["SecParticle"].toObject();
                   int isp = MaterialCollection->findOrCreateParticle(jsp);
                   //qDebug() << "Secondary:"<<isp;
                   MatParticle[ip].Terminators[iTerm].GeneratedParticles[is] = isp;
                   parseJson(jsecpart, "energy", MatParticle[ip].Terminators[iTerm].GeneratedParticleEnergies[is]);
                 }
               //using branchings to calculate partical cross sections (can be changed in the future!)
               MatParticle[ip].Terminators[iTerm].PartialCrossSectionEnergy = MatParticle[ip].InteractionDataX;
               MatParticle[ip].Terminators[iTerm].PartialCrossSection.resize(0);
               for (int i=0; i<MatParticle[ip].InteractionDataF.size(); i++)
                 MatParticle[ip].Terminators[iTerm].PartialCrossSection.append(MatParticle[ip].Terminators[iTerm].branching * MatParticle[ip].InteractionDataF[i]);
             }
         }
    }  
  return true;
}
