#ifndef MATERIALCOLLECTIONCLASS_H
#define MATERIALCOLLECTIONCLASS_H

#include <QVector>
#include <QString>
#include "classes.h"

class TGeoMaterial;
class TGeoMedium;
class TH1D;
class GeneralSimSettings;
class OpticalOverrideClass;

struct NeutralTerminatorStructure //descriptor for the interaction scenarios for neutral particles
{
    int ReactionType;  //0 - photoelectric;  1 - Compton; 2 - capture

    QVector<double> PartialCrossSection;
    QVector<double> PartialCrossSectionEnergy;
    double branching; //for neutrons - cross sections do not depend on energy, can scale using total

    QVector<int> GeneratedParticles;
    QVector<double> GeneratedParticleEnergies;
};

struct MatParticleStructure  //each paticle have this entry in MaterialStructure
{
    MatParticleStructure() {Terminators.resize(0); PhYield = 0; InteractionDataF.resize(0); InteractionDataX.resize(0); TrackingAllowed = false; MaterialIsTransparent = true;}

    bool TrackingAllowed;
    bool MaterialIsTransparent;
    double PhYield;         // Photon yield of the primary scintillation

    QVector<double> InteractionDataX; //energy in keV
    QVector<double> InteractionDataF; //stopping power (for charged) or total interaction cross-section (neutrals)

    QVector<NeutralTerminatorStructure> Terminators;

    //--methods--
    bool CalculateTotal() //true - success, false - mismatch in binning of the data
      {
        InteractionDataX = Terminators[0].PartialCrossSectionEnergy;
        InteractionDataF = Terminators[0].PartialCrossSection;
        for (int i=1; i<Terminators.size(); i++)
        {
            for (int j=0; j<Terminators[0].PartialCrossSectionEnergy.size(); j++)
            {
                if (Terminators[i].PartialCrossSectionEnergy[j] != Terminators[0].PartialCrossSectionEnergy[j]) return false;// mismatch!
                InteractionDataF[j] += Terminators[i].PartialCrossSection[j];
            }
        }
        return true;
      }    
};

class MaterialCollectionClass;

struct MaterialStructure
{
    QString name;
    double density; //in g/cm3
    double atomicDensity; //in atoms/cm3
    double p1,p2,p3; //parameters for TGeoManager
    double n;   //refractive index for monochrome
    double abs; //exp absorption per mm   for monochrome    (I = I0*exp(-abs*length[mm]))    
    double rayleighMFP; //0 - no rayleigh scattering
    double rayleighWave;
    QVector<double> rayleighBinned;//regular step (WaveStep step, WaveNodes bins)
    double e_driftVelocity;
    double W; //default W
    double SecYield;  // ph per secondary electron
    double PriScintDecayTime;
    double SecScintDecayTime;
    QString Comments;

    QVector<MatParticleStructure> MatParticle; //material properties related to individual particles

    //QVector<double> ReflectionOverride;  //0 = no override,
    //QVector<double> LossOverride;  //0 = no direct loss on a border, otherwise gives the probability (0 to 1] that a photon is lost
    //QVector<double> ScatterOverride;  //0 = no difusde reflection on a border, otherwsie gives the probability (0 to 1]
    //QVector<int> ScatterModel;  //0 = in 4Pi, 1 = in 2Pi
      //for the remaining probability fraction : use Fresnel equation to calculate reflection probability on a border
    QVector<OpticalOverrideClass*> OpticalOverrides; //NULL - override not defined


    QVector<double> nWave_lambda;
    QVector<double> nWave;
    QVector<double> nWaveBinned; //regular step (WaveStep step, WaveNodes bins)
    QVector<double> absWave_lambda;
    QVector<double> absWave;
    QVector<double> absWaveBinned;//regular step (WaveStep step, WaveNodes bins)

    QVector<double> PrimarySpectrum_lambda;
    QVector<double> PrimarySpectrum;
    TH1D* PrimarySpectrumHist;    //pointer!
    QVector<double> SecondarySpectrum_lambda;
    QVector<double> SecondarySpectrum;
    TH1D* SecondarySpectrumHist;  //pointer!

    TGeoMaterial* GeoMat; //pointer, but it is taken care of by TGEoManager
    TGeoMedium* GeoMed;   //pointer, but it is taken care of by TGEoManager

    MaterialStructure()
        {PrimarySpectrumHist=0; SecondarySpectrumHist=0; GeoMat=0; GeoMed=0; rayleighMFP = 0; rayleighWave = 500; Comments = "";} //setting all pointers to 0

    void clear();
    void writeToJson(QJsonObject &json, QVector<ParticleDescriptor*>* ParticleCollection); //does not save overrides!
    bool readFromJson(QJsonObject &json, MaterialCollectionClass *MaterialCollection, QVector<ParticleDescriptor*> *ParticleCollection);
};

//----------------------- MaterialCollection ------------------------

class MaterialCollectionClass
{
  public:
    MaterialStructure tmpMaterial; //all pointers are 0 on start -see default constructor
    bool fLogLogInterpolation;

  private:
    QVector<MaterialStructure*> MaterialCollectionData;    
    QVector<ParticleDescriptor*>* ParticleCollection;  //pointer to the ParticleCollection
    double WaveFrom, WaveTo, WaveStep;
    int WaveNodes;
    bool WavelengthResolved;

  public:
    //configuration
    void SetParticleCollection(QVector<ParticleDescriptor*>* particleCollection);
    void SetWave(bool wavelengthResolved, double waveFrom, double waveTo, double waveStep, int waveNodes);
    void updateWaveProperties(GeneralSimSettings *SimSet);

    //info requests
    MaterialStructure* operator[](int i) {return MaterialCollectionData[i]; } //get pointer to material with index i
    int size() const {return MaterialCollectionData.size();}
    void getFirstOverridenMaterial(int &ifrom, int &ito);
    double convertWaveIndexToWavelength(int index) {return WaveFrom + WaveStep * index;}

    //MaterialCollection handling
    void Clear();
    void AddNewMaterial();
    void UpdateMaterial(int index, QString name, double density, double atomicDensity, double n, double abs, double PriScintDecayTime, double W, double SecYield, double SecScintDecayTime, double e_driftVelocity, double p1, double p2, double p3);
    int FindMaterial(QString name); //if not found, returns -1; if found, returns material index
    void RegisterNewParticle(); //has to be called after a particle was added to material collection
    bool DeleteMaterial(int imat); //takes care of overrides of materials with index larger than imat!
    void UpdateWaveResolvedProperties(int imat); //updates wavelength-resolved material properties

    //tmpMaterial - related
    void ClearTmpMaterial(); //deletes all objects pointed by the class pointers!!!
    void CopyTmpToMaterialCollection(); //creates a copy of all pointers // true is new material was added to material collection    
    void CopyMaterialToTmp(int imat);
    void RecalculateCrossSections(int particleId); //for neutrons - using branchings, calculate cross section vectors

    //json write/read handling
    void writeToJson(QJsonObject &json);
    void writeMaterialToJson(int imat, QJsonObject &json);
    bool readFromJson(QJsonObject &json);
    void AddNewMaterial(QJsonObject &json);

    //save/load
    bool SaveMaterialJSON(MaterialStructure *mat, QString fileName);
    bool LoadMaterialJSON(QString fileName);
      //old style - obsolete - compatibility only!!!
    int SaveMaterialData(MaterialStructure *tmpMaterial, QString outputFilename);
    int LoadMaterialData(QString fileName);  //data are loaded to tmpMaterial  !!!

  public:
    MaterialCollectionClass();
    MaterialCollectionClass(QVector<ParticleDescriptor*>* particleCollection);
    ~MaterialCollectionClass();

    void ConvertToStandardWavelengthes(QVector<double> *sp_x, QVector<double> *sp_y, QVector<double> *y);    
    int FindCreateParticle(QString Name, ParticleType Type, int Charge, double Mass, bool fOnlyFind = false);  //copy of function of the main window!
    int findOrCreateParticle(QJsonObject &json);

    int CheckMaterial(MaterialStructure* mat, int iPart); //0 - check passed, see cpp file for possible error codes
    int CheckMaterial(int iMat, int iPart); //0 - check passed, see cpp file for possible error codes
    int CheckMaterial(int iMat); //0 - check passed, see cpp file for possible error codes
    int CheckTmpMaterial(); //0 - check passed, see cpp file for possible error codes
    QString getErrorString(int iError);

    int CheckParticleEnergyInRange(int iPart, double Energy); //check all materials - if this particle is tracable and mat is not-tansparent,
        //check that the particle energy is withing the defined energy range of the total interaction.
        //if not in range, the first material index with such a problem is returned.
        // -1 is returned if there are no errors

private:
    int ConflictingMaterialIndex; //used by CheckMaterial function
    int ConflictingParticleIndex; //used by CheckMaterial function    
};

#endif // MATERIALCOLLECTIONCLASS_H
