#ifndef PARTICLESOURCESCLASS_H
#define PARTICLESOURCESCLASS_H

#include <QString>
#include <QVector>
#include "TRandom2.h"
#include "TVector3.h"
#include "TH1D.h"

class ParticleDescriptor;
class MaterialCollectionClass;
class QJsonObject;

struct GunParticleStruct
{
  int ParticleId;
  double StatWeight;
  double energy;
  TH1D* spectrum; //energy spectrum
  bool Individual; // true = individual particle; false = linked
  int LinkedTo; // index of the "parent" particle this one is following
  double LinkingProbability;  //probability to be emitted after the parent particle
  int LinkingDirection; // -1 - random direction; otherwise index of the reference particle, this linked particle is emitted in the opposite direction in respect to the reference particle

  GunParticleStruct()
    {
      spectrum = 0;
      Individual = true;
      LinkedTo = 0;//-1;
      LinkingProbability = 0;
      LinkingDirection = -1;
    }

  ~GunParticleStruct()
    {
      if (spectrum) delete spectrum;
    }
};

struct ParticleSourceStructure
{
  //name
  QString name;
  //source type (geometry)
  int index; //shape
  //position
  double X0;
  double Y0;
  double Z0;
  //orientation
  double Phi;
  double Theta;
  double Psi;
  //size
  double size1;
  double size2;
  double size3;
  //collimation
  double CollPhi;
  double CollTheta;
  double Spread;

  //Activity
  double Activity;

  //particles
  QVector<GunParticleStruct*> GunParticles;

  //constructor - creates a default source with _no_ particles!
  ParticleSourceStructure()
    {
      name = "Underfined_name";
      Activity = 1;
      index = 0;
      X0 = 0; Y0 = 0; Z0 = 0;
      size1 = 10; size2 = 10; size3 = 10;
      Phi = 0; Theta = 0; Psi = 0;
      CollPhi = 0; CollTheta = 0;
      Spread = 45;

      GunParticles.resize(0);
    }

  //destructor - takes care of dynamic GunParticles
  ~ParticleSourceStructure()
    {
      for (int i=0; i<GunParticles.size(); i++) delete GunParticles[i];
    }
};

class GeneratedParticleStructure
{
public:
    int ParticleId;
    double Energy;
    double Position[3];
    double Direction[3];
};

class LinkedParticleStructure
{
public:
  int iParticle; //indexed according to GunParticles index
  int LinkedTo;  //index of particle it is linked to

  LinkedParticleStructure() {}
  LinkedParticleStructure(int iparticle, int linkedto = -1) {iParticle = iparticle; LinkedTo = linkedto;}
};

class ParticleSourcesClass
{
public:
  //configuration
  void connectDataContainers(const QVector<ParticleDescriptor *> *particleCollection, MaterialCollectionClass* MaterialCollection);
  void connectRandomGenerator(TRandom2 *randGen) {RandGen = randGen;}
  void setNameAddon(TString nameID) {NameID = nameID;}

  //MAIN usage
  void Init(); // !!! has to be called before the first use of "GenerateEvent"  NOT thread safe!
  QVector<GeneratedParticleStructure>* GenerateEvent(); //see Init!!!

  //requests
  int size() {return ParticleSourcesData.size();}
  double getTotalActivity();
  ParticleSourceStructure &operator[] (int iSource) {return *ParticleSourcesData[iSource];}
  ParticleSourceStructure &last() {return *ParticleSourcesData.last();}

  //Source handling - after handling is finished, requires Init() !!!
  void append(ParticleSourceStructure* gunParticle);
  void remove(int iSource);
  void clear();

  //check consistency of data
  int CheckSource(int isource); // 0 - no errors
  QString getErrorString(int error);

  // save
  bool writeSourceToJson(int iSource, QJsonObject &json); //only one source
  bool writeToJson(QJsonObject &json); //all config
  bool SaveSourceToJsonFile(QString fileName, int iSource);
  // load
  bool readSourceFromJson(int iSource, QJsonObject &json);
  bool readFromJson(QJsonObject &json); // all config
  bool LoadSourceFromJsonFile(QString fileName, int iSource);
  bool LoadGunEnergySpectrum(int iSource, int iParticle, QString fileName);

  //constructor / destructor
  ParticleSourcesClass();
  ~ParticleSourcesClass();

  TVector3 GenerateRandomDirection();
private:
  QVector<ParticleSourceStructure*> ParticleSourcesData;
  const QVector<ParticleDescriptor*>* ParticleCollection;  //pointer to the ParticleCollection
  MaterialCollectionClass* MaterialCollection;
  TRandom2 *RandGen;
  TString NameID; //use to make unique hist name in multithread environment

  //QVector<GeneratedParticleStructure> GeneratedParticles;
  QVector<double> TotalParticleWeight;  
  double TotalActivity;
  QVector< QVector< QVector<LinkedParticleStructure> > > LinkedPartiles; //[isource] [iparticle] []  (includes the record of the particle iteslf!!!)
                              //full recipe of emission builder (containes particles linked to particles etc up to the top level individual particle)

  QVector<TVector3> CollimationDirection; //[isource] collimation direction 
  QVector<double> CollimationProbability; //[isource] collimation probability: solid angle inside cone / 4Pi
  //QVector<int> CollimationSpreadProduct; //[isource] vector product of collimation and normal (used to check is the new vector inisde collimation cone)

  //utilities
  void CalculateTotalActivity();
  void GeneratePosition(int isource, double *R);
  void AddParticleInCone(int isource, int iparticle, QVector<GeneratedParticleStructure> *GeneratedParticles); //QVector - only pointer is transferred!
};

#endif // PARTICLESOURCESCLASS_H
