//ANTS
#include "photon_tracker.h"
#include "functions.h"
#include "pms.h"
#include "materialcollectionclass.h"
#include "generalsimsettings.h"
#include "opticaloverrideclass.h"
#include "detectionstatisticsclass.h"

//Qt
#include <QtGui>

//ROOT
#include "TGeoManager.h"
#include "TVirtualGeoTrack.h"
#include "TGeoTrack.h"
#include "TMath.h"

#define c_in_vac 2.997925e2 //speed of light in mm/ns

Photon_Tracker::Photon_Tracker(TGeoManager *geoManager, TRandom2 *RandomGenerator, MaterialCollectionClass* materialCollection, pms* Pms, const QVector<OpticalGridStructure> *Grids, QObject *parent) :
    QObject(parent)
{    
    RandGen = RandomGenerator;
    GeoManager = geoManager;
    MaterialCollection = materialCollection;
    PMs = Pms;
    grids = Grids;
    BuildTracks = false;
}

void Photon_Tracker::configure(const GeneralSimSettings *simSet, OneEventClass* oneEvent, bool fBuildTracks, QVector<TrackHolderClass*> *tracks)//bool fWave, bool fAngle,  bool fArea, int MaxTrans, bool fTracks, bool fFastTracks, bool fAccelQE, double maxQE, QVector<double> *MaxQEwave)
{
   SimSet = simSet;
   OneEvent = oneEvent;
   BuildTracks = fBuildTracks;
   Tracks = tracks;
}

void Photon_Tracker::TrackPhoton(One_Photon* Photon)
{
  //qDebug() << "----accel:"<<SimSet->MaxQE<<SimSet->fQEaccelerator;
  //accelerators;
  double rnd;
  if (SimSet->fQEaccelerator)
    {
      rnd = RandGen->Rndm();
      if (SimSet->fWaveResolved && Photon->WaveIndex != -1)
        {
          if (rnd > PMs->getMaxQEvsWave(Photon->WaveIndex))
            {
#ifdef advanced_photon_tracing
         OneEvent->DetStat->TracingSkipped++;
#endif
              return; //no need to trace this photon
            }
        }
      else
        {
           if (rnd > PMs->getMaxQE())
             {
#ifdef advanced_photon_tracing
         OneEvent->DetStat->TracingSkipped++;
#endif
               return; //no need to trace this photon
             }
        }
    }

   //=====inits=====
   navigator = GeoManager->GetCurrentNavigator();
   navigator->SetCurrentPoint(Photon->r);
   navigator->SetCurrentDirection(Photon->v);
   navigator->FindNode();

   if (navigator->IsOutside())
     {
//       qDebug()<<"Generated outside geometry!";
       return;
     }

   //current vector:     
   K[0] = Photon->v[0];
   K[1] = Photon->v[1];
   K[2] = Photon->v[2];
//   qDebug()<<"on start:"<<K[0]<<K[1]<<K[2]<<K[0]*K[0]+K[1]*K[1]+K[2]*K[2] ;

//   qDebug()<<"Photon starts from:";
//   qDebug()<<GeoManager->GetPath();
//   qDebug()<<"material name: "<<GeoManager->GetCurrentVolume()->GetMaterial()->GetName();
//   qDebug()<<"material index: "<<GeoManager->GetCurrentVolume()->GetMaterial()->GetIndex();

   TrackHolderClass* track;
   bool missPM = true;
   if (BuildTracks)
     {
       if (Tracks->size()<1000)
         {
           track = new TrackHolderClass();
           track->Nodes.append(TrackNodeStruct(Photon->r, Photon->time));
         }
       else BuildTracks = false;
     }

   TGeoNode* next;
   int PreviousMatIndex = navigator->GetCurrentVolume()->GetMaterial()->GetIndex();
   int ThisMatIndex;

   int Counter = 0; //number of photon transitions - there is a limit on this set by user
   //---------------------------------------------=====cycle=====-----------------------------------------
   while (Counter < SimSet->MaxNumTrans)
     {
     Counter++;

     MaterialStructure* PreviousMaterial = (*MaterialCollection)[PreviousMatIndex]; //this is the material photon is currently in!

     navigator->FindNextBoundary();
     double Step = navigator->GetStep();
     //                 qDebug()<<"Step:"<<Step;
     navigator->Step(kTRUE, kFALSE); //stopped right before the crossing
     //putting this location/state to the stack
     navigator->PushPoint(); //DO NOT FORGET TO CLEAN IT IF NOT USED!!!

     next = navigator->FindNextBoundaryAndStep(); //next - the node we enter after crossing the boundary
     //this method MOVES the current position! different from FindNextBoundary method, which only calculates the step
     //now current point is inside the next volume!

     //-----Bulk Absorption----- AND -----Rayleigh scattering-----
     //Optimized assuming rare use of Rayleigh and very rare use when both abs and Rayleigh are defined!
     //if both active, extract distance at which process would happen, then select the process with the shortest path
     bool DoAbsorption;
      //prepare for abs
     double AbsPath, AbsCoeff;
     if (Photon->WaveIndex == -1) AbsCoeff = (*MaterialCollection)[PreviousMatIndex]->abs;
     else
       {
         if ((*MaterialCollection)[PreviousMatIndex]->absWaveBinned.isEmpty()) AbsCoeff = (*MaterialCollection)[PreviousMatIndex]->abs;
         else AbsCoeff = (*MaterialCollection)[PreviousMatIndex]->absWaveBinned[Photon->WaveIndex]; //if defined
       }

     if (AbsCoeff > 0)
       {
         AbsPath = -log(RandGen->Rndm())/AbsCoeff;
         if (AbsPath < Step) DoAbsorption = true;
         else DoAbsorption = false;
       }
     else DoAbsorption = false;

      //prepare Rayleigh
     bool DoRayleigh;
     double RayleighPath;
     if ((*MaterialCollection)[PreviousMatIndex]->rayleighMFP == 0) DoRayleigh = false;      // == 0 - means undefined
     else
       {
         double RayleighMFP;
         if (Photon->WaveIndex == -1) RayleighMFP = (*MaterialCollection)[PreviousMatIndex]->rayleighMFP;
         else RayleighMFP = (*MaterialCollection)[PreviousMatIndex]->rayleighBinned[Photon->WaveIndex];

         RayleighPath = -RayleighMFP * log(RandGen->Rndm());
         if (RayleighPath < Step) DoRayleigh = true;
         else DoRayleigh = false;
       }

     //checking abs and rayleigh
     if (DoAbsorption || DoRayleigh)
       {
         if (DoAbsorption && DoRayleigh)
           {
             //slecting the one having the shortest path
             if (AbsPath<RayleighPath) DoRayleigh = false;
             else DoAbsorption = false;
           }

         if (DoAbsorption)
           {
             //qDebug()<<"Absorption was triggered!";
             if (BuildTracks)
               {
                 double compPath = Step - AbsPath;
                 //track->AddPoint(GeoManager->GetCurrentPoint()[0]-K[0]*compPath,GeoManager->GetCurrentPoint()[1]-K[1]*compPath,GeoManager->GetCurrentPoint()[2]-K[2]*compPath,Photon->time);
                 Double_t point[3];
                 point[0] = navigator->GetCurrentPoint()[0]-K[0]*compPath;
                 point[1] = navigator->GetCurrentPoint()[1]-K[1]*compPath;
                 point[2] = navigator->GetCurrentPoint()[2]-K[2]*compPath;
                 track->Nodes.append(TrackNodeStruct(point, Photon->time));
                }
             navigator->PopDummy();//clean up the stack
#ifdef advanced_photon_tracing
             double refIndex;
             if (Photon->WaveIndex == -1) refIndex = PreviousMaterial->n;
             else
               {
                 if (PreviousMaterial->nWaveBinned.isEmpty()) refIndex = PreviousMaterial->n;
                 else refIndex = PreviousMaterial->nWaveBinned[Photon->WaveIndex];
               }
             Photon->time += AbsPath/c_in_vac*refIndex;
             OneEvent->DetStat->Absorbed++;
#endif
             goto force_stop_tracing; //finished with this photon
           }
         else if (DoRayleigh)
           {
             //qDebug()<<"Scattering was triggered";
               //interaction position
             double compPath = Step - RayleighPath;
             double R[3];
             R[0] = navigator->GetCurrentPoint()[0]-K[0]*compPath;
             R[1] = navigator->GetCurrentPoint()[1]-K[1]*compPath;
             R[2] = navigator->GetCurrentPoint()[2]-K[2]*compPath;
             navigator->SetCurrentPoint(R);
             navigator->FindNode();
               //new direction
             double Kold[3];
             Kold[0] = K[0]; Kold[1] = K[1]; Kold[2] = K[2];
             double dotProduct;
             do
               {                 
                 //choosing random direction for this photon
                 Photon_Tracker::RandomDir();
                 dotProduct = K[0]*Kold[0] + K[1]*Kold[1] + K[2]*Kold[2];
               }
             while ( (dotProduct*dotProduct + 1.0) < RandGen->Rndm(2.0));
             navigator->SetCurrentDirection(K);

             //updating track if needed
             if (BuildTracks)
               {
                 //track->AddPoint(R[0], R[1], R[2],Photon->time);
                 track->Nodes.append(TrackNodeStruct(R, Photon->time));
               }
             navigator->PopDummy();//clean up the stack
#ifdef advanced_photon_tracing
             double refIndex;
             if (Photon->WaveIndex == -1) refIndex = PreviousMaterial->n;
             else
               {
                 if (PreviousMaterial->nWaveBinned.isEmpty()) refIndex = PreviousMaterial->n;
                 else refIndex = PreviousMaterial->nWaveBinned[Photon->WaveIndex];
               }
             Photon->time += RayleighPath/c_in_vac*refIndex;
#endif
             continue; //next step
           }
       }

     //check if after the border the photon is outside the defined world
     if (navigator->IsOutside()) //outside of geometry - end tracing
       {
         const double *point = navigator->GetCurrentPoint();
         if (BuildTracks)
           {
             //track->AddPoint(point[0],point[1],point[2],Photon->time);
             track->Nodes.append(TrackNodeStruct(point, Photon->time));
           }
         navigator->PopDummy();//clean up the stack
#ifdef advanced_photon_tracing
         OneEvent->DetStat->Escaped++;
#endif
         goto force_stop_tracing; //finished with this photon
       }

     //-------------------------
     //can make the track now - definitely made it to the other border
     if (BuildTracks)
       {
         const double *point = navigator->GetCurrentPoint();
         track->Nodes.append(TrackNodeStruct(point, Photon->time));
       }
#ifdef advanced_photon_tracing
     double refIndex;
     if (Photon->WaveIndex == -1) refIndex = PreviousMaterial->n;
     else
       {
         if (PreviousMaterial->nWaveBinned.isEmpty()) refIndex = PreviousMaterial->n;
         else refIndex = PreviousMaterial->nWaveBinned[Photon->WaveIndex];
       }
     Photon->time += Step/c_in_vac*refIndex;
#endif

     //new volume info
     TGeoVolume* ThisVolume = next->GetVolume();     
     ThisMatIndex = ThisVolume->GetMaterial()->GetIndex();     
     flagHaveNormal = false;

//     qDebug()<<"Found border with another volume: "<<ThisVolume->GetName();
//     qDebug()<<"new mat index: "<<ThisMatIndex<<" Previous mat index: "<<PreviousMatIndex;
//     qDebug()<<"coordinates: "<<GeoManager->GetCurrentPoint()[0]<<GeoManager->GetCurrentPoint()[1]<<GeoManager->GetCurrentPoint()[2];

     //-----Checking overrides-----
     bool flagDoFresnel;
     OpticalOverrideClass* ov = PreviousMaterial->OpticalOverrides[ThisMatIndex];
     if (ov)
       {
         //qDebug() << "Overrides defined! Model = "<<ov->getType();
         N = navigator->FindNormal(kFALSE);
         flagHaveNormal = true;
         OpticalOverrideClass::OpticalOverrideResultEnum result = ov->calculate(RandGen, K, N, Photon->WaveIndex);

         switch (result)
           {
           case OpticalOverrideClass::Absorbed:
               navigator->PopDummy(); //clean up the stack
#ifdef advanced_photon_tracing
               OneEvent->DetStat->OverrideLoss++;
#endif
               goto force_stop_tracing; //finished with this photon
           case OpticalOverrideClass::Back:
               navigator->PopPoint();  //remaining in the original volume
               navigator->SetCurrentDirection(K); //updating direction
               continue; //send to the next iteration
           case OpticalOverrideClass::Forward:
               navigator->SetCurrentDirection(K); //updating direction
               flagDoFresnel = false; //stack cleaned afterwards
               break; //switch break
           case OpticalOverrideClass::NotTriggered:
               flagDoFresnel = true;
               break; //switch break
           default:
               qCritical() << "override reported an error!";
           }
       }
     else flagDoFresnel = true;

     double n1, n2; //refractive indexes
     if (flagDoFresnel)
       {
//         qDebug()<<"Performing fresnel";
         //-----Reflection-----
           //Fresnel equations //http://en.wikipedia.org/wiki/Fresnel_equations
         //normal vector?
         N = navigator->FindNormal(kFALSE);
//         qDebug()<<"Normal length is:"<<sqrt(N[0]*N[0]+N[1]*N[1]+N[2]*N[2]);
//         qDebug()<<"Dir vector length is:"<<sqrt(K[0]*K[0]+K[1]*K[1]+K[2]*K[2]);
         flagHaveNormal = true;

         double prob; //reflection probability
         double NK = 0;
         for (int i=0; i<3; i++) NK += N[i]*K[i]; // NK = cos of the angle of incidence = cos1
         //double NK = N[0]*K[0] + N[1]*K[1] + N[2]*K[2]; //this one is slower!
         double cos1 = fabs(NK);

         if (Photon->WaveIndex == -1)
           { //WavelengthResolved is OFF or no data for emission spectrum
             n1 = PreviousMaterial->n;
             n2 = (*MaterialCollection)[ThisMatIndex]->n;
           }
         else
           {
             if (PreviousMaterial->nWaveBinned.isEmpty()) n1 = PreviousMaterial->n;
             else n1 = PreviousMaterial->nWaveBinned[Photon->WaveIndex];

             MaterialStructure* ThisMaterial = (*MaterialCollection)[ThisMatIndex];            
             if (ThisMaterial->nWaveBinned.isEmpty()) n2 = ThisMaterial->n;
             else n2 = ThisMaterial->nWaveBinned[Photon->WaveIndex];
           }

//         qDebug()<<"Photon wavelength"<<Photon->wavelength<<"WaveIndex:"<<Photon->WaveIndex<<"n1 and n2 are: "<<n1<<n2;       
         double sin1 = sqrt(1.0 - NK*NK);
         double sin2 = n1/n2*sin1;

//         qDebug()<<"cos1 sin1 sin2 are:"<<cos1<<sin1<<sin2;
         if (fabs(sin2)>1.0)
           {
             //total internal!
//             qDebug()<<"total internal, forcing reflection";
             prob = 1.0;
           }
         else
           {
             double cos2 = sqrt(1-sin2*sin2);
             double Rs = (n1*cos1-n2*cos2) / (n1*cos1+n2*cos2);
             Rs *= Rs;
             double Rp = (n1*cos2-n2*cos1) / (n1*cos2+n2*cos1);
             Rp *= Rp;
             prob = 0.5*(Rs + Rp);
//             qDebug()<<"reflection prob: "<<prob;
           }

         if (RandGen->Rndm() < prob)
           {
//             qDebug()<<"Fresnel - reflection!";
             navigator->PopPoint(); //restore the point before the border
             PerformReflection();         
             //photon remains in the same volume! -> next iteration
             continue;
           } //else transmission
       }

//         qDebug()<<"Photon entered new volume";
         //Photon entered another volume
         //could be scattering or fresnel - for the beginning we ignore refraction!
         navigator->PopDummy(); //clean up the stack

         //entered =====PM=====?        
         const char* VolName = ThisVolume->GetName();
         if (VolName[0] == 'P' &&  VolName[1] == 'M') // Starts with PM - then comes the type number
                 {
                      //int PMnumber = GeoManager->GetCurrentNode()->GetNumber();
                      int PMnumber = next->GetNumber();
//                      qDebug()<<"PM hit:"<<VolName<<PMnumber;

                      bool SiPM = PMs->isSiPM(PMnumber);
                      Double_t local[3];
                      if (SimSet->fAreaResolved || SiPM)
                        {
                          const Double_t *global = navigator->GetCurrentPoint();
                          // Go from MARS to local coordinates:
                          navigator->MasterToLocal(global,local);
                          //qDebug()<<local[0]<<local[1];
                        }//if no area dep - local[0] and [1] are undefined

                      double cosAngle;
                      if (SimSet->fAngResolved)
                        {
                          //since we check vs cos of _refracted_:
                          if (flagDoFresnel) Refraction( n1 / n2); // true - successful
                          if (!flagHaveNormal) N = navigator->FindNormal(kFALSE);
//                                   qDebug()<<N[0]<<N[1]<<N[2]<<"Normal length is:"<<sqrt(N[0]*N[0]+N[1]*N[1]+N[2]*N[2]);
//                                   qDebug()<<K[0]<<K[1]<<K[2]<<"Dir vector length is:"<<sqrt(K[0]*K[0]+K[1]*K[1]+K[2]*K[2]);
                          cosAngle = 0;
                          for (int i=0; i<3; i++) cosAngle += N[i]*K[i];
//                          qDebug()<<"cos() = "<<cosAngle;                      
                        }

                      if (!SimSet->fQEaccelerator) rnd = RandGen->Rndm(); //else already calculated
                      if (SiPM) OneEvent->CheckSiPMhit(PMnumber, Photon->time, Photon->WaveIndex, local[0], local[1], cosAngle, Counter, rnd);
                      else OneEvent->CheckPMThit(PMnumber, Photon->time, Photon->WaveIndex, local[0], local[1], cosAngle, Counter, rnd);

                      missPM = false;
#ifdef advanced_photon_tracing
                      OneEvent->DetStat->HitPM++;
#endif
                      goto force_stop_tracing; //finished with this photon
                 }         
         else if (VolName[0] == 'd' && VolName[1] == 'P' &&  VolName[2] == 'M')
           {
             // dummy PM hit
#ifdef advanced_photon_tracing
             OneEvent->DetStat->HitDummy++;
#endif
             goto force_stop_tracing; //finished with this photon
           }
         else if (!grids->isEmpty() && VolName[0] == 'G' && VolName[1] == 'R') //optical grid hit?
           {
             int igrid = next->GetNumber();
             if (igrid < grids->size())
               {
                 //model:
                 //theta - angle of incidence
                 //phi - in-plain angle with direction of wires
                 //w-distance between wires, d - diameter
                 //normal ansorption probability: Pnorm=d/w
                 //for arb angle: P = Pnorm * (sqrt(cos2Theta*cos2phi+sin2Phi)) / cosTheta
//                 qDebug() << "Grid hit!  name:"<< VolName << " index:" << next->GetNumber();

                 //calculating incidence angle
                 if (!flagHaveNormal) N = navigator->FindNormal(kFALSE);
                 double cosTheta = 0;
                 for (int i=0; i<3; i++) cosTheta += N[i]*K[i];
//                 qDebug()<<"cos of incidence angle = "<<cosTheta;
                 //calculating local vector
//                 qDebug()<<"Global photon vector:"<<K[0]<<K[1]<<K[2];
                 double k[3];
                 navigator->MasterToLocalVect(K,k);
//                 qDebug()<<"Local photon vector:"<<k[0]<<k[1]<<k[2];
                 //norm to unitary length in plain
                 double norm = sqrt(k[0]*k[0] + k[1]*k[1]);
                 if (norm<1.0e-10)
                   {
                     k[0] = 0;
                     k[1] = 1.0;
                   }
                 else
                   {
                     k[0] /= norm;
                     k[1] /= norm;
                   }
//                 qDebug() << "After norm:"<<k[0]<<k[1];

                 //calculating intercept probability                
                 double absProb = 0;                
                 if (grids->at(igrid).Pitch>1.0e-10)
                   {
                     double normAbsProb = grids->at(igrid).Diameter / grids->at(igrid).Pitch;
                     //qDebug() << "NormAbsProb="<<normAbsProb;
                     for (int ig=0; ig<grids->at(igrid).Type+1; ig++)
                       {
                         //if type==0 - parallel wires, one pass; if type==1, do two passes->crossed wires
                         //qDebug()<<"->pass:"<<ig;
                         double angle = grids->at(igrid).WireAngle;
                         if (ig == 1) angle += 90.0; //second system of wires
                         //qDebug() << "Wire angle:"<<angle;

                         //rotating local vector to get elative in-plane angle phi of the wires
                         angle *= -3.1415926535/180.0; //clockwise!
                         double cosA = cos(angle);
                         double sinA = sin(angle);
                         double cosPhi = k[0]*cosA - k[1]*sinA;//xcos-ysin
                         double sinPhi = k[0]*sinA + k[1]*cosA;//xsin+ycos
                         //qDebug() << "cosPhi, sinPhi="<<cosPhi<<sinPhi;
                         if (cosTheta == 0)
                           {
                             absProb = 1.0;
                             break; //local break
                           }
                         double delta = normAbsProb * sqrt(cosTheta*cosTheta*cosPhi*cosPhi + sinPhi*sinPhi) / cosTheta;
                         absProb += delta;
                         //qDebug() << "AbsProb:"<<absProb;

                         //compensate for two times taken areas for crossed wires
                         if (ig == 1)
                           {
                             //P = P * {1 - d*d/w/w * (sqrt(cos2Theta*cos2phi+sin2Phi)) / cosTheta}
                             absProb -= absProb * normAbsProb * delta;
                             //qDebug() << "AbsProb after cross correction:"<<absProb;
                           }
                       }
                   }
                 else absProb = 1.0;

                 //checking: is photon hits the grid
                 if (RandGen->Rndm() < absProb)
                   {
//                   qDebug() << "Grid was hit!";
#ifdef advanced_photon_tracing
                     OneEvent->DetStat->LossOnGrid++;
#endif
                     goto force_stop_tracing; //finished with this photon (stack already cleaned!)
                   }

//                 qDebug()<<"Passed the grid!";

                 //continuing tracking - have to check refraction in case grid volume is just after the border
                 if (flagDoFresnel)
                   {
                     bool ok = Refraction( n1 / n2); // true - successful
                     // true - successful, false - forbidden -> considered that the photon is absorbed at the surface! Should not happen
                     if (!ok) qWarning()<<"Error in photon tracker: problem with transmission!";
                   }
               }
           }
         else
                 {
                    //if doing fresnel, perform refraction
                    if (flagDoFresnel)
                      {
                        //-----Refraction-----
                        bool ok = Refraction( n1 / n2); // true - successful
                        // true - successful, false - forbidden -> considered that the photon is absorbed at the surface! Should not happen
                        if (!ok) qWarning()<<"Error in photon tracker: problem with transmission!";
                      }
                 }

         PreviousMatIndex = ThisMatIndex;
    } // finished with this photon   

#ifdef advanced_photon_tracing
     if (Counter == SimSet->MaxNumTrans) OneEvent->DetStat->MaxCyclesReached++;
#endif

force_stop_tracing:
   if (BuildTracks)     
     {       
       //if (SimSet->fTracksOnPMsOnly && missPM || GeoManager->GetListOfTracks()->GetEntries() > 1000 ) track->Delete();
       if ( SimSet->fTracksOnPMsOnly && missPM ) delete track;
       else
         {
           track->UserIndex = 22;
           //track->SetLineWidth(1);
           track->Width = 1;
           if (missPM)
             {
               switch (Photon->scint_type)
                 {
                   case 1: { //primary
                      //track->SetLineColor(kTeal);
                      track->Color = 7;//kTeal;
                      break;
                   }
                   case 2: { //secondary
                      //track->SetLineColor(kMagenta);
                      track->Color = kMagenta;
                      break;
                   }
                   default: { //just in case
                      //track->SetLineColor(kGray);
                      track->Color = kGray;
                   }
                 }
             }
           else
             {
               //track->SetLineColor(kRed);
               track->Color = 2;//kRed
             }
             //registering track with GeoManager
           //GeoManager->AddTrack(track); will register later!
           Tracks->append(track);
         }
     }
   //   qDebug()<<"Finished with the photon";
}

bool Photon_Tracker::Refraction(Double_t nn) // nn = nFrom / nTo
{
    //qDebug()<<"refractiontest";
    //N - normal, K - origial direction
    // nn = n(in)/n(tr)
    // (NK) - scalar product
    // T = -( nn*(NK) - sqrt(1-nn*nn*[1-(NK)*(NK)]))*N + nn*K

    //if (!flagHaveNormal) N = GeoManager->FindNormal(kFALSE);
    if (!flagHaveNormal) N = navigator->FindNormal(kFALSE);
    //Double_t NK = 0;
    //for (int i=0; i<3; i++) NK += N[i]*K[i];
    Double_t NK = N[0]*K[0] + N[1]*K[1] + N[2]*K[2];

    Double_t UnderRoot = 1.0 - nn*nn*(1.0 - NK*NK);
    if (UnderRoot<0)
      {
//        qDebug()<<"total internal detected - will assume this photon is abosrbed at the surface";
        return false;    //total internal reflection! //previous reflection test should catch this situation
      }
    Double_t tmp = nn*NK - sqrt(UnderRoot);

    //for (int i=0; i<3; i++) K[i] = -tmp*N[i] + nn*K[i];
    K[0] = -tmp*N[0] + nn*K[0]; K[1] = -tmp*N[1] + nn*K[1]; K[2] = -tmp*N[2] + nn*K[2];

    ///GeoManager->SetCurrentDirection(K);
    navigator->SetCurrentDirection(K);
    return true;
}

void Photon_Tracker::PerformReflection()
{
  if (!flagHaveNormal)
    {
      ///N = GeoManager->FindNormal(kFALSE);
      N = navigator->FindNormal(kFALSE);
      flagHaveNormal = true;
    }

  //rotating the vector

  //K = K - 2*(NK)*N
  //double NK=0;
  //for (int i=0; i<3; i++) NK += N[i]*K[i];
  double NK = N[0]*K[0]; NK += N[1]*K[1];  NK += N[2]*K[2];

  //for (int i=0; i<3; i++) K[i] -= 2*NK*N[i];
  K[0] -= 2.0*NK*N[0]; K[1] -= 2.0*NK*N[1]; K[2] -= 2.0*NK*N[2];

  ///GeoManager->SetCurrentDirection(K);
  navigator->SetCurrentDirection(K);
}

void Photon_Tracker::RandomDir()
{
  //Sphere function of Root:
  double a=0,b=0,r2=1;
  while (r2 > 0.25)
    {
        a  = RandGen->Rndm() - 0.5;
        b  = RandGen->Rndm() - 0.5;
        r2 =  a*a + b*b;
    }
  K[2] = ( -1.0 + 8.0 * r2 );
  double scale = 8.0 * TMath::Sqrt(0.25 - r2);
  K[0] = a*scale;
  K[1] = b*scale;
}
