#include "primaryparticletracker.h"
#include "functions.h"
#include "materialcollectionclass.h"
#include "generalsimsettings.h"

#include <QtGui>
#include "QMessageBox"
#include "QtAlgorithms"

//#include "TROOT.h"
#include "TMath.h"
#include "TGeoManager.h"
#include "TVirtualGeoTrack.h"
#include "TRandom2.h"
#include "TGeoTrack.h"

PrimaryParticleTracker::PrimaryParticleTracker(TGeoManager *geoManager, TRandom2 *RandomGenerator, const QVector<ParticleDescriptor *> *particleCollection, MaterialCollectionClass* materialCollection, QVector<ParticleOnStack*>* particleStack, QVector<EnergyDepositionCell*>* energyVector, QVector<EventHistoryStructure*>* eventHistory, QObject *parent) :
    QObject(parent)
{    
     GeoManager = geoManager;
     RandGen = RandomGenerator;
     ParticleCollection = particleCollection;
     MaterialCollection = materialCollection;
     ParticleStack = particleStack;
     EnergyVector = energyVector;
     EventHistory = eventHistory;

     BuildTracks = true;
     RemoveTracksIfNoEnergyDepo = true;  
     AddColorIndex = 0;

     counter = -1;
}

void PrimaryParticleTracker::configure(const GeneralSimSettings* simSet, bool fbuildTracks, QVector<TrackHolderClass*> *tracks, bool fremoveEmptyTracks)
{
  SimSet = simSet;  
  BuildTracks = fbuildTracks;
  RemoveTracksIfNoEnergyDepo = fremoveEmptyTracks;
  AddColorIndex = simSet->TrackColorAdd;
  Tracks = tracks;
}

bool PrimaryParticleTracker::TrackParticlesInStack(int eventId)
{
   double InitialEnergy;

   while (ParticleStack->size())
   {//-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/ beginning cycle over the elements of the stack
       //reading data for the top particle at the stack
       counter ++; //new particle      
       Id = ParticleStack->at(0)->Id; //id of the particle
       //qDebug()<<"Tracking new particle "<<counter;
       //qDebug()<<"Particle Id:"<<Id<<" (Collection size = "<<ParticleCollection->size()<<")";
       //qDebug()<<"Part name, charge, mass"<<ParticleCollection->at(Id)->ParticleName<<ParticleCollection->at(Id)->charge<<ParticleCollection->at(Id)->mass;
       //qDebug()<<"Part type:"<<ParticleCollection->at(Id)->type;

       EventHistoryStructure* History;
       if (SimSet->fLogsStat)
         {
           //initializing history entry - at end will be appended to EventHistory QVector
           History = new EventHistoryStructure;
           History->index = counter;
           History->ParticleId = Id;
           History->flagSecondary = ParticleStack->at(0)->flagSecondary;
           History->dx = ParticleStack->at(0)->v[0];
           History->dy = ParticleStack->at(0)->v[1];
           History->dz = ParticleStack->at(0)->v[2];
         }

       // interaction type for this particle
       int InteractionType;
       if (ParticleCollection->at(Id)->type == _charged_) InteractionType = 1;
       else InteractionType = 0;
                            //0 - point
                            //1 - continuous
       //qDebug()<<"Int type (0-point, 1-continuous):"<<InteractionType;
       for (int i=0; i<3; i++) r[i] = ParticleStack->at(0)->r[i];
       for (int i=0; i<3; i++)  v[i] = ParticleStack->at(0)->v[i];
       MyNormalizeVector(v); //normalization of the staring vector       
       InitialEnergy = ParticleStack->at(0)->Energy;
       energy = InitialEnergy;
       time = ParticleStack->at(0)->t;

       //initializing GeoManager
       TGeoNavigator *navigator = GeoManager->GetCurrentNavigator();
       ///GeoManager->SetCurrentPoint(r);
       navigator->SetCurrentPoint(r);
       ///GeoManager->SetCurrentDirection(v);
       navigator->SetCurrentDirection(v);
       ///GeoManager->FindNode();
       navigator->FindNode();

       //is it created outside of Top?
       ///if (GeoManager->IsOutside())
       if (navigator->IsOutside())
         {
           if (SimSet->fLogsStat)
             {
               History->Termination = 7;
               EventHistory->append(History);
             }

           //delete the record and remove from the stack
           delete ParticleStack->at(0);
           ParticleStack->remove(0);
           return true;
         }

       //What is the starting material?
       ///int MatId = GeoManager->GetCurrentVolume()->GetMaterial()->GetIndex();
       int MatId = navigator->GetCurrentVolume()->GetMaterial()->GetIndex();
//       qDebug()<<"First material id: "<<MatId;

       int flagContinue = 0; //0 - continue

                             //1 - escaped
                             //2 - all energy dissipated
                             //3 - photoeffect
                             //4 - compton
                             //5 - capture
                             //6 - error - undefined termination
                             //7 - created outside defined geometry
                             //8 - found untrackable material

       //TVirtualGeoTrack *track;
       TrackHolderClass *track;
       if (BuildTracks)
         {
          if (Tracks->size()<1000)
            {
              //Int_t track_index = GeoManager->AddTrack(1,22); //  Here track_index is the index of the newly created track in the array of primaries. One can get the pointer of this track and make it known as current track by the manager class:
              //TVirtualGeoTrack *track = GeoManager->GetTrack(track_index);
              //GeoManager->SetCurrentTrack(track);

              //track = new TGeoTrack(1,22);
              track = new TrackHolderClass();
              track->UserIndex = 22;
              //track->SetLineColor(Id+1+ AddColorIndex);
              track->Color = Id+1+ AddColorIndex;
              track->Width = 2;
              //track->AddPoint(r[0], r[1], r[2], time); //GeoManager->GetCurrentTrack()->AddPoint(r[0], r[1], r[2], time);
              track->Nodes.append(TrackNodeStruct(r, time));
            }
          else BuildTracks = false;
         }

       //will do a loop for tracking-allowed materials on the path
       do {//--------------------------------------------------------------------
//           qDebug()<<"cycle starts for the new material: "<<GeoManager->GetPath();
           double distanceHistory = 0; //for diagnostics - travelled distance and deposited energy in THIS material
           double energyHistory = 0;

           ///for (int j=0; j<3; j++) rCur[j]=GeoManager->GetCurrentPoint()[j]; //current position
           for (int j=0; j<3; j++) rCur[j]=navigator->GetCurrentPoint()[j]; //current position

           ///if (GeoManager->IsOutside())
           if (navigator->IsOutside())
             {
//              qDebug()<<"Escaped from the defined geometry!";
               flagContinue = 1;
               break; //do-break
             }

           if ( !(*MaterialCollection)[MatId]->MatParticle[Id].TrackingAllowed )
           {
 //              qDebug()<<"Found medium where tracking is not allowed!";
               flagContinue = 8;
               break; //do-break
           }           

           //Maximum allowed range?
           ///GeoManager->FindNextBoundary();
           navigator->FindNextBoundary();
           ///Double_t MaxLength  = GeoManager->GetStep();
           Double_t MaxLength  = navigator->GetStep();

           if ( !(*MaterialCollection)[MatId]->MatParticle[Id].MaterialIsTransparent && (*MaterialCollection)[MatId]->MatParticle[Id].InteractionDataX.size() > 1 )
          { // if no interaction data are defined, pass this medium (one point is also not enough)


//           qDebug()<<"Distance to the next border: "<<MaxLength;

           //checking are we in range of available interaction data
           int DataPoints = (*MaterialCollection)[MatId]->MatParticle[Id].InteractionDataX.size();
           if (energy >  (*MaterialCollection)[MatId]->MatParticle[Id].InteractionDataX[DataPoints-1]) //largest value is at the end - sorted data!
             {  //outside!
//             qDebug()<<"energy: "<<energy;
               message("energy is outside of the supplied data!");
               return false; //fail
             }

           double Density = (*MaterialCollection)[MatId]->density;
           double AtomicDensity = (*MaterialCollection)[MatId]->atomicDensity;

           //Next processing depends on the interaction type!
           if (InteractionType == 0)
           {// /////////neutral particle/////////////   - photoeffect/compton/capture
               //how many processes?
               int processes = (*MaterialCollection)[MatId]->MatParticle[Id].Terminators.size();
//               qDebug()<<"defined processes:"<<processes;
               if (processes != 0)
               {
                   double SoFarShortestId = 0;
                   double SoFarShortest = 1.0e10;
                   for (int i=0; i<processes; i++)
                   {
//                       qDebug()<<"process #:"<<i;
                       //calculating (random) how much this particle would travel before this kind of interaction happens
                       double InteractionCoefficient = InteractionValue(energy, &(*MaterialCollection)[MatId]->MatParticle[Id].Terminators[i].PartialCrossSectionEnergy,
                                                                                &(*MaterialCollection)[MatId]->MatParticle[Id].Terminators[i].PartialCrossSection,
                                                                                MaterialCollection->fLogLogInterpolation);
//                       qDebug()<<energy<<InteractionCoefficient;

                       double MeanFreePath;                     
                       if (ParticleCollection->at(Id)->type == _neutron_)
                         { //neutron! - using atomic density and cross-section in barns
//                           qDebug()<<InteractionCoefficient<<AtomicDensity;
                           MeanFreePath = 10.0/InteractionCoefficient/AtomicDensity;  //1/(cm2)/(1/cm3) - need in mm (so that 10.)
                         }
                       else MeanFreePath = 10.0/InteractionCoefficient/Density;  //1/(cm2/g)/(g/cm3) - need in mm (so that 10.)
//                       qDebug()<<"MeanFreePath="<<MeanFreePath;                      
                       double Step = -MeanFreePath * log(RandGen->Rndm());
//                       qDebug()<< "Generated length:"<<Step;
                       if (Step < SoFarShortest)
                       {
                          SoFarShortest = Step;
                          SoFarShortestId = i;
                       }
                   }
//                   qDebug()<<SoFarShortest<<SoFarShortestId;
                   //shortest vs MaxLength?
                   if (SoFarShortest >= MaxLength)
                   {
                     //nothing was triggered in this medium
//                       qDebug()<<"Passed the volume - no interaction";
                       distanceHistory = MaxLength;
                       energyHistory = 0;
                       flagContinue = 0;
                   }
                   else
                   {
                     //termination scenario Id was triggered at the distance SoFarShortest
                       //doing the proper step
                       ///GeoManager->SetStep(SoFarShortest);
                       navigator->SetStep(SoFarShortest);
                       ///GeoManager->Step(kTRUE, kFALSE);
                       navigator->Step(kTRUE, kFALSE);
                       ///for (int j=0; j<3; j++) rCur[j] = GeoManager->GetCurrentPoint()[j];  //new current point
                       for (int j=0; j<3; j++) rCur[j] = navigator->GetCurrentPoint()[j];  //new current point

                       //triggering the process
                       switch ( (*MaterialCollection)[MatId]->MatParticle[Id].Terminators[SoFarShortestId].ReactionType)
                       {
                       case (0):
                         {
                           //Photoelectric
//                           qDebug()<<"Photoelectric";
                             EnergyDepositionCell* tc = new EnergyDepositionCell();
                             tc->eventId = eventId;
                             tc->dE = energy;
                             tc->cellLength = 0;
                             for (int j=0; j<3; j++) tc->r[j] = rCur[j];
                             tc->time = time;
                             tc->index = counter;
                             tc->ParticleId = Id;
                             tc->MaterialId = MatId;
                             EnergyVector->append(tc);

                             flagContinue = 3;
                             distanceHistory = SoFarShortest;
                             energyHistory = tc->dE;
                            break; //switch-break
                         }
                       case (1):
                         {
                           //Compton
//                           qDebug()<<"Compton";
                             GammaStructure G0;
                             G0.energy = energy;
                             TVector3 dir (v);
                             G0.direction = dir;
                             //qDebug()<<dir[0]<<dir[1]<<dir[2];

                             GammaStructure G1 = Compton(&G0, RandGen); // new gamma
                             //qDebug()<<"energy"<<G1.energy<<" "<<G1.direction[0]<<G1.direction[1]<<G1.direction[2];

                             EnergyDepositionCell* tc = new EnergyDepositionCell();
                             tc->eventId = eventId;
                             tc->cellLength = 0;
                             tc->dE = G0.energy - G1.energy;
                             for (int j=0; j<3; j++) tc->r[j] = rCur[j];
                             tc->time = time;
                             tc->index = counter;
                             tc->ParticleId = Id;
                             tc->MaterialId = MatId;
                             EnergyVector->append(tc);

                             //creating particles and putting them on stack
                             //creating gamma                           

                        //     int IdGamma = (*MaterialCollection)[MatId]->MatParticle[Id].Terminators[SoFarShortestId].GeneratedParticles[0];
                        //     qDebug()<<"------"<<MaterialCollection->at(MatId)->MatParticle[Id].Terminators[SoFarShortestId].GeneratedParticles.size()<<MaterialCollection->at(MatId)->MatParticle[Id].Terminators[SoFarShortestId].GeneratedParticles[0];
                        //     ParticleOnStack *tmp = new ParticleOnStack(IdGamma, rCur[0],rCur[1],rCur[2], G1.direction[0], G1.direction[1], G1.direction[2], time, G1.energy, 1);
                             ParticleOnStack *tmp = new ParticleOnStack(Id, rCur[0],rCur[1],rCur[2], G1.direction[0], G1.direction[1], G1.direction[2], time, G1.energy, 1);
                             ParticleStack->append(tmp);

                             //creating electron
                             //int IdElectron =
                             //double ElectronEnergy = G0.energy - G1.energy;
                             //TVector3 eDirection = G0.energy*G0.direction - G1.energy*G1.direction;
                             //eDirection = eDirection.Unit();
                             //tmp = new ParticleOnStack(IdElectron, rCur[0],rCur[1],rCur[2], eDirection[0], eDirection[1], eDirection[2], time, ElectronEnergy, 1);
                             //ParticleStack->append(tmp);

                             flagContinue = 4;
                             distanceHistory = SoFarShortest;
                             energyHistory = tc->dE;
                            break;//switch-break
                         }
                       case (2):
                         {
                           //capture
//                           qDebug()<<"capture";
                             //nothing is added to the EnergyVector for this reaction!
                             int numGenParticles = (*MaterialCollection)[MatId]->MatParticle[Id].Terminators[SoFarShortestId].GeneratedParticles.size();

                             //adding first particle to the stack
                             //random direction
                             double vv[3];                             
                             //Sphere function of Root:
                             double a=0,b=0,r2=1;
                             while (r2 > 0.25) {
                                   a  = RandGen->Rndm() - 0.5;
                                   b  = RandGen->Rndm() - 0.5;
                                   r2 =  a*a + b*b;
                                }
                             vv[2] = ( -1. + 8.0 * r2 );
                             double scale = 8.0 * TMath::Sqrt(0.25 - r2);
                             vv[0] = a*scale;
                             vv[1] = b*scale;

                             int Particle1 = (*MaterialCollection)[MatId]->MatParticle[Id].Terminators[SoFarShortestId].GeneratedParticles[0];
                             double energy1 = (*MaterialCollection)[MatId]->MatParticle[Id].Terminators[SoFarShortestId].GeneratedParticleEnergies[0];
  //                         qDebug()<<"Adding to stack particle with id: "<<Particle1<<" energy:"<<energy;
                             ParticleOnStack* pp = new ParticleOnStack(Particle1, rCur[0], rCur[1], rCur[2], vv[0], vv[1], vv[2], time, energy1, 1);
                             ParticleStack->append(pp);

                             // adding second particle    - TO ADD: creation rules
                             //  for the moment, the creation rule is: opposite direction to the first particle    <- TO CHANGE
                             if (numGenParticles > 1)
                               {
                                 int Particle2 = (*MaterialCollection)[MatId]->MatParticle[Id].Terminators[SoFarShortestId].GeneratedParticles[1];
                                 double energy2 = (*MaterialCollection)[MatId]->MatParticle[Id].Terminators[SoFarShortestId].GeneratedParticleEnergies[1];
//                                 qDebug()<<"Adding to stack particle with id: "<<Particle2<<" energy:"<<energy;
                                 pp = new ParticleOnStack(Particle2, rCur[0], rCur[1], rCur[2], -vv[0], -vv[1], -vv[2], time, energy2, 1);
                                 ParticleStack->append(pp);
                               }
                             if (numGenParticles > 2)
                                 for (int ipart=2; ipart<numGenParticles; ipart++)
                                  {
//                                   //all particle after 2nd - random direction                                     
                                     //Sphere function of Root:
                                     double a=0,b=0,r2=1;
                                     while (r2 > 0.25) {
                                           a  = RandGen->Rndm() - 0.5;
                                           b  = RandGen->Rndm() - 0.5;
                                           r2 =  a*a + b*b;
                                        }
                                     vv[2] = ( -1. + 8.0 * r2 );
                                     double scale = 8.0 * TMath::Sqrt(0.25 - r2);
                                     vv[0] = a*scale;
                                     vv[1] = b*scale;

                                     Particle1 = (*MaterialCollection)[MatId]->MatParticle[Id].Terminators[SoFarShortestId].GeneratedParticles[ipart];
                                     energy1 = (*MaterialCollection)[MatId]->MatParticle[Id].Terminators[SoFarShortestId].GeneratedParticleEnergies[ipart];
      //                             qDebug()<<"Adding to stack particle with id: "<<Particle1<<" energy:"<<energy;
                                    ParticleOnStack* pp = new ParticleOnStack(Particle1, rCur[0], rCur[1], rCur[2], vv[0], vv[1], vv[2], time, energy1, 1);
                                    ParticleStack->append(pp);
                                  }

                             flagContinue = 5;
                             distanceHistory = SoFarShortest;
                             energyHistory = 0;
                            break; //switch-break
                         }
                       }
                   }

               }
               else
               {
                  //no processes - error!
                   qDebug()<<"Error - no processes are defined for this particle in this material!" ;
                   return false;
               }

           }///////////END: neutral particle///////////
           else
           {///////////charged particle/////////////   - continuous deposition

               //historyDistance - travelled inside this material; energy - current energy
               bool flagDone = false;
               do
               {
                   //dE/dx [keV/mm] = Density[g/cm3] * [cm2/g*keV] * 0.1  //0.1 since local units are mm, not cm
                   double dEdX = 0.1*Density * InteractionValue(energy, &(*MaterialCollection)[MatId]->MatParticle[Id].InteractionDataX,
                                                                        &(*MaterialCollection)[MatId]->MatParticle[Id].InteractionDataF,
                                                                        MaterialCollection->fLogLogInterpolation);

                   //recommended step: (RecFraction - suggested decrease of energy per step)
                   double RecStep = SimSet->dE * energy/dEdX; //*** TO DO zero control
//                   qDebug()<<dEdX<<RecStep;
                   //vs absolute limits?
                   if (RecStep > SimSet->MaxStep) RecStep = SimSet->MaxStep;
                   if (RecStep < SimSet->MinStep) RecStep = SimSet->MinStep;

                   //will pass to another volume?
                   if (distanceHistory + RecStep > MaxLength - SimSet->Safety)
                     {
//                       qDebug()<<"boundary reached!";
                       RecStep = MaxLength - distanceHistory - SimSet->Safety;
                       flagDone = true;
                       flagContinue = 0; //should enter the next material after "do" is over
                     }
                   //doing the step
                   ///GeoManager->SetStep(RecStep);
                   navigator->SetStep(RecStep);
                   ///GeoManager->Step(kTRUE, kFALSE);
                   navigator->Step(kTRUE, kFALSE);
 //                  qDebug() << "Asked for:"<<RecStep<<"Did:"<<navigator->GetStep();  //actual step is less by a fixed value (safity?)
                   RecStep = navigator->GetStep();
                   //energy loss?
                   double dE = dEdX*RecStep;
//                   qDebug()<<"     "<<RecStep<<dE;
                   if (dE > energy) dE = energy;
                   energy -= dE;

                   //all energy dissipated?
                   if (energy < SimSet->MinEnergy)
                     {
                       flagDone = true;
                       flagContinue = 2;
                       dE += energy;
                     }

                   EnergyDepositionCell* tc = new EnergyDepositionCell();
                   tc->eventId = eventId;
                   tc->dE = dE;
                   tc->cellLength = RecStep;
//                   qDebug()<<"Step"<<RecStep<<"dE"<<dE<<"New energy"<<energy;
                   for (int j=0; j<3; j++)
                     {
                       ///rCur[j] = GeoManager->GetCurrentPoint()[j];  //new current point
                       rCur[j] = navigator->GetCurrentPoint()[j];  //new current point
                       tc->r[j] = rCur[j];
                     }
                   tc->time = time;
                   tc->index = counter;
                   tc->ParticleId = Id;
                   tc->MaterialId = MatId;
                   EnergyVector->append(tc);

                   distanceHistory += RecStep;
                   energyHistory += dE;
               }
               while (!flagDone);

           }///////////END: charged particle///////////

          } //end block - interaction data defined
           else
           {  //interaction data were not defined
              distanceHistory = MaxLength;
           }


            if (SimSet->fLogsStat) History->Deposition.append(MaterialHistoryStructure(MatId,energyHistory,distanceHistory)); //if continue to the next volume

            //was particle killed in this material?
            if (flagContinue != 0) break;  //do-break

            //stepping to check the next medium
            ///GeoManager->FindNextBoundaryAndStep();
            navigator->FindNextBoundaryAndStep();
            ///MatId = GeoManager->GetCurrentVolume()->GetMaterial()->GetIndex();
            MatId = navigator->GetCurrentVolume()->GetMaterial()->GetIndex();
//            qDebug()<<"Next material id: "<<MatId;

       //-------------------------------------------------------------------
       } while (flagContinue == 0); //repeat if tracking is OK


       //particle escaped, stopped or captured     do-breaks are collected here!
       //if (BuildTracks) GeoManager->GetCurrentTrack()->AddPoint(rCur[0], rCur[1], rCur[2], time);
       if (BuildTracks)
         {
           //track->AddPoint(rCur[0], rCur[1], rCur[2], time);
           track->Nodes.append(TrackNodeStruct(rCur, time));
           TrackCandidates.append(track);
         }

       //finalizing diagnostics
       if (SimSet->fLogsStat)
         {
           History->Termination = flagContinue;
           EventHistory->append(History);
         }

       //delete the record and remove from the stack
       delete ParticleStack->at(0);
       ParticleStack->remove(0);
   }//-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/ end while cycle over the elements of the stack

   //stack is empty, no errors found

   if (TrackCandidates.size()>0)
     {
       if (RemoveTracksIfNoEnergyDepo && EnergyVector->isEmpty() )
         {
           //clear all particle tracks - there were no energy deposition - nothing will be added to Geomanager later
           for (int i=0; i<TrackCandidates.size(); i++)
             delete TrackCandidates[i];
         }
       else
         {
           for (int i=0; i<TrackCandidates.size(); i++)
             Tracks->append(TrackCandidates.at(i)); // delete later when create Geomanager tracks
         }
       TrackCandidates.clear();
     }

   return true; //success
}
