#ifndef PROCESSORCLASS_H
#define PROCESSORCLASS_H

#include <QObject>

class ReconstructionManagerClass;
class ReconstructionSettings;
class DetectorClass;
class DynamicPassivesHandler;
class EventsDataClass;
struct ReconstructionStruct;
namespace ROOT { namespace Minuit2 { class Minuit2Minimizer; } }
namespace ROOT { namespace Math { class Functor; } }

class ProcessorClass : public QObject
{
  Q_OBJECT
public:
  ProcessorClass(ReconstructionManagerClass *recManager, int eventsFrom, int eventsTo);
  virtual ~ProcessorClass();

    // thread control
  char Progress; //char - hope its completely safe to retrieve from thread
  int eventsProcessed;
  bool fFinished;
  int Id;
  bool fStopRequested;

    // local object - for static use should be public
  DynamicPassivesHandler *DynamicPassives;

    // pointers: for static use with minimizer, should be public:
  ReconstructionSettings *RecSet;
  DetectorClass *Detector;
  EventsDataClass *EventsDataHub;

public slots:
  virtual void execute() = 0;
signals:
  void finished();

protected:
  ReconstructionManagerClass *RecManager;
  int EventsFrom, EventsTo;
  double calculateChi2NoDegFree(int iev, ReconstructionStruct *rec);
  double calculateMLfactor(int iev, ReconstructionStruct *rec);
};

/// Center of Gravity
class CoGReconstructorClass : public ProcessorClass
{
  Q_OBJECT
public:
  CoGReconstructorClass(ReconstructionManagerClass *recManager, int eventsFrom, int eventsTo)
    : ProcessorClass(recManager, eventsFrom, eventsTo) {}
  ~CoGReconstructorClass(){}
public slots:
  virtual void execute();
};

/// Contracting grids on CPU
class CGonCPUreconstructorClass : public ProcessorClass
{
  Q_OBJECT
public:
  CGonCPUreconstructorClass(ReconstructionManagerClass *recManager, int eventsFrom, int eventsTo)
    : ProcessorClass(recManager, eventsFrom, eventsTo) {}
  ~CGonCPUreconstructorClass(){}
public slots:
  virtual void execute();
};

/// Root minimizer (Migrad2 or Simplex) for point events
class RootMinReconstructorClass : public ProcessorClass
{
  Q_OBJECT
public:
  RootMinReconstructorClass(ReconstructionManagerClass *recManager, int eventsFrom, int eventsTo);
  ~RootMinReconstructorClass();
 //public for static use
 double LastMiniValue;
 const QVector< float >* PMsignals;
public slots:
  virtual void execute();
private:
  ROOT::Math::Functor *FunctorLSML;
  ROOT::Minuit2::Minuit2Minimizer* RootMinimizer;
};

/// Root minimizer (Migrad2 or Simplex) with double events
/// RecSet->MultipleEventOption: 0-cannot be here, 1-only doubles+chi2, 2-does doubles+chi2 then copmares with already provided rec results for singles
class RootMinDoubleReconstructorClass : public ProcessorClass
{
  Q_OBJECT
public:
  RootMinDoubleReconstructorClass(ReconstructionManagerClass *recManager, int eventsFrom, int eventsTo);
  ~RootMinDoubleReconstructorClass();
 //public for static use
 double LastMiniValue;
 const QVector< float >* PMsignals;
public slots:
  virtual void execute();
private:
  ROOT::Math::Functor *FunctorLSML;
  ROOT::Minuit2::Minuit2Minimizer* RootMinimizer;
  double calculateChi2DoubleEvent(const double *result);
};

/// Claculates chi2 for all events
class Chi2calculatorClass : public ProcessorClass
{
  Q_OBJECT
public:
  Chi2calculatorClass(ReconstructionManagerClass *recManager, int eventsFrom, int eventsTo)
    : ProcessorClass(recManager, eventsFrom, eventsTo){}
  ~Chi2calculatorClass(){}
public slots:
  virtual void execute();
};

/// Checks all filters which can be multithreaded
class EventFilterClass : public ProcessorClass
{
  Q_OBJECT
public:
    EventFilterClass(ReconstructionManagerClass *recManager, int eventsFrom, int eventsTo)
        : ProcessorClass(recManager, eventsFrom, eventsTo) {}
    ~EventFilterClass(){}
public slots:
  virtual void execute();
};

//static functions to use with Root minimizer
double Chi2static(const double *p);
double Chi2staticDouble(const double *p);
double MLstatic(const double *p);
double MLstaticDouble(const double *p);
#endif // PROCESSORCLASS_H
