#include "reconstructionmanagerclass.h"
#include "eventsdataclass.h"
#include "detectorclass.h"
#include "CorrelationFilters.h"
#include "reconstructionsettings.h"
#include "sensorlrfs.h"
#include "functions.h"
#include "processorclass.h"

#ifdef ANTS_FLANN
#include "nnmoduleclass.h"
#endif
#ifdef __USE_ANTS_CUDA__
#include "cudamanagerclass.h"
#endif
#ifdef ANTS_FANN
#include "neuralnetworksmodule.h"
#endif

#include <QThread>
#include <QtGui>
#include <QTime>
#include <QJsonObject>
#include <QJsonArray>

#include "TGeoManager.h"

ReconstructionManagerClass::ReconstructionManagerClass(EventsDataClass *eventsDataHub, DetectorClass *detector)
{
  EventsDataHub = eventsDataHub;
  Detector = detector;

#ifdef ANTS_FANN
  ANNmodule = new NeuralNetworksModule(Detector->PMs, EventsDataHub);
  //qDebug() << "->NeuralNetworks module created";
#endif

#ifdef ANTS_FLANN
  KNNmodule = new NNmoduleClass(EventsDataHub, Detector); //fast nearest neighbour module
  //qDebug() << "->Nearest neighbour module created";
#endif
}

ReconstructionManagerClass::~ReconstructionManagerClass()
{
#ifdef ANTS_FLANN
  delete KNNmodule;
  //qDebug() << "  Nearest neighbour module deleted";
#endif

#ifdef ANTS_FANN
  delete ANNmodule;
  //qDebug() << "  NeuralNetworks module deleted";
#endif
}

bool ReconstructionManagerClass::reconstructAll(QJsonObject &json)
{
  //qDebug() << "==> Reconstruct all events triggered";
  bool fOK = ReconstructionManagerClass::fillSettingsAndVerify(json);
  if (!fOK) return false; //something wrong with configuration
  //EventsDataHub->createDefaultReconstructionData(); // also does clear
  EventsDataHub->resetReconstructionData(); //does some clear too, keeps EventId

  QTime timer;
  timer.start();

  //pre-phase - doing CoG for all events
  //qDebug() << "--> Doing pre-phase CoG";
  QList<ProcessorClass*> todo;
  distributeWork(0, todo);
  fOK = run(todo); //on finish, todo is empty

  //main algorithm
  //qDebug() << "--> Performing reconstruction with the algorithm:"<<RecSet.ReconstructionAlgorithm;
  bool fMultiThreaded = RecSet.NumThreads>0 && ( RecSet.ReconstructionAlgorithm == 1 || RecSet.ReconstructionAlgorithm == 2);
  if (fMultiThreaded)
    {      
      todo.clear();
      distributeWork(RecSet.ReconstructionAlgorithm, todo);
      fOK = run(todo);
    }
  else
    {
      switch (RecSet.ReconstructionAlgorithm)
        {
        case 0:
          { //just copy cog data to positions
            for (int iev=0; iev<EventsDataHub->Events.size(); iev++)
              {
                EventsDataHub->ReconstructionData[iev]->Points[0].r[0] = EventsDataHub->ReconstructionData[iev]->xCoG;
                EventsDataHub->ReconstructionData[iev]->Points[0].r[1] = EventsDataHub->ReconstructionData[iev]->yCoG;
                EventsDataHub->ReconstructionData[iev]->Points[0].r[2] = EventsDataHub->ReconstructionData[iev]->zCoG;
                EventsDataHub->ReconstructionData[iev]->Points[0].energy = 1.0; //recently - to protect LRF recon with "energy on" option
                EventsDataHub->ReconstructionData[iev]->chi2 = 0;
              }            
            break;
          }
        case 3:
#ifdef ANTS_FANN
         ANNmodule->Configure(&RecSet);
         for (int iev=0; iev<EventsDataHub->Events.size(); iev++)
           if (EventsDataHub->ReconstructionData[iev]->ReconstructionOK) //CoG OK
             ANNmodule->ReconstructEvent(iev, -1, &EventsDataHub->ReconstructionData[iev]->Points[0].r, &EventsDataHub->ReconstructionData[iev]->Points[0].energy);
#else
         ErrorString = "FANN was not configured in .pro";
         return false;
#endif
          break;
        case 4:
          {
#ifdef __USE_ANTS_CUDA__

            CudaManagerClass cm(Detector, EventsDataHub, RecSet.CGonCUDAsettings);
            QString Method = (*Detector->SensLRF)[0]->type();//  "Axial" "XY" "Sliced3D" "ComprAxial" "Composite"
            bool ok = cm.RunCuda(Method);
            if (!ok)
              {
                ErrorString = cm.getLastError();//"CUDA module reports an error during reconstruction";
                return false;
              }
            qDebug() << "Rate (us/event) reported by cuda manager module:" << cm.getUsPerEvent();
#else
            ErrorString = "CUDA was not configured in .pro";
            return false;
#endif
          }
          break;
        case 5:
#ifdef ANTS_FLANN
          {
           KNNmodule->Reconstructor.readFromJson(RecSet.kNNrecSet);
           bool ok = KNNmodule->Reconstructor.reconstructEvents();
           if (!ok)
             {
               ErrorString = "kNN module reports error in reconstruction";
               return false;
             }
           for (int iev=0; iev<EventsDataHub->Events.size(); iev++)
             {
               EventsDataHub->ReconstructionData[iev]->Points[0].r[2] = EventsDataHub->ReconstructionData[iev]->zCoG;
               EventsDataHub->ReconstructionData[iev]->Points[0].energy = 1.0; //recently - to protect LRF recon with "energy on" option
               EventsDataHub->ReconstructionData[iev]->chi2 = 0;
             }
          }
#else
            ErrorString = "kNN was not configured in .pro";
            return false;
#endif
          break;
        case 1:
          //not supported - easy to add but have to insert gui update
        case 2:
          //not supported - easy to add but have to insert gui update: RootMinReconstructorClass rmr(this, 0, numEvents); rmr.execute();
        default:
          ErrorString = "Unsupported algorithm or threading option";
          return false;
        }
    }
  if (fStopRequested)
    {
      //EventsDataHub->createDefaultReconstructionData(); // to do clear properly
      EventsDataHub->resetReconstructionData(); // to do clear properly
      return false;
    }
  int msElapsed = timer.elapsed();
  usPerEvent = (numEvents>0) ? 1000.0*msElapsed/(double)numEvents : 0;

  //chi2 calculation
  bool fDoChi2Calc = Detector->SensLRF->isAllLRFsDefined()
      && RecSet.ReconstructionAlgorithm != 0
      && RecSet.ReconstructionAlgorithm != 4
      && RecSet.ReconstructionAlgorithm != 5 //might make sense to activate!
      && !(RecSet.ReconstructionAlgorithm == 2 && RecSet.MultipleEventOption == 1);
  if (fDoChi2Calc)
    {
      //qDebug() << "--> Calculating Chi2";
      todo.clear();
      distributeWork(10, todo);
      fOK = run(todo);
    }
  //else qDebug() << "--> skipping Chi2 reconstruction";

  //if multiple option = 2, have to do double rec and compare sigma with single
  if (RecSet.MultipleEventOption == 2)
    {
      //qDebug() << "--> Running double event reconstruction";
      todo.clear();
      distributeWork(12, todo);
      fOK = run(todo);
    }

  EventsDataHub->RecSettings = RecSet;
  EventsDataHub->fReconstructionDataReady = true;
  //event filtering
  //qDebug() << "--> Running event filters";
  todo.clear();
  distributeWork(11, todo);
  fOK = run(todo);
  //additional filters which has to be checked in single thread  
  singleThreadEventFilters();
  return true;
}

void ReconstructionManagerClass::distributeWork(int Algorithm, QList<ProcessorClass*> &todo)
// Algorithm options:
//0 - CoG reconstruction, 1 - MG, 2 - RootMini
//10 - Calculate Chi2, 11 - process event filters
{ 
  numEvents = EventsDataHub->Events.size();
  int perOne = numEvents / RecSet.NumThreads;
  if (perOne*RecSet.NumThreads < numEvents) perOne++;
  int from = 0;
  int to;
  int Id = 0;
  do
    {
      to = from + perOne;
      if (to>numEvents) to = numEvents;
      //qDebug()<<"Adding thread with Id"<<Id<<"to process events from"<<from<<" to "<<to-1;
      switch (Algorithm)
        {
        case 0:
          todo << new CoGReconstructorClass(this, from, to);
          break;
        case 1:
          todo << new CGonCPUreconstructorClass(this, from, to);
          break;
        case 2:
          if (RecSet.MultipleEventOption == 1) //in the case of 2(chose best single or double), first single is calculated
               todo << new RootMinDoubleReconstructorClass(this, from, to);
          else todo << new RootMinReconstructorClass(this, from, to);
          break;
        case 10:
          todo << new Chi2calculatorClass(this, from, to);
          break;
        case 11:
          todo << new EventFilterClass(this, from, to);
          break;
        case 12:
          todo << new RootMinDoubleReconstructorClass(this, from, to);
          break;
        }

      todo.last()->Id = Id;
      Id++;
      from += perOne;
    }
  while (from<numEvents);
}

bool ReconstructionManagerClass::fillSettingsAndVerify(QJsonObject &json)
{
  if (EventsDataHub->isEmpty())
    {
      ErrorString = "No data to reconstruct";
      return false;
    }
  if (!json.contains("ReconstructionConfig"))
    {
      ErrorString = "Reconstruction settings not found in Json file";
      return false;
    }
  QJsonObject js = json["ReconstructionConfig"].toObject();

  bool fOK = RecSet.readFromJson(js);
  if (!fOK)
    {
       ErrorString = RecSet.ErrorString;
       return false;
    }
  fOK = ReconstructionManagerClass::configureFilters(js);
  if (!fOK) return false;


  int Alg = RecSet.ReconstructionAlgorithm; // 0 - CoG, 1 -CGonCPU, 2-RootMin, 3 - ANN, 4-CUDA
  if (Alg!=0 && Alg!=3 && Alg!=5) // CoG, ANN and KNN do not need LRFs
    if (!Detector->SensLRF->isAllLRFsDefined())
     {
       ErrorString = "LRFs are not defined";
       return false;
     }
  if (RecSet.MultipleEventOption > 0 && (Alg!=2))
    {
        ErrorString = "Multiple events can be reconstructed onnly with Root-based minimizer method";
        return false;
    }
  //if start option is scan, scanData should exist
  if (Alg == 1 && RecSet.CGstartOption == 2 && EventsDataHub->isScanEmpty())
    {
      ErrorString = "Start option is set to 'True XY' but scan data are empty";
      return false;
    }
  if (Alg == 2 && RecSet.RMstartOption == 2 && EventsDataHub->isScanEmpty())
    {
      ErrorString = "Start option is set to 'True XY' but scan data are empty";
      return false;
    }

  if (RecSet.Zstrategy == 1)
    {
      //Scan should contain data for all events!
      if (EventsDataHub->Scan.size() != EventsDataHub->Events.size())
        {
          ErrorString = "Z strategy is set to use loaded/simulated data, but there are no such data!";
          return false;
        }
    }

#ifdef ANTS_FANN
  if (Alg == 3)
    {
      if (!ANNmodule->isFANN() || !ANNmodule->checkPMsDimension())
        {
          ErrorString = "No FANN library is loaded or wrong number of PM inputs";
          return false;
        }
    }
#endif

  if (RecSet.fRMsuppressConsole) gErrorIgnoreLevel = kWarning;
  ErrorString = "";
  return true;
}

bool ReconstructionManagerClass::configureFilters(QJsonObject &json)
{
  if ( !json.contains("FilterOptions") )
    {
      qWarning() << "||| Json config file does not contain filter settings!";
      return false;
    }

  QJsonObject js = json["FilterOptions"].toObject();

  fEventNumberFilter = js.contains("EventNumber");
  if (fEventNumberFilter)
    {
      QJsonObject njson = js["EventNumber"].toObject();
      parseJson(njson, "Min", EventNumberFilterMin);
      parseJson(njson, "Max", EventNumberFilterMax);
    }

  fSumCutOffFilter = js.contains("SumSignal");
  if (fSumCutOffFilter)
    {
      QJsonObject ssjson = js["SumSignal"].toObject();
      parseJson(ssjson, "UseGains", fSumCutUsesGains);
      parseJson(ssjson, "UsePassives", fCutOffsForPassivePMs);
      parseJson(ssjson, "Min", SumCutOffMin);
      parseJson(ssjson, "Max", SumCutOffMax);
    }

  fCutOffFilter = js.contains("IndividualPMSignal");
  if (fCutOffFilter)
    {
      QJsonObject sjson = js["IndividualPMSignal"].toObject();
      parseJson(sjson, "UsePassives", fCutOffsForPassivePMs);
    }

  fEnergyFilter = js.contains("ReconstructedEnergy");
  if (fEnergyFilter)
    {
      QJsonObject rejson = js["ReconstructedEnergy"].toObject();
      parseJson(rejson, "Min", EnergyFilterMin);
      parseJson(rejson, "Max", EnergyFilterMax);
    }

  LoadedEnergyFilter = js.contains("LoadedEnergy");
  if (LoadedEnergyFilter)
    {
      QJsonObject lejson = js["LoadedEnergy"].toObject();
      parseJson(lejson, "Min", LoadedEnergyFilterMin);
      parseJson(lejson, "Max", LoadedEnergyFilterMax);
    }

  fChi2Filter = js.contains("Chi2");
  if (fChi2Filter)
    {
      QJsonObject cjson = js["Chi2"].toObject();
      parseJson(cjson, "Min", Chi2FilterMin);
      parseJson(cjson, "Max", Chi2FilterMax);
    }

  fSpF_PrScint = js.contains("SpatialToPrimary");

  fSpF_custom = js.contains("SpatialCustom");
  if (fSpF_custom)
    {
      QJsonObject spfCjson = js["SpatialCustom"].toObject();
      SpF_RecOrScan = 0; //compatibility
      parseJson(spfCjson, "RecOrScan", SpF_RecOrScan);
      parseJson(spfCjson, "Shape", SpF_shape);
      parseJson(spfCjson, "SizeX", SpF_SizeX);
      parseJson(spfCjson, "SizeY", SpF_SizeY);
      parseJson(spfCjson, "Diameter", SpF_Diameter);
      parseJson(spfCjson, "Side", SpF_Side);
      parseJson(spfCjson, "Angle", SpF_Angle);
      parseJson(spfCjson, "AllZ", fSpF_allZ);
      parseJson(spfCjson, "Zfrom", SpF_Zfrom);
      parseJson(spfCjson, "Zto", SpF_Zto);
      SpF_X0 = 0; //compatibility
      parseJson(spfCjson, "X0", SpF_X0);
      SpF_Y0 = 0; //compatibility
      parseJson(spfCjson, "Y0", SpF_Y0);
      SpF_CutOutsideInside = 0; //compatibility
      parseJson(spfCjson, "CutOutsideInside", SpF_CutOutsideInside);

      SpF_polygon.clear();
      if (spfCjson.contains("Polygon"))
        {
          QJsonArray polyarr = spfCjson["Polygon"].toArray();
          for (int i=0; i<polyarr.size(); i++)
            {
              double x = polyarr[i].toArray()[0].toDouble();
              double y = polyarr[i].toArray()[1].toDouble();
              SpF_polygon.append(QPointF(x,y));
            }
        }
      //need to calculate..
      switch (SpF_shape)
        {
        case 0:
          SpF_halfSizeX = 0.5*SpF_SizeX;
          SpF_halfSizeY = 0.5*SpF_SizeY;
          break;
        case 1:
          SpF_radius2 = 0.5*SpF_Diameter;
          SpF_radius2 *= SpF_radius2;
          //qDebug() << "Rad2"<<SpF_radius2;
          break;
        case 2:
          for (int i=0; i<6; i++)
            {
              double angle = SpF_Angle + 60.0*i;
              angle = angle/180.0*3.1415926536;
              SpF_polygon << QPointF(SpF_Side*sin(angle), SpF_Side*cos(angle));
            }
          SpF_polygon << QPointF(SpF_Side*sin(SpF_Angle/180.0*3.1415926536), SpF_Side*cos(SpF_Angle/180.0*3.1415926536)); //to close it
          break;
        case 3:
          if (SpF_polygon.size()<4)
            {
              ErrorString = "Custom spatial filter: polygon not provided or not valid";
              return false;
            }
          break;
        default:
          ErrorString = "Unknown shape of custom spatial filter";
          return false;
        }
    }

  fCorrelationFilters = js.contains("Correlation");
  if (fCorrelationFilters)
    {
      QJsonObject corjson = js["Correlation"].toObject();
      for (int i=0; i<CorrelationFilters.size(); i++) delete CorrelationFilters[i];
      CorrelationFilters.clear();

      QJsonArray jsonArray = corjson["Set"].toArray();
      if (jsonArray.isEmpty())
        {
          ErrorString = "Correlation filters active, but json does not contain their config";
          return false;
        }
      for (int i=0; i<jsonArray.size(); i++)
        {
          QJsonObject js = jsonArray[i].toObject();
          CorrelationFilterStructure* tmp = CorrelationFilterStructure::createFromJson(js, Detector, EventsDataHub);
          if (!tmp)
            {
              ErrorString = "Error in json config of correlation filter";
              return false;
            }
          CorrelationFilters.append(tmp);
        }
    }

   fKNNfilter = js.contains("kNN");
   if (fKNNfilter)
     {
       QJsonObject knnjson = js["kNN"].toObject();
       parseJson(knnjson, "Min", KNNfilterMin);
       parseJson(knnjson, "Max", KNNfilterMax);
       parseJson(knnjson, "AverageOver", KNNfilterAverageOver);
     }
   return true;
}

bool ReconstructionManagerClass::run(QList<ProcessorClass *> reconstructorList)
{
  fStopRequested = false;
  QList<QThread*> threads;  
  for (int ithread = 0; ithread<reconstructorList.size(); ithread++)
    {
      threads.append(new QThread());
      QObject::connect(threads.last(), SIGNAL(started()), reconstructorList[ithread], SLOT(execute()));
      QObject::connect(reconstructorList[ithread], SIGNAL(finished()), threads.last(), SLOT(quit()));
      QObject::connect(threads.last(), SIGNAL(finished()), threads.last(), SLOT(deleteLater()));
      reconstructorList[ithread]->moveToThread(threads.last());
      threads.last()->start();
    }

  int StartingTreads = threads.size();
  QTime RefresherTimer;
  RefresherTimer.start();
  int TotalMsPassed = 0;
  while (threads.size()>0)
    {
      for (int i=threads.size()-1; i>-1; i--)
        {
          //finished thread?
          if (reconstructorList[i]->fFinished)
            {
              //qDebug() << "Thread"<< reconstructorList[i]->Id << " reports finished status";
              delete reconstructorList[i];
              reconstructorList.removeAt(i);
              threads.removeAt(i);
            }
        }      
      int msPassed = RefresherTimer.elapsed();
      if (msPassed > 500)
        {
          qApp->processEvents();
          //reporting progress
          float ProgressF = 100.0*(StartingTreads - threads.size()); //of finished threads
          int EventsDone = 0;
          for (int i=0; i<threads.size(); i++)
            {
              int tp = reconstructorList[i]->Progress;
              //qDebug() << "Thread ["<<reconstructorList[i]->Id<< "] progress:"<<tp<<StartingTreads<<threads.size();
              ProgressF += tp;
              EventsDone += reconstructorList[i]->eventsProcessed;
            }
          int TotalProgress = ProgressF/StartingTreads;
          //qDebug() << "Total progress:"<<TotalProgress;

          TotalMsPassed += msPassed;
          //double MsPerEv =  100.0 * 1000 * TotalMsPassed / ( numEvents*(TotalProgress+0.00001));
          double MsPerEv = (EventsDone>0) ? 1000.0 * TotalMsPassed / (double)EventsDone : 0;
          emit UpdateReady(TotalProgress, MsPerEv);

          //"stop" user request handling
          if (fStopRequested)
            {
              for (int i=0; i<reconstructorList.size(); i++)
                reconstructorList[i]->fStopRequested = true;
            }

          RefresherTimer.restart();
        }
    }
  //qDebug() << "All threads finished!";
  return !fStopRequested;
}

void ReconstructionManagerClass::filterEvents(QJsonObject &jsonFilters)
{
  //qDebug() << "Updating status of all events triggered";
//  int numEvents = EventsDataHub->Events.size();
  bool fOK = ReconstructionManagerClass::configureFilters(jsonFilters);
  if (!fOK) return;
  RecSet.NumThreads = 1;
  parseJson(jsonFilters, "NumThreads", RecSet.NumThreads);

  //if ReconstructionDta is empty, create default containers
  if (EventsDataHub->ReconstructionData.isEmpty())
    {
      //qDebug() << "Creating default reconstruction data containers";
      EventsDataHub->createDefaultReconstructionData();

      //have to fill original fail status for bad CoG events
      //qDebug() << "Running CoG to fill created ReconstructionData";
      QList<ProcessorClass*> todo;
      distributeWork(0, todo);
      fOK = run(todo);
    }

  //qDebug() << "Running normal filters for all events";
  QList<ProcessorClass*> todo;
  distributeWork(11, todo);
  fOK = run(todo);

  //qDebug() << "Running monothread filters";
  singleThreadEventFilters();
}

void ReconstructionManagerClass::singleThreadEventFilters()
{
  int numEvents = EventsDataHub->Events.size();  
#ifdef ANTS_FLANN
  if (fKNNfilter)
    {
      qDebug() << "  Preparing kNN module..";
      KNNmodule->Filter.prepareNNfilter(KNNfilterAverageOver+1); //reuses old data set or calculates new
        //+1 - module assumes 0 is the point itself, therefore incrementing by 1
      qDebug() << "  Done!";
    }
#endif

  for (int iev=0; iev<numEvents; iev++)
    {
      ReconstructionStruct* rec = EventsDataHub->ReconstructionData[iev];
      if (!rec->ReconstructionOK) continue;

      //PrimScint - custom spatialfilter already checked
      if (fSpF_PrScint && rec->FilterPosition != 2)
      {
          Detector->GeoManager->SetCurrentPoint(rec->Points[0].r);
          Detector->GeoManager->FindNode();
          TString VolName = Detector->GeoManager->GetCurrentVolume()->GetName();
          //qDebug()<<VolName;
          if (VolName == "PrScint") rec->FilterPosition = 1;
          else rec->FilterPosition = 2;
        }
      //else do nothing - status is already set by custom spatial filter

      //correlation
      if (fCorrelationFilters)
        {
//              qDebug()<<"Corr filters="<<CorrelationFilters->size();
          bool pass = true;
          for (int ifilter = 0; ifilter<CorrelationFilters.size(); ifilter++)
            if (CorrelationFilters.at(ifilter)->Active)
             {
                /*
              qDebug()<<"First:"<<CorrelationFilters->at(ifilter)->getCorrelationUnitX()->getType()<<CorrelationFilters->at(ifilter)->getCorrelationUnitX()->Channels;
              qDebug()<<"Second:"<<CorrelationFilters->at(ifilter)->getCorrelationUnitX()->getType()<<CorrelationFilters->at(ifilter)->getCorrelationUnitX()->Channels;
              qDebug()<<"Cut:"<<CorrelationFilters->at(ifilter)->getCut()->getType()<<CorrelationFilters->at(ifilter)->getCut()->Data;
              qDebug()<<"Current event:"<<CurrentEvent;
              qDebug()<<"Adress of Events in X:"<<CorrelationFilters->at(ifilter)->getCorrelationUnitX()->Events;
              qDebug()<<"Actual Events address:"<<&Events;
              qDebug()<<"Events in X:"<<*(CorrelationFilters->at(ifilter)->getCorrelationUnitX()->Events);
              qDebug()<<"Val_X:"<<CorrelationFilters->at(ifilter)->getCorrelationUnitX()->getValue(CurrentEvent);
              qDebug()<<"Val_Y:"<<CorrelationFilters->at(ifilter)->getCorrelationUnitY()->getValue(CurrentEvent);
              */
              if (!CorrelationFilters.at(ifilter)->getCut()->filter( CorrelationFilters.at(ifilter)->getCorrelationUnitX()->getValue(iev),
                                                                      CorrelationFilters.at(ifilter)->getCorrelationUnitY()->getValue(iev) ))
                {
                  pass = false;
                  break;
                }
//                  qDebug()<<"   #"<<ifilter<<" :"<<pass;
             }
          if (pass) rec->FilterCorrelation = 1; else rec->FilterCorrelation = 2;
//              qDebug()<<"  overall:"<<rec->FilterCorrelation;
        }
      else rec->FilterCorrelation = 0;

      //kNN
#ifdef ANTS_FLANN
          //kNNfilter - can be activated only if ANTS_FLANN is defined
          if (fKNNfilter)
            {
              float dist = KNNmodule->Filter.getAverageDist(iev);
              if (dist<KNNfilterMin || dist>KNNfilterMax) rec->FilterkNN = 2;
              else rec->FilterkNN = 1;
            }
          else rec->FilterkNN = 0;
#endif ANTS_FLANN

      //updating goodevent
      if (rec->FilterPosition == 2 ||
#ifdef ANTS_FLANN
          rec->FilterkNN == 2 ||
#endif
          rec->FilterCorrelation == 2) rec->GoodEvent = false;
    }
}
