#include "scriptinterfaces.h"
#include "functions.h"
#include "genericscriptwindowclass.h"
#include "mainwindow.h"
#include "detectoraddonswindow.h"
#include "checkupwindowclass.h"

#include <QDateTime>
#include <QFileInfo>
#include <QFile>
#include <QtGui>

QString BaseInterfaceClass::help(QString method)
{
  if (method.endsWith("()")) method.remove("()");
  if (method.endsWith("(")) method.remove("(");
  if (!H.contains(method)) return "Help text is missing";

  return H[method];
}

CoreInterfaceClass::CoreInterfaceClass(GenericScriptWindowClass *SW) : SW(SW)
{
  Description = "Unit handles general-purpose opeartions: abort script, basic text output and file save/load";

  H["print"] = "Print the argument (string) on the script output text field";
  H["clearText"] = "Clear the script output text field";
  H["abort"] = "Abort skript execution.\nOptional string argument is a message to be shown on the script output text field";
  H["save"] = "Add string (second argument) to the file with the name given by the first argument.\n"
              "Script is aborted if file does not exist.";
  H["createFile"] = "Create a file (or clear an existent file) with the name given by the string argument.\n"
                    "If the second argument is false, skript is aborted in case the file already exists";
  H["isFileExists"] = "Return true if file exists";
  H["loadColumn"] = "Load a column with ascii numeric data from the file.\nSecond argument is the column number.";
}

void CoreInterfaceClass::abort(QString message)
{
  qDebug() << "In-script abort triggered!";
  SW->abortEvaluation(message);
}

void CoreInterfaceClass::sleep(int ms)
{
  if (ms == 0) return;
  QTime t;
  t.restart();
  do qApp->processEvents();
  while (t.elapsed()<ms);
}

void CoreInterfaceClass::print(QString text)
{
  SW->ShowText(text);
}

void CoreInterfaceClass::clearText()
{
  SW->ClearText();
}

QString CoreInterfaceClass::GetTimeStamp()
{
  return QDateTime::currentDateTime().toString("H:m:s");
}

QString CoreInterfaceClass::GetDateTimeStamp()
{
  return QDateTime::currentDateTime().toString("d/M/yyyy H:m:s");
}

bool CoreInterfaceClass::save(QString fileName, QString str)
{
  if (!QFileInfo(fileName).exists())
    {
      abort("File does not exist: " + fileName);
      return false;
    }

  QFile file(fileName);
  if ( !file.open(QIODevice::Append ) )
  {
      abort("Cannot open file for appending:" + fileName);
      return false;
  }

  //qDebug() << str;
  str.replace("<br>", "\n");
  QTextStream outstream(&file);
  outstream << str;
  return true;
}

QVariant CoreInterfaceClass::loadColumn(QString fileName, int column)
{
  if (column<0 || column>2)
    {
      abort ("Supported loadColumn with column # 0, 1 and 2");
      return QVariant();
    }

  if (!QFileInfo(fileName).exists())
  {
    abort("File does not exist: " + fileName);
    return QVariant();
  }

  QVector<double> v1, v2, v3;
  int res;
  if (column == 0)
     res = LoadDoubleVectorsFromFile(fileName, &v1);
  else if (column == 1)
     res = LoadDoubleVectorsFromFile(fileName, &v1, &v2);
  else if (column == 2)
     res = LoadDoubleVectorsFromFile(fileName, &v1, &v2, &v3);

  if (res != 0)
      {
        abort("Error reading from file: "+fileName);
        return QVariant();
      }
  QList<QVariant> l;
  for (int i=0; i<v1.size(); i++)
    {
      if (column == 0) l.append(v1[i]);
      else if (column == 1) l.append(v2[i]);
      else if (column == 2) l.append(v3[i]);
    }

  return l;
}

QString CoreInterfaceClass::GetWorkDir()
{
  return SW->LastOpenDir;
}

QString CoreInterfaceClass::GetScriptDir()
{
  return SW->LibScripts;
}

QString CoreInterfaceClass::GetExamplesDir()
{
  return SW->ExamplesDir;
}

bool CoreInterfaceClass::createFile(QString fileName, bool AbortIfExists)
{
  if (QFileInfo(fileName).exists())
    {
      if (AbortIfExists)
        {
          abort("File already exists: " + fileName);
          return false;
        }
    }

  //create or clear content of the file
  QFile file(fileName);
  if ( !file.open(QIODevice::WriteOnly) )
  {
      abort("Cannot open file: "+fileName);
      return false;
  }
  return true;
}

bool CoreInterfaceClass::isFileExists(QString fileName)
{
  return QFileInfo(fileName).exists();
}


//--------------------------------------------------------------
InterfaceToAddObjScript::InterfaceToAddObjScript(MainWindow *MW)
  : MW(MW)
{
  H["Box"] = "Adds to geometry a volume 'name', which is a box with sizes (sizeX,sizeY,sizeZ)\n"
             "filled with material id 'material', and positioned inside 'container' object\n"
             "at local coordinates (x,y,z) with orientation angles (phi,thetha,psi)";
  H["Cylinder"] = "Adds to geometry a volume 'name', which is a cylinder with given diameter and height\n"
                  "filled with material id 'material', and positioned inside 'container' object\n"
                  "at local coordinates (x,y,z) with orientation angles (phi,thetha,psi)";
  H["Cone"] =  "Adds to geometry a volume 'name', which is a cone with given top and bottom diameters and height\n"
               "filled with material id 'material', and positioned inside 'container' object\n"
               "at local coordinates (x,y,z) with orientation angles (phi,thetha,psi)";
  H["Sphere"] = "Adds to geometry a volume 'name', which is a sphere of a given diameter\n"
                "filled with material id 'material', and positioned inside 'container' object\n"
                "at local coordinates (x,y,z) with orientation angles (phi,thetha,psi)";
  H["ClearAll"] = "Remove all AddGeoObjects existing in the current geometry";
  H["UpdateGeometry"] = "Update geometry and do optional check for collisions.\n"
                        "Performed automatically for standalone script";
}

void InterfaceToAddObjScript::Box(QString name, double Lx, double Ly, double Lz, int iMat, QString container, double x, double y, double z, double phi, double theta, double psi)
{
  AdditionalGeoObjectClass tmp = AdditionalGeoObjectClass(name,0, Lx,Ly,Lz, iMat, container, x,y,z, phi,theta,psi);
  tmp.fActivated = true;
  GeoObjects.append(tmp);
}

void InterfaceToAddObjScript::Cylinder(QString name, double D, double h, int iMat, QString container, double x, double y, double z, double phi, double theta, double psi)
{
  AdditionalGeoObjectClass tmp = AdditionalGeoObjectClass(name,1, D,D,h, iMat, container, x,y,z, phi,theta,psi);
  tmp.fActivated = true;
  GeoObjects.append(tmp);
}

void InterfaceToAddObjScript::Cone(QString name, double Dtop, double Dbot, double h, int iMat, QString container, double x, double y, double z, double phi, double theta, double psi)
{
  AdditionalGeoObjectClass tmp = AdditionalGeoObjectClass(name,2, Dtop,Dbot,h, iMat, container, x,y,z, phi,theta,psi);
  tmp.fActivated = true;
  GeoObjects.append(tmp);
}

void InterfaceToAddObjScript::Sphere(QString name, double D, int iMat, QString container, double x, double y, double z, double phi, double theta, double psi)
{
  AdditionalGeoObjectClass tmp = AdditionalGeoObjectClass(name,3, D,D,D, iMat, container, x,y,z, phi,theta,psi);
  tmp.fActivated = true;
  GeoObjects.append(tmp);
}

void InterfaceToAddObjScript::ClearAll()
{
  //emit clearRequested();
  //GeoObjects.clear();

  GeoObjects.clear();
  MW->Detector->AdditionalGeoObjects.clear();
  MW->ReconstructDetector();
}

//--------------------------------------------------------------
void InterfaceToAddObjScript::UpdateGeometry(bool CheckOverlaps)
{
  //checkup
  for (int i=0; i<GeoObjects.size(); i++)
    {
      QString name = GeoObjects[i].Name;
      bool fNotUnique = MW->DAwindow->isNameInUse(name);
      if (fNotUnique)
        {
          abort("Add geo object: Not unique object name encountered: "+name);
          GeoObjects.clear();
          return;
        }
      if (GeoObjects[i].Material > MW->Detector->MaterialCollection->size()-1)
        {
          abort("Add geo object: Wrong material index for object: "+name);
          GeoObjects.clear();
          return;
        }
      for (int j=0; j<GeoObjects.size(); j++)
        {
          if (i==j) continue;
          if (name == GeoObjects[j].Name)
            {
              abort("Add geo object: At least two objects have the same name: "+name);
              GeoObjects.clear();
              return;
            }
        }
      QString cont = GeoObjects[i].Container;
      bool fOK = MW->DAwindow->isNameInUse(cont);
      if (!fOK)
        {
          abort("Add geo object: Container does not exist: "+cont);
          GeoObjects.clear();
          return;
        }
    }

  //adding objects
  for (int i=0; i<GeoObjects.size(); i++)
      MW->Detector->AdditionalGeoObjects.append(GeoObjects[i]);
  GeoObjects.clear();

  MW->ReconstructDetector();

  //checking overlaps
  int opt = 2; //0-no overlaps, 1-overlaps detected, 2-overlaps not checked
  if (CheckOverlaps)
    {
      int nooverlaps = MW->CheckUpWindow->CheckGeoOverlaps();
      if (nooverlaps > 0)
        {
          MW->CheckUpWindow->show();
          abort("Add geo object: Overlap detected!");
          opt = 1;
        }
      else
        opt = 0;
    }  
  MW->DAwindow->UpdateAddObjIndicationAfterScript(opt);
}


double MathInterfaceClass::abs(double val)
{
  return std::abs(val);
}

double MathInterfaceClass::acos(double val)
{
  return std::acos(val);
}

double MathInterfaceClass::asin(double val)
{
  return std::asin(val);
}

double MathInterfaceClass::atan(double val)
{
  return std::atan(val);
}

double MathInterfaceClass::atan2(double y, double x)
{
  return std::atan2(y, x);
}

double MathInterfaceClass::ceil(double val)
{
  return std::ceil(val);
}

double MathInterfaceClass::cos(double val)
{
  return std::cos(val);
}

double MathInterfaceClass::exp(double val)
{
  return std::exp(val);
}

double MathInterfaceClass::floor(double val)
{
  return std::floor(val);
}

double MathInterfaceClass::log(double val)
{
  return std::log(val);
}

double MathInterfaceClass::max(double val1, double val2)
{
  return std::max(val1, val2);
}

double MathInterfaceClass::min(double val1, double val2)
{
  return std::min(val1, val2);
}

double MathInterfaceClass::pow(double val, double power)
{
  return std::pow(val, power);
}

double MathInterfaceClass::sin(double val)
{
  return std::sin(val);
}

double MathInterfaceClass::sqrt(double val)
{
  return std::sqrt(val);
}

double MathInterfaceClass::tan(double val)
{
  return std::tan(val);
}

InterfaceToHolesScript::InterfaceToHolesScript()
{
  H["Tube"] = "Adds a round hole with the given diameter at the provided x,y coordinates";
  H["Box"] = "Adds a rectangular hole with the given dx and dy sizes at the provided x,y coordinates.\n"
             "The last (optional) arguments specify the rotation angle (in degrees).";
}
