#ifndef SCRIPTINTERFACES_H
#define SCRIPTINTERFACES_H

//#include "classes.h"
#include "detectorclass.h"

#include <QVector>
#include <QVariant>

class GenericScriptWindowClass;
class MainWindow;

//basic object to inherit for any interface object: adds help system
class BaseInterfaceClass : public QObject
{
  Q_OBJECT

public:
  BaseInterfaceClass() {}
  virtual bool InitOnRun() {return true;} // automatically called before script evaluation
  virtual void ForceStop() {} // called when abort was triggered - used to e.g. abort simulation or reconstruction

  QString Description;  // used as help for the unit
  QString UnitName;     // used as default name of the unit

public slots:
  QString help(QString method);  //automatically requested to obtain help strings

signals:
  void AbortScriptEvaluation(QString); //abort request is automatically linked to abort slot of core unit

protected:
  QHash<QString, QString> H;

  void abort(QString message = "Aborted!") {emit AbortScriptEvaluation(message);}
};

class CoreInterfaceClass : public BaseInterfaceClass
{
  Q_OBJECT

public:
  CoreInterfaceClass(GenericScriptWindowClass* SW);

public slots:
  //abort execution of the script
  void abort(QString message = "Aborted!");

  //sleep
  void sleep(int ms);

  //output part of the script window
  void print(QString text);
  void clearText();

  //time stamps
  QString GetTimeStamp();
  QString GetDateTimeStamp();

  //save to file
  bool createFile(QString fileName, bool AbortIfExists = true);
  bool isFileExists(QString fileName);
  bool save(QString fileName, QString str);
  //load column of doubles from file and return it as an array
  QVariant loadColumn(QString fileName, int column = 0);

  //dirs
  QString GetWorkDir();
  QString GetScriptDir();
  QString GetExamplesDir();

private:
  GenericScriptWindowClass* SW;
};

class MathInterfaceClass : public BaseInterfaceClass
{
  Q_OBJECT

public:
  MathInterfaceClass() {}

public slots:
  double abs(double val);
  double acos(double val);
  double asin(double val);
  double atan(double val);
  double atan2(double y, double x);
  double ceil(double val);
  double cos(double val);
  double exp(double val);
  double floor(double val);
  double log(double val);
  double max(double val1, double val2);
  double min(double val1, double val2);
  double pow(double val, double power);
  double sin(double val);
  double sqrt(double val);
  double tan(double val);


private:

};

class InterfaceToHolesScript : public BaseInterfaceClass
{
  Q_OBJECT
public:
  QVector<MaskElementStructure> Holes;

  InterfaceToHolesScript();

public slots:
  void Tube(double x, double y, double diameter) {Holes.append(MaskElementStructure(x, y, diameter));}
  void Box(double x, double y, double dx, double dy, double angle = 0) {Holes.append(MaskElementStructure(x,y,dx,dy,angle));}
};

class InterfaceToAddObjScript : public BaseInterfaceClass
{
  Q_OBJECT
public:
  InterfaceToAddObjScript(MainWindow* MW);

  QVector<AdditionalGeoObjectClass> GeoObjects;

public slots:
  void Box(QString name, double Lx, double Ly, double Lz, int iMat, QString container, double x, double y, double z, double phi, double theta, double psi);
  void Cylinder(QString name, double D, double h, int iMat, QString container, double x, double y, double z, double phi, double theta, double psi);
  void Cone(QString name, double Dtop, double Dbot, double h, int iMat, QString container, double x, double y, double z, double phi, double theta, double psi);
  void Sphere(QString name, double D, int iMat, QString container, double x, double y, double z, double phi, double theta, double psi);
  void ClearAll();

  void UpdateGeometry(bool CheckOverlaps = true);

signals:
  void clearRequested();

private:
  MainWindow* MW;
};

#endif // SCRIPTINTERFACES_H
