#include "configscriptinterface.h"
#include "scriptfunctions.h"

#include <QDebug>
#include <QDir>
#include <QVariant>
#include <QScriptValue>
#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonArray>

const QChar ConfigScriptInterface::pathSep = QDir::separator();

static QJsonObject mergeJsonObject(const QJsonObject &base, const QJsonObject &overlay)
{
    QJsonObject result;
    for(QJsonObject::const_iterator item = overlay.constBegin(); item != overlay.constEnd(); item++)
    {
        if(base[item.key()].isObject() && item.value().isObject())
        {
            QJsonObject overlayitem = item.value().toObject();
            QJsonObject baseitem = base[item.key()].toObject();

            result[item.key()] = mergeJsonObject(baseitem, overlayitem);;
        }
        else
            result[item.key()] = item.value();
    }
    for(QJsonObject::const_iterator item = base.constBegin(); item != base.constEnd(); item++)
    {
        if(result.contains(item.key()))
            continue;
        result[item.key()] = item.value();
    }
    return result;
}

static QJsonObject wrapJsonObject(const QStringList &levels, QJsonObject obj)
{
    for(int j = levels.count()-1; j >= 0; j--)
    {
        QJsonObject tmp;
        tmp[levels[j]] = obj;
        obj = tmp;
    }
    return obj;
}

ConfigScriptInterface::ConfigScriptInterface()
{
    config = QJsonObject();
    cwd = "";
    fRemakeDetector = false;
}

QString ConfigScriptInterface::get(QString setting)
{
    QStringList pathtree = this->normalizepath(setting);
    if(pathtree.isEmpty())
    {
        //qDebug()<<"Returning all config";
        return QJsonDocument(this->config).toJson();
    }

    QJsonObject result = this->config;
    for(int i = 0; i < pathtree.count()-1; i++)
    {
        if(!result[pathtree[i]].isObject())
        {
            qDebug()<<result[pathtree[i]]<<"is not object";
            //qDebug()<<"Returning undefined";
            return "null";
        }
        result = result[pathtree[i]].toObject();
    }
    if(result.contains(pathtree.last()))
    {
      QJsonValue val = result[pathtree.last()];
      switch(val.type())
      {
        case QJsonValue::Bool: return val.toBool() ? "true" : "false";
        case QJsonValue::Double: return QString::number(val.toDouble());
        case QJsonValue::String: return val.toString();
        case QJsonValue::Array: return QJsonDocument(val.toArray()).toJson();
        case QJsonValue::Object: return QJsonDocument(val.toObject()).toJson();
        case QJsonValue::Null:
        case QJsonValue::Undefined:
        default: return "null";
      }
    }
    else
    {
        qDebug()<<pathtree.last()<<"doesn't exist";
        return "null";
    }
}

bool ConfigScriptInterface::loadFile(QString filename)
{
    qDebug()<<"loadFile("<<filename<<"):";
    QByteArray data = Functions::readAllFile(filename);
    if(data.isEmpty()) return false;

    this->config = QJsonDocument::fromJson(data).object();
    this->cwd = "";
    this->fRemakeDetector = true;
    return true;
}

bool ConfigScriptInterface::saveFile(QString filename)
{
  return Functions::saveJson(filename, config);
}

bool ConfigScriptInterface::set(QString setting, QVariant value)
{
    QStringList path = normalizepath(setting);
    QJsonObject json;
    if(path.isEmpty())
    {
        QJsonValue val = QJsonValue::fromVariant(value);
        if(!val.isObject())
        {
            qWarning()<<"Can't set / to a non-object value";
            return false;
        }
        json = val.toObject();
    }
    else
    {
        json[path.last()] = QJsonValue::fromVariant(value);
        path.removeLast();
        json = wrapJsonObject(path, json);
    }
    if(json.contains("DetectorConfig"))
      this->fRemakeDetector = true;
    this->config = mergeJsonObject(this->config, json);
    //qDebug()<<"Config after setting"<<setting<<"="<<value<<":"<<this->config;
    return true;
}

bool ConfigScriptInterface::cd(QString path)
{
    QStringList pathtree = this->normalizepath(path);
    if(pathtree.isEmpty())
    {
        //qDebug()<<"cd'ing to root";
        this->cwd = "";
        return true;
    }
    this->cwd = pathtree.join(pathSep);

    //Navigate through already existing paths
    int pathlevel;
    QJsonValue curcfg = config;
    //qDebug()<<"Entered /";
    for(pathlevel = 0; pathlevel < pathtree.count(); pathlevel++)
    {
        QJsonObject subbranch = curcfg.toObject();
        if(!subbranch.contains(pathtree[pathlevel]))
        {
            //qDebug()<<pathtree[pathlevel]<<"sub-object doesn't exist.";
            break;
        }
        curcfg = subbranch[pathtree[pathlevel]];
        if(!curcfg.isObject())
        {
            qWarning()<<pathtree[pathlevel]<<"is not a json object.";
            return false;
        }
        //qDebug()<<"Entered"<<pathtree[pathlevel];
    }
    //qDebug()<<"Current config path:"<<this->cwd;
    return true;
}

QString ConfigScriptInterface::pwd()
{
    return this->cwd;
}

void ConfigScriptInterface::clear()
{
    this->cwd = "";
    this->config = QJsonObject();
}

QStringList ConfigScriptInterface::normalizepath(QString path)
{
    //Check for relative directory
    if(path[0] != pathSep)
        path = this->cwd+pathSep+path;
    path = QDir::cleanPath(path);
    //qDebug()<<"Cleaned up path:"<<path;

    QStringList pathtree = path.split(pathSep, QString::SkipEmptyParts);
    //qDebug()<<"Path Tree:"<<pathtree;

    //Can't go up from root dir, and Qt doesn't care about it
    while(pathtree.count() && pathtree[0] == "..")
        pathtree.removeFirst();

    //qDebug()<<"Path Tree after removing leading '../'s' :"<<pathtree;
    return pathtree;
}
