#include "mainscriptinterface.h"
#include "scriptmode.h"
#include "scriptfunctions.h"
#include "configscriptinterface.h"

#include "eventsdataclass.h"
#include "detectorclass.h"
#include "reconstructionmanagerclass.h"
#include "simulationmanager.h"
#include "sensorlrfs.h"
#include "genericscriptwindowclass.h"

#include <QVariant>
#include <QScriptValue>
#include <QJsonObject>
#include <QMetaMethod>

#include <TTree.h>
#include <TFile.h>

#include <QFile>
#include <QScriptEngine>

MainScriptInterface::MainScriptInterface(ScriptMode *batch, QScriptEngine *engine)
{
  construct(batch);

  // Attach to engine
  QScriptValue interfacerVal = engine->newQObject(this, QScriptEngine::QtOwnership);
  QScriptValue configInterfaceVal = engine->newQObject(configInterface, QScriptEngine::QtOwnership);
  interfacerVal.setProperty("Config", configInterfaceVal);
  engine->globalObject().setProperty("Ants", interfacerVal);

  int methods = this->metaObject()->methodCount();
  for(int i = 0; i < methods; i++)
  {
    const QMetaMethod &m = this->metaObject()->method(i);
    bool fSlot = (m.methodType() == QMetaMethod::Slot);
    bool fSignal = (m.methodType() == QMetaMethod::Signal);
    bool fPublic = (m.access() == QMetaMethod::Public);
    if((fSlot || fSignal) && fPublic)
      engine->evaluate(m.name()+" = Ants."+m.name());
  }
  engine->evaluate("Config = Ants.Config");
}

#if 0
BatchScriptInterface::BatchScriptInterface(Batcher *batch, GenericScriptWindowClass *window)
{
  construct(batch, engine);

  window->SetInterfaceObject(this);
  window->SetInterfaceObject(configInterface, "Config");
}
#endif

MainScriptInterface::~MainScriptInterface()
{
  delete configInterface;
}

QJsonObject &MainScriptInterface::getConfig()
{
  return configInterface->getConfig();
}

QByteArray MainScriptInterface::loadFile(QString filename)
{
  return Functions::readAllFile(filename);
}

void MainScriptInterface::construct(ScriptMode *batch)
{
  this->detector = batch->detector;
  this->eventsHub = batch->eventsHub;
  this->reconMan = batch->reconMan;
  this->simMan = batch->simMan;

  configInterface = new ConfigScriptInterface();
}

void MainScriptInterface::setupDetector()
{
  if(configInterface->pleaseRemakeDetector())
  {
    qDebug() << "Setting up Detector";
    detector->makeDetector(configInterface->getConfig());
    eventsHub->clear();
    detector->SensLRF->clear(detector->pmCount());
    configInterface->weRemadeDetector();
    qApp->processEvents();
  }
}

void MainScriptInterface::print(QVariant msg)
{
  qDebug()<<"Script says:";
  if(msg.type() == QVariant::String)
    qDebug()<<"\t"<<msg.toString();
  else
    qDebug()<<msg;
}

void MainScriptInterface::simulate(int threadCount)
{
    setupDetector();
    simMan->setup(configInterface->getConfig(), threadCount);
    simMan->simulate();
    onNewSimlationData();
}

int MainScriptInterface::loadSimulation(QString fileName)
{
  setupDetector();
  int loadedEventsCount = eventsHub->loadSimulatedEventsFromTree(fileName, detector->PMs);
  if(loadedEventsCount > 0)
    onNewSimlationData();
  return loadedEventsCount;
}

bool MainScriptInterface::saveSimulation(QString fileName)
{
    qDebug() << "SaveSimulation("<<fileName<<"): ";
    if (eventsHub->Events.isEmpty())
    {
        qDebug() << "No data to save!";
        return false;
    }

    bool ok = eventsHub->saveSimulationAsTree(fileName);
    if (ok) qDebug() << "Wrote to " << fileName << " successfully!";
    else qDebug() << "Error writing to file " << fileName;
    return ok;
}

void MainScriptInterface::clearEvents()
{
  eventsHub->clear();
}

bool MainScriptInterface::reconstruct()
{
  setupDetector();
  bool ok = reconMan->reconstructAll(configInterface->getConfig());
  if(!ok) qDebug()<<"Reconstruction failed:"<<reconMan->getErrorString();
  qApp->processEvents();
  return ok;
}

bool MainScriptInterface::saveReconstruction(QString fileName)
{
  return eventsHub->saveReconstructionAsTree(fileName, detector->PMs);
}

void MainScriptInterface::filterEvents()
{
  reconMan->filterEvents(configInterface->getConfig());
  qApp->processEvents();
}

bool MainScriptInterface::makeLRFs(QVariant lrfcreationconfig)
{
  QJsonValue cfg = QJsonValue::fromVariant(lrfcreationconfig);
  if(!cfg.isObject())
  {
    qDebug()<<"LRF creation config must be a JsonObject (not value, like 5, or \"6\")";
    return false;
  }
  QJsonObject json = cfg.toObject();
  bool ok = detector->SensLRF->makeLRFs(json, eventsHub, detector->PMs);
  if(!ok) qDebug()<<detector->SensLRF->getLastError();
  qApp->processEvents();
  return ok;
}

bool MainScriptInterface::saveLRFs(QString fileName, int iter)
{
    QJsonObject json;
    if (!detector->SensLRF->saveIteration(iter, json))
    {
        qDebug()<<"Failed to saveIteration("<<iter<<", json):\n"<<detector->SensLRF->getLastError();
        return false;
    }

    return Functions::saveJson(fileName, json);
}

bool MainScriptInterface::loadLRFs(QVariant lrfconfig)
{
  QJsonObject json;

  switch(lrfconfig.type())
  {
    case QVariant::String:
    {
      QByteArray saveData = Functions::readAllFile(lrfconfig.toString());
      if(saveData.isEmpty()) return false;
      json = QJsonDocument::fromJson(saveData).object();
      break;
    }
    default:
    {
      QJsonValue val = QJsonValue::fromVariant(lrfconfig);
      if(!val.isObject())
      {
        qDebug()<<"LRF config must be a JsonObject (not value, like 5, or \"6\")";
        return false;
      }
      json = val.toObject();
      break;
    }
  }

  setupDetector();
  if(!detector->SensLRF->loadAll(json))
  {
      qDebug()<<"Failed to load LRFs:"<<detector->SensLRF->getLastError();
      return false;
  }
  return true;
}

void MainScriptInterface::onNewSimlationData()
{
  eventsHub->clearReconstruction();
  if(eventsHub->Events.size() == 0) return;
  QJsonObject recOptions = configInterface->getConfig()["ReconstructionConfig"].toObject();
  reconMan->filterEvents(recOptions);
  qApp->processEvents();
}

bool MainScriptInterface::generateScintillation()
{
  qDebug()<<"MainScriptInterface::generateScintillation() is not implemented yet";
  return false;
#if 0
  eventsHub->clear();
  detector->GeoManager->ClearTracks();

  //============ prepare config ============
  QJsonObject &config = configInterface->getConfig();
  config["Mode"] = "StackSim";
  config["DoGuiUpdate"] = false;
  GeneralSimSettings simSettings;
  simSettings.readFromJson(config);
  simSettings.fLogsStat = true; //force to report logs
  detector->makeDetector(config);

  //============ prepare simulator ============
  ParticleSourceSimulator *pss = new ParticleSourceSimulator(detector, "TestSimulator");
  pss->setSimSettings(&simSettings);
  pss->setup(config);
  //pss->setRngSeed(rng->Rndm()*1000000);

  //============ simulate ============
  /*bool fOK = pss->standaloneGenerateLight(&EnergyVector);

    if (!fOK)
        out << pss->getErrorString();
    else
        pss->appendToDataHub(eventsHub);
    delete pss;
    reconMan->filterEvents(config);
    return fOK;*/
#endif
}
