#include "scriptminimizer.h"

#include "mainwindow.h"
#include "genericscriptwindowclass.h"

#include <QScriptEngine>

#include "TMath.h"
#include "Math/Functor.h"
#include "Minuit2/Minuit2Minimizer.h"

static double bestResult = 1e30;
static bool fAbort = false;
static int numVariables = 0;
static QString FunctName = "";

double ScriptFunctor(const double *p)
//last parameter contains the pointer to MainWindow object
{
  if (fAbort) return 1e30;

  void *thisvalue;
  memcpy(&thisvalue, &p[numVariables], sizeof(void *));
  MainWindow* MW = (MainWindow*)thisvalue;

//  QString str;
//  for (int i=0; i<numVariables; i++)
//    str += QString::number(p[i])+"  ";
//  qDebug() << "Functor call with parameters:"<<str;

  QScriptValueList input;
  for (int i=0; i<numVariables; i++)
    input << p[i];

  QScriptValue sv = MW->GenScriptWindow->engine.globalObject().property(FunctName);
  double result = sv.call(QScriptValue(), input).toNumber();

//  qDebug() << "Minimization parameter value obtained:"<<result;
  return result;
}

void InterfaceToMinimizerScript::ForceStop()
{
  fAbort = true;
  qDebug() << "Abort requested for minimization procedure";
}

//void InterfaceToMinimizerScript::AbortMinimization()
//{
//  fAbort = true;
//  qDebug() << "Aborting...";
//}

void InterfaceToMinimizerScript::Clear()
{
  Name.clear();
  Start.clear();
  Step.clear();
  Min.clear();
  Max.clear();
  FunctorName = "";
}

void InterfaceToMinimizerScript::SetFunctorName(QString name)
{
  FunctorName = FunctName = name;
}

void InterfaceToMinimizerScript::AddVariable(QString name, double start, double step, double min, double max)
{
  Name << name;
  Start << start;
  Step << step;
  Min << min;
  Max << max;
}

void InterfaceToMinimizerScript::ModifyVariable(int varNumber, double start, double step, double min, double max)
{
  int size = Name.size();
  if (varNumber > size-1)
    {
      abort("Wrong variable number: " + QString::number(varNumber) + "  There are "+ QString::number(size) +" variables defined");
      return;
    }
  Start.replace(varNumber, start);
  Step.replace(varNumber, step);
  Min.replace(varNumber, min);
  Max.replace(varNumber, max);
}

QString InterfaceToMinimizerScript::Run()
{
  qDebug() << "Optimization run called";
  numVariables = Name.size();
  if (numVariables == 0)
    {
      qDebug() << "No variables defined!";
      abort("Variables are not defined!");
      return "";
    }

  QScriptValue sv = MW->GenScriptWindow->engine.globalObject().property(FunctName);
  if (!sv.isFunction())
    {
      qDebug() << "Minimization function not defined!";
      abort("Minimization function is not defined!");
      return "";
    }

  //making a copy of det and sim config
  QJsonObject json;
  MW->writeDetectorToJson(json);
  MW->writeSimSettingsToJson(json, true);

  //Creating ROOT minimizer
  ROOT::Minuit2::Minuit2Minimizer *RootMinimizer = new ROOT::Minuit2::Minuit2Minimizer(ROOT::Minuit2::kSimplex);//(ROOT::Minuit2::kMigrad);
  RootMinimizer->SetMaxFunctionCalls(500);
  RootMinimizer->SetMaxIterations(500);
 // RootMinimizer->SetTolerance(0.01);
  RootMinimizer->SetPrintLevel(1);

  // 1 standard
  // 2 try to improve minimum (slower)
  RootMinimizer->SetStrategy(1);

  ROOT::Math::Functor *Funct = new ROOT::Math::Functor(&ScriptFunctor, numVariables+1);
  RootMinimizer->SetFunction(*Funct);
  //setting up variables   -  start step min max
  for (int i=0; i<numVariables; i++)
      RootMinimizer->SetLimitedVariable(i, Name[i].toLatin1().data(),  Start[i], Step[i], Min[i], Max[i]);
  //prepare to transfer pointer to MainWindow
  double dPoint;
  void *thisvalue = MW;
  memcpy(&dPoint, &thisvalue, sizeof(void *));
  //We need to fix for the possibility that double isn't enough to store void*
  RootMinimizer->SetFixedVariable(numVariables, "p", dPoint);
  qDebug() << "Minimizer created and configured";

  // do the minimization
  fAbort = false;
  bool fOK = RootMinimizer->Minimize();
  fOK = fOK && !fAbort;

  //report results
  qDebug()<<"Minimization success? "<<fOK;

  //loading back settings of det and sim
  MW->fSimDataNotSaved = false;
  MW->readDetectorFromJson(json);
  MW->readSimSettingsFromJson(json);

  return "Done!";
}
